/*
 * Copyright 1995, Silicon Graphics, Inc.
 * ALL RIGHTS RESERVED
 *
 * UNPUBLISHED -- Rights reserved under the copyright laws of the United
 * States.   Use of a copyright notice is precautionary only and does not
 * imply publication or disclosure.
 *
 * U.S. GOVERNMENT RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in FAR 52.227.19(c)(2) or subparagraph (c)(1)(ii) of the Rights
 * in Technical Data and Computer Software clause at DFARS 252.227-7013 and/or
 * in similar or successor clauses in the FAR, or the DOD or NASA FAR
 * Supplement.  Contractor/manufacturer is Silicon Graphics, Inc.,
 * 2011 N. Shoreline Blvd. Mountain View, CA 94039-7311.
 *
 * THE CONTENT OF THIS WORK CONTAINS CONFIDENTIAL AND PROPRIETARY
 * INFORMATION OF SILICON GRAPHICS, INC. ANY DUPLICATION, MODIFICATION,
 * DISTRIBUTION, OR DISCLOSURE IN ANY FORM, IN WHOLE, OR IN PART, IS STRICTLY
 * PROHIBITED WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SILICON
 * GRAPHICS, INC.
 */

/*
 * IOC3 10/100 Mbps Fast Ethernet Device
 * initial speedracer/lego ioc3 driver
 *
 * history
 *
 * 02/95 initial code - nn
 *
 */

#define EF_DEBUG 1

#include <tcp-param.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/sysmacros.h>
#include <sys/cmn_err.h>
#include <sys/debug.h>
#include <sys/edt.h>
#include <sys/errno.h>
#include <sys/immu.h>
#include <sys/invent.h>
#include <sys/kopt.h>
#include <sys/mbuf.h>
#include <sys/sbd.h>
#include <sys/socket.h>
#include <sys/cpu.h>
#include <net/if.h>
#include <net/raw.h>
#include <net/soioctl.h>
#include <ether.h>
#include <sys/kmem.h>
#include <sys/PCI/ioc3.h>
#include <sys/if_ef.h>

/*
 * Declarations for variables in master.d/if_ef
 */
extern uint_t	ef_maxif;
extern uint_t	ef_nrbuf10;
extern uint_t	ef_nrbuf100;
extern uint_t	ef_nrbuf10_promisc;
extern uint_t	ef_nrbuf100_promisc;
extern uint_t	ef_rxdelay;
extern uint_t	ef_nrbufthreshdiv;
extern uint_t	ef_ntxd;
extern uint_t	ef_fd_ipgt;
extern uint_t	ef_fd_ipgr1;
extern uint_t	ef_fd_ipgr2;
extern uint_t	ef_hd_ipgt;
extern uint_t	ef_hd_ipgr1;
extern uint_t	ef_hd_ipgr2;
extern int	ef_halfduplex[];
extern int	ef_10mbps[];

/*
 * Other externs.
 */
extern void	qprintf(char *, ...);

/*
 * Prototypes.
 */
void 		if_efedtinit(struct edt *edtp);
static void	ef_tunesane(void);
static int	ef_allocef(struct ef_info *ei);
static void	ef_deallocef(struct ef_info *ei);
static void	ef_reclaim(struct ef_info *ei);
static int	ef_init(struct etherif *eif, int flags);
static int	ef_ioc3init(struct ef_info *ei, int flags);
static int	ef_phyinit(struct ef_info *ei, int flags);
static void	ef_phyerr(struct ef_info *ei);
static int	ef_phyget(struct ef_info *ei, int reg);
static void	ef_phyput(struct ef_info *ei, int reg, u_short val);
static void	ef_physet(struct ef_info *ei, int reg, u_short bits);
static void	ef_phyclear(struct ef_info *ei, int reg, u_short bits);
static void	ef_fill(struct ef_info *ei);
static void	ef_reset(struct etherif *eif);
static void	ef_ioc3reset(struct ef_info *ei);
static void	ef_watchdog(int unit);
static void	ef_getcdc(struct ef_info *ei);
static int	ef_intr(struct ef_info *ei);
static void	ef_recv(struct ef_info *ei);
static int	ef_transmit(struct etherif *eif, struct etheraddr *edst,
			struct etheraddr *esrc, u_short type,
			struct mbuf *m0);
static int	ef_ioctl(struct etherif *eif, int cmd, caddr_t data);
static int	ef_lafhash(u_char *addr, int len);
static void	ef_dump(int unit);
static void	ef_dumpif(struct ifnet *ifp);
static void	ef_dumpregs(struct ef_info *ei);
static void 	ef_hexdump(char *msg, char *cp, int len);

/*
 * Temp defines, for emulation only.
 */
 
#define KTOIOADDR(x) ((__uint64_t)(x))

/*
 * Misc defines.
 */

/* XXX enable this until bringup has completed */
#define DBG_PRINT(a)	/* if (ei->ei_if.if_flags & IFF_DEBUG) */ printf a

/*
 * Non-rx, non-tx, "other" interrupt conditions
 */
#define	EISR_OTHER \
	(EISR_RXOFLO | EISR_RXBUFOFLO | EISR_RXMEMERR | EISR_RXPARERR | \
	EISR_TXRTRY | EISR_TXEXDEF | EISR_TXLCOL | EISR_TXGIANT | \
	EISR_TXBUFUFLO | EISR_TXCOLLWRAP | EISR_TXDEFERWRAP | \
	EISR_TXMEMERR | EISR_TXPARERR)

#define	NEXTTXD(i)	(((i)+1) & (ei->ei_ntxd - 1))
#define	NEXTRXD(i)	(((i)+1) & (NRXD - 1))

#define	offsetof(s,m)	(size_t)(&(((s*)0)->m))

/*
 * Since the ioc3 compares the rx consume and produce indices
 * modulo-16, we increase number of receive buffers by 15.
 */
#define	NRBUFADJ(n)	(n + 15)

#define	TXDBUFSZ	(ETXD_DATALEN - sizeof (struct ether_header))

#define	EFMAXPHY	(3)			/* max phy to probe */

#define	ETHER_HDRLEN	(sizeof (struct ether_header))
#define	ETHER_CRCSZ	(4)			/* # bytes ethernet crc */

#define	EFDOG		(IFNET_SLOWHZ)		/* watchdog timeout */

#define	M2ALIGN(m)	((mtod((m), uint_t) & 1) == 0)
#define	M2LEN(m)	(((m)->m_len > 0) && ((m)->m_len & 1) == 0)
#define	M2(m)		(M2ALIGN(m) && M2LEN(m))
#define	M128ALIGN(m)	((mtod((m), uint_t) & 127) == 0)
#define	M1PAGE(m)	((mtod((m), uint_t) & ~(NBPP-1)) == (mtod((m), uint_t) & ~(NBPP-1)))

#define LAF_TSTBIT(laf, bit) \
	((laf)->laf_vec[(bit) >> 4] & (1 << ((bit) & 0xf)))
#define LAF_SETBIT(laf, bit) \
	((laf)->laf_vec[(bit) >> 4] |= (1 << ((bit) & 0xf)))
#define LAF_CLRBIT(laf, bit) \
	((laf)->laf_vec[(bit) >> 4] &= ~(1 << ((bit) & 0xf)))
#define LAF_SETALL(laf) \
	((laf)->laf_vec[0] = (laf)->laf_vec[1] = \
	 (laf)->laf_vec[2] = (laf)->laf_vec[3] = 0xFFFF)


static struct efrxbuf ef_dummyrbuf;

/*
 * Pointer to array of per-interface data structures.
 */
static struct ef_info *ef_info = NULL;

/*
 * Opsvec for ether.[ch] interface.
 */
static struct etherifops efops =
	{ ef_init, ef_reset, ef_watchdog, ef_transmit,
	  (int (*)(struct etherif*,int,void*))ef_ioctl };

#define	PATTERN	0x5555
static void
ssram_discovery(struct ef_info *ei)
{
	volatile __uint32_t *ssram_addr0 = &ei->ei_ssram[0];
	volatile __uint32_t *ssram_addr1 =
		&ei->ei_ssram[IOC3_SSRAM_LEN / (2 * sizeof(__uint32_t))];
	
	/*
	 * the ssram can be either 64kb or 128kb so we use
	 * a write-write-read sequence to find out which.
	 */

	/* assume the larger size SSRAM and enable parity checking */
	ei->ei_regs->emcr |= EMCR_BUFSIZ | EMCR_RAMPAR;

	*ssram_addr0 = PATTERN;
	/* first word in the second half of the larger size SSRAM */
	*ssram_addr1 = ~PATTERN & IOC3_SSRAM_DM;

	if ((*ssram_addr0 & IOC3_SSRAM_DM) != PATTERN ||
	    (*ssram_addr1 & IOC3_SSRAM_DM) != (~PATTERN & IOC3_SSRAM_DM)) {
		ei->ei_ssram_bits = EMCR_RAMPAR;
		ei->ei_regs->emcr &= ~EMCR_BUFSIZ;
	} else
		ei->ei_ssram_bits = EMCR_BUFSIZ | EMCR_RAMPAR;
}
	

/*
 * XXX Pretend we're a GIO device for emulation environment, update needed
 * for use with dynamic PCI configuration
 */
/* ARGSUSED */
void
if_efedtinit(struct edt *edtp)
{
	struct ef_info *ei;
	struct etheraddr ea;
	int s;
	int unit;
	int size;
#if defined(IFSIM) || defined(SABLE)
	char *ioc3_test;
#endif	/* IFSIM */

	if (ef_info)
		return;

	ef_tunesane();

	/* allocate ef_info[] */
	size = ef_maxif * sizeof(struct ef_info);
	if ((ef_info = (struct ef_info*) kmem_zalloc(size, KM_SLEEP)) == NULL) {
		cmn_err(CE_ALERT, "if_efedtinit: ef_info[] alloc failed!\n");
		return;
	}

	s = splimp();

	for (unit = 0; unit < ef_maxif; unit++) {
		ei = &ef_info[unit];

		/* this is NOT an Everest driver - but to allow a bit 
		 * of testing we allocate some memory to act as IOC3
		 * registers & SSRAM
		 */
		if (unit == 0) {
#if defined(IFSIM) || defined(SABLE)
			ioc3_test = kmem_zalloc(512*1024, KM_SLEEP);
			if (ioc3_test == NULL) {
                                cmn_err(CE_ALERT, "if_efedtinit: ioc3_test alloc failed!\n");
                                return;
			}
			ei->ei_regs = (volatile ioc3_eregs_t *)(ioc3_test+IOC3_REG_OFF);
			ei->ei_ssram = (__uint32_t *)(ioc3_test+IOC3_RAM_OFF);
#else	/* ndef IFSIM */
			ei->ei_regs = (volatile ioc3_eregs_t *)PHYS_TO_K1(IOC3_PCI_DEVIO_BASE + IOC3_REG_OFF);
			ei->ei_ssram = (volatile __uint32_t *)PHYS_TO_K1(IOC3_PCI_DEVIO_BASE + IOC3_RAM_OFF);
#endif	/* IFSIM */
                } else {
                	continue;
                }

		/* For Lego PCI devices we should be called with a field
		 * edtp->e_base which points to the IOC3 address space.
		 * Therefore, instead of looping through and probing for
		 * a unit, we should loop to find a free entry
		 * in ef_info[], and install ourselves there.
		 */

                /* Set the ERBAR register, to enable the barrier bit in
                 * xtalk packets from bridge.  This really should be handled
                 * in generic routines since all DMA PCI devices will likely
                 * need this.  Caution: these bits are XOR'd with the 64 bit
                 * DMA address so do not set a bit in both places!
                 */
                ei->ei_erbar = ERBAR_BARRIER_BIT << ERBAR_RXBARR_SHIFT;
                
#ifdef IFSIM
		/* note: IFSIM is for nn's simulation environment */
		if (unit == 0) {
			ei->ei_regs = &ereg;
			ether_copy(&ethersrcaddr, &ea);
		} else
			continue;
#endif
                /* we have a choice of 128 or 512 transmit descriptors,
                 * for now use 128, should be tunable, probably should
                 * be 128 if we later discover that we're on a 10 Mbit
                 * link XXX.
                 */
		ei->ei_ntxd = NTXDSML;
		if (ef_ntxd != 0)
			ei->ei_ntxd = ef_ntxd;

                ssram_discovery(ei);

		ef_ioc3reset(ei);

		/* XXX todo: load ether addr into "ea" here */

		if (showconfig)
			printf("ef%d: hardware ethernet address %s\n",
			   	unit, ether_sprintf(ea.ea_vec));

		ei->ei_unit = unit;

		if (ef_allocef(ei)) {
			cmn_err(CE_ALERT, "ef%d: kmem_alloc failed!\n",unit);
			splx(s);
			return;
		}

		/* XXX todo: register interrupt handler here */

		ether_attach(&ei->ei_eif, "ef", unit, (caddr_t)ei, &efops,
			&ea, INV_ETHER_EF, 0);
	}

	splx(s);
	return;
}

static void
ef_tunesane()
{
	if (ef_nrbufthreshdiv < 2) {
		cmn_err(CE_ALERT, "ef_nrbufthreshdiv bogus\n");
		ef_nrbufthreshdiv = 3;
	}
	if ((ef_ntxd != 0) && (ef_ntxd != NTXDSML) && (ef_ntxd != NTXDBIG)) {
		cmn_err(CE_ALERT, "ef_ntxd bogus\n");
		ef_ntxd = 0;
	}
}

/*
 * Allocate ef_info fields.  Return 0 or errno.
 */
static int
ef_allocef(struct ef_info *ei)
{
	int	size;
        int	npgs;
        pgno_t	pgno;
        
	if (ei->ei_txd)
		return 0;

	/* allocate transmit descriptor ring.
	 * must be 16k aligned for small (128 entry) ring
	 * must be 64k aligned for large (512 entry) ring
	 * we align to 64k for now...
	 */
	size = ei->ei_ntxd * TXDSZ;
	npgs = (size+NBPP-1)/NBPP;
	if ((pgno = contig_memalloc(npgs, ETBR_ALIGNMENT/NBPP, VM_DIRECT)) == NULL)
		goto fail;
	ei->ei_txd = (struct eftxd*)small_pfntova_K0(pgno);
	
	/* allocate transmit pending-mbuf vector */
	size = ei->ei_ntxd * sizeof (struct mbuf*);
	if ((ei->ei_txm = (struct mbuf**) kmem_zalloc(size, KM_SLEEP)) == NULL)
		goto fail;

	/* allocate receive descriptor ring, must be 4k aligned */
	size = NRXD * RXDSZ;
	npgs = (size+NBPP-1)/NBPP;
	if ((pgno = contig_memalloc(npgs, 1, VM_DIRECT)) == NULL)
		goto fail;
	ei->ei_rxd = (__uint64_t*)small_pfntova_K0(pgno);
	
	/* allocate receive pending-mbuf vector */
	size = NRXD * sizeof (struct mbuf*);
	if ((ei->ei_rxm = (struct mbuf**) kmem_zalloc(size, KM_SLEEP)) == NULL)
		goto fail;

	return (0);

fail:
	/* can be called even if only some were successful */
	ef_deallocef(ei);

	return (ENOMEM);
}

/*
 * Free ef_info fields prior to unloading.
 */
static void
ef_deallocef(struct ef_info *ei)
{
	int	size;

	size = ei->ei_ntxd * TXDSZ;
	if (ei->ei_txd)
          ;			/* XXX no contig_memalloc_free yet */
	
	size = ei->ei_ntxd * sizeof (struct mbuf*);
	if (ei->ei_txm)
          kmem_free((void*)ei->ei_txm, size);
	
	size = NRXD * RXDSZ;
	if (ei->ei_rxd)
          ;			/* XXX no contig_memalloc_free yet */
	
	size = NRXD * sizeof (struct mbuf*);
	if (ei->ei_rxm)
	  kmem_free((void*)ei->ei_rxm, size);
}

/*
 * Reclaim all the completed tx descriptors/mbufs.
 */
static void
ef_reclaim(struct ef_info *ei)
{
	int	i;
	int	consume;

	ASSERT(IFNET_ISLOCKED(&ei->ei_if));

	consume = (ei->ei_regs->etcir & ETCIR_TXCONSUME_MASK)/TXDSZ;
	ASSERT(consume < ei->ei_ntxd);

	for (i = ei->ei_txtail; i != consume; i = NEXTTXD(i)) {
		GOODMP(ei->ei_txm[i]);
		m_freem(ei->ei_txm[i]);
		ei->ei_txm[i] = NULL;
	}

	ei->ei_txtail = i;
}

/*
 * Initialize the interface.  Interrupts must already be off.
 * Return 0 or errno.
 */
static int
ef_init(struct etherif *eif,
	int flags)
{
	struct ef_info *ei;
	int unit;
	int rc;

	ei = eiftoei(eif);
	unit = ei->ei_unit;

	ASSERT(IFNET_ISLOCKED(&ei->ei_if));

	/* reset the chip */
	ef_reset(eif);

	/* clear transmit descriptor ring. */
	bzero((caddr_t) ei->ei_txd, ei->ei_ntxd * TXDSZ);
	ei->ei_txhead = ei->ei_txtail = 0;

	/* clear receive descriptor ring. */
	bzero((caddr_t) ei->ei_rxd, NRXD * RXDSZ);
	ei->ei_rxhead = ei->ei_rxtail = 0;

	/* IOC3 initialization */
	if (rc = ef_ioc3init(ei, flags))
		return (rc);

	/* post receive buffers */
	ef_fill(ei);

	/* set quick pointer to head receive buffer */
	if (ei->ei_rxm[ei->ei_rxhead])
		ei->ei_rb = mtod(ei->ei_rxm[ei->ei_rxhead], struct efrxbuf*);
	else
		ei->ei_rb = &ef_dummyrbuf;

	/* enable IOC3 */
	ei->ei_regs->emcr |= EMCR_TXEN | EMCR_TXDMAEN | EMCR_RXEN | EMCR_RXDMAEN;

	ei->ei_if.if_timer = EFDOG;	/* turn on watchdog */
	
	DBG_PRINT(("ef%d: ef_init completed\n",unit));
	return 0;
}

static int
ef_ioc3init(struct ef_info *ei,
	int flags)
{
	volatile ioc3_eregs_t	*regs = ei->ei_regs;
	int	phyreg1;
	int	mode100;
	__uint32_t	w;
	int	rc;

	/* phy initialization first */
	if (rc = ef_phyinit(ei, flags))
		return (rc);

	phyreg1 = ef_phyget(ei, 1);

	if (phyreg1 & (MII_R1_100TXFD | MII_R1_100TXHD))
		mode100 = 1;
	else
		mode100 = 0;

	/* determine number of rbufs to post */
	if (flags & IFF_PROMISC)
		ei->ei_nrbuf = mode100? ef_nrbuf100_promisc: ef_nrbuf10_promisc;
	else
		ei->ei_nrbuf = mode100? ef_nrbuf100: ef_nrbuf10;
	ei->ei_nrbuf = NRBUFADJ(ei->ei_nrbuf);

	/* set ipg values */
	if (phyreg1 & (MII_R1_100TXFD | MII_R1_10TXFD))	/* full duplex */
		w = (ef_fd_ipgr2 << ETCSR_IPGR2_SHIFT)
			| (ef_fd_ipgr1 << ETCSR_IPGR1_SHIFT)
			| ef_fd_ipgt;
	else
		w = (ef_hd_ipgr2 << ETCSR_IPGR2_SHIFT)
			| (ef_hd_ipgr1 << ETCSR_IPGR1_SHIFT)
			| ef_hd_ipgt;
	regs->etcsr = w;

	/* set receive barrier register */
	regs->erbar = ei->ei_erbar;

	/* set receive interrupt threshold */
	regs->ercsr = (ei->ei_nrbuf / ef_nrbufthreshdiv) & ERCSR_THRESH_MASK;

	/* set receive interrupt delay - XXX fixed for now */
	regs->ertr = ef_rxdelay;

	/* XXX not yet using rx timer register */

	/* set our individual ethernet address */
	w = (ei->ei_curaddr[3] << 24) | (ei->ei_curaddr[2] << 16)
		| (ei->ei_curaddr[1] << 8) | ei->ei_curaddr[0];
	regs->emar_l = w;
	w = (ei->ei_curaddr[5] << 8) | ei->ei_curaddr[4];
	regs->emar_h = w;

	/* set the multicast filter */
	/* XXX is this right? */
	if (flags & IFF_ALLMULTI) {
		regs->ehar_l = 0xffffffff;
		regs->ehar_h = 0xffffffff;
	} else {
		w = (ei->ei_laf.laf_vec[3] << 24) | (ei->ei_laf.laf_vec[2] << 16)
		| (ei->ei_laf.laf_vec[1] << 8) | ei->ei_laf.laf_vec[0];
		regs->ehar_l = w;
		w = (ei->ei_laf.laf_vec[7] << 24) | (ei->ei_laf.laf_vec[6] << 16)
		| (ei->ei_laf.laf_vec[5] << 8) | ei->ei_laf.laf_vec[4];
		regs->ehar_h = w;
	}

	/* set the receive ring base address */
	w = (__uint32_t) KTOIOADDR((caddr_t)ei->ei_rxd);
	ASSERT((w & ERBR_L_RXRINGBASE_MASK) == w);
	regs->erbr_l = w;
	regs->erbr_h = KTOIOADDR((caddr_t)ei->ei_rxd) >> 32;

	/* set the transmit ring base address */
	w = (__uint32_t) KTOIOADDR((caddr_t)ei->ei_txd);
	ASSERT((w & ETBR_L_TXRINGBASE_MASK) == w);
	regs->etbr_l = w;
	regs->etbr_h = KTOIOADDR((caddr_t)ei->ei_txd) >> 32;

	/* set the transmit ring size */
	regs->etbr_l |= (ei->ei_ntxd == NTXDBIG)? ETBR_L_RINGSZ512 : ETBR_L_RINGSZ128;
	
	/*
	 * Initialize the MAC CSR register to reflect the
	 * phy autonegotiated values and previously-determined
	 * sram parity and sram size bits.
	 * Leave the tx and rx channel bits OFF for now.
	 */
	w = 0;
	ASSERT(phyreg1 & (MII_R1_100TXFD | MII_R1_10TXFD));
	w |= EMCR_DUPLEX;
	if (flags & IFF_PROMISC)
		w |= EMCR_PROMISC;
	w |= EMCR_PADEN;
	w |= (offsetof(struct efrxbuf,rb_ebh.ebh_ether) << EMCR_RXOFF_SHIFT);
	w |= ei->ei_ssram_bits;
	regs->emcr = w;

	/* enable all interrupts except TXEMPTY */
	regs->eier = EISR_RXTIMERINT | EISR_RXTHRESHINT | EISR_RXOFLO
		| EISR_RXBUFOFLO | EISR_RXMEMERR | EISR_RXPARERR
		| EISR_TXRTRY | EISR_TXEXDEF | EISR_TXLCOL
		| EISR_TXGIANT | EISR_TXBUFUFLO | EISR_TXEXPLICIT
		| EISR_TXCOLLWRAP | EISR_TXDEFERWRAP | EISR_TXMEMERR
		| EISR_TXPARERR;

	return (0);
}

static int
ef_phyinit(struct ef_info *ei,
	int flags)
{
	int	unit;
	int	i;

	unit = ei->ei_unit;

	/* probe to find phy address */
	for (i = 0; i <= EFMAXPHY; i++)
			break;
	if (i == EFMAXPHY) {
		cmn_err(CE_ALERT, "ef%d: phyinit: can't find PHY\n", unit);
		return (ENXIO);
	}
	ei->ei_phyunit = i;

	DBG_PRINT(("ef%d: phy %d found, flags 0x%x\n", unit, ei->ei_phyunit, flags));

	/*
	 * Initialize PHY.
	 */

	/* first reset it */
	ef_phyput(ei, 0, MII_R0_RESET);
	
	/* XXX if run on Everest with a fake set of regs, hang here */	
	while (ef_phyget(ei, 0) & MII_R0_RESET)
		DELAYBUS(10);	/* XXX just infinite loop for now */
	/*
	 * If we have autonegotiate capability, then use it.
	 * Otherwise, manually set speed and duplex based
	 * on the master.d variables.
	 */
	if (ef_phyget(ei, 1) & MII_R1_AUTOCONFIG) {
		ef_physet(ei, 0, MII_R0_AUTOEN | MII_R0_RESTARTAUTO);
		while (!(ef_phyget(ei, 1) & MII_R1_AUTODONE))
			DELAYBUS(10);
	} else {
		DBG_PRINT(("ef%d: no phy autonegotiate capability", unit));
		if (ef_halfduplex[unit] <= 0)
			ef_phyclear(ei, 0, MII_R0_DUPLEX);
		else
			ef_physet(ei, 0, MII_R0_DUPLEX);
		if (ef_10mbps[unit] <= 0)
			ef_phyclear(ei, 0, MII_R0_SPEEDSEL);
		else
			ef_physet(ei, 0, MII_R0_SPEEDSEL);
			
	}
	DBG_PRINT(("ef%d: %dMbps %sduplex\n",
		unit,
		(ef_phyget(ei, 0) & MII_R0_SPEEDSEL)? 100: 10,
		(ef_phyget(ei, 0) & MII_R0_DUPLEX)? "full": "half"));

	/* make sure the isolate bit is not set */
	ef_phyclear(ei, 0, MII_R0_ISOLATE);

	/* check phy for errors */
	ef_phyerr(ei);

	DBG_PRINT(("ef%d: phyinit: 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n",
		unit,
		ef_phyget(ei, 0), ef_phyget(ei, 1), ef_phyget(ei, 2),
		ef_phyget(ei, 3), ef_phyget(ei, 4), ef_phyget(ei, 5),
		ef_phyget(ei, 6)));
	return (0);
}

/*
 * Check PHY register 1 for errors.
 *
 * We maintain a last-known-value for RemoteFault, Jabber,
 * and LinkStatus and use this to detect a change in the state.
 * We don't want to check the link status *during* phy
 * initialization, only afterwards.
 */
static void
ef_phyerr(struct ef_info *ei)
{
	int unit;
	int reg1;

	unit = ei->ei_unit;
	reg1 = ef_phyget(ei, 1);

	if (reg1 & MII_R1_REMFAULT) {
		if (ei->ei_remfaulting == 0)
			cmn_err(CE_ALERT, "ef%d: remote hub fault\n", unit);
		ei->ei_remfaulting = 1;
	} else {
		if (ei->ei_remfaulting)
			DBG_PRINT(("ef%d: remote hub ok\n", unit));
		ei->ei_remfaulting = 0;
	}

	if (reg1 & MII_R1_JABBER) {
		if (ei->ei_jabbering == 0)
			cmn_err(CE_ALERT, "ef%d: jabbering\n", unit);
		ei->ei_jabbering = 1;
	} else {
		if (ei->ei_jabbering)
			DBG_PRINT(("ef%d: jabber ok\n", unit));
		ei->ei_jabbering = 0;
	}

	if (reg1 & MII_R1_LINKSTAT) {
		if (ei->ei_linkdown == 0)
			cmn_err(CE_ALERT, "ef%d: link fail - check ethernet cable\n");
		ei->ei_linkdown = 1;
	} else {
		if (ei->ei_linkdown)
			cmn_err(CE_ALERT, "ef%d: link ok\n", unit);
		ei->ei_linkdown = 0;
	}
}

/*
 * Read a PHY register.
 *
 * Several error bits in register 1 are read-clear.
 * We don't care about these bits until after phy
 * initialization is over, then we explicit test for
 * them.
 */
static int
ef_phyget(struct ef_info *ei,
	int reg)
{
	volatile ioc3_eregs_t	*regs = ei->ei_regs;

#if EVEREST || SABLE
	return 0;
/*NOTREACHED*/
#endif
	/* XXX infinite loop on busy for now */
	while (regs->micr & MICR_BUSY) {
		DELAYBUS(1);
	}

	regs->micr = MICR_READTRIG | (ei->ei_phyunit << MICR_PHYADDR_SHIFT)
		| reg;

	while (regs->micr & MICR_BUSY)
		DELAYBUS(1);
#ifdef IFSIM
	return (phyreg[reg]);
#else
	return (regs->midr);
#endif
}

/*
 * Write a PHY register.
 */
static void
ef_phyput(struct ef_info *ei,
	int reg,
	u_short val)
{
	volatile ioc3_eregs_t	*regs = ei->ei_regs;

	while (regs->micr & MICR_BUSY)
		DELAYBUS(1);

	regs->midr = val;
	regs->micr = (ei->ei_phyunit << MICR_PHYADDR_SHIFT) | reg;

#ifdef IFSIM
	phyreg[reg] = val;
#endif

	while (regs->micr & MICR_BUSY)
		DELAYBUS(1);
}

/*
 * Set a PHY register bit or bits.
 */
static void
ef_physet(struct ef_info *ei,
	int reg,
	u_short bits)
{
	int	v;

	v = ef_phyget(ei, reg);
	v |= bits;
	ef_phyput(ei, reg, (u_short)v);
}

/*
 * Clear a PHY register bit or bits.
 */
static void
ef_phyclear(struct ef_info *ei,
	int reg,
	u_short bits)
{
	int	v;

	v = ef_phyget(ei, reg);
	v &= ~bits;
	ef_phyput(ei, reg, (u_short)v);
}

/*
 * Allocate and post receive buffers.
 */
static void
ef_fill(struct ef_info *ei)
{
	volatile struct efrxbuf *rb;
	struct mbuf *m;
	int	head, tail;
	int	n;
	int	i;

	/*
	 * Determine how many receive buffers we're lacking
	 * from the full complement, allocate, initialize,
	 * and post them, then update the chip rx produce index.
	 * Deal with m_vget() failure.
	 */

	head = ei->ei_rxhead;
	tail = ei->ei_rxtail;

	if (head <= tail)
		n = ei->ei_nrbuf - (tail - head);
	else
		n = ei->ei_nrbuf - (NRXD - head) - tail;

	for (i = 0; i < n; i++) {
		if ((m = m_vget(M_DONTWAIT, sizeof (struct efrxbuf),
			MT_DATA)) == NULL)
			break;
		ASSERT(M128ALIGN(m));

		rb = mtod(m, struct efrxbuf*);
		rb->rb_rxb.w0 = 0;
		rb->rb_rxb.err = 0;
		ei->ei_rxm[tail] = m;
		ei->ei_rxd[tail] = KTOIOADDR((caddr_t)rb);

		tail = NEXTRXD(tail);
	}

	ei->ei_rxtail = tail;

	ei->ei_regs->erpir = ERPIR_ARM | (tail * RXDSZ);
}

/*
 * Reset the interface, leaving it disabled.
 */
static void
ef_reset(struct etherif *eif)
{
	struct ef_info *ei;
	int i;
	
	ei = eiftoei(eif);

	ASSERT(IFNET_ISLOCKED(&ei->ei_if));

	/* stop watchdog */
	eiftoifp(eif)->if_timer = 0;

	/* remember to snarf the collision and defer counts */
	ef_getcdc(ei);

	/* reset the ioc3 ethernet state */
	ef_ioc3reset(ei);

	/* free any pending transmit mbufs */
	for (i = 0; i < ei->ei_ntxd; i++)
		if (ei->ei_txm[i]) {
			m_freem(ei->ei_txm[i]);
			ei->ei_txm[i] = NULL;
		}

	/* free the posted rbufs */
	for (i = 0; i < NRXD; i++)
		if (ei->ei_rxm[i]) {
			m_freem(ei->ei_rxm[i]);
			ei->ei_rxm[i] = NULL;
		}

	ei->ei_rb = &ef_dummyrbuf;
}

/*
 * Reset ethernet portions of IOC3 chip.
 */
static void
ef_ioc3reset(struct ef_info *ei)
{
	ei->ei_regs->emcr = EMCR_RST;
	ei->ei_regs->emcr;
	DELAY(4);
	ei->ei_regs->emcr = 0x0;

	ASSERT(ei->ei_regs->emcr == 0);
}

/*
 * Do miscellaneous periodic things.
 */
static void
ef_watchdog(int unit)
{
	struct ef_info *ei;
	int s;

	ei = &ef_info[unit];

	s = splimp();

	ASSERT(IFNET_ISLOCKED(&ei->ei_if));

	/* update our collision and defer counts */
	ef_getcdc(ei);

	/* reclaim any completed tx buffers */
	ef_reclaim(ei);

	/* refill any missing receive buffers */
	ef_fill(ei);

	/* recover from earlier outofmbufs condition */
	if ((ei->ei_rb == &ef_dummyrbuf) && ei->ei_rxm[ei->ei_rxhead])
		ei->ei_rb = mtod(ei->ei_rxm[ei->ei_rxhead], struct efrxbuf*);

	/* check for phy errors */
	ef_phyerr(ei);

	ei->ei_if.if_timer = EFDOG;

	splx(s);

	/* check for missed interrupt */
	IFNET_UNLOCKNOSPL(&ei->ei_if);
	(void) ef_intr(ei);
	IFNET_LOCKNOSPL(&ei->ei_if);
}
	
/*
 * Read chip collision and defer counters and update ifnet stats.
 */
static void
ef_getcdc(struct ef_info *ei)
{
	int etcdc;

	etcdc = ei->ei_regs->etcdc;
	ei->ei_if.if_collisions += (etcdc & ETCDC_COLLCNT_MASK);
	ei->ei_defers += (etcdc >> ETCDC_DEFERCNT_SHIFT);
}

/*
 * generic ethernet interface interrupt.  Come here at splimp().
 */
static int
ef_intr(register struct ef_info *ei)
{
	int	unit;
	int	isr;
	int	fatal;
	int	s;

	unit = ei->ei_unit;

	s = splimp();
	IFNET_LOCKNOSPL(&ei->ei_if);

	if (ei->ei_txd == NULL)
		goto bad;
	if (iff_dead(ei->ei_if.if_flags))
		goto bad;

	/* read and clear all interrupt bits */
	isr = ei->ei_regs->eisr;
	ei->ei_regs->eisr = isr;
#ifdef IFSIM
	ei->ei_regs->eisr &= ~ei->ei_regs->eisr;	/* clear bits */
#endif

	if (ei->ei_rb->rb_rxb.w0 & ERXBUF_V)
		ef_recv(ei);

	if ((isr & EISR_OTHER) == 0) {	/* if nothing else interesting */
		IFNET_UNLOCKNOSPL(&ei->ei_if);
		splx(s);
		return (1);
	}

	fatal = 0;

	if (isr & EISR_RXOFLO) {
		ei->ei_drop++;
		ei->ei_if.if_ierrors++;
		fatal = 1;
		DBG_PRINT(("ef%d: dropped packets\n", unit));
	}

	if (isr & EISR_RXBUFOFLO) {
		ei->ei_ssramoflo++;
		ei->ei_if.if_ierrors++;
		fatal = 1;
		DBG_PRINT(("ef%d: receive overflow\n", unit));
	}

	if (isr & EISR_RXMEMERR) {
		ei->ei_memerr++;
		ei->ei_if.if_ierrors++;
		fatal = 1;
		DBG_PRINT(("ef%d: receive PCI error\n", unit));
	}

	if (isr & EISR_RXPARERR) {
		ei->ei_parityerr++;
		ei->ei_if.if_ierrors++;
		fatal = 1;
		cmn_err(CE_ALERT, "ef%d:  rx SSRAM hardware parity error\n", unit);
	}

	if (isr & EISR_TXRTRY) {
		ei->ei_rtry++;
		ei->ei_if.if_oerrors++;
		DBG_PRINT(("ef%d: excessive collisions\n", unit));
	}

	if (isr & EISR_TXEXDEF) {
		ei->ei_exdefer++;
		ei->ei_if.if_oerrors++;
		DBG_PRINT(("ef%d: excessive defers\n", unit));
	}

	if (isr & EISR_TXLCOL) {
		ei->ei_lcol++;
		ei->ei_if.if_oerrors++;
		DBG_PRINT(("ef%d: late collision\n", unit));
	}

	if (isr & EISR_TXGIANT) {
		ei->ei_txgiant++;
		ei->ei_if.if_oerrors++;
		DBG_PRINT(("ef%d: transmit giant packet error\n", unit));
	}

	if (isr & EISR_TXBUFUFLO) {
		ei->ei_ssramuflo++;
		ei->ei_if.if_oerrors++;
		fatal = 1;
		DBG_PRINT(("ef%d: transmit underflow\n", unit));
	}

	if (isr & (EISR_TXCOLLWRAP | EISR_TXDEFERWRAP)) {
		ef_getcdc(ei);
		DBG_PRINT(("ef%d: collision or defer count wrapped\n", unit));
	}

	if (isr & EISR_TXMEMERR) {
		ei->ei_memerr++;
		ei->ei_if.if_oerrors++;
		fatal = 1;
		DBG_PRINT(("ef%d: transmit PCI error\n", unit));
	}

	if (isr & EISR_TXPARERR) {
		ei->ei_parityerr++;
		ei->ei_if.if_oerrors++;
		fatal = 1;
		cmn_err(CE_ALERT, "ef%d:  tx SSRAM hardware parity error\n", unit);
	}

	if (fatal)
		(void) ef_init(&ei->ei_eif, ei->ei_if.if_flags);

	IFNET_UNLOCKNOSPL(&ei->ei_if);
	splx(s);

	return (1);

bad:
	ether_stop(&ei->ei_eif, ei->ei_if.if_flags);
	IFNET_UNLOCKNOSPL(&ei->ei_if);
	splx(s);

	return (0);
}

/*
 * Process received packets.
 */
static void
ef_recv(struct ef_info *ei)
{
	struct efrxbuf *rb;
	struct ether_header *eh;
	struct mbuf *m;
	int snoopflags;
	int unit;
	int head, tail;
	int w0;
	int err;
	int rlen;
	int i;

	head = ei->ei_rxhead;
	tail = ei->ei_rxtail;

	unit = ei->ei_unit;

	/*
	 * Process all the received packets.
	 */
	for (i = head; i != tail; i = NEXTRXD(i)) {
		m = ei->ei_rxm[i];
		GOODMP(m);
		rb = mtod(m, struct efrxbuf*);
		w0 = rb->rb_rxb.w0;
		err = rb->rb_rxb.err;

		if ((w0 & ERXBUF_V) == 0)
			break;

		ei->ei_rxm[i] = NULL;

		rlen = ((w0 & ERXBUF_BYTECNT_MASK) >> ERXBUF_BYTECNT_SHIFT) - ETHER_CRCSZ;
		m->m_off += sizeof (struct ioc3_erxbuf);

		IF_INITHEADER(&rb->rb_ebh, &ei->ei_if,
			sizeof(struct etherbufhead));

		eh = &rb->rb_ebh.ebh_ether;

#if EF_DEBUG
		ef_hexdump("ef_received_header",(char *)eh,64);
#endif

		/* XXX todo: ip checksum here */

		/* The mac110 doesn't hear its own packets */
                /* if it did, we would check for our own packets here */
                
		ei->ei_if.if_ipackets++;
		ei->ei_if.if_ibytes += rlen;

		if ((rlen < ETHER_HDRLEN) || err)
			goto error;

		m->m_len = (rlen - ETHER_HDRLEN) + sizeof (struct etherbufhead);

		ether_input(&ei->ei_eif, 0, m);

		continue;

error:
		snoopflags = SN_ERROR;

		if (rlen < ETHER_HDRLEN) {
			snoopflags |= SNERR_TOOSMALL;
			ei->ei_runt++;
			DBG_PRINT(("ef%d: runt packet\n", unit));
			rlen = ETHER_HDRLEN;
		}

		if (err & ERXBUF_CRCERR) {
			if (rlen + ETHER_CRCSZ == GIANT_PACKET_SIZE) {
				snoopflags |= SNERR_TOOBIG;
				ei->ei_giant++;
				DBG_PRINT(("ef%d: giant packet received\n", unit));
				rlen = ETHERMTU + ETHER_HDRLEN;
			}
			else {
				snoopflags |= SNERR_CHECKSUM;
				ei->ei_crc++;
				DBG_PRINT(("ef%d: crc error\n", unit));
			}
		}

		if (err & ERXBUF_FRAMERR) {
			snoopflags |= SNERR_FRAME;
			ei->ei_frame++;
			DBG_PRINT(("ef%d: framing error\n", unit));
		}

		if (err & ERXBUF_CODERR) {
			snoopflags |= SNERR_FRAME;	/* XXX ? */
			ei->ei_code++;
			DBG_PRINT(("ef%d: coding error\n", unit));
		}

		if (err & ERXBUF_INVPREAMB) {
			snoopflags |= SNERR_FRAME;	/* XXX ? */
			ei->ei_invpreamb++;
			DBG_PRINT(("ef%d: invalid preamble\n", unit));
		}

		m->m_len = (rlen - ETHER_HDRLEN) + sizeof (struct etherbufhead);

		ether_input(&ei->ei_eif, snoopflags, m);
	}

	/* update rxd head */
	ei->ei_rxhead = i;

	/* post any necessary receive buffers */
	ef_fill(ei);

	/* update quick rbuf pointer */
	if (ei->ei_rxm[i])
		ei->ei_rb = mtod(ei->ei_rxm[i], struct efrxbuf*);
	else
		ei->ei_rb = &ef_dummyrbuf;
}

static int
ef_transmit(
	struct etherif *eif,
	struct etheraddr *edst,
	struct etheraddr *esrc,
	u_short type,
	struct mbuf *m0)
{
	struct ef_info *ei;
	struct mbuf *m;
	struct mbuf *m1, *m2, *m3;
	volatile struct eftxd	*txd;
	struct ether_header *eh;
	int mlen;
	int mcnt;
	int total;

	ei = eiftoei(eif);

	/*
	 * The basic performance idea is to keep a dense,
	 * linear fastpath and throw rare conditions out via "goto".
	 */

	if (ei->ei_linkdown)
		goto linkdown;

top:
	/* allocate tx descriptor */
	txd = &ei->ei_txd[ei->ei_txhead];
	if (NEXTTXD(ei->ei_txhead) == ei->ei_txtail)
		goto outoftxd;

	/*
	 * Each 128byte IOC3 tx descriptor describes a single
	 * packet and contains up to 104 bytes of data locally,
	 * plus two pointers to noncontiguous data.
	 *
	 * The IOC3 tx has the following restrictions:
	 * - if data0 and buf1 are used, then data0 must
	 *   be an even number of bytes in length.
	 * - if used, buf1 must always start on an even byte address.
	 * - if used, buf2 must start on a 128byte address
	 *   and buf1 must be an even number of bytes in length
	 * - neither buf1 nor buf2 may cross a page boundary
	 */

	mlen = mcnt = 0;
	for (m = m0; m; m = m->m_next) {
		mlen += m->m_len;
		mcnt++;
	}

	if (mlen > ETHERMTU)
		goto toobig;

	/* create ether header */
	eh = (struct ether_header*) &txd->eh;
	*(struct etheraddr*)eh->ether_dhost = *edst;
	*(struct etheraddr*)eh->ether_shost = *esrc;
	eh->ether_type = type;

	total = mlen + sizeof (*eh);

	if (mlen <= TXDATASZ) {	/* if small, copy it into txd */
		m_datacopy(m0, 0, mlen, (caddr_t)txd->data);
		txd->cmd = ETXD_D0V | total;
		txd->bufcnt = mlen + sizeof(*eh);
	}
	else if (mcnt == 1) {	/* point buf1 at it */
		if (!M2ALIGN(m0) || !M1PAGE(m0))
			goto copyall;
		txd->cmd = ETXD_D0V | ETXD_B1V | total;
		txd->bufcnt = (m0->m_len << ETXD_B1CNT_SHIFT) | sizeof (*eh);
		txd->p1 = KTOIOADDR((caddr_t)mtod(m0, caddr_t));
	}
	else if (mcnt == 2) {	/* point buf1 and buf2 at it */
		m1 = m0;
		m2 = m1->m_next;
		if (!M2ALIGN(m1) || !M2LEN(m1) || !M1PAGE(m1)
			|| !M128ALIGN(m2) || !M1PAGE(m2))
			goto copyall;
		txd->cmd = ETXD_D0V | ETXD_B1V | ETXD_B2V | total;
		txd->bufcnt = (m1->m_len << ETXD_B1CNT_SHIFT) |
				(m2->m_len << ETXD_B2CNT_SHIFT) |
				sizeof (*eh);
		txd->p1 = KTOIOADDR((caddr_t)mtod(m1, caddr_t));
		txd->p2 = KTOIOADDR((caddr_t)mtod(m2, caddr_t));
	}
	else if (mcnt == 3) {	/* copy m1, point buf1 and buf2 */
		m1 = m0;
		m2 = m1->m_next;
		m3 = m2->m_next;
		if ((m1->m_len > TXDATASZ) || !M2LEN(m1)
			|| !M2ALIGN(m2) || !M2LEN(m2) || !M1PAGE(m2)
			|| !M128ALIGN(m3) || !M1PAGE(m3))
			goto copyall;
		m_datacopy(m1, 0, m1->m_len, (caddr_t)txd->data);
		txd->cmd = ETXD_D0V | ETXD_B1V | ETXD_B2V | total;
		txd->bufcnt = (m2->m_len << ETXD_B1CNT_SHIFT) |
				(m3->m_len << ETXD_B2CNT_SHIFT) |
				m1->m_len + sizeof (*eh);
		txd->p1 = KTOIOADDR((caddr_t)mtod(m2, caddr_t));
		txd->p2 = KTOIOADDR((caddr_t)mtod(m3, caddr_t));
	}
	else {	/* copy it all into a single new large mbuf */
copyall:
		if ((m = m_vget(M_DONTWAIT, mlen, MT_DATA)) == NULL)
			goto outofmbufs;
		ASSERT(M2ALIGN(m));
		m_datacopy(m0, 0, mlen, mtod(m, caddr_t));
		m->m_len = mlen;
		m_freem(m0);
		m0 = m;
		txd->cmd = ETXD_D0V | ETXD_B1V | total;
		txd->bufcnt = (m0->m_len << ETXD_B1CNT_SHIFT) | sizeof (*eh);
		txd->p1 = KTOIOADDR((caddr_t)mtod(m0, caddr_t));
	}

	/* XXX todo: fill in txd checksum fields */

	/* save mbuf chain to free later */
	ASSERT(ei->ei_txm[ei->ei_txhead] == 0);
	ei->ei_txm[ei->ei_txhead] = m0;

	/* increment index of txd head */
	ei->ei_txhead = NEXTTXD(ei->ei_txhead);

	/* go */
	ei->ei_regs->etpir = ei->ei_txhead * TXDSZ;

	ei->ei_if.if_obytes += total;

	/* check for received packets */
	if (ei->ei_rb->rb_rxb.w0 & ERXBUF_V)
		ef_recv(ei);

	if (RAWIF_SNOOPING(&ei->ei_rawif))
		goto snoop;

	return (0);

snoop:
	/* XXX todo: can the ioc3 hear itself when transmitting
	 * in regular and promiscuous modes?
	 */
	return (0);

toobig:
	m_freem(m0);
	DBG_PRINT(("ef%d: ef_transmit: invalid mbuf len: %d\n", ei->ei_unit, mlen));
	return (E2BIG);

outoftxd:
	ef_reclaim(ei);
	if (NEXTTXD(ei->ei_txhead) != ei->ei_txtail)
		goto top;
	m_freem(m0);
	DBG_PRINT(("ef%d: ef_transmit: out of tx descriptors\n", ei->ei_unit));
	return (ENOBUFS);

outofmbufs:
	m_freem(m0);
	DBG_PRINT(("ef%d: ef_transmit: out of mbufs\n", ei->ei_unit));
	return (ENOBUFS);

linkdown:
	m_freem(m0);
	DBG_PRINT(("ef%d: ef_transmit: link down\n", ei->ei_unit));
	return (EIO);
}

static int
ef_ioctl(struct etherif *eif,
	int cmd,
	caddr_t data)
{
	struct ef_info *ei;
	struct mfreq *mfr;
	union mkey *key;
	int rc;

	ei = eiftoei(eif);

	ASSERT(IFNET_ISLOCKED(&ei->ei_if));

	mfr = (struct mfreq*) data;
	key = mfr->mfr_key;

	switch (cmd) {
	case SIOCADDMULTI:
		mfr->mfr_value = ef_lafhash(key->mk_dhost,
			sizeof(key->mk_dhost));
		if (LAF_TSTBIT(&ei->ei_laf, mfr->mfr_value)) {
			ASSERT(mfhasvalue(&eif->eif_mfilter, mfr->mfr_value));
			break;
		}
		ASSERT(!mfhasvalue(&eif->eif_mfilter, mfr->mfr_value));
		LAF_SETBIT(&ei->ei_laf, mfr->mfr_value);
		if (ei->ei_nmulti == 0)
			ei->ei_if.if_flags |= IFF_FILTMULTI;
		ei->ei_nmulti++;
		if (rc = ef_init(eif, ei->ei_if.if_flags))
			return (rc);
		break;

	case SIOCDELMULTI:
		if (mfr->mfr_value 
		    != ef_lafhash(key->mk_dhost, sizeof(key->mk_dhost)))
			return (EINVAL);
		if (mfhasvalue(&eif->eif_mfilter, mfr->mfr_value))
			break;
		/*
	 	 * If this multicast address is the last one to map
		 * the bit numbered by mfr->mfr_value in the filter,
		 * clear that bit and update the chip.
	 	 */
		LAF_CLRBIT(&ei->ei_laf, mfr->mfr_value);
		ei->ei_nmulti--;
		if (ei->ei_nmulti == 0)
			ei->ei_if.if_flags &= ~IFF_FILTMULTI;
		if (rc = ef_init(eif, ei->ei_if.if_flags))
			return (rc);
		break;

	default:
		return (EINVAL);;
	}

	return (0);
}

/*
 * Given a multicast ethernet address, this routine calculates the 
 * address's bit index in the logical address filter mask
 */
#define CRC_MASK	0xEDB88320

static int
ef_lafhash(u_char *addr, int len)
{
	u_int crc;
	u_char byte;
	int bits;
	u_int temp = 0;

	for (crc = ~0; --len >= 0; addr++) {
		byte = *addr;
		for (bits = 8; --bits >= 0; ) {
			if ((byte ^ crc) & 1)
				crc = (crc >> 1) ^ CRC_MASK;
			else
				crc >>= 1;
			byte >>= 1;
		}
	}

	crc &= 0x3f;	/* bit reverse lowest 6 bits for hash index */
	for (bits = 6; --bits >= 0; ) {
		temp <<= 1;
		temp |= crc & 0x1;
		crc >>= 1;
	}

	return temp;
}

static void
ef_dump(int unit)
{
	struct ef_info *ei;

	if ((unit < 0) || (unit >= ef_maxif)) {
		qprintf("no ef controller %d\n", unit);
		return;
	}

	ei = &ef_info[unit];

	ef_dumpif(&ei->ei_if);
	qprintf("ac_enaddr %s\n", ether_sprintf(ei->ei_curaddr));
	qprintf("eif_rawif 0x%x unit %d phyunit %d\n",
		&ei->ei_rawif, ei->ei_unit, ei->ei_phyunit);
	ef_dumpregs(ei);
	qprintf("ntxd %d txd 0x%x txhead %d txtail %d txm 0x%x linkdown %d\n",
		ei->ei_ntxd, ei->ei_txd, ei->ei_txhead, ei->ei_txtail,
		ei->ei_txm, ei->ei_linkdown);
	qprintf("nrbuf %d rxd 0x%x rxhead %d rxtail %d rb 0x%x rxm 0x%x\n",
		ei->ei_nrbuf, ei->ei_rxd, ei->ei_rxhead, ei->ei_rxtail,
		ei->ei_rb, ei->ei_rxm);
	qprintf("laf 0x%x nmulti %d erbar 0x%x ssram_bits 0x%x\n",
		ei->ei_laf, ei->ei_nmulti, ei->ei_erbar, ei->ei_ssram_bits);
	qprintf("remfaulting %d jabbering %d\n",
		ei->ei_remfaulting, ei->ei_jabbering);
	qprintf("defers %d runt %d crc %d frame %d giant %d code %d\n",
		ei->ei_defers, ei->ei_runt, ei->ei_crc, ei->ei_frame,
		ei->ei_giant, ei->ei_code);
	qprintf("invpreamb %d drop %d ssramoflo %d ssramuflo %d memerr %d\n",
		ei->ei_invpreamb, ei->ei_drop, ei->ei_ssramoflo,
		ei->ei_ssramuflo, ei->ei_memerr);
	qprintf("parityerr %d rtry %d exdefer %d lcol %d txgiant %d\n",
		ei->ei_parityerr, ei->ei_rtry, ei->ei_exdefer,
		ei->ei_lcol, ei->ei_txgiant);
}

static void
ef_dumpif(struct ifnet *ifp)
{
	qprintf("if_name \"%s\", if_unit %d if_mtu %d if_flags 0x%x if_timer %d\n",
		ifp->if_name, ifp->if_unit, ifp->if_mtu, ifp->if_flags, ifp->if_timer);
	qprintf("ifq_len %d ifq_maxlen %d ifq_drops %d ifq_lock 0x%x\n",
		ifp->if_snd.ifq_len, ifp->if_snd.ifq_maxlen,
		ifp->if_snd.ifq_drops, ifp->if_snd.ifq_lock);
	qprintf("if_ipackets %d if_ierrors %d if_opackets %d if_oerrors %d\n",
		ifp->if_ipackets, ifp->if_ierrors,
		ifp->if_opackets, ifp->if_oerrors);
	qprintf("if_collisions %d if_ibytes %d if_obytes %d if_odrops %d\n",
		ifp->if_collisions, ifp->if_ibytes,
		ifp->if_obytes, ifp->if_odrops);
}

static void
ef_dumpregs(struct ef_info *ei)
{
	volatile ioc3_eregs_t *regs = ei->ei_regs;

	qprintf("ei 0x%x ", ei);
	qprintf("ei_regs 0x%x\n", regs);
	qprintf("\temcr 0x%x eisr 0x%x eier 0x%x\n",
		regs->emcr, regs->eisr, regs->eier);
	qprintf("\tercsr 0x%x erbr_l 0x%x erbr_h 0x%x erbar 0x%x\n",
		regs->ercsr, regs->erbr_l, regs->erbr_h, regs->erbar);
	qprintf("\tercir 0x%x erpir 0x%x ertr 0x%x etcsr 0x%x etcdc 0x%x\n",
		regs->ercir, regs->erpir, regs->ertr,
		regs->etcsr, regs->etcdc);
	qprintf("\tetbr_l 0x%x etbr_h 0x%x etcir 0x%x etpir 0x%x\n",
		regs->etbr_l, regs->etbr_h, regs->etcir, regs->etpir);
	qprintf("\tebir 0x%x emar_l 0x%x emar_h 0x%x ehar_l 0x%x ehar_h 0x%x\n",
		regs->ebir, regs->emar_l, regs->emar_h, regs->ehar_l, regs->ehar_h);
	qprintf("\tmicr 0x%x\n", regs->micr);
}

static void
ef_hexdump(char *msg, char *cp, int len)
{
	register int idx;
	register int qqq;
	char qstr[512];
	int maxi = 128;
	static char digits[] = "0123456789abcdef";

	if (len < maxi)
		maxi = len;
	for (idx = 0, qqq = 0; qqq<maxi; qqq++) {
		if ((qqq%16) == 0)
			qstr[idx++] = '\n';
		qstr[idx++] = digits[cp[qqq] >> 4];
		qstr[idx++] = digits[cp[qqq] & 0xf];
		qstr[idx++] = ' ';
	}
	qstr[idx] = 0;
	printf("%s:%s\n", msg, qstr);
}

