/*
 * Driver for the CMC ENP-10 ethernet board.
 *
 * Copyright 1988, 1990, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */

/* do not compile on machines without VME buses */
#if !defined(IP20) && !defined(IP22) && !defined(EVEREST)

#ident "$Revision: 3.102 $"

#include "tcp-param.h"
#include "sys/param.h"
#include "sys/sysmacros.h"
#include "sys/systm.h"
#include "sys/cmn_err.h"
#include "sys/debug.h"
#include "sys/edt.h"
#include "sys/vmereg.h"
#include "sys/errno.h"
#include "sys/immu.h"
#include "sys/pda.h"
#include "sys/kopt.h"
#include "sys/mbuf.h"
#include "sys/sbd.h"
#include "sys/socket.h"
#include "sys/invent.h"
#include "sys/par.h"
#include "sys/pio.h"
#include "sys/cpu.h"
#include "net/if.h"
#include "net/if_types.h"
#include "net/netisr.h"
#include "net/multi.h"
#include "net/raw.h"
#include "net/soioctl.h"
#include "netinet/in.h"
#include "netinet/in_systm.h"
#include "netinet/ip.h"
#include "netinet/if_ether.h"
#include "mclance.h"
#include "bstring.h"
#include "misc/ether.h"


extern  struct ifnet loif;
extern	short enp_intval;		/* default interrupt interval=none */
extern	char enp_recvdmamask;
extern	char enp_xmitdmamask;

/*
 * Debugging levels
 */
#define DBG_LVL_PRINT	  1		/* enable most printfs */
#define DBG_LVL_ADDR_CMD  2		/* enable address-command printfs */
#define DBG_LVL_MONITOR   3		/* enable xmit-recv monitor printfs */

#ifndef DBG_LVL
#define DBG_LVL 0
#endif

#if DBG_LVL >= DBG_LVL_PRINT
# define DP(a)		printf a
#else
# define DP(a)
#endif

#define MAXBD	4			/* allow 4 boards */

#if defined(R2300) || defined(lint)
#define WB() wbflush()			/* flush to preserve order */
#else
#define WB() 1
#endif


#define DOG	(2*IFNET_SLOWHZ)	/* check this often after alive */
#define R_DOG	15			/* seconds reset/restart may take */
#define ENP_R_CLK 20			/* interval in HZ for poking board */

extern time_t	lbolt;

/* XXX move this to some header file, for all drivers */
#define EMSG_DELAY	(HZ*60*1)	/* suppress messages this long */


/*
 * MAX_ADDR_TIMEOUTS is the maximum number of AddrCmdTimer timeouts that
 * can occur before we consider the board to be dead.
 */
#define MAX_ADDR_TIMEOUTS	((10*HZ)/ENP_R_CLK)

#define CRC_LEN	4			/* bytes/ethernet CRC */

/* maximum & minimum packet sizes */
#define MAX_TPKT 1514			/* transmit len--7990 adds cksum */
#define MIN_TPKT 60
#define MAX_RPKT (MAX_TPKT+CRC_LEN)	/* receive length */
#define MIN_RPKT (MIN_TPKT+CRC_LEN)


/*
 * Receive buffer padding, to ether header and to data.
 */
#define	RCVPAD \
	(sizeof(struct ifheader) + RAW_BUFPAD(sizeof(struct ether_header)))
#define RCVDATAPAD \
	(RCVPAD + sizeof(struct ether_header))

#define	HDRPAD	RAW_HDRPAD(sizeof(struct ether_header))
#define	MAXRBUF	(RCVPAD + MAX_RPKT)


/*
 * Definitions to force 16 bit references
 */
typedef volatile struct {
	u_short hi;
	u_short lo;
} BD_INT;
#define	STO_BD_INT(i, val) { \
	BD_INT *sto_tmpp = &(i);		\
	u_int sto_tmp = (u_int)(val);		\
	sto_tmpp->hi = sto_tmp>>16; WB();	\
	sto_tmpp->lo = sto_tmp; WB();		\
}

#define	FET_BD_INT(type, dest, i) { \
	BD_INT *fet_tmpp = &(i); \
	(dest) = (type)((fet_tmpp->hi << 16) | fet_tmpp->lo); \
}

#define EADDR_LEN 6			/* Size of an ethernet address.
					 * DON'T change this value! */
typedef struct {
	u_char	b[EADDR_LEN];
} ETHER_ADDR;

/* ---------------------------------------------------------------------- */

/*
 * ETHLIST is used by the PROM to return the hardware ethernet address.
 * In versions of the ENP10 firmware supplied by CMC, this struct contained
 * an (unused) array of 16 ethernet addresses following e_baseaddr.
 * In SGI-modified ENP10 firmware, the space occupied by the 96-byte array
 * is used to store a version number and a data structure to communicate
 * address-related change commands to the link-level driver in the PROM.
 *
 * To maintain compatibility with CMC PROMs, ETHLIST's e_listsize and
 * e_baseaddr fields must not change.
 */

typedef struct {
	BD_INT	e_listsize;		/* ignored. */
	ETHER_ADDR e_baseaddr;		/* addr lance is working with */
} ETHLIST;

/*
 * ENP_VERSION is used to identify ENP PROMs with the SGI-modified link-level
 * driver. The version number is changed whenever a new feature or bug fix
 * is made in the PROM link-level driver.
 */

#define PROM_VERSION_MAGIC		0xFEEDBABE
typedef struct {
	BD_INT  magic;		/* A magic number to identify a SGI-modified PROM.
			 * If the number is not present, the PROM contains
			 * old code. */
    short num;			/* Version # of the PROM (defined below). */
} ENP_VERSION;

/*
 * Version # required in SGI-modified proms to handle multicasting,
 * promiscuity, and the ability to change the address.
 */
#define PROM_VERSION_MCAST	4


typedef struct {
	short	eac_cmd;	/* Flags that describe the command(s) to be
				 * executed by the ENP. */
#define EAC_PROMISC	0x1	/* The eac_promisc field should be examined
				 * to enable/disable promiscuous mode. */
#define EAC_SET_ADDR	0x2	/* The eac_addr field contains a valid
				 * address. */
#define EAC_SET_FILTER	0x4	/* The eac_filter field contains a valid
				 * address filter. */
#define EAC_DMA_ECHO	0x8	/* dma cmd. see DMA-CMD control block */
#define EAC_INTVAL	0x10	/* receive interrupt interval. */
#define EAC_ALL	(EAC_PROMISC|EAC_SET_ADDR|EAC_SET_FILTER|EAC_INTVAL)
#define EAC_SET_ALLMULTI 0x20	/* Set all bits in eac_filter */

	short	eac_unused;	/* For future use. */
	short	eac_promisc;	/* TRUE=receive every packet */
	ETHER_ADDR eac_addr;   /* Receive all packets addressed to this
				* address. This address overrides the one in
				* the board PROM. */
	short	eac_filter[4];	/* Logical address filter for doing
				 * multicasting. On the LANCE chip the filter
				 * is 64 bits.  If the filter is all 0s, it
				 * is not used.
				 * N.B. The LANCE performs imperfect filtering
				 * so a multicast address must be examined
				 * by this driver before it is accepted.
				 */
	/* DMA-CMD control block */
	short	eac_stat;	/* DMA-CMD status */
	short	eac_mode;	/* DMA-CMD mode */
#define EAC_DMA_RR	0	/* memoty to memory dma */
#define EAC_DMA_RB	1	/* memoty to board dma */
#define EAC_DMA_BR	2	/* board to memory dma */
#define EAC_DMA_BB	3	/* board to board dma */
	short	eac_num;	/* number of bytes for dma */
	BD_INT	eac_dst;	/* destination address for DMA-CMD */
	BD_INT	eac_src;	/* source address for DMA-CMD */
	u_short	eac_space[9];	/* general purpose buffer */

	/* Start of RING area */
	BD_INT	eac_bcbstart;	/* start address of RINGs */

	/* interrupt TIMER control block */
	BD_INT	eac_tlink;	/* link address */
	short	eac_treq;	/* request code */
#define TCIB_DONE	0x8000
#define TCIB_ERR	0x4000
#define TCIB_RESERVED	0x3ffc
#define TCIB_CLEAR	0x2
#define TCIB_SET	0x1
	short	eac_ttime;	/* time value. 1tic = 1ms */
	BD_INT	eac_thndlr;	/* handler address */
	short	eac_tevent;	/* event code */
	short	eac_treserved;	/* reserved */
} ENP_ADDR_CMD_BLOCK;




#define RING_SIZE	32	/* entries/ring */
#define RINGGAP		2	/* prom alocates RING_SIZE-RINGGAP bcbs */
#define RINGSPERUNIT	((RING_SIZE-RINGGAP)*2)
typedef volatile struct {
	short	r_rdidx;
	short	r_wrtidx;
	short	r_size;
	short	r_pad;
	BD_INT	r_slot[RING_SIZE];
} RING;


#define RAM_SIZE (123*1024+512)		/* size of visible RAM */

/* ENP Ram data layout
 */
typedef volatile struct {
	char	enp_ram_rom[4*1024];	/* either RAM or ROM */
	union {
		char	enp_ram[RAM_SIZE];	/* RAM */
		struct {
			unsigned short reg_go;	/* 0x8080 to start board */
			unsigned short reg_csr;	/* board control/status reg */
			BD_INT	reg_pstart;
		} regs;
		struct {			/* link-level structures */
			char	pad1[0x100];
			u_short	e_enpstate;	/* =S_ENPRUN after init */
			short	e_enpmode;
			BD_INT	e_enpbase;	/* =&enp_u.enp_ram[0] */
			short	e_enprun;	/* 'I am alive' counter */
			short	e_intvector;	/* alternate interrupt */

			RING	h_toenp;
			RING	h_hostfree;
			RING	e_tohost;
			RING	e_enpfree;

			BD_INT e_xmit_ok;	/* Successful xmits */
			BD_INT e_mult_retry;	/* multiple retries on xmit */
			BD_INT e_one_retry;	/* single retries */

			BD_INT e_fail_retry;	/* too many retries */
			BD_INT e_deferrals;	/* xmit delayed--cable busy */
			BD_INT e_xmit_buff_err;	/* impossible */
			BD_INT e_silo_underrun;	/* transmit underrun */
			BD_INT e_late_coll;	/* collision after xmit */
			BD_INT e_lost_carrier;
			BD_INT e_babble;	/* xmit length > 1518 */
			BD_INT e_no_heartbeat;	/* transceiver mismatch */
			BD_INT e_xmit_mem_err;

			BD_INT e_rcv_ok;	/* good receptions */
			BD_INT e_rcv_missed;	/* no recv buff available */
			BD_INT e_crc_err;	/* checksum failed */
			BD_INT e_frame_err;	/* surely implies CRC error */
			BD_INT e_rcv_buff_err;	/* 'impossible' */
			BD_INT e_silo_overrun;	/* receive overrun */
			BD_INT e_rcv_mem_err;

			ETHLIST			e_addr;

			ENP_VERSION	e_version; /* A version # of -1 means
						    * the ENP board has old
						    * proms. */

			ENP_ADDR_CMD_BLOCK	e_addrCmdBlk;
		} iface;
	} enp_u;

	char	pad2[255];
	char	enp_interrupt;		/* R or W interrupts ENP */
	char	pad3[255];
	char	enp_reset;		/* R or W resets ENP */
} ENPDEVICE;

#define enp_go		enp_u.regs.reg_go
#define enp_state	enp_u.iface.e_enpstate
#define enp_mode	enp_u.iface.e_enpmode
#define enp_base	enp_u.iface.e_enpbase
#define enp_toenp	enp_u.iface.h_toenp
#define enp_hostfree	enp_u.iface.h_hostfree
#define enp_tohost	enp_u.iface.e_tohost
#define enp_enpfree	enp_u.iface.e_enpfree
#define enp_baseaddr	enp_u.iface.e_addr.e_baseaddr
#define enp_version	enp_u.iface.e_version
#define enp_addrCmdBlk	enp_u.iface.e_addrCmdBlk

#define ADDR_CHECK(ei, foo) \
    if ((u_int)foo < (u_int)&ei->ei_addr->enp_u.enp_ram[0]) { \
       cmn_err(CE_PANIC, \
	 "enp%d: ring address %x below ENP ram (%x) at line %d\n", \
		    ei->ei_if.if_unit, foo, \
		    &ei->ei_addr->enp_u.enp_ram[0], __LINE__); \
    } else if((u_int)foo > (u_int)&ei->ei_addr->enp_u.enp_ram[RAM_SIZE-1]){\
      cmn_err(CE_PANIC,\
	 "enp%d: ring address %x above ENP ram (%x) at line %d\n", \
		    ei->ei_if.if_unit, foo, \
		    &ei->ei_addr->enp_u.enp_ram[RAM_SIZE-1], __LINE__); \
    }

/* enp_u.regs.reg_csr bits
 */
#define	ENPCSR_ONLINE	0x2		/* entered application firmware */
#define	ENPCSR_RDY	0x4		/* reset complete */
#define	ENPCSR_OVPROM	0x8		/* don't transfer to prom on reset */
#define	ENPCSR_IE	0x40		/* interrupt enable for bus debug */
#define	ENPCSR_ERR	0x8000		/* enp detected error */

/* state bits in enp_state
 */
#define S_ENPRESET	01		/* enp is in reset state */
#define S_ENPRUN	02		/* enp is in run state */

/*
 * mode bits in enp_mode
 */
#define E_SWAP16	0x1		/* swap two octets within 16 */
#define E_SWAP32	0x2		/* swap 16s within 32 */
#define E_SWAPRD	0x4		/* swap on read */
#define E_SWAPWRT	0x8		/* swap on write */
#define E_DMA		0x10		/* enp does data moving */
#define E_NOTERM	0x80		/* suppress enp debugging terminal */


/* The ENP Data Buffer Structure
 */
typedef volatile struct bcb {
/* b_link must be BD_INT. for word alignment. */
	BD_INT	b_link;
	short	b_stat;
	short	b_len;
	BD_INT	b_addr;
	short	b_msglen;
	short	b_reserved;
} BCB;

/* BCB b_stat field defines
 */
#define	BCB_DONE	0x8000		/* buffer complete */
#define	BCB_ERR		0x4000		/* error summary */
#define	BCB_FRAME	0x2000		/* framing error */
#define	BCB_OFLO	0x1000		/* silo overflow */
#define	BCB_CRC		0x0800		/* crc error */
#define	BCB_RXBUF	0x0400		/* rx buffer err */
#define	BCB_STP		0x0200		/* start of packet */
#define	BCB_ENP		0x0100		/* end of packet */
#define	BCB_MEMERR	0x0002		/* memory error */
#define	BCB_MISSED	0x0001		/* missed packet */

#define ENPMAXPADDR	0xfe0000	/* MAX phys-mem viewed by enpbd */
#define ENPMINPADDR	0x1000		/* MIN phys-mem viewed by enpbd */
#define PROM_VERSION_DMA	5		/* min version for dma */
#define ENPMAXCHAIN	4		/* max chain len */
#define ENPCLSIZ	ETHERMTU + sizeof(caddr_t)	/* local buf size */
#define ENPLOW8M	0x800000	/* low 8M of physical memory */
#define ENPTHRESHOLD	512		/* DMA threshold */

#define NMENPCL		76		/* max num of local buf */

/* data chain entry for dma */
typedef struct enpdbuf {
	BD_INT	d_len;			/* data length */
	BD_INT	d_pbp;			/* physical data addr */
} ENPDBUF;

/* Buffer Control Block Overlay for dma & backward compatibility */
typedef struct dmabcb {
	BCB	dbcb_bcb;		/* buffer control block */
	struct  ether_header dbcb_ethdr;/* ether header */
	char	dbcb_data[ETHERMTU];	/* data area */
	BD_INT	dbcb_crc;		/* crc */
	BD_INT	dbcb_vmp;		/* virtual mbuf ptr */
	BD_INT	dbcb_numchain;		/* num of entries in dma chain */
	ENPDBUF	dbcb_chain[ENPMAXCHAIN];/* dma data chain */
}DMABCB;/* total = 16+14+1500+4+4+4+32=1574 */

/* local static buffer */
typedef struct enpcl {
	caddr_t	enp_next;		/* next enpcl */
	int	enp_refcnt;		/* ref count */
	char	enp_data[ENPCLSIZ];	/* data */
}ENPCL;

/* Ethernet info per interface.
 */
struct enp_info {
	struct etherif	ei_eif;		/* common ethernet structures */
	ETHER_ADDR ei_boardaddr;	/* Ethernet address from PROM */
	u_char	ei_enaddrInitialized;	/* TRUE if ei_enaddr is valid */

	u_char	ei_vec;			/* VME interrupt vector */
	u_char	ei_brl;			/* VME bus request level */
	int	ei_vme_adapter;		/* which VME adapater */

	u_char	ei_sick;		/* 2=reset board on next watchdog */
	u_char	ei_rdidx;		/* recent output ring index */

	u_char	ei_state;		/* reset state */
#define EI_IDLE 0			/*	unknown */
#define EI_START1 1			/*	reseting in order to start */
#define EI_START2 2			/*	starting */
#define EI_OK 3				/*	started */
#define EI_SICK 99			/*	sick board */
	int	ei_timer_id;
	u_short	ei_timer;		/* reset and start timer */

	short	ei_addrCmds;		/* multicast or promiscuous set */

	u_char	ei_addrCmdInProg;	/* TRUE=doing address change */
	u_char	ei_inPromiscMode;	/* TRUE=LANCE chip is promiscuous */
	u_char	ei_dorecvdma;		/* TRUE=do recv dma */
	u_char	ei_doxmitdma;		/* TRUE=do xmit dma */

	char*	ei_terr;		/* recent transmit error message */
	time_t	ei_terr_lbolt;		/* when to forget it */

	int	ei_lostintrs;		/* count lost interrupts */

	struct {			/* error statistics */
		int e_mult_retry;	/*	These track board counts */
		int e_one_retry;
		int e_deferrals;

		int e_fail_retry;
		int e_xmit_buff_err;
		int e_silo_underrun;
		int e_late_coll;
		int e_lost_carrier;
		int e_babble;
		int e_xmit_mem_err;

		int e_rcv_missed;
		int e_crc_err;
		int e_frame_err;
		int e_rcv_buff_err;
		int e_silo_overrun;
		int e_rcv_mem_err;
	} ei_stat;

	ENPDEVICE	*ei_addr;	/* physical address */

	short		ei_version;	/* PROM Version # */

	struct mfilter	ei_mcastTable;	/* multicast filter table */
	Mcast_Filter	ei_laf;		/* hardware logical address filter */

	struct rawif	ei_rawif;	/* raw protocol private data */
} enp_info[MAXBD];


#define ei_ac		ei_eif.eif_arpcom /* passthru to common arpcom struct */
#define ei_if		ei_ac.ac_if	/* network-visible interface */
#define ei_enaddr	ei_ac.ac_enaddr	/* hardware ethernet address */


/*
 * handle ENP-10 LINK-10 'rings'
 */
#define ringempty(rp) ((rp)->r_rdidx == (rp)->r_wrtidx)

#define ringfull(rp) ((((rp)->rwrtidx+1) & ((rp)->r_size-1)) == (rp)->r_rdidx)

static int	VersionCheck(int, int, int, char *);
static int	IsCmdDone(struct enp_info *ei);
static void	AddrCmdTimer(struct enp_info *ei);
static void	UpdateBoardParams(int, ENPDEVICE*, struct enp_info *);
static int	enpread(struct enp_info *ei, BCB *bcbp);
static int 	enpput(struct etherif*, struct etheraddr*, struct etheraddr*,
			u_short, struct mbuf*);
int if_enpintr(int);

static int enp_poll(struct enp_info*, int, int);


/*
 *	Attach a mbuf to BufferControlBlock(BCB).
 *
 *	ASSUME: 1. # of mbufs <= ENPMAXCHAIN.
 */
static void
enp_attobd(struct mbuf *m,
	DMABCB *bcbp,
	int overhead)
{
	int i, len;
	u_char *mcp;
	ENPDBUF *pmp;
	int s = splimp();

	if (0 == bcbp)
		cmn_err(CE_PANIC, "enp_attobd: bcbp=0\n");

	STO_BD_INT(bcbp->dbcb_vmp, m);
	for (i = 0; m; m = m->m_next) {
		if (m->m_len) {
			pmp = &(bcbp->dbcb_chain[i]);
			mcp = mtod(m, u_char *);
			len = m->m_len;
			if (i == 0) {
				ASSERT(len >= overhead);
				mcp += overhead;
				len -= overhead;
			}
			STO_BD_INT(pmp->d_len, (u_int)len);
			STO_BD_INT(pmp->d_pbp, kvtophys(mcp));
#ifdef IP17
			/*
			 * Write back the cache line containing the mbuf
			 * header to keep it from being destroyed by input
			 * DMA.  Flush the rest of the buffer for input DMA.
			 *
			 * Write back any cached data for output DMA.
			 *
			 * This could be done faster, but the ENP board is
			 * so slow that it does not matter.
			 */
			dki_dcache_wbinval(mcp, len);
#endif
			i++;
			if (i >= ENPMAXCHAIN)
				cmn_err(CE_PANIC, "enp: chain overflow.\n");
		}
	}
	STO_BD_INT(bcbp->dbcb_numchain, i);

	splx(s);
}

static int
ringput(RING *rp,
	BCB *bcbp)
{
	register int idx;

	idx = (rp->r_wrtidx + 1) & (rp->r_size-1);
	if (idx != rp->r_rdidx) {
		STO_BD_INT(rp->r_slot[rp->r_wrtidx], bcbp);
		rp->r_wrtidx = idx; WB();
		if ((idx -= rp->r_rdidx) < 0 )
			idx += rp->r_size;
		return (idx);		/* num ring entries */
	}
	return (0);
}

static BCB *
ringget(RING *rp)
{
	BCB *p;

	if (rp->r_rdidx != rp->r_wrtidx) {
		FET_BD_INT(BCB*, p, rp->r_slot[rp->r_rdidx]);
		rp->r_rdidx = (rp->r_rdidx + 1) & (rp->r_size-1);
		WB();
	} else {
		p = 0;
	}
	return p;
}

/*
 * reset the board
 */
static void
enpreset(register struct enp_info *ei)
{
	register ENPDEVICE *addr;
	struct mbuf *mbrcvrtbl[RINGSPERUNIT];  /* mbuf recover table */
	register int s = splimp();

	DP(("enpreset%d:\n", ei->ei_if.if_unit));
	addr = ei->ei_addr;
	ASSERT(0 != addr);

	/* save mbuf-ptrs lent if any
	 * But freeing mbufs must be done after
	 * hardware reset to avoid race condition.
	 */
	addr->enp_mode = E_NOTERM; WB();    /* disable dma. */

	if ((ei->ei_dorecvdma) || (ei->ei_doxmitdma)) {
		register u_int dbcb;
		register int i;

		FET_BD_INT(u_int, dbcb, addr->enp_addrCmdBlk.eac_bcbstart);
		if (dbcb != 0) {
			for (i = 0; i < RINGSPERUNIT; i++) {
				FET_BD_INT(struct mbuf *,mbrcvrtbl[i],
					   ((DMABCB *)dbcb)->dbcb_vmp);
				dbcb += ((sizeof(DMABCB) + 2*sizeof(int) - 1)
					 / sizeof(int) * sizeof(int));
			}
		}
	}
	addr->enp_addrCmdBlk.eac_cmd = 0;
	addr->enp_u.regs.reg_csr = 0; WB();
	addr->enp_state = 0; WB();
	addr->enp_reset = 1; WB();

	ei->ei_sick = 0;		/* stop watchdog resets */
	ei->ei_terr_lbolt = 0;		/* force error messages */
	ei->ei_if.if_timer = 0;		/* turn off watchdog */
	ei->ei_addrCmdInProg = 0;	/* not changing address right now */

	if (ei->ei_timer_id != 0)
		untimeout(ei->ei_timer_id);
	NETPAR(NETSCHED, NETWAKINGTKN, NETPID_NULL,
		 NETEVENT_LINKDOWN, NETCNT_NULL, NETRES_BUSY);
	wakeup((caddr_t)&ei->ei_state);
	ei->ei_state = EI_IDLE;

	if ((ei->ei_dorecvdma) || (ei->ei_doxmitdma)) {
		register int i;

		for (i = 0; i < RINGSPERUNIT; i++)
			if (mbrcvrtbl[i])
				m_freem(mbrcvrtbl[i]);
	}
	ei->ei_doxmitdma = 0;
	ei->ei_dorecvdma = 0;

	splx(s);
}

/*
 * see if the board has started itself yet
 */
static void
enpstart2(struct enp_info *ei)
{
	register int s1 = spl1();
	register int s = splimp();

	if (ei->ei_addr->enp_state == S_ENPRUN) {
	    ei->ei_state = EI_OK;
	    NETPAR(NETSCHED, NETWAKINGTKN, NETPID_NULL,
			 NETEVENT_LINKDOWN, NETCNT_NULL, NETRES_BUSY);
	    wakeup((caddr_t)&ei->ei_state);
	    DP(("enpstart2%d: state = OK\n", ei->ei_if.if_unit));
	    /*
	     * Tell the board to use the latest address info, if necessary.
	     */
	    if (ei->ei_version >= PROM_VERSION_MCAST) {
		if (ei->ei_version >= PROM_VERSION_DMA)
			ei->ei_addrCmds |= EAC_INTVAL;
		if (ei->ei_addrCmds != 0)
			UpdateBoardParams(ei->ei_addrCmds, ei->ei_addr, ei);
	    }
	} else if (++ei->ei_timer >= R_DOG*HZ/ENP_R_CLK) {
		printf("enp%d: firmware failed to start\n",
		       ei->ei_if.if_unit);
		ei->ei_state = EI_SICK;
		NETPAR(NETSCHED, NETWAKINGTKN, NETPID_NULL,
			 NETEVENT_LINKDOWN, NETCNT_NULL, NETRES_BUSY);
		wakeup((caddr_t)&ei->ei_state);

	} else {
		ei->ei_timer_id = cputimeout(enpstart2,(caddr_t)ei,ENP_R_CLK);
		DP(("enpstart2%d: timeout()\n", ei->ei_if.if_unit));
	}
	splx(s); splx(s1);
}

/*
 * see if board has reset itself yet
 */
static void
enpstart1(struct enp_info *ei)
{
	register u_short stat;
	register ENPDEVICE *addr;
#if DBG_LVL > 0
	static int count;
#endif /* DBG_LVL */
	register int s1 = spl1();
	register int s = splimp();

	addr = ei->ei_addr;
	stat = addr->enp_u.regs.reg_csr;
	if (stat & ENPCSR_ERR) {
		printf("enp%d: detected error on reset\n",
		       ei->ei_if.if_unit);
		ei->ei_state = EI_SICK;
		NETPAR(NETSCHED, NETWAKINGTKN, NETPID_NULL,
			 NETEVENT_LINKDOWN, NETCNT_NULL, NETRES_BUSY);
		wakeup((caddr_t)&ei->ei_state);

	} else if (stat & ENPCSR_RDY) {
		addr->enp_mode = E_NOTERM;  /* we do DMA; no terminal */
		WB();
		if (ei->ei_version >= PROM_VERSION_DMA) {
			char dmamask;

			dmamask = (1 << ei->ei_if.if_unit);
			ei->ei_dorecvdma = enp_recvdmamask & dmamask;
			ei->ei_doxmitdma = enp_xmitdmamask & dmamask;
			addr->enp_mode |= E_DMA; WB();
		} else {
			ei->ei_dorecvdma = 0;
			ei->ei_doxmitdma = 0;
		}

		addr->enp_u.iface.e_intvector = ei->ei_vec;
		WB();			/* set interrupt vector */

		/* let controller know where its "all_ram" region
		 * begins in our address space so it can present
		 * addresses correctly to us.
		 */
		STO_BD_INT(addr->enp_base,
			   (u_int)&addr->enp_u.enp_ram[0]);


		addr->enp_state = 0; WB();  /* start firmware */
		addr->enp_go = 0x8080; WB();

		DP(("enp%d: enpstart1: setting vector to 0x%x\n",
			ei->ei_if.if_unit, ei->ei_vec));

		ei->ei_state = EI_START2;
		ei->ei_addrCmdInProg = 0;  /* not changing the address
					    * right now. */

		ei->ei_timer_id = cputimeout(enpstart2,(caddr_t)ei,ENP_R_CLK);
		DP(("enpstart1%d: timeout(enpstart2) > %d <\n",
			ei->ei_if.if_unit, count));
#if DBG_LVL > 0
		count = 0;
#endif /* DBG_LVL */

	} else if (++ei->ei_timer >= R_DOG*HZ/ENP_R_CLK) {
		printf("enp%d: timed-out waiting for reset\n",
		       ei->ei_if.if_unit);
		ei->ei_state = EI_SICK;
		NETPAR(NETSCHED, NETWAKINGTKN, NETPID_NULL,
			 NETEVENT_LINKDOWN, NETCNT_NULL, NETRES_BUSY);
		wakeup((caddr_t)&ei->ei_state);

	} else {
		ei->ei_timer_id = cputimeout(enpstart1,(caddr_t)ei,ENP_R_CLK);
#if DBG_LVL > 0
		count++;
#endif /* DBG_LVL */
	}

	splx(s); splx(s1);
}

/*
 * initialize the board
 */
static
enpbdinit(register struct enp_info *ei)
{
	register int s = splimp();

	enpreset(ei);			/* reset it */

	ei->ei_timer = 0;		/* then start it */
	ei->ei_state = EI_START1;
	ei->ei_timer_id = cputimeout(enpstart1,(caddr_t)ei,ENP_R_CLK);

	bzero(&ei->ei_stat,		/* clear error counters	*/
	      sizeof(ei->ei_stat));

	splx(s);
}

/*
 * wait until the board has initialized itself
 */
static int				/* 0 or error number */
enp_winit(register struct enp_info *ei)
{
	while (ei->ei_state == EI_START1
	       || ei->ei_state == EI_START2) {
		NETPAR(NETSCHED, NETSLEEPTKN, NETPID_NULL,
			 NETEVENT_LINKDOWN, NETCNT_NULL, NETRES_BUSY);
		(void)sleep((caddr_t)&ei->ei_state,PZERO);
		NETPAR(NETSCHED, NETWAKEUPTKN, NETPID_NULL,
			 NETEVENT_LINKDOWN, NETCNT_NULL, NETRES_BUSY);
	}

	/* get ethernet address */
	hwcpin(&ei->ei_addr->enp_baseaddr,
	       &ei->ei_boardaddr,
	       sizeof(ei->ei_boardaddr));

	if (!ei->ei_enaddrInitialized) {
		bcopy(&ei->ei_boardaddr, ei->ei_enaddr,
		      sizeof(ei->ei_enaddr));
		ei->ei_enaddrInitialized = 1;
	}

	return (ei->ei_state == EI_OK ? 0 : EIO);
}

/*
 * initialize the interface
 */
static int
enpifinit(struct etherif *eif, int flags)

{
	register struct enp_info *ei = (struct enp_info *)eif->eif_private;
	int error;

	enpbdinit(ei);
	error = enp_winit(ei);	/* wait for board to act alive (or dead) */
	if (error)
		return(error);

	ei->ei_if.if_flags |= IFF_RUNNING;
	ei->ei_if.if_timer = DOG;	/* turn on watchdog */
	if ((flags & IFF_ALLMULTI) != 0) {
		ei->ei_addrCmds |= EAC_SET_ALLMULTI;
	} else if ((flags & IFF_FILTMULTI) != 0) {
		ei->ei_addrCmds |= EAC_SET_FILTER;
	}
	if (( (flags & IFF_PROMISC) && !ei->ei_inPromiscMode) ||
	    (!(flags & IFF_PROMISC) &&  ei->ei_inPromiscMode)) {
		ei->ei_inPromiscMode = !ei->ei_inPromiscMode;
		ei->ei_addrCmds |= EAC_PROMISC;
	}

	ei->ei_addrCmds |= EAC_SET_ADDR;
	UpdateBoardParams(ei->ei_addrCmds, ei->ei_addr, ei);
	return(0);
}

/*
 * Periodically poll the controller in case an interrupt gets lost.
 */
static void
enpwdog(int unit)
{
	struct enp_info *ei;
	ENPDEVICE *addr;

	ASSERT(unit >= 0 && unit < MAXBD);
	ei = &enp_info[unit];
	addr = ei->ei_addr;
	ASSERT(0 != addr);

	ASSERT(IFNET_ISLOCKED(&ei->ei_if));
	if ((ei->ei_state == EI_OK) &&
	   ((ei->ei_if.if_flags & (IFF_UP|IFF_RUNNING))
		== (IFF_UP|IFF_RUNNING))) {
		u_short tmp;

		tmp = addr->enp_toenp.r_rdidx;
		if (enp_poll(ei, unit, 0)) {
			ei->ei_lostintrs++;
			ei->ei_sick = 0;	/* notice successful output */
		} else if (ei->ei_rdidx != (u_char)tmp) {
			ei->ei_sick = 0;	/* notice successful output */
		} else if (ei->ei_sick >= 2) {
			printf("enp%d: died and restarted(%x)\n",
			       unit, addr->enp_u.regs.reg_csr);
			enpbdinit(ei);
			ei->ei_if.if_oerrors++;
		} else if (!ringempty(&addr->enp_toenp)) { /* have output? */
			ei->ei_sick++;		/* yes, we may be sick */
		}
		ei->ei_rdidx = (u_char)tmp;

		/* collect error statistics */
#define FET_STAT(nm) FET_BD_INT(int, i, addr->enp_u.iface.nm);
#define MUL_STAT(nm,cnt,mul) { \
		int i; \
		FET_STAT(nm); \
		ei->ei_if.cnt += (i - ei->ei_stat.nm)*(mul); \
		ei->ei_stat.nm = i; \
		}
#define ADD_STAT(nm,cnt) MUL_STAT(nm,cnt,1)
#define CMN_ERR_STAT(nm,cnt,mul,msg) { \
		int i,j; \
		static char *emsg = "enp%d: " # msg "\n"; \
		FET_STAT(nm); \
		j = (i - ei->ei_stat.nm)*(mul); \
		ei->ei_if.cnt += j; \
		ei->ei_stat.nm = i; \
		if (j > 0 \
		    && (emsg != ei->ei_terr \
			|| ei->ei_terr_lbolt <= lbolt)) { \
			cmn_err(CE_ALERT, ei->ei_terr = emsg, ei->ei_if.if_unit); \
			ei->ei_terr_lbolt = lbolt+EMSG_DELAY; \
			ei->ei_terr = emsg; \
		} }
#define PRINT_STAT(nm,cnt,mul,msg) { \
		int i,j; \
		static char *emsg = "enp%d: " # msg "\n"; \
		FET_STAT(nm); \
		j = (i - ei->ei_stat.nm)*(mul); \
		ei->ei_if.cnt += j; \
		ei->ei_stat.nm = i; \
		if (j > 0 \
		    && (emsg != ei->ei_terr \
			|| ei->ei_terr_lbolt <= lbolt)) { \
			printf(ei->ei_terr = emsg, ei->ei_if.if_unit); \
			ei->ei_terr_lbolt = lbolt+EMSG_DELAY; \
			ei->ei_terr = emsg; \
		} }

		MUL_STAT(e_mult_retry,	if_collisions,2);
		ADD_STAT(e_one_retry,	if_collisions);

		ADD_STAT(e_fail_retry,	if_oerrors);
		ADD_STAT(e_xmit_buff_err,if_oerrors);
		ADD_STAT(e_silo_underrun,if_oerrors);
		PRINT_STAT(e_late_coll,	if_oerrors,1,late collision);
		CMN_ERR_STAT(e_lost_carrier,if_oerrors,1,no carrier: check Ethernet cable);
		ADD_STAT(e_babble,	if_oerrors);
		ADD_STAT(e_xmit_mem_err,if_oerrors);

		ADD_STAT(e_rcv_missed,	if_ierrors);
		ADD_STAT(e_crc_err,	if_ierrors);
		ADD_STAT(e_frame_err,	if_ierrors);
		ADD_STAT(e_rcv_buff_err,if_ierrors);
		ADD_STAT(e_silo_overrun,if_ierrors);
		ADD_STAT(e_rcv_mem_err,	if_ierrors);
#undef FET_STAT
#undef MUL_STAT
#undef ADD_STAT
#undef PRINT_STAT
	}

	ei->ei_if.if_timer = DOG;
}

/*
 * Ethernet interface interrupt.
 *	Called at splvme(), returns 0 if no work found, 1 otherwise.
 */
if_enpintr(int unit)
{
	struct enp_info *ei;
	ENPDEVICE *addr;
	int found;
#if defined(IP6) || defined(IP12) || defined(IP5) || defined(IP17) || defined(EVEREST)
	register int s = splimp();
#define INT_SPLX(s) splx(s)
#else
#define INT_SPLX(s)
#endif

	ASSERT(unit >= 0 && unit < MAXBD);
	NETPAR(NETSCHED, NETEVENTKN, NETPID_NULL,
		 NETEVENT_LINKUP, NETCNT_NULL, NETRES_INT);

	ei = &enp_info[unit];
	addr = ei->ei_addr;

	if (!addr) {
		ei->ei_terr_lbolt = 0;
		printf("enp%d: stray interrupt\n", unit);
		INT_SPLX(s);
		return(0);
	}
	if (ei->ei_enaddrInitialized) 
		IFNET_LOCKNOSPL(&ei->ei_if);
	found = enp_poll(ei, unit, 1);
	if (ei->ei_enaddrInitialized) 
		IFNET_UNLOCKNOSPL(&ei->ei_if);
	INT_SPLX(s);
	return(found);
}

static int				/* 0=did nothing, 1=found work */
enp_poll(struct enp_info *ei, int unit, int yield)
{
	BCB *bcbp;
	ENPDEVICE *addr;
	int found = 0;
	int yield_len = 0;

	ASSERT(unit >= 0 && unit < MAXBD);
	ASSERT(!ei->ei_enaddrInitialized || IFNET_ISLOCKED(&ei->ei_if));
	NETPAR(NETSCHED, NETEVENTKN, NETPID_NULL,
		 NETEVENT_LINKUP, NETCNT_NULL, NETRES_INT);

	addr = ei->ei_addr;

	if (ei->ei_state == EI_START2)	/* force wait for reset to complete */
		enpstart2(ei);
	if (ei->ei_state != EI_OK) {	/* kill dead but interrupting board */
		enpreset(ei);
		return(found);
	} else if (!IsCmdDone(ei)) {
		return(found);
	}

	while ((bcbp = ringget(&addr->enp_tohost)) != 0) {
		ADDR_CHECK(ei, bcbp);
		found = 1;
		if ((ei->ei_if.if_flags & (IFF_UP|IFF_RUNNING))
		    == (IFF_UP|IFF_RUNNING)) {
			yield_len += enpread(&enp_info[unit], bcbp);
		}
		bcbp->b_msglen = 0;	/* XXX just paranoia */
		(void)ringput(&addr->enp_enpfree, bcbp);

		if (yield && yield_len > 1000) {
			vme_yield(ei->ei_brl, ei->ei_vme_adapter);
			yield_len = 0;
		}
	}

	return(found);
}

/*
 * Copy a packet from ENP into mbufs.
 */
static struct mbuf *
enpget(
	register u_char *hp,
	register int hlen,
	register struct enp_info *ei,
	DMABCB *bcbp)
{
	struct mbuf *smp;	/* src mbuf */
	struct mbuf *dmp;	/* dst mbuf */
	u_char *dbp;		/* dst buffer */

#if DBG_LVL >= DBG_LVL_MONITOR
	DP(("R"));
#endif /* DBG_LVL */

	if (!ei->ei_dorecvdma) {
		smp = 0;
		dmp = 0;
		goto enpskip_getdma; /* smp=0, dmp=0 */
	}

	FET_BD_INT(struct mbuf *, smp, bcbp->dbcb_vmp);
	if (!smp) {
		dmp = 0;
		goto enpskip_getdma; /* smp=0, dmp=0 */
	}

	hp = mtod(smp, u_char *) + RCVDATAPAD;

#ifndef IP17
	dki_dcache_wbinval((caddr_t)hp, hlen);
#endif
	if ((dmp = m_vget(M_DONTWAIT, MAXRBUF, MT_DATA)) &&
	   ((u_int)kvtophys(mtod(dmp, caddr_t))<(u_int)ENPLOW8M)) {
		IF_INITHEADER(mtod(dmp, caddr_t), &ei->ei_if, RCVDATAPAD);
		/* smp = xxx, dmp = xxx */
		enp_attobd(dmp, bcbp, RCVDATAPAD);
		smp->m_len = RCVDATAPAD + hlen;
		return(smp);
	}

	/* smp=xxx, dmp = 0 or xxx */

enpskip_getdma:
	/* Direct mbuf dma and lend case handled above already
	 * Thus, at this case, only trailer, no dma, or bcopy
	 * cases to be handled.
	 */
	if (!dmp) {
		dmp = m_vget(M_DONTWAIT, RCVPAD+hlen, MT_DATA);
		if (!dmp) {
			DP(("enpget%d: no buff\n",ei->ei_if.if_unit));
			return(0);
		}
	} else
		m_adj(dmp, (RCVPAD+hlen-MAXRBUF));

	IF_INITHEADER(mtod(dmp, caddr_t), &ei->ei_if, RCVDATAPAD);
	dbp = mtod(dmp, u_char *) + RCVPAD;

	if (smp) {
		if (hlen)
			bcopy(hp, dbp, hlen);
		return(dmp);
	}

	if (hlen)
		hwcpin(hp, dbp, hlen);

	/* Prepare dmabuf for next recv. */
	if (ei->ei_dorecvdma
	    && (smp = m_vget(M_DONTWAIT, MAXRBUF, MT_DATA))) {
		if ((u_int)kvtophys(mtod(smp,caddr_t))<ENPLOW8M) {
			IF_INITHEADER(mtod(smp,caddr_t),
				      &ei->ei_if,RCVDATAPAD);
			enp_attobd(smp, bcbp, RCVDATAPAD);
		} else {
			m_freem(smp);
		}
	}
	return(dmp);
}

/*
 * receive packet from interface
 */
static int
enpread(ei, bcbp)
	register struct enp_info *ei;
	register BCB *bcbp;
{
	register u_char *hp;
	register struct mbuf *m;
	register int hlen;
	register struct ether_header *eh;
	int snoopflags = 0;
	static struct snoopflagmap errorMap[] = {
		{ BCB_FRAME,	SNERR_FRAME },
		{ BCB_OFLO,	SNERR_OVERFLOW },
		{ BCB_CRC,	SNERR_CHECKSUM },
		{ BCB_RXBUF,	SNERR_NOBUFS },
		{ BCB_MEMERR,	SNERR_MEMORY },
		{ BCB_MISSED,	SNERR_NOBUFS },
		{ 0, }
	};

	ei->ei_if.if_ipackets++;

	if (bcbp->b_stat & BCB_ERR) {	/* got a bad packet */
#if DBG_LVL >= DBG_LVL_PRINT
		printf("enpread%d: BCB_ERR = %x\n",
		       ei->ei_if.if_unit, bcbp->b_stat);
#endif /* DBG_LVL */
		ei->ei_if.if_ierrors++;
		snoopflags = snoop_mapflags(errorMap, (u_short)bcbp->b_stat);
	}

	hlen = bcbp->b_msglen;
	ei->ei_if.if_ibytes += hlen;
	if (hlen < MIN_RPKT) {
		ei->ei_terr_lbolt = 0;
		printf("enp%d: packet too small (length = %d)\n",
		       ei->ei_if.if_unit, hlen);
		ei->ei_if.if_ierrors++;
		if (hlen < (CRC_LEN + sizeof *eh)) {
			DP(("enpread%d: hlen<=0\n",ei->ei_if.if_unit));
			hlen = CRC_LEN + sizeof(struct ether_header);
		}
		snoopflags |= SNERR_TOOSMALL|SN_ERROR;
	} else if (hlen > MAX_RPKT) {
		ei->ei_terr_lbolt = 0;
		printf("enp%d: packet too large (length = %d)\n",
		       ei->ei_if.if_unit, hlen);
		ei->ei_if.if_ierrors++;
		snoopflags |= SNERR_TOOBIG|SN_ERROR;
		hlen = MAX_RPKT;
	}

	/*
	 * Get the ethernet header from the input buffer.
	 */
	FET_BD_INT(u_char *, hp, bcbp->b_addr);
	ADDR_CHECK(ei, hp);
	hlen -= CRC_LEN;

	/* convert to mbuf(s) */
	m = enpget(hp, hlen, ei, (DMABCB*)bcbp);
	if (m == 0) {
		ei->ei_if.if_iqdrops++;
		DP(("enpread%d: no buff\n",ei->ei_if.if_unit));
		return(0);
	}
	/*
	 * Pass m to the ethernet decapsulator.
	 */
	ether_input((struct etherif *)&ei->ei_if, snoopflags, m);
	return(hlen);
}

/*
 * copy from mbuf chain to transmitter buffer in ENP memory.
 */
static int				/* 0 or errno */
enpput(struct etherif *eif,		/* on this device */
	struct etheraddr *edst,		/* with these addresses */
	struct etheraddr *esrc,
	u_short type,			/* of this type */
	struct mbuf *m0)		/* send this chain */
{
	register struct enp_info *ei;
	register BCB *bcbp;
	register ENPDEVICE *addr;
	register u_char *bp;
	register u_int totlen;
	register struct mbuf * tm;
	register struct mbuf * bdmp;
	register u_int len;
	register int i;

	ei = &enp_info[eiftoifp(eif)->if_unit];

#if DBG_LVL >= DBG_LVL_MONITOR
	DP(("T"));
#endif /* DBG_LVL */
	ei->ei_if.if_opackets++;
	if (ETHER_ISGROUP(edst))
		ei->ei_if.if_omcasts++;

	if (ei->ei_state != EI_OK) {
		DP(("enpput%d: state = %x\n",
			ei->ei_if.if_unit, ei->ei_state));
		ei->ei_if.if_oerrors++;
		m_freem(m0);
		return EIO;
	}

	addr = ei->ei_addr;

	bcbp = ringget(&addr->enp_hostfree);
	if (!bcbp) {
		DP(("enpput%d: !bcbp\n", ei->ei_if.if_unit));
		NETPAR(NETFLOW, NETDROPTKN, NETPID_NULL,
			 NETEVENT_LINKDOWN, NETCNT_NULL, NETRES_NULL);
		ei->ei_if.if_odrops++;
		m_freem(m0);
		return ENOBUFS;
	}
	ADDR_CHECK(ei, bcbp);

#define ETHERHDR ((struct ether_header*)bp)
	FET_BD_INT(u_char *, bp, bcbp->b_addr);
	ADDR_CHECK(ei, bp);
	ETHERHDR->ether_type = type;
	hwcpout((u_char *)edst, ETHERHDR->ether_dhost,
		 sizeof(ETHERHDR->ether_dhost));
	hwcpout((u_char *)esrc, ETHERHDR->ether_shost,
		 sizeof(ETHERHDR->ether_shost));

	bp += sizeof(*ETHERHDR);
	totlen = sizeof(*ETHERHDR);
#undef ETHERHDR

	/* Assert the total len first! */
	len = 0;
	for (tm = m0; tm; tm = tm->m_next)
		len += tm->m_len;
	totlen += len;
	ASSERT(totlen <= MAX_TPKT);
	if (totlen < MIN_TPKT)
		totlen = MIN_TPKT;
	ei->ei_if.if_obytes += totlen;

	if (RAWIF_SNOOPING(&ei->ei_rawif)) {
		struct ether_header *eh;
		u_long snoopbuf[SNOOP_FILTERLEN];
		register struct mbuf *n;

		eh = RAW_HDR(snoopbuf, struct ether_header);
		*(ETHER_ADDR *)eh->ether_dhost = *(ETHER_ADDR *)edst;
		*(ETHER_ADDR *)eh->ether_shost = *(ETHER_ADDR *)esrc;
		eh->ether_type = type;
		(void) m_datacopy(m0, 0,
				  sizeof snoopbuf - (HDRPAD + sizeof *eh),
				  (caddr_t)(eh + 1));
		if (snoop_match(&ei->ei_rawif, (caddr_t) eh, len)) {
			/*
			 * Tailored version of ether_selfsnoop for enp.
			 */
			n = m_vget(M_DONTWAIT, RCVDATAPAD + len, MT_DATA);
			if (n == 0) {
				snoop_drop(&ei->ei_rawif, SN_PROMISC,
					   (caddr_t) eh, len);
			} else {
				caddr_t cp;

				cp = mtod(n, caddr_t);
				IF_INITHEADER(cp, &ei->ei_if, RCVDATAPAD);
				m_datacopy(m0, 0, len, cp + RCVDATAPAD);
				cp += RCVPAD;
				*(struct ether_header *)cp = *eh;
				snoop_input(&ei->ei_rawif, SN_PROMISC, cp, n,
					    totlen - sizeof *eh);
			}
		}
	}

	/* check dma cases */
	if (!ei->ei_doxmitdma) {
		bdmp = 0;
		goto enpskip_putdma;
	}

	FET_BD_INT(struct mbuf *, bdmp, ((DMABCB *)bcbp)->dbcb_vmp);
	if (!bdmp) {
		bdmp = m_vget(M_DONTWAIT, len, MT_DATA);
		if (!bdmp)
			goto enpskip_putdma;
		if ((u_int)kvtophys(mtod(bdmp,caddr_t)) >= ENPLOW8M) {
			m_freem(bdmp);
			bdmp = 0;
			goto enpskip_putdma;
		}

		enp_attobd(bdmp, (DMABCB*)bcbp, 0);
		bp = mtod(bdmp, u_char *);
		goto enpskip_putdma;
	}

	/* bdmp = xxx */
	/* check if current mbuf is dma-able */
	for (tm=m0, i=0; tm && i<ENPMAXCHAIN; tm=tm->m_next, i++) {
		caddr_t dt;

		/* Bcopy the split mbuf(k2) */
		dt = mtod(tm, caddr_t);
		if (((u_int)kvtophys(dt)+(u_int)tm->m_len) >= ENPLOW8M
		    || IS_KSEG2(dt))
			break;
	}
	if ((!tm) && (i < ENPMAXCHAIN)) {   /* m0 is dma-able */
		m_freem(bdmp);		/* release old bdmp */
		enp_attobd(m0, (DMABCB*)bcbp, 0);
		bcbp->b_len = totlen; WB();
		NETPAR(NETFLOW, NETFLOWTKN, NETPID_NULL,
		       NETEVENT_LINKDOWN, totlen, NETRES_NULL);
		if (ringput(&addr->enp_toenp, bcbp) == 1) {
			addr->enp_addrCmdBlk.eac_cmd = 0;
			WB();
			addr->enp_interrupt = 0xff;
		}
		return(0);
	}

	/* at this point, bdmp undefined, but m0 is not dmable */
	if (bdmp && (bdmp->m_len >= len)) { /* bdmp reusable */
		STO_BD_INT(((DMABCB *)bcbp)->dbcb_numchain,1);
		STO_BD_INT(((DMABCB *)bcbp)->dbcb_chain[0].d_len,len);
		bp = mtod(bdmp, u_char *);
		goto enpskip_putdma;
	}

	/* at this point, m0 is not dmable and bdmp not reusable*/
	m_freem(bdmp);
	if (bdmp = m_vget(M_DONTWAIT, len, MT_DATA)) {
		if ((u_int)kvtophys(mtod(bdmp, caddr_t)) < ENPLOW8M) {
			bp = mtod(bdmp, u_char *);
		} else {
			m_freem(bdmp);
			bdmp = 0;
		}
	}				/* bdmp = 0, and bm = xxx */
	enp_attobd(bdmp, (DMABCB*)bcbp, 0);

enpskip_putdma:
	/* handle non-dma and non-dma-able mbuf cases */
	if (bdmp) {
		for (tm = m0; tm; tm = tm->m_next) {
			if ((len = tm->m_len) == 0)
				continue;
			bcopy(mtod(tm, u_char *), bp, len);
			bp += len;
		}
	} else {
		for (tm = m0; tm; tm = tm->m_next) {
			if ((len = tm->m_len) == 0)
				continue;
			hwcpout(mtod(tm, u_char *), bp, len);
			bp += len;
		}
	}

	m_freem(m0);

	NETPAR(NETFLOW, NETFLOWTKN, NETPID_NULL,
		 NETEVENT_LINKDOWN, totlen, NETRES_NULL);
	bcbp->b_len = totlen; WB();
	if (ringput(&addr->enp_toenp, bcbp) == 1) {
		addr->enp_addrCmdBlk.eac_cmd = 0;
		WB();
		addr->enp_interrupt = 0xff;
	}
	return 0;
}

/*
 * Process an ioctl request.
 */
static int
enpioctl(
	struct etherif *eif,
	int cmd,
	caddr_t data)
{
	register struct ifnet *ifp = eiftoifp(eif);
	struct enp_info *ei = &enp_info[ifp->if_unit];
	int error = 0;
	ENPDEVICE *addr = ei->ei_addr;
	struct mfreq *mfr = (struct mfreq *)data;
	union mkey *key = mfr->mfr_key;

	switch (cmd) {

	    case SIOCADDMULTI: {
		int doChange;

		error = VersionCheck(ifp->if_unit, ei->ei_version,
				PROM_VERSION_MCAST, "add multicast");
		if (error)
			break;

		DP(("enp%d SIOCADDMULTI:", ei->ei_if.if_unit));


		while (!IsCmdDone(ei)) {
			NETPAR(NETSCHED, NETSLEEPTKN, NETPID_NULL,
				 NETEVENT_LINKDOWN, NETCNT_NULL, NETRES_BUSY);
			(void)sleep((caddr_t)&ei->ei_addrCmdInProg,PZERO);
			NETPAR(NETSCHED, NETWAKEUPTKN, NETPID_NULL,
				 NETEVENT_LINKDOWN, NETCNT_NULL, NETRES_BUSY);
		}

		DP((" <> "));

		if (eiftoifp(eif)->if_flags & IFF_ALLMULTI) {
			/* Want to receive all multicasts */
			ifp->if_flags |= IFF_ALLMULTI;
			UpdateBoardParams(EAC_SET_ALLMULTI, addr, ei);
		} else {
			error = AddMulti(key, &eif->eif_mfilter,
					 ei->ei_laf, &doChange);
			if (error == 0 && doChange &&
			    !(ifp->if_flags & IFF_ALLMULTI))
				UpdateBoardParams(EAC_SET_FILTER, addr, ei);
		}
		DP(("... done:\n"));
		break;
	    }

	    case SIOCDELMULTI: {
		error = VersionCheck(ifp->if_unit, ei->ei_version,
				PROM_VERSION_MCAST, "del. multicast");
		if (error)
			break;

		DP(("enp%d SIOCDELMULTI:", ei->ei_if.if_unit));

		while (!IsCmdDone(ei)) {
			NETPAR(NETSCHED, NETSLEEPTKN, NETPID_NULL,
				 NETEVENT_LINKDOWN, NETCNT_NULL, NETRES_BUSY);
			(void)sleep((caddr_t)&ei->ei_addrCmdInProg,PZERO);
			NETPAR(NETSCHED, NETWAKEUPTKN, NETPID_NULL,
				 NETEVENT_LINKDOWN, NETCNT_NULL, NETRES_BUSY);
		}

		DP((" <> "));

		if (eiftoifp(eif)->if_flags & IFF_ALLMULTI) {
			ifp->if_flags &= ~IFF_ALLMULTI;
			/* Must restore former mcast filter */
			UpdateBoardParams(EAC_SET_FILTER, addr, ei);
		} else if (DeleteMulti(key, &eif->eif_mfilter, ei->ei_laf)
			    && !(ifp->if_flags & IFF_ALLMULTI)) {
			UpdateBoardParams(EAC_SET_FILTER, addr, ei);
		}
		DP(("... done:\n"));
		break;
	    }

	    default:
		error = EINVAL;
	}

	if (error == EIO) {
		ifp->if_flags &= ~(IFF_UP|IFF_RUNNING);
		if_down(ifp);
	}
	return (error);
}

char	XXXetheraddr[6];
char	XXXetheraddr_initialized;

struct etherifops enpops = {
	enpifinit, (void (*)(struct etherif *))enpreset, enpwdog, enpput,
	(int (*)(struct etherif*, int, void*))enpioctl
};


/*
 * probe and attach a single board
 */
void
if_enpedtinit(edtp)
	register struct edt *edtp;
{
	uint unit;
	struct enp_info *ei;
	struct vme_intrs *intp;
	ENPDEVICE *addr;
	int magic;
	struct etheraddr ea;
	piomap_t *pmap;

	unit = 99;
	if (0 == (intp = (vme_intrs_t *)edtp->e_bus_info)
	    || MAXBD <= (unit = edtp->e_ctlr)) {
		printf("enp%u: bad EDT entry\n", unit);
		return;
	}

	ei = &enp_info[unit];
	if (ei->ei_addr != 0) {
		printf("enp%u: duplicate EDT entry\n", unit);
		return;
	}

	ei->ei_vme_adapter = edtp->e_adap;
        pmap = pio_mapalloc(edtp->e_bus_type, edtp->e_adap, &edtp->e_space[0],
                PIOMAP_FIXED,"enp");
        if (!pmap) {
		printf("enp%u: PIO map failed\n", unit);
                return;
	}
	addr = (ENPDEVICE*)pio_mapaddr(pmap, edtp->e_iobase);
	if (badaddr(&addr->enp_u.enp_ram[0], sizeof(short))) {
		pio_mapfree(pmap);
		printf("enp%u: missing\n", unit);
		return;
	}
	ei->ei_addr = addr;		/* say the board is here */

	/* allocate an interrupt vector */
	if (!intp->v_vec) {
		DP(("enp%u: allocating IVEC\n", unit));
		intp->v_vec = vme_ivec_alloc(ei->ei_vme_adapter);
		if (!intp->v_vec) {
			DP(("enp%u: waisting NULL-IVEC\n", unit));
			intp->v_vec = vme_ivec_alloc(ei->ei_vme_adapter);
			if (!intp->v_vec) {
				return;
			}
		}
	}
	DP(("enp%u: IVEC=%x\n", unit, intp->v_vec));

	if (intp->v_vec == (unsigned char)-1) {
		printf("enp%u: no interrupt vector\n", unit);
		pio_mapfree(pmap);
		return;
	}
	vme_ivec_set(ei->ei_vme_adapter, intp->v_vec, if_enpintr, unit);
	DP(("enp%u: IVEC set!\n", unit));

	ei->ei_vec = intp->v_vec;	/* remember interrupt vector */
	ei->ei_brl = intp->v_brl;	/* and VME bus request level */
	ei->ei_inPromiscMode = 0;
	ei->ei_addrCmds = 0;
	ei->ei_enaddrInitialized = 0;

	enpbdinit(ei);		/* try to start the board */
	(void)enp_winit(ei);	/* wait for board to act alive (or dead) */

	ea = *(struct etheraddr *)ei->ei_enaddr;

	/*
	 * enp_winit sets up this info in ei
	 */
	if ( !XXXetheraddr_initialized ) {
		bcopy(ei->ei_enaddr, XXXetheraddr, sizeof(ei->ei_enaddr));
		XXXetheraddr_initialized = 1;
	}

	/*
	 * Read the version number from the prom. If the magic number
	 * doesn't match, then the proms contain CMC-derived code.
	 */

	FET_BD_INT(int, magic, addr->enp_version.magic);
	if (magic == PROM_VERSION_MAGIC) {
		ei->ei_version = addr->enp_version.num;
	} else {
		ei->ei_version = 0;
	}

	if (ei->ei_version >= PROM_VERSION_MCAST) {
		ei->ei_if.if_flags |= IFF_MULTICAST;
	}

	if (ei->ei_version >= PROM_VERSION_DMA) {
		char dmamask;

		dmamask = (1 << ei->ei_if.if_unit);
		ei->ei_dorecvdma = enp_recvdmamask & dmamask;
		ei->ei_doxmitdma = enp_xmitdmamask & dmamask;
		ei->ei_addr->enp_mode |= E_DMA;
		WB();
	} else {
		ei->ei_dorecvdma = 0;
		ei->ei_doxmitdma = 0;
	}

	if (showconfig) {
		printf("enp%u: ethernet address %s,  firmware version ",
		       unit, ether_sprintf(ei->ei_enaddr));

		if (ei->ei_version == 0) {
			printf("0 (CMC)\n");
		} else {
			printf("%d\n", ei->ei_version);
		}
	}

	ether_attach(&ei->ei_eif, "enp", unit, (caddr_t)ei, &enpops,
			&ea, INV_ETHER_ENP, (long)ei->ei_version);

}

/*
 *  IsCmdDone --
 *
 *	Checks the board to see if the previous address-change command has
 *	completed. Returns 1 if the command has completed.
 */

static int
IsCmdDone(struct enp_info *ei)
{
	int done = 1;		/* Assume the command has finished. */
	int s = splimp();

	if (ei->ei_addrCmdInProg) {
		/*
		 * The board clears enp_addrCmdBlk.eac_cmd when it is done.
		 */
		if (ei->ei_addr->enp_addrCmdBlk.eac_cmd == 0) {

#if DBG_LVL >= DBG_LVL_ADDR_CMD
			u_int base;

			FET_BD_INT(u_int, base, ei->ei_addr->enp_base);
			printf("IsCmdDone%d: %d   enpstate %d, base %x\n",
			       ei->ei_if.if_unit, ei->ei_addrCmdInProg,
			       ei->ei_addr->enp_state, base);
#endif /* DBG_LVL */

			ei->ei_addrCmdInProg = 0;
			if (ei->ei_timer_id != 0) {
				untimeout(ei->ei_timer_id);
			}
			NETPAR(NETSCHED, NETWAKINGTKN, NETPID_NULL,
				 NETEVENT_LINKDOWN, NETCNT_NULL, NETRES_BUSY);
			wakeup((caddr_t)&ei->ei_addrCmdInProg);
		} else {
			done = 0;	/* Still not finished... */
		}
	}

	(void) splx(s);
	return(done);
}

/*
 * AddrCmdTimer --
 *
 * This is a timeout routine to check if the board has completed
 * an address-change command (e.g., set multicast filter).  If the
 * command has not finished, another timeout is set up to call this
 * routine again.  A limit on the number of timeouts is used to
 * detect when the board has died.
 */

static void
AddrCmdTimer(struct enp_info *ei)
{
	int s1 = spl1();
	int s = splimp();

	if (ei->ei_addrCmdInProg) {
		if (ei->ei_addr->enp_addrCmdBlk.eac_cmd == 0) {
			/*
			 * The board has completed the command (it cleared
			 * eac_cmd) so wakeup interested parties.
			 */
			ei->ei_addrCmdInProg = 0;
			NETPAR(NETSCHED, NETWAKINGTKN, NETPID_NULL,
			       NETEVENT_LINKDOWN, NETCNT_NULL, NETRES_BUSY);
			wakeup((caddr_t)&ei->ei_addrCmdInProg);
		} else {
			/*
			 * Still not done. Reset the board if we made
			 * too many callbacks.
			 */
			ei->ei_addrCmdInProg++;
			if (ei->ei_addrCmdInProg > MAX_ADDR_TIMEOUTS) {
				ei->ei_addrCmdInProg = 0;
				printf("enp%d: address command failed\n",
				       ei->ei_if.if_unit);
				enpbdinit(ei);
			}
			ei->ei_timer = 0;
			ei->ei_timer_id = cputimeout(AddrCmdTimer,
						     (caddr_t)ei, ENP_R_CLK);
		}
	}
	splx(s); splx(s1);
}


/*
 * This routine copies the relevant address-change info to the
 * board's RAM and then interrupts the board so it will make the
 * necessary changes. A timeout routine is set up to detect when
 * the command has finished.
 */
static void
UpdateBoardParams(int cmd,		/* OR of EAC_SET_ADDR, EAC_PROMISC,
					 * EAC_SET_FILTER */
		  ENPDEVICE *addr,	/* Board RAM address. */
		  struct enp_info *ei)
{
	int i;

	/*
	 * In promiscous mode, always tell the board in case it got reset.
	 * Also, tell the board if cmd has the EAC_PROMISC bit set.
	 */
	if (cmd & EAC_PROMISC || ei->ei_inPromiscMode) {
		addr->enp_addrCmdBlk.eac_promisc = ei->ei_inPromiscMode;
		WB();
		cmd |= EAC_PROMISC;
	}

	/*
	 * Change the ethernet address.
	 */
	if (cmd & EAC_SET_ADDR) {
		for (i = 0; i < 6; i++) {
			addr->enp_addrCmdBlk.eac_addr.b[i] = ei->ei_enaddr[i];
			WB();
		}
	}

	/*
	 * Change the logical multicast address filter on the board.
	 */
	if (cmd & EAC_SET_FILTER) {
		for (i = 0; i < 4; i++) {
			addr->enp_addrCmdBlk.eac_filter[i] = ei->ei_laf[i];
			WB();
		}
	}

	/*
	 * Special case of EAC_SET_FILTER: cause reception of all multicast
	 * pkts.
	 */
	if (cmd & EAC_SET_ALLMULTI) {
		cmd |= EAC_SET_FILTER;
		for (i = 0; i < 4; i++) {
			addr->enp_addrCmdBlk.eac_filter[i] = 0xFFFF;
			WB();
		}
	}

	if (cmd & EAC_INTVAL) {
		addr->enp_addrCmdBlk.eac_treq = TCIB_SET;
		addr->enp_addrCmdBlk.eac_ttime = enp_intval;
		WB();
	}

	cmd &= EAC_ALL;
	ei->ei_addrCmds |= cmd;
	ei->ei_addrCmdInProg = 1;
	addr->enp_addrCmdBlk.eac_cmd = cmd;
	WB();
	addr->enp_interrupt = 0xff;	/* interrupt the board */
	WB();
	AddrCmdTimer(ei);
}


/*
 * Check the ENP PROMs' version to see if the operation is supported.
 */
static int
VersionCheck(int unit,			/* ENP unit # */
	     int promVersion,		/* version # in the PROM. */
	     int version,		/* operation requires at least this */
	     char *op)			/* Operation name */
{
	if (promVersion < version) {
		printf("enp%d: %s failed: no firmware support\n", unit, op);
		return EINVAL;
	}
	return 0;
}
#endif /* !IP20 && !IP22 !EVEREST */
