/*
 * Copyright 1988 Silicon Graphics, Inc.  All rights reserved.
 *
 * Driver for Silicon Graphics Integral Clover2 Ethernet Controller.
 */
#ident "$Revision: 3.78 $"

/*
 * Compile for IP5 only.
 */
#ifdef IP5

/*
 * Be noisy about some arbitrary things:
 *   0  = quiet
 *   1  = print "dropped" packets msgs.
 *   2  = print debug msgs.
 */
#ifndef NOISE
#define NOISE 1
#endif

#if NOISE >= 2
# define DBG_PRINT(a)		printf a
#else
# define DBG_PRINT(a)
#endif

#include "tcp-param.h"
#include "sys/types.h"
#include "sys/cpu.h"
#include "sys/sbd.h"
#include "sys/immu.h"
#include "sys/pda.h"
#include "sys/param.h"
#include "sys/systm.h"
#include "sys/cmn_err.h"
#include "sys/debug.h"
#include "sys/edt.h"
#include "sys/pio.h"
#include "sys/errno.h"
#include "sys/invent.h"
#include "sys/kopt.h"
#include "sys/mbuf.h"
#include "sys/socket.h"
#include "sys/vmereg.h"
#ifndef IRIX33
#include "bstring.h"
#endif
#include "net/if.h"
#include "net/raw.h"

extern struct ifnet	loif;

#ifdef OS_METER
/* keep count of errors that don't print a msg. on the console */
u_int	et_istat_frame;
u_int	et_istat_crc;
u_int	et_istat_oflo;
u_int	et_istat_missed;
#endif /* OS_METER */

#include "lance.h"

/* first byte of SGI's assigned ethernet addresses */
#define SGI_ETHER_ADDR 0x08


/* this generates some extra code to check problems in the board */
#define CK_ADDR	1

#define DOG	(2*IFNET_SLOWHZ)  	/* watchdog duration in seconds */

extern time_t	lbolt;

/* XXX move this to some header file, for all drivers */
#define EMSG_DELAY	(HZ*60*1)	/* suppress messages this long */


/* XXX NUM_RMD and NUM_TMD don't use full 1Mb RAM */
#define RAM_SIZE (1024*1024)		/* total RAM on a board */

#define LOG_RMD		6		/* log2 # receive memory descriptors */
#define LOG_TMD		7		/* log2 # transmit memory descriptors */

#define NUM_RMD		(1<<LOG_RMD)	/* # of receive buffers */
#define NUM_TMD		(1<<LOG_TMD)	/* # of transmit buffers */

/* convert from 32-bit VME address to 24-bit on-board address */
#define LOC_ADDR(a,base)	((caddr_t)(a) - (caddr_t)(base))

/* generate host-accessible address from an on-board address */
#define HOST_ADDR(lo,hi,base)	(MK_ADDR((lo),(hi)) + (u_char*)(base))

#define IDPROM_LEN	32
/*
 * Shape of I/O space of a single board.
 */
typedef volatile struct {		/* start at 0xb7fc0002 */
	u_short	io_rdp;			/* 0x00002 7990 register data port */
	u_short	io_pad1[3];
	u_short	io_rap;			/* 0x0000a 7990 register address port */
	u_short io_pad2[0x17FFB];
	struct {			/* 0x30002-0x30102 enet address/stat */
		u_char 	ad_byte;	/* enet addr */
		u_char 	ad_stat;	/* enet stat */
		u_short ad_pad1;
		u_int	ad_pad2;
	} io_ad[IDPROM_LEN];		/* 1st 6 are enet addr */
} *ETIO;
#define io_stat io_ad[0].ad_stat	/* status of enet */

#define BASE	((ETMEM)0)		/* start board data structures here */

/* 7990 CSR 1 & 2 values */
#define CSR1		LO_ADDR(&BASE->ib)
#define CSR2		HI_ADDR(&BASE->ib)

/*
 * Shape of local memory on a single board.
 */
struct etbuf {
	struct etherbufhead	ebh;
	char			data[MAX_RDAT];
};

typedef volatile struct {
	struct ib	ib;
	struct rmd	rmd[NUM_RMD];
	struct tmd	tmd[NUM_TMD];
	struct etbuf	rbuf[NUM_RMD];	/* packet buffers */
	struct etbuf	tbuf[NUM_TMD];
} *ETMEM;

/*
 * Each interface is represented by a network interface structure, et_info,
 * which the routing code uses to locate the interface.
 */
struct et_info {
	struct lanceif	ei_lif;		/* common ethernet structures */

	char		*ei_terr;	/* recent transmit error msg */
	time_t		ei_terr_lbolt;	/* when to forget it */

	u_int		ei_ract;	/* next active receive buffer */
	u_int		ei_tact;	/* 1st active transmit buffer */
	u_int		ei_tnum;	/* # of active transmit buffers */
	u_int		ei_tfree;	/* next free transmit buffer */

	u_int		ei_numtdr;	/* # of TDR samples */
	u_long		ei_tdr;		/* sum of all TDR samples */

	ETIO		ei_io;		/* I/O addresses */
	ETMEM		ei_mem;		/* start of memory */
} et_info;

#define	ei_eif		ei_lif.lif_eif
#define	ei_ib		ei_lif.lif_ib
#define	ei_laf		ei_lif.lif_laf
#define	ei_ac		ei_eif.eif_arpcom
#define ei_if           ei_ac.ac_if		/* network-visible interface */
#define ei_curaddr	ei_ac.ac_enaddr		/* current ethernet address */
#define	ei_addr		ei_eif.eif_addr
#define	ei_rawif	ei_eif.eif_rawif
#define	ei_lostintrs	ei_eif.eif_lostintrs
#define	ei_resets	ei_eif.eif_resets
#define	ei_sick		ei_eif.eif_sick

#define	ET_INFO(eif)	((struct et_info *)(eif)->eif_private)

/* these macros assume that there is only 1 unit */
#define ET_RESET()		{ HWREG_USELOCK(); HWREG_LOCK(); \
				*(volatile char *)(MODE_ADDR+3) &= ~ET_NRESET; \
				DELAYBUS(5); \
				*(volatile char *)(MODE_ADDR+3) |= ET_NRESET; \
				DELAYBUS(5); \
				HWREG_UNLOCK(); }
#define ET_PARON()		{ HWREG_USELOCK(); HWREG_LOCK(); \
				*(volatile char *)(MODE_ADDR+3) &= ~ET_PARENA; \
				DELAYBUS(5); \
				*(volatile char *)(MODE_ADDR+3) |= ET_PARENA; \
				DELAYBUS(5); \
				HWREG_UNLOCK(); }

/*
 * Initialize a board. 	Interrupts must already be off.
 */
static int				/* 0 or an errno */
et_init(eif, flags)
	struct etherif *eif;
	int flags;
{
	int error;

	error = et_start(ET_INFO(eif), flags);
	if (!error)
		eif->eif_arpcom.ac_if.if_timer = DOG;
	return error;
}

static int
et_start(ei, flags)
	register struct et_info *ei;
	int flags;
{
	register int i;
	register u_int baddr;
	register volatile struct rmd *rmd;
	register volatile struct tmd *tmd;
	register ETIO io;
	register ETMEM mem;

	io = ei->ei_io;
	mem = ei->ei_mem;

	ET_RESET(); 			/* reset LAN chip */
	ET_PARON();			/* enable parity checking */

	ei->ei_sick = 0;		/* stop watchdog resets */
	ei->ei_terr_lbolt = 0;		/* force error messages */

	ei->ei_ract = 0;		/* start with the 1st buffers */
	ei->ei_tact = 0;
	ei->ei_tnum = 0;
	ei->ei_tfree = 0;

	/* setup initialization block */
	lance_init(&ei->ei_lif, flags, (u_int)&BASE->rmd[0],
		(u_int)&BASE->tmd[0], LOG_RMD, LOG_TMD);

	/* setup receive descriptors */
	bzero((char*)&ei->ei_mem->rmd[0], sizeof(ei->ei_mem->rmd));
	baddr = (u_int) &BASE->rbuf[0].ebh.ebh_ether;
	for (i = 0, rmd = &mem->rmd[0];
	     i < NUM_RMD;
	     i++, rmd++, baddr += sizeof(mem->rbuf[0])) {
		rmd->rmd_ladr = LO_ADDR(baddr);
		rmd->rmd_hadr = HI_ADDR(baddr);
		rmd->rmd_bcnt = -MAX_RPKT;
		rmd->rmd_own = 1;
	}

	/* setup transmit descriptors */
	bzero((char*)&ei->ei_mem->tmd[0], sizeof(ei->ei_mem->tmd));
	baddr = (u_int) &BASE->tbuf[0].ebh.ebh_ether;
	for (i = 0, tmd = &mem->tmd[0];
	     i < NUM_TMD;
	     i++, tmd++, baddr += sizeof(mem->tbuf[0])) {
		tmd->tmd_ladr = LO_ADDR(baddr);
		tmd->tmd_hadr = HI_ADDR(baddr);
		tmd->tmd_stp_enp = 3;
	}

	/* start the chip */
	io->io_rap = 1; DELAYBUS(0);
	io->io_rdp = CSR1; DELAYBUS(0);

	io->io_rap = 2; DELAYBUS(0);
	io->io_rdp = CSR2; DELAYBUS(0);

	io->io_rap = 3; DELAYBUS(0);
	io->io_rdp = CSR3; DELAYBUS(0);

	io->io_rap = 0; DELAYBUS(0);
	io->io_rdp = CSR0_INEA|CSR0_INIT|CSR0_STRT;

	return 0;
}

/*
 * Reset the board.
 */
static void
et_reset(eif)
	struct etherif *eif;
{
	register struct et_info *ei;

	ei = ET_INFO(eif);
	ASSERT(ei->ei_io);
	ei->ei_if.if_timer = 0;		/* turn off watchdog */
	ET_RESET();
}

/*
 * Ethernet interface interrupt.
 */
if_etintr(int unit)
{
	ETIO io;
	struct et_info *ei;
	register int s = splimp();	/* get the spin lock */

	ASSERT(unit == 0);
	ei = &et_info;
	io = ei->ei_io;

	if (io == 0) {				/* ignore early interrupts */
		printf("et0: early interrupt\n");
		splx(s);
		return 1;
	}
	IFNET_LOCKNOSPL(&ei->ei_if);
	et_poll(ei, 1);
	IFNET_UNLOCKNOSPL(&ei->ei_if);
	splx(s);
}

/* ARGSUSED */
et_poll(struct et_info *ei, int yield)
{
	ETMEM mem;
	ETIO io;
	volatile struct rmd *rmd;
	volatile struct tmd *tmd;
	int i;
	int found = 0;
	int yield_len = 0;

	ASSERT(IFNET_ISLOCKED(&ei->ei_if));
	ei = &et_info;
	mem = ei->ei_mem;
	io = ei->ei_io;

	if (iff_dead(ei->ei_if.if_flags)) {	/* if down don't process */
		ether_stop(&ei->ei_eif, ei->ei_if.if_flags);
		printf("et0: interrupt during down\n");
		return 1;
	}
	if (et_err()) {
		printf("et0: memory parity\n");
		(void) et_start(ei, ei->ei_if.if_flags);
		ei->ei_if.if_oerrors++;
		return(1);
	}

	io->io_rap = 0; DELAYBUS(0);

	while ((i = io->io_rdp) & CSR0_INTR) {
		found = 1;
		ei->ei_if.if_timer = DOG; /* yes, postpone watchdog */

		if (i & (CSR0_RINT|CSR0_TINT)) {	/* note data int. */
			io->io_rdp = CSR0_INEA|CSR0_RINT|CSR0_TINT;
		} else if (i & CSR0_MERR) {
			printf("et0: memory timeout\n");
			(void) et_start(ei, ei->ei_if.if_flags);
			ei->ei_if.if_ierrors++;
			return(1);
		} else if (i & CSR0_BABL) {
			printf("et0: babbling\n");
			(void) et_start(ei, ei->ei_if.if_flags);
			ei->ei_if.if_oerrors++;
			return(1);
		} else if (i & CSR0_MISS) {	/* count dropped packets */
			io->io_rdp = CSR0_INEA|CSR0_MISS;
			ei->ei_if.if_idrops++;
			METER(et_istat_missed++);
		} else if (i & CSR0_IDON) {	/* forget about init. int. */
			io->io_rdp = CSR0_INEA|CSR0_IDON;
		} else {
			printf("et0: unknown interrupt\n");
			(void) et_start(ei, ei->ei_if.if_flags);
			ei->ei_if.if_oerrors++;
			return(1);
		}
	}


	rmd = &mem->rmd[ei->ei_ract];	/* process received packets */
	while (!rmd->rmd_own) {
		found = 1;
		ei->ei_if.if_ipackets++;

		if (yield && yield_len > 2000) {
			/* 
			 * used to do splx, splimp here but the handler is
			 * called at splimp, so that won't do any good
			 */
#ifdef IRIX33
			vme_yield(SCSI_ET_REQLV);
#else
			vme_yield(SCSI_ET_REQLV,0);
#endif
			yield_len = 0;
		}

		if (rmd->rmd_stp_enp != (STP|ENP)) {
			/* part of chain of packets? */
			if (rmd->rmd_stp_enp == STP) {
				/* count error on 1st packet */
				ei->ei_terr_lbolt = 0;
				printf("et0: receive packet overflow\n");
				ei->ei_if.if_ierrors++;
			}
			rmd->rmd_stp_enp = STP|ENP;
		} else {
			register int len;
			register struct ether_header *eh;
			struct mbuf *m;
			struct etherbufhead *ebh;
			int snoopflags = 0;

			if (rmd->rmd_buff) {
				ei->ei_terr_lbolt = 0;
				printf("et0: receive buffer error\n");
				ei->ei_if.if_ierrors++;
				(void) et_start(ei, ei->ei_if.if_flags);
				break;
			}

			if (rmd->rmd_err) {
				snoopflags = SN_ERROR;
				if (rmd->rmd_fram) {
					snoopflags |= SNERR_FRAME;
					METER(et_istat_frame++);
				}
				if (rmd->rmd_crc)  {
					snoopflags |= SNERR_CHECKSUM;
					METER(et_istat_crc++);
				}
				if (rmd->rmd_oflo)  {
					snoopflags |= SNERR_OVERFLOW;
					METER(et_istat_oflo++);
				}
			}

			/*
			 * Make sure the packet length is valid.
			 */
			len = rmd->rmd_mcnt;
			if (len < MIN_RPKT) {
				ei->ei_terr_lbolt = 0;
				printf("et0: packet too small (%d)\n", len);
				snoopflags |= SNERR_TOOSMALL | SN_ERROR;
				if (len < EHDR_LEN + CRC_LEN)
					len = EHDR_LEN + CRC_LEN;
			} else if (len > MAX_RPKT) {
				ei->ei_terr_lbolt = 0;
				printf("et0: packet too large (%d)\n", len);
				snoopflags |= SNERR_TOOBIG | SN_ERROR;
				len = MAX_RPKT;
			}

			eh = (struct ether_header *)
				HOST_ADDR(rmd->rmd_ladr,rmd->rmd_hadr,mem);
#ifdef CK_ADDR
			if (eh != &mem->rbuf[ei->ei_ract].ebh.ebh_ether) {
				printf(
				 "et0: broken rmd rbadr@%x=%x; should be=%x\n",
					LOC_ADDR(rmd,mem),
					MK_ADDR(rmd->rmd_ladr, rmd->rmd_hadr),
					&BASE->rbuf[ei->ei_ract].ebh.ebh_ether);
				ei->ei_if.if_ierrors++;
				(void) et_start(ei, ei->ei_if.if_flags);
				break;
			}
			if (rmd->rmd_bcnt != -MAX_RPKT) {
				printf(
				 "et0: broken rmd bcnt@%x=%x; should be=%x\n",
					LOC_ADDR(rmd,mem),
					rmd->rmd_bcnt, -MAX_RPKT);
				ei->ei_if.if_ierrors++;
				(void) et_start(ei, ei->ei_if.if_flags);
				break;
			}
#endif
			/*
			 * Trim off the tail & CRC from the incoming packet.
			 * and copy the packet into mbuf and pass it to
			 * ether_input.
			 */
			len -= CRC_LEN;
			m = m_vget(M_DONTWAIT,sizeof*ebh-EHDR_LEN+len,MT_DATA);
			if (m == 0) {
				DBG_PRINT(("et0: out of mbufs\n"));
				break;
			}

			ebh = mtod(m, struct etherbufhead *);
			IF_INITHEADER(ebh, &ei->ei_if, sizeof *ebh);

			bcopy(eh, &ebh->ebh_ether, len);

			ether_input(&ei->ei_eif, snoopflags, m);
			ei->ei_if.if_ibytes += len;
			yield_len += len;
		}
		rmd->rmd_own = 1;
		if (++ei->ei_ract >= NUM_RMD) {
			ei->ei_ract = 0;
			rmd = &mem->rmd[0];
		} else {
			rmd++;
		}
	}


	while (ei->ei_tnum != 0) {	/* process completed transmissions */
		tmd = &mem->tmd[ei->ei_tact];
		if (tmd->tmd_own)	/* only completed packets */
			break;

		ei->ei_sick = 0;	/* stop watchdog resets */
		found = 1;
		if (tmd->tmd_buff) {
			ei->ei_terr_lbolt = 0;
			printf("et0: transmit buffer error\n");
			(void) et_start(ei, ei->ei_if.if_flags);
			ei->ei_if.if_oerrors++;
			break;
		}

		if (tmd->tmd_uflo) {
			ei->ei_terr_lbolt = 0;
			printf("et0: transmit underflow\n");
			(void) et_start(ei, ei->ei_if.if_flags);
			ei->ei_if.if_oerrors++;
			break;
		}

		if (tmd->tmd_err) {
			register char *terr = NULL;

			ei->ei_if.if_oerrors++;
			if (tmd->tmd_rtry) {
				ei->ei_if.if_collisions += 16;
				ei->ei_tdr += tmd->tmd_tdr;
				ei->ei_numtdr++;
			}
			if (tmd->tmd_lcol)
				terr = "et0: late collision\n";
			else if (tmd->tmd_lcar) {
                /*
                 * Added some code so that we print out this message
                 * separately from the rest of the 'terr' messages.
                 * It's easier to resolve here, and we supposedly don't
                 * fall into the loop below that does dri_printf(terr).
                 */
				terr = "et0: no carrier: check Ethernet cable\n";
				cmn_err(CE_ALERT, "%s", terr);
				ei->ei_terr_lbolt = lbolt+EMSG_DELAY;
				ei->ei_terr = terr;
			}
			if (terr != NULL
			    && (terr != ei->ei_terr
				|| ei->ei_terr_lbolt <= lbolt)) {
				printf(terr);
				ei->ei_terr_lbolt = lbolt+EMSG_DELAY;
				ei->ei_terr = terr;
			}
		} else {
			ei->ei_if.if_collisions += tmd->tmd_try;
		}

		ei->ei_tact = (ei->ei_tact+1) % NUM_TMD;
		ei->ei_tnum--;		/* advance to next ring entry */
	}

	return found;
}

/*
 * Periodically poll the controller for input packets in case
 * an interrupt gets lost.
 */
/* ARGSUSED */
static
et_watchdog(int unit)
{
	struct et_info *ei = &et_info;

	ASSERT(IFNET_ISLOCKED(&ei->ei_if));
	if (et_poll(ei, 0))
		ei->ei_lostintrs++;

	if (ei->ei_sick > 1) {
		printf("et0: died and restarted\n");
		(void)et_init(&ei->ei_eif, ei->ei_if.if_flags);
		ei->ei_if.if_oerrors++;
	} else if (ei->ei_tnum != 0) {	/* still have output? */
		ei->ei_sick++;		/* yes, we may be sick */
		ei->ei_if.if_timer = DOG;
	}
}


/*
 * Copy from mbuf chain to transmitter buffer in VMEbus memory.
 */
static int					/* 0 or errno */
et_transmit(eif, edst, esrc, type, m0)
	register struct etherif *eif;		/* on this interface */
	struct etheraddr *edst, *esrc;		/* with these addresses */
	u_short type;				/* of this type */
	struct mbuf *m0;			/* send this chain */
{
	struct et_info *ei;
	ETMEM mem;
	volatile struct tmd *tmd;
	struct ether_header *eh;
	int len;

	ei = ET_INFO(eif);

	/*
	 * We keep at least one free buffer to ensure that we walk around
	 * the ring in step with the 7990.
	 */
	if (ei->ei_tnum >= NUM_TMD-1) {
#if NOISE >= 1
		ei->ei_terr_lbolt = 0;
		printf("et0: dropped packet\n");
#endif
		IF_DROP(&ei->ei_if.if_snd);
		m_freem(m0);
		return ENOBUFS;
	}

	mem = ei->ei_mem;
	tmd = &mem->tmd[ei->ei_tfree];
	if (tmd->tmd_own) {
		printf("et0: broken tmd ownership\n");
		IF_DROP(&ei->ei_if.if_snd);
		m_freem(m0);
		return ENOBUFS;
	}

	eh = (struct ether_header *)HOST_ADDR(tmd->tmd_ladr,tmd->tmd_hadr,mem);
#ifdef CK_ADDR
	if (eh != &mem->tbuf[ei->ei_tfree].ebh.ebh_ether) {
		printf("et0: broken tmd tbadr@%x=%x; should be=%x\n",
			    LOC_ADDR(tmd,mem),
			    MK_ADDR(tmd->tmd_ladr, tmd->tmd_hadr),
			    &BASE->tbuf[ei->ei_tfree].ebh.ebh_ether);
		(void) et_start(ei, ei->ei_if.if_flags);
		m_freem(m0);
		return ENOBUFS;	/* XXX */
	}
#endif

	ei->ei_tfree = (ei->ei_tfree + 1) % NUM_TMD;
	ei->ei_tnum++;

	*(struct etheraddr *)eh->ether_dhost = *edst;
	*(struct etheraddr *)eh->ether_shost = *esrc;
	eh->ether_type = type;

	len = m_datacopy(m0, 0, MAX_TPKT, (caddr_t)(eh + 1));

	if (RAWIF_SNOOPING(&ei->ei_rawif)
	    && snoop_match(&ei->ei_rawif, (caddr_t) eh, len)) {
		ether_selfsnoop(&ei->ei_eif, eh, m0, 0, len);
	}
	m_freem(m0);

	len += sizeof *eh;
	if (len < MIN_TPKT_U)
		len = MIN_TPKT_U;
	eiftoifp(eif)->if_obytes += len;
	tmd->tmd_bcnt = -len;

	tmd->tmd_own = 1;		/* finish the packet */

	ei->ei_io->io_rap = 0; DELAYBUS(0);
	ei->ei_io->io_rdp = CSR0_INEA|CSR0_TDMD;	/* poke the board */
	return 0;
}


/* routine to check out ET hardware */
static int
et_diag(struct et_info *ei)
{
	int i;
	long *p;

	/* test on-board RAM a little */
	for (i = 0, p = (long*)ei->ei_mem; i < RAM_SIZE/sizeof(long); i++, p++)
		*p = 0x12345678;
	if (bcmp((char*)ei->ei_mem, sizeof(long)+(char*)ei->ei_mem,
		 RAM_SIZE - sizeof(long))) {
		printf("et0: IO2 bad RAM\n");
		return(1);
	}

	for (i = 0, p = (long*)ei->ei_mem; i < RAM_SIZE/sizeof(long); i++, p++)
		*p = ~0x12345678;
	if (bcmp((char*)ei->ei_mem, sizeof(long)+(char*)ei->ei_mem,
		 RAM_SIZE - sizeof(long))) {
		printf("et0: IO2 bad RAM\n");
		return(1);
	}

	return(0);
}

char	XXXetheraddr[6];
char	XXXetheraddr_initialized;

struct etherifops etops =
    { et_init, et_reset, et_watchdog, et_transmit, lance_ioctl };

/*
 * early init for ethernet
 */
if_etedtinit(edtp)
	register struct edt *edtp;
{
	struct et_info *ei;
	struct etheraddr ea;
	register int i;
	register int s;
	piomap_t *piomap[2];

	if (!etpresent()) { 
		if (showconfig)
			printf("et0: IO2 missing\n");
		return;
	}

	ei = &et_info;
	if (ei->ei_io) {
		printf("et0: duplicate IO2 EDT entry\n");
		return;
	}


	/* map I/O address space */

	for (i = 0; i < 2; i++) {
		piomap[i] = pio_mapalloc(edtp->e_bus_type, edtp->e_adap, 
			&edtp->e_space[i], PIOMAP_FIXED, "et0");
		edtp->e_space[i].ios_vaddr = 
			pio_mapaddr(piomap[i], edtp->e_space[i].ios_iopaddr);
	}
	ei->ei_mem = (ETMEM)edtp->e_base;
	ei->ei_io = (ETIO)edtp->e_base2;

	ET_RESET(); 			/* reset LAN chip */
	ET_PARON();			/* enable parity checking */

	s = splimp();
	if (et_diag(ei) != 0) {
		pio_mapfree(piomap[0]);
		pio_mapfree(piomap[1]);
		splx(s);
		return;
	}

	/* get ethernet address from board, 1 byte at a time */
	for (i = 0; i < sizeof ea.ea_vec; i++)
		ea.ea_vec[i] = ei->ei_io->io_ad[ETHERADDRLEN-i-1].ad_byte;

	if (showconfig) {
		printf("et0: hardware ethernet address %s\n",
			   ether_sprintf(ea.ea_vec));
	}

	if (ea.ea_vec[0] != SGI_ETHER_ADDR) {
		printf("et0: board has bad ethernet address: %s\n",
			   ether_sprintf(ea.ea_vec));
		pio_mapfree(piomap[0]);
		pio_mapfree(piomap[1]);
		splx(s);
		return;
	}

	if (!XXXetheraddr_initialized) {
		bcopy(ea.ea_vec, XXXetheraddr, sizeof ea.ea_vec);
		XXXetheraddr_initialized = 1;
	}

	ether_attach(&ei->ei_eif, "et", 0, (caddr_t)ei, &etops, &ea,
		     INV_ETHER_ET, 2);
	ei->ei_ib = &ei->ei_mem->ib;
	splx(s);
}

/*
 * et_err()
 *	check the INTSTATUS VEC0 for ENETERR_MASK
 */
et_err()
{
	register volatile struct mc1vector *mc1vector;
	register int stat;

	mc1vector = (struct mc1vector *)(MC1INTVECTOR);
	stat = mc1vector->statword & ENETERR_MASK;
	if (!stat)
		return(ENETERR_MASK);
	return(0);
}

#endif /* IP5 */
