/*
 * GIO bus Token-Ring Driver
 *
 * Copyright 1992 Silicon Graphics, Inc.  All rights reserved.
 */
#ident "$Revision: 1.63 $"

/*
 * Compile only for Indigo(And Magnum since we can't tell the difference)
 */
#if defined(IP12) || defined(IP20) || defined(IP22)

#ifndef _IRIX4
#define	HOSTLLC	1
#endif
#define STEP		1
#define	QDBG		1
/*#define	ONCHIPLLC	1*/
/*#define IFRMDBG 	1*/
/*#define IWAITDBG 	1*/
/* delay 1000usec for iframe xmit */
#define SLOWDOWN
#define SYIELD		/* s1 = spl1() */
#define EYIELD		/* splx(s1) */

/* This may cause problem for MP but no MP for now */
/*#define ISTRINTR	1 */
#define NEWIFRAME	1

#ifdef IWAITDBG
int     babble = 0;
#define IWMSG(a)	printf a
#else
#define IWMSG(a)
#endif

#include "tcp-param.h"
#include "sys/types.h"
#include "sys/sysmacros.h"
#include "sys/cmn_err.h"
#include "sys/debug.h"
#include "sys/cpu.h"
#include "sys/vmereg.h"
#include "sys/immu.h"
#include "sys/pda.h"
#include "sys/param.h"
#include "sys/systm.h"
#include "sys/mbuf.h"
#include "sys/socket.h"
#include "sys/errno.h"
#include "sys/edt.h"
#include "sys/kopt.h"
#include "sys/invent.h"
#include "sys/major.h"
#include "bstring.h"

#include "sys/time.h"
#include "sys/sbd.h"
#include "sys/stream.h"
#include "sys/strmp.h"

#include "net/soioctl.h"
#include "net/if.h"
#include "net/if_types.h"
#include "net/netisr.h"
#include "sys/par.h"
#include "net/raw.h"
#include "net/multi.h"
#include "net/route.h"

#include "netinet/in_systm.h"
#include "netinet/in.h"
#include "netinet/ip.h"
#include "netinet/if_ether.h"
#include "netinet/ip_var.h"
#include "netinet/in_pcb.h"
#include "netinet/udp.h"
#include "netinet/tcp.h"
#include "netinet/udp_var.h"
#include "netinet/tcpip.h"

#include "ether.h"
#include "sys/tr.h"
#include "sys/llc.h"
#include "sys/fddi.h"
#include "sys/tms380.h"
#include "sys/tr_user.h"
#include "sys/if_gtr.h"
#ifndef _IRIX4
#include "sys/dlsap_register.h"
#endif
#include "sys/capability.h"

extern struct ifqueue ipintrq;		/* ip packet input queue */

#ifdef ONCHIPLLC
#define LDP(a)
#endif

#ifdef QDBG
struct tr_qs	gtrqs;
struct tr_ints	gtrints;
struct tr_ifx	gtrifx;

#define QDBGUP(a,b)	(a)+=(b)
#define QDBGDOWN(a,b)	(a)-=(b)
#define QDBGHW(a,b)	if ((a) > (b)) { (b) = (a); }
#define QDBGMSG(a)	printf a
#else
#define QDBGUP(a,b)
#define QDBGDOWN(a,b)
#define QDBGMSG(a)
#endif

#if defined(IP20) || defined(IP22)
#define WBCACHE				/* IP17 style R4000 cache */
#else
#undef WBCACHE
#endif

extern time_t lbolt;
extern struct ifnet loif;

/* lboot variables */
extern gtr_mtu;
extern int gtr_batype;
extern u_short gtr_bc_mask;
extern int gtr_timer_t1_1, gtr_timer_t2_1, gtr_timer_ti_1;
extern int gtr_timer_t1_2, gtr_timer_t2_2, gtr_timer_ti_2;
extern int gtr_cksum;
extern int gtr_low_latency;

static char gtr_baddr_c[6] = {0xc0, 0x00, 0xff, 0xff, 0xff, 0xff};
static char gtr_baddr_f[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

struct gtr_info;			/* forward reference */

#ifdef ONCHIPLLC
static int gtr_dcmd(struct gtr_info*,caddr_t,u_long,caddr_t,u_long,u_long);
static int gtr_open_ip_sap(struct gtr_info*);
static void gtr_dcmdintr(struct gtr_info*, u_long, SSB*, TR_ICB*);
static int gtr_xllc(struct gtr_info*, TR_XMIT_PARMS*, u_short);
static int gtr_llc_stat(struct gtr_info*);
static int gtr_irearm(struct gtr_info*);
static int dcmd_hndl(struct gtr_info *);
#endif

static void gtr_initring(struct gtr_info *gtr, int buffers_only);
static void gtr_hang(struct gtr_info*);
static void gtr_poolov(struct gtr_info *gtr);
static int poolpeek(BPOOL*);
static int poolput(BPOOL*, struct mbuf*);
static struct mbuf* poolget(struct gtr_info*,BPOOL*, u_long);
static void rrfree(SLOT*);
static SLOT* rrnxt(RING*);
static SLOT* xrnxt(RING*);
static int xrfree(RING*);
static void ringinit(struct ringhdr*, int);
static int errmsg(struct gtr_info*);
static int gtr_recv(struct gtr_info*, SLOT*);
static int gtr_xcmd(struct gtr_info*,u_long,caddr_t,u_long,caddr_t,u_long,u_long);
static int gtrsend(struct gtr_info*, struct mbuf*, u_short, XLOB*);
static int gtr_close(struct gtr_info*, int);
static int gtr_insert(struct gtr_info*);
static int gtr_reset(struct gtr_info*, int);
static int gtr_init(struct gtr_info*);
static int gtr_nibplay(struct gtr_info*, int, u_long);
static int gtr_ready(struct gtr_info*);
static int gtr_chkvers(struct gtr_info*);
static int gtr_iwakeup(struct gtr_info*);

static int gtr_valid_unit(int);
static char *gtr_get_sa(int);
static int gtr_lli_output(u_short, SCB*);

/* flags to distinguish between IP and SNA packets */
#define	TR_IP_PACKET	0x01
#define	TR_SNA_PACKET	0x02
#define TR_DROP_PKT	0x04

/*
 * Token Ring internal data structure
 */
struct gtr_info {
	struct trif	gtr_tif;	/* common TR interface structure */
#define gtr_ac		gtr_tif.tif_arpcom
#define gtr_rawif	gtr_tif.tif_rawif
#define gtr_bia		gtr_tif.tif_bia
#ifdef XXX
	struct arpcom	gtr_ac;		/* common interface structure */
	struct rawif	gtr_rawif;
	char	gtr_bia[TR_MAC_ADDR_SIZE];
#endif
	u_char *dma1;			/* uncached dma area #1 */
	u_char *dma2;			/* uncached dma area #2 */
	struct mfilter	mfilter;


	GTR_HC	*hc;			/* host cmd area - 128 */
	GTR_CNTS *stat;			/* status counters - 120 */
	RING	*xr;			/* xmit ring - 2192 */
	RING	*rr;			/* recv ring - 2192 */
	BPOOL	*sml;			/* small mbuf pool - 272 */
	BPOOL	*med;			/* medium mbuf pool - 272 */
	BPOOL	*big;			/* large mbuf pool - 272 */

#define DOG	(2*IFNET_SLOWHZ)	/* check this often after alive */

#define DELAY_NIBBLE	50		/* usec delay for each nibble */
#define DELAY_OP	1000		/* most operations */
#define DELAY_BUD	(9*1000*1000)	/* 0.5*6*3=9 sec for BUD completion */
#define DELAY_OPEN	(12*1000*1000)	/* 11 sec for OPEN + 1sec margin */
#define DELAY_COPY	((128*1024*2/16)*3) /* time to copy EPROM */
#define DELAY_COPY_DBG	(5*1000*1000)	/* 5 sec for emulator */
#define DELAY_RESET	(DELAY_COPY+DELAY_BUD+DELAY_OP)
#define DELAY_RESET_DBG	(DELAY_COPY_DBG+DELAY_BUD+DELAY_OP)


#define NORM_CHECK	500
#define START_CHECK	100
#define OP_CHECK	15		/* check command done this often */

#define BADBD(gtr,ex,bail) if (!(ex)) {				\
	printf("gtr%d: failure \"ex\"\n", (gtr)->gif_unit);	\
	(void)gtr_reset(gtr,0);					\
	bail;							\
}
#define CMDBD(gtr,c) {int tmp; (gtr)->hc->cmd = (c);		\
		      (gtr)->cmd_line = __LINE__;		\
		      (gtr)->hc->cmd_id = ++(gtr->xid);		\
		      tmp = (gtr)->hc->cmd_id; /* flush wb */	\
		      *(gtr->gio) = GTR_GIO_H2B_INTR;}
	volatile unsigned long *gio;	/* look for the board here */

#define CMD_OK(gtr,id)	((gtr)->hc->cmd_ack == id)
	u_long		inum;
	u_long		xid;
	u_long		cmd_line;
	u_int		flags;		/* miscellaneous flags */
	int		addrset;	/* MAC_SETADDR flag */
	int		naddrset;	/* IP_SETADDR flag */

	TR_GRPADDR	tr_grp_addr;
	TR_FCNADDR	tr_fcn_addr;

#define GTR_STATE_RESET		0x00	/* device is resetting */
#define GTR_STATE_INIT		0x01	/* device is initializing */
#define GTR_STATE_OPEN		0x02	/* device is being opened */
#define GTR_STATE_OPENSAP	0x03	/* IP sap is being opened */
#define GTR_STATE_OK		0x04	/* device is xmitting & receiving */
#define GTR_STATE_CLOSE		0x05	/* device is closed */
#define GTR_STATE_SICK		0x80	/* device is broken */
	u_int		bd_state;	/* useful state bits */
	int		zapped;
	int		rtry;		/* # of init retries */

#define TR_INCR_TIMEOUT         HZ
#define TR_RESET_TIMEOUT        35
#define TR_OPEN_TIMEOUT         20
	int		reset_tid;
	int		init_tid;
	time_t		resettimo;	/* timeout time of last rcv */
	int		inittimo;

#define	CMD_BUSY	0x1
#define	CMD_CANWAIT	0x2
#define	CMD_PENDING	0x4
#define	CMD_HANG	0x8
	u_long		cmd_Flags;	/* pending wait */
	u_long		cmd_Status;	/* in progress wait */

	TMS_INTERNAL_ADDR ia;		/* internal addrs */

	u_long		ip_sid;
	u_long		vers;

	u_char		gtr_hash[256/8];/* DA filter */

	int		first_xmit;
	int		rcvpad;
	u_long		alarm;
	struct mbuf	*alarm_mbuf;

	u_char	mcl[10];		/* Micro Code Level */
	u_char	pid[18];		/* Product ID */
	u_char  sid[4];			/* Station ID */

	u_short	nullsri;		/* default bcast sri */

#ifdef ONCHIPLLC
	SSB 		*ssb;		/* Direct LLC status for ONCHIPLLC */
	u_long		*sifsts;	/* Direct LLC SIFSTS */
	u_char		*ssb_parm;	/* Direct LLC status param */
	u_short		dcmd;		/* direct cmd pending for status */
	u_short		ir_sid;		/* iframe rearm sid */
	int		ir_tid;
	u_long		send_seq;	/* send frame seq num */
	TR_LLC_STAT_DATA llcstat;	/* Direct cmd LLC status */
#endif
	u_long		cmdhiwat;
	u_long		sndhiwat;
	u_long		smlhiwat;
	u_long		mdehiwat;
	u_long		bighiwat;

	int		firstsml;
	int		firstmed;
	int		firstbig;

	time_t		xmit_hz;
} gtr_info[GTR_MAXBD];
#define gif		gtr_ac.ac_if
#define gif_enaddr	gtr_ac.ac_enaddr
#define gif_unit	gif.if_unit
#define gif_mtu		gif.if_mtu

typedef struct tokenbufhead {
	struct ifheader         tbh_ifh;	/* 8 bytes */
	struct snoopheader      tbh_snoop;	/* 16 bytes */
} TOKENBUFHEAD;

#ifdef DEBUG_FW
typedef struct tr_dbg_info {
	int tlen;
	int nbufs;
	int b1len;
	int b2len;
} TRDBGINFO;

TRDBGINFO tr_dbg_info[0x100];
int trdbginfo_ix;
int even_worse;
int fwbug_hit;
#endif


/* overheads */
#define TRPAD		sizeof(TOKENBUFHEAD)
#define LLCOVHD		(sizeof(*mh)+sizeof(LLC1)+sizeof(SNAP))
#define SRCROVHD	32

static void
gtr_hexdump(char *cp, int len)
{
	register int idx;
	register int qqq;
	char qstr[512];
	int maxi = 128;
	static char digits[] = "0123456789abcdef";

	if (len < maxi)
		maxi = len;
	for (idx = 0, qqq = 0; qqq<maxi; qqq++) {
		if ((qqq%16) == 0)
			qstr[idx++] = '\n';
		qstr[idx++] = digits[cp[qqq] >> 4];
		qstr[idx++] = digits[cp[qqq] & 0xf];
		qstr[idx++] = ' ';
	}
	qstr[idx] = 0;
	printf("%s\n", qstr);
}

static int
gtr_iwakeup(struct gtr_info *gtr)
{
	gtr->init_tid = 0;
	wakeup((caddr_t)&gtr->bd_state);
	return(0);
}

static void
gtr_setnullsri(struct gtr_info *gtr)
{
	if (gtr_bc_mask == 0)
		gtr->nullsri = (TR_RIF_ALL_ROUTES|0x200);
	else
		gtr->nullsri = (gtr_bc_mask|0x200);

	if (gtr->gif_mtu >= 17800)
		gtr->nullsri |= TR_RIF_LF_17800;
	else if (gtr->gif_mtu >= 11407)
		gtr->nullsri |= TR_RIF_LF_11407;
	else if (gtr->gif_mtu >= 8144)
		gtr->nullsri |= TR_RIF_LF_8144;
	else if (gtr->gif_mtu >= 4472)
		gtr->nullsri |= TR_RIF_LF_4472;
	else if (gtr->gif_mtu >= 2052)
		gtr->nullsri |= TR_RIF_LF_2052;
	else if (gtr->gif_mtu >= 1470)
		gtr->nullsri |= TR_RIF_LF_1470;
	else if (gtr->gif_mtu >= 516)
		gtr->nullsri |= TR_RIF_LF_516;
	else
		gtr->nullsri |= TR_RIF_LF_INITIAL;
}

/*
 *  Initialize the board.
 *	- if_gtrinit is called in splimp already.
 */
static
if_gtrinit(int unit)
{
	struct gtr_info *gtr = &gtr_info[unit];
	struct ifnet *ifp = &gtr->gif;
	int secs = 0;

	if ((ifp->if_flags & IFF_RUNNING) != 0) {
		DP(("if_gtrinit%d: already running\n", gtr->gif_unit));
		return(0);
	}

	if (gtr->init_tid) {
		untimeout(gtr->init_tid);
		gtr->init_tid = 0;
	}
	while (gtr->bd_state < GTR_STATE_OK) {
		gtr->init_tid = cputimeout((void (*)())gtr_iwakeup,
					(caddr_t)gtr,TR_INCR_TIMEOUT);
		DP(("gtr%d: sleeping for open\n",gtr->gif_unit));
		sleep((caddr_t)&gtr->bd_state, PZERO|PCATCH);
		DP(("gtr%d: wakenup for open, secs=%d\n",gtr->gif_unit,secs));
		if (++secs>=TR_OPEN_TIMEOUT)
			break;
	}
	if (gtr->bd_state != GTR_STATE_OK) {
#ifdef STASH
		int s;
		gtr->bd_state = GTR_STATE_SICK;
		s = splimp();
		gtr_reset(gtr,0);
		splx(s);
#endif
		return(EIO);
	}
	ifp->if_flags |= (IFF_UP|IFF_RUNNING);
	gtr->gif.if_timer = DOG;	/* turn on watchdog */
	return(0);
}

#ifdef STASH
/* This code is stashed because it's essentially putting the chip
 * into promiscuous mode and filtering the received packets in
 * software. Seems cheaper to turn on a functional address bit
 * instead, and let the chip do the address decoding in hw.
 */

/* Set the MAC multicast mode, address filters, and so on
 *	Interrupts must be off.
 */
static void
gtr_set_mcast(struct gtr_info *gtr)
{
	struct hc_mult arg;
	struct mfent *mfe;

	arg.mode = gtr->flags & (GTR_PROMISC|GTR_ALLMULTI|GTR_MULTI);
	bcopy(gtr->gtr_hash, arg.hash, sizeof(arg.hash));
	(void)gtr_xcmd(gtr,XCMD_MADDR,(char*)&arg,sizeof(arg),0,0, 0);
}

/* compute DA hash value
 */
static mval_t
gtr_dahash(char *addr)
{
	u_char hv;

	/* This is not called often, and so clarity is more important than
	 *	speed.
	 */

	hv = addr[0] ^ addr[1] ^ addr[2] ^ addr[3] ^ addr[4] ^ addr[5];

	return((mval_t)hv);
}


/* turn on a DA hash bit
 */
static
gtr_hash1(struct gtr_info *gtr, mval_t b)
{
	u_char *hp = &gtr->gtr_hash[0];

	hp[b/8] |= (((unsigned char)0x80) >> (b % 8));
}



/* turn off a DA hash bit
 */
static
gtr_hash0(struct gtr_info *gtr, mval_t b)
{
	u_char *hp = &gtr->gtr_hash[0];

	hp[b/8] &= ~(((unsigned char)0x80) >> (b % 8));
}
/* Add a destination address.
 */
static int
gtr_add_da(struct gtr_info *gtr, union mkey *key, int ismulti)
{
	struct mfreq mfr;
 
	if (mfmatchcnt(&gtr->mfilter, 1, key, 0))
		return 0;

	mfr.mfr_key = key;
	mfr.mfr_value = gtr_dahash(key->mk_dhost);
	if (!mfadd(&gtr->mfilter, key, mfr.mfr_value)) {
		return ENOMEM;
	}

	gtr_hash1(gtr, mfr.mfr_value);	/* tell the board */

	gtr_set_mcast(gtr);

	return 0;
}


/* Delete an address filter.  If key is unassociated, do nothing.
 * Otherwise delete software filter first, then hardware filter.
 */
static int
gtr_del_da(struct gtr_info *gtr, union mkey *key, int ismulti)
{
	struct mfreq mfr;

	if (!mfmatchcnt(&gtr->mfilter, -1, key, &mfr.mfr_value))
		return 0;
	mfdel(&gtr->mfilter, key);

	/* Remove bit from the table if no longer in use.
	 */
	if (!mfhasvalue(&gtr->mfilter, mfr.mfr_value)) {
		gtr_hash0(gtr, mfr.mfr_value);
	}

	gtr_set_mcast(gtr);

	return 0;
}
#endif /* STASH */

/* Add a functional address.
 */
static int
gtr_add_fa(struct gtr_info *gtr, struct sockaddr * sa)
{
	int error;

	gtr->tr_fcn_addr.a[0] |= *(u_short *) &(sa->sa_data[2]);
	gtr->tr_fcn_addr.a[1] |= *(u_short *) &(sa->sa_data[4]);
#if DBG_LVL >= DBG_LVL_DRV
	printf ("gtr0: Setting functional address %x-%x.\n",
		gtr->tr_fcn_addr.a[0], gtr->tr_fcn_addr.a[1]);
#endif
	error = gtr_xcmd(gtr,XCMD_FADDR, (char*)&gtr->tr_fcn_addr,
			sizeof(gtr->tr_fcn_addr),0,0,0);
	return error;
}


/* Delete a functional address.
 */
static int
gtr_del_fa(struct gtr_info *gtr, struct sockaddr * sa)
{
	int error;

	gtr->tr_fcn_addr.a[0] &= ~(*(u_short *) &(sa->sa_data[2]));
	gtr->tr_fcn_addr.a[1] &= ~(*(u_short *) &(sa->sa_data[4]));

	error = gtr_xcmd(gtr,XCMD_FADDR, (char*)&gtr->tr_fcn_addr,
			sizeof(gtr->tr_fcn_addr),0,0,0);
	return error;
}

/*
 * Convert an Internet multicast socket address into a 
 * token ring MAC address.
 */
gtr_cvtmulti(register struct sockaddr *sa, int *all)
{
    *all = 0;
    switch (sa->sa_family) {
	case AF_UNSPEC:
	    break;

	case AF_INET:
	    if (((struct sockaddr_in *)sa)->sin_addr.s_addr == INADDR_ANY) {
		*all = 1;
	    }
	    /* 
	     * Per RFC 1469, all IP mcast addresses 
	     * map to C0-00-00-04-00-00.
	     */
	    sa->sa_data[0] = 0xC0;
	    sa->sa_data[1] = 0;
	    sa->sa_data[2] = 0;
	    sa->sa_data[3] = 4;
	    sa->sa_data[4] = 0;
	    sa->sa_data[5] = 0;
	    sa->sa_family = AF_UNSPEC;
	    break;

	case AF_RAW:
	    sa->sa_family = AF_UNSPEC;
	    break;

	default:
	    return EAFNOSUPPORT;
    }

    /*
     * Verify that we have a valid tokenring multicast addresses.
     */
    if (sa->sa_data[0] != 0xC0 || sa->sa_data[1] != 0) {
	    return EINVAL;
    }
    /* Firmware has no support for group addresses, only for 
     * functional addresses. */
    if (sa->sa_data[2] & 0x80)
	    return EOPNOTSUPP;

    return(0);
}

/*
 *  Process an ioctl request.
 */
static int
if_gtrioctl(struct ifnet *ifp, int cmd, caddr_t data)
{
	int i;
	int error = 0;
	int s = splimp();
	TR_FCNADDR *fa = (TR_FCNADDR *)data;
	struct gtr_info *gtr = &gtr_info[ifp->if_unit];

	ASSERT(&gtr->gtr_ac == (struct arpcom*)ifp);
	switch (cmd) {
	case SIOCSIFADDR: {
		struct sockaddr *addr = _IFADDR_SA(data);
		int flags = ((struct ifreq *)data)->ifr_flags;
		int naddrset =	gtr->naddrset;

		switch (addr->sa_family) {
	    	case AF_INET:
			DP(("gtr%d:SIOCSIFADDR-INET\n",gtr->gif_unit));
			/*
			 * Set IP addr no matter what, so that null ipaddr
			 * can be avoided.
			 */
			gtr->gtr_ac.ac_ipaddr = ((struct sockaddr_in*)addr)->sin_addr;
			((struct arpcom*)ifp)->ac_ipaddr=((struct sockaddr_in*)data)->sin_addr;

			gtr->naddrset = 1;
			if (naddrset) {
				/* force hard reset except first time */
				DP(("gtr%d:SIFADDR:FORCE\n",gtr->gif_unit));
				if (gtr->reset_tid) {
					DP(("gtr%d:SIFADDR:FORCE:stopped other reset\n", gtr->gif_unit));
					untimeout(gtr->reset_tid);
					gtr->reset_tid = 0;
				}
				if (gtr->init_tid) {
					DP(("gtr%d:SIFADDR:FORCE: stopped other init\n",gtr->gif_unit));
					gtr->bd_state = FSTATE_SANITY;
					(void)gtr_iwakeup(gtr);
				}
				error = gtr_reset(gtr,1);
				if (error) {
					DP(("gtr%d: SIOCSIFADDR: FORCE: reset failed\n",gtr->gif_unit));
					break;
				}
				DP(("gtr%d:SIFADDR:FORCE:reset done\n",gtr->gif_unit));
			} else if (gtr->zapped) {
				DP(("gtr%d: SIOCSIFADDR: zap\n",gtr->gif_unit));
				/* Board was zapped */
				if (gtr_chkvers(gtr)) {
					DP(("gtr%d:SIFADDR: vers failed\n",gtr->gif_unit));
					error = EIO ;
					break;
				}
				DP(("gtr%d:SIFADDR: vers checked\n",gtr->gif_unit));
				gtr_initring(gtr, 1);
				gtr->zapped = 0;
				if ((error = gtr_init(gtr)) != 0) {
					DP(("gtr%d: SIOCSIFADDR: init failed\n",gtr->gif_unit));
					error = EIO ;
					break;
				}
				DP(("gtr%d:SIFADDR:init done\n",gtr->gif_unit));
			} else if ((gtr->flags&GTR_EDTINIT) == 0) {
				/* practically useless */
				DP(("gtr%d:SIFADDR:no EDTINIT yet\n",gtr->gif_unit));
				if ((error = if_gtrinit(ifp->if_unit)) != 0) {
					DP(("gtr%d: SIOCSIFADDR: no EDTINIT: gtrinit failed\n",gtr->gif_unit));
					break;
				}
			}

			break;

		case AF_RAW:
			DP(("gtr%d: SIOCSIFADDR-RAW\n",gtr->gif_unit));
			bcopy(addr->sa_data, gtr->gif_enaddr,
		      		sizeof(gtr->gif_enaddr));
			if (showconfig) {
				printf("gtr%d: tokenring address set to %s\n",
				    gtr->gif_unit, ether_sprintf(gtr->gif_enaddr));
			}
			gtr->addrset = 1;
			ifp->if_flags &= ~(IFF_RUNNING | IFF_UP);
			if (gtr_close(gtr,1)) {
				DP(("gtr%d: SIFADDR: close failed\n",gtr->gif_unit));
				error = EIO; break;
			}
			break;
		}

		/* here, wait for open */
		if (gtr->bd_state == GTR_STATE_OK) {
			DP(("gtr%d: SIOCSIFADDR: GTR_STATE_OK already\n",gtr->gif_unit));
			gtr->gif.if_flags |= (IFF_RUNNING|IFF_UP);
		} else {
			if (gtr->bd_state == GTR_STATE_SICK) {
				DP(("gtr%d: SIOCSIFADDR: GTR_STATE_SICK resetting\n",gtr->gif_unit));
			 	error = gtr_reset(gtr, 1);
			} else if (gtr->bd_state == GTR_STATE_CLOSE) {
				DP(("gtr%d: SIOCSIFADDR: GTR_STATE_CLOSE initting\n",gtr->gif_unit));
				error = gtr_init(gtr);
			}
			if (error != 0) {
				DP(("gtr%d: SIOCSIFADDR: ERROR=%d\n",gtr->gif_unit, error));
				break;
			}
			if ((error = if_gtrinit(ifp->if_unit)) != 0) {
				DP(("gtr%d: SIOCSIFADDR: if_gtrinit ERROR=%d\n",gtr->gif_unit, error));
				break;
			}
			DP(("gtr%d: SIOCSIFADDR: if_gtrinit DONE\n",gtr->gif_unit));
		}
		arpwhohas((struct arpcom*)ifp,
				&((struct arpcom *)ifp)->ac_ipaddr);
	} break;

	case SIOCSIFFLAGS: {
		int doreset = 0;
		int oflag = ifp->if_flags;
		struct ifreq *ifr = (struct ifreq *)data;

		DP(("gtr%d: SIOCSIFFLAGS: n=%x o=%x\n",
			gtr->gif_unit,ifr->ifr_flags,oflag));
		ifp->if_flags = ifr->ifr_flags;
  		if ((ifp->if_flags&IFF_PROMISC) && !(gtr->flags&GTR_PROMISC)){
#ifdef ONCHIPLLC
			if (gtr->flags&GTR_ONCHIPLLC) {
				DP(("gtr%d: SIOCSIFADDR: invalid SNOOP req\n",
							gtr->gif_unit));
				error = EINVAL;
				ifp->if_flags = oflag;
				break;
			}
#endif
			gtr->flags |= GTR_PROMISC;
			doreset = 1;
		} else if (!(ifp->if_flags&IFF_PROMISC)&&(gtr->flags&GTR_PROMISC)){
			gtr->flags &= ~GTR_PROMISC;
			doreset = 1;
		}
		if (doreset) {
			u_long parg;
			DP(("gtr%d: PROMISC change\n",gtr->gif_unit));
			parg = gtr->flags & GTR_PROMISC;
			error = gtr_xcmd(gtr,XCMD_PROMISC,
					(char*)&parg,sizeof(parg),0,0,0);
			if (error)
				break;

			if (ifp->in_ifaddr != 0) {
				arpwhohas((struct arpcom*)ifp,
					&((struct arpcom *)ifp)->ac_ipaddr);
			}
		}
  		if (!(ifp->if_flags & IFF_UP) && (oflag&IFF_UP)) {
			DP(("gtr%d: IFF_DOWN\n", gtr->gif_unit));
			if (gtr_close(gtr,0)) {
				DP(("gtr%d: SIOCSIFFLAGS: close failed\n",gtr->gif_unit));
				error = EIO; break;
			}
			gtr->gif.if_flags &= ~(IFF_RUNNING|IFF_UP);
			if_down(ifp);
		} else if ((ifp->if_flags&IFF_UP) && !(oflag&IFF_UP)) {
			DP(("gtr%d: IFF_UP\n", gtr->gif_unit));
			if (gtr->zapped) {
				/* Board was zapped */
				if (gtr_chkvers(gtr)) {
					DP(("gtr%d: SIOCSIFFLAGS: checkvers failed\n",gtr->gif_unit));
					error = EIO;
					break;
				}
				gtr_initring(gtr, 1);
				gtr->zapped = 0;
				if ((error = gtr_init(gtr)) != 0) {
					DP(("gtr%d: SIOCSIFFLAGS: init failed\n",gtr->gif_unit));
					break;
				}
			} else {
				if ((error = gtr_insert(gtr)) != 0) {
					DP(("gtr%d:SIOCSIFFLAGS: insert fail\n",
						gtr->gif_unit));
					break;
				}
			}
		 	if ((error = if_gtrinit(ifp->if_unit)) != 0) {
				DP(("gtr%d: SIOCSIFFLAGS: wait for insert failed\n",gtr->gif_unit));
				break;
			}
			/* if we were using functional addresses, 
			 * reinstate them. */
			if (gtr->tr_fcn_addr.a[0] || gtr->tr_fcn_addr.a[1])
				gtr_xcmd (gtr,XCMD_FADDR, 
					  (char*)&gtr->tr_fcn_addr,
					  sizeof(gtr->tr_fcn_addr),0,0,0);

			if (ifp->in_ifaddr != 0) {
				arpwhohas((struct arpcom*)ifp,
					&((struct arpcom *)ifp)->ac_ipaddr);
			}
		}
		DP(("gtr%d: SIOCSIFFLAGS done\n", gtr->gif_unit));
	} break;

	case SIOCADDMULTI:
	case SIOCDELMULTI: {
		int allmulti;

		error = gtr_cvtmulti((struct sockaddr *)data, &allmulti);

		if (0 == error) {
			s = splimp();
			if (SIOCADDMULTI == cmd) {
				error = gtr_add_fa(gtr, (struct sockaddr *)data);
			} else {
				error = gtr_del_fa(gtr, (struct sockaddr *)data);
			}
			if ((error == 0) && (allmulti)) {
				if (SIOCADDMULTI == cmd) {
					ifp->if_flags |= IFF_ALLMULTI;
				} else {
					ifp->if_flags &= ~IFF_ALLMULTI;
				}
			}
			splx(s);
		}
	} break;

	case SIOC_TR_GETMULTI:
		*fa = gtr->tr_grp_addr;
		break;

        case SIOC_TR_ADDMULTI:
	case SIOC_TR_DELMULTI:
		error = EINVAL;
		break;

	case SIOC_TR_GETFUNC:
		*fa = gtr->tr_fcn_addr;
		break;

	case SIOC_TR_ADDFUNC:
	DP(("gtr%d: FA_ADD %x %x\n",gtr->gif_unit,(int)fa->a[0],(int)fa->a[1]));
		/* Assumed only 1 agent requests */
		for (i = 0; i < (sizeof(*fa)/sizeof(short)); i++)
			gtr->tr_fcn_addr.a[i] |= fa->a[i];
		error = gtr_xcmd(gtr,XCMD_FADDR, (char*)&gtr->tr_fcn_addr,
					sizeof(gtr->tr_fcn_addr),0,0,0);
		break;

	case SIOC_TR_DELFUNC:
	DP(("gtr%d: FA_DEL %x %x\n",gtr->gif_unit,(int)fa->a[0],(int)fa->a[1]));
		for (i = 0; i < (sizeof(*fa)/sizeof(short)); i++)
			gtr->tr_fcn_addr.a[i] &= ~(fa->a[i]);
		error = gtr_xcmd(gtr,XCMD_FADDR, (char*)&gtr->tr_fcn_addr,
					sizeof(gtr->tr_fcn_addr),0,0,0);
		break;

	case SIOCADDSNOOP:
	case SIOCDELSNOOP:
		/* Force promisc always */
		error = EINVAL;
		break;

	case SIOC_TR_GETCONF: {
		TR_CONFIG trc;
		TR_SIOC *sioc = (TR_SIOC*)data;
		int len = MIN(sizeof(trc), sioc->len);

		bcopy(gtr->gif_enaddr,trc.caddr, TR_MAC_ADDR_SIZE);
		bcopy(gtr->gtr_bia, trc.bia, TR_MAC_ADDR_SIZE);
		bcopy(gtr->mcl, trc.mcl, sizeof(trc.mcl));
		bcopy(gtr->pid, trc.pid, sizeof(trc.pid));
		bcopy(gtr->sid, trc.sid, sizeof(trc.sid));

		if (copyout((caddr_t)&trc, (caddr_t)sioc->tr_ptr_u.conf, len))
			error = EFAULT;
		} break;

	case SIOC_TR_GETST: {
		TR_SIOC *sioc = (TR_SIOC*)data;
		GTR_CNTS *trs = gtr->stat;
		int len = MIN(sizeof(*trs), sioc->len);

		error = gtr_xcmd(gtr,XCMD_COUNT,0,0,0,0,0);
		if (error)
			break;
		if (copyout((caddr_t)trs, (caddr_t)sioc->tr_ptr_u.st, len))
			error = EFAULT;
		} break;

	case SIOC_TR_ARM: {
		TR_SIOC *sioc = (TR_SIOC*)data;

		if (!_CAP_ABLE(CAP_NETWORK_MGT)) {
			error = EPERM;
			break;
		}
		if (0 == gtr->alarm_mbuf) {
			gtr->alarm_mbuf = m_get(M_WAIT, MT_DATA);
			if (0 == gtr->alarm_mbuf) {
				error = ENOBUFS;
				break;
			}
		}
		sioc->tr_ptr_u.alarm = gtr->alarm;
		gtr->alarm = 0;
		} break;

	case SIOC_GTR_STO:
#define DWN ((struct gtr_dwn*)data)
		if (!_CAP_ABLE(CAP_NETWORK_MGT)) {
			error = EPERM;
			break;
		}
		if (!iff_dead(gtr->gif.if_flags)
		    || DWN->cnt == 0 || DWN->cnt > GTR_DWN_LEN) {
			error = EINVAL;
			break;
		}
#ifdef STASH
		if ((error = gtr_ready(gtr)) != 0)
			break;
#endif
		gtr->zapped = 1;
		error = gtr_xcmd(gtr,XCMD_STO,(char*)DWN,sizeof(*DWN), 0,0,1);
		if (error) {
			gtr->zapped = 0;
			gtr->gif.if_flags |= IFF_UP;
			gtr->bd_state = GTR_STATE_SICK;
		}
		break;

	case SIOC_GTR_EXEC:
		if (!_CAP_ABLE(CAP_NETWORK_MGT)) {
			error = EPERM;
			break;
		}
		if (!iff_dead(gtr->gif.if_flags)) {
			error = EINVAL;
			break;
		}
#ifdef STASH
		if ((error = gtr_ready(gtr)) != 0) {
			break;
		}
#endif
		gtr->bd_state = FSTATE_SANITY;
		gtr->zapped = 1;
		gtr->hc->sign = 0;
		error = gtr_xcmd(gtr,XCMD_EXEC,(char*)DWN,sizeof(*DWN), 0,0,1);
		if (error) {
			DP(("gtr: EXEC failed(%x) but ignoring\n",error));
			gtr->zapped = 0;
			gtr->gif.if_flags |= IFF_UP;
			gtr->bd_state = GTR_STATE_SICK;
		}
		break;

	case SIOC_GTR_FET:
		if (!_CAP_ABLE(CAP_NETWORK_MGT)) {
			error = EPERM;
			break;
		}
		if (DWN->cnt == 0 || DWN->cnt > GTR_DWN_LEN) {
			error = EINVAL;
			break;
		}

#ifdef STASH
		if (gtr->bd_state != GTR_STATE_OK &&
		   ((error = gtr_ready(gtr)) != 0)) {
			DP(("gtr: FET: failed\n"));
			break;
		}
#endif
		error = gtr_xcmd(gtr,XCMD_FET,(char*)DWN,sizeof(*DWN),
					(char*)DWN,sizeof(*DWN),1);
#undef DWN
		break;

	case SIOC_GTR_VERS:
		/* check that the board is alive */
		error = gtr_xcmd(gtr,XCMD_NOP,0,0,0,0,1);
		if (error)
			break;
		*(u_long*)data = gtr->vers;
		break;

	case SIOC_GTR_SIGNAL:
		if (!_CAP_ABLE(CAP_NETWORK_MGT)) {
			error = EPERM;
			break;
		}
		if (!iff_dead(gtr->gif.if_flags) || *(u_long*)data > 0xffff) {
			error = EINVAL;
			break;
		}
		if (*(u_long*)data != 0)
			(void)gtr_reset(gtr,0);
		if (gtr_nibplay(gtr,DELAY_RESET,(u_long)0))
			error = EIO;
		break;

	case SIOC_TR_GETLLCSTAT: {
#ifdef ONCHIPLLC
		TR_SIOC *sioc = (TR_SIOC*)data;
		TR_LLC_STAT_DATA *trs = &gtr->llcstat;
		int len = MIN(sizeof(*trs), sioc->len);

		if (copyout((caddr_t)trs, (caddr_t)sioc->tr_ptr_u.st, len)) {
			error = EFAULT;
			break;
		}
		error = gtr_llc_stat(gtr);
#else
		error = EIO;
#endif
		} break;
	}
	splx(s);
	return(error);
}

#ifdef ONCHIPLLC
static int
dcmd_hndl(struct gtr_info *gtr)
{
	SSB ssb;
	char *cp;
	TR_ICB icb;
	u_long *sifp;
	u_long sifsts;

	sifp = (u_long *)&(((char*)gtr->ssb)[sizeof(SSB)]);
	if (!IS_KSEG1(gtr->hc)) {
		sifp = (u_long*)K0_TO_K1(sifp);
	}

	sifsts = *sifp;
	if ((gtr->ssb->ssb_command == 0) && (sifsts == 0))
		return(0);

	bcopy((char*)gtr->ssb, (char*)&ssb, sizeof(ssb));
	bcopy((char*)gtr->ssb_parm, (char*)&icb, sizeof(icb));
	sifsts = *sifp;
	cp = (char*)&ssb;
	LDP(("AI: %x %x %x %x %x %x %x %x\n",
			(int)cp[0],
			(int)cp[1],
			(int)cp[2],
			(int)cp[3],
			(int)cp[4],
			(int)cp[5],
			(int)cp[6],
			(int)cp[7]));
	*sifp = 0;
	gtr->ssb->ssb_command = 0;
	if (ssb.ssb_command == gtr->dcmd)
		gtr->dcmd = 0;		/* free dcmd queue */
	gtr_dcmdintr(gtr, sifsts, &ssb, &icb);
	return(1);
}
#endif

#if DBG_LVL > 0
SLOT *gtr_cslot = 0;
#endif
/*
 * Process an interrupt from GIO TokenRing board.
 *
 *	comes at splgio{0|1} = splimp.
 *	But splgio2 is actually higher, so must be taken care of
 *	when such thing is ready.
 */
void
if_gtrintr(struct gtr_info *gtr)
{
	SLOT *sl;
	u_long inum;
	int found, i;
	int s = splimp();

	ASSERT(gtr >= &gtr_info[0] && gtr < &gtr_info[GTR_MAXBD]);
	ASSERT(gtr == &gtr_info[gtr->gif_unit]);

	QDBGUP(gtrints.tot,1);
	found = 0;
	*(gtr->gio) = GTR_GIO_B2H_CLR;	/* clear intr no matter what */

	inum = gtr->hc->inum;
	if (inum == gtr->inum) {
		splx(s);
		QDBGUP(gtrints.stray,1);
		return;		/* false alarm */
	}

#ifdef ONCHIPLLC
	while (dcmd_hndl(gtr) != 0) {
		found++;
	}
#endif
	if (gtr->bd_state != GTR_STATE_OK) {
		QDBGUP(gtrints.stray,1);
		splx(s);
		return;		/* false alarm */
	}

	gtr->inum = inum;

	/* XXX: TBD interrupt dealy */

	/* check status */
	if (gtr->hc->state&FSTATE_SICK) {
		/* let wdog or app take action */
		gtr->bd_state = GTR_STATE_SICK;
		(void)errmsg(gtr);
	}

	/* recv */
	i = 0;
	while ((sl = rrnxt(gtr->rr)) != 0) {
#if DBG_LVL > 0
		gtr_cslot = sl;
#endif
		i++;
		found += gtr_recv(gtr, sl);
		rrfree(sl);
	}
	if (i > 0) {
		QDBGUP(gtrints.recv,1);
#ifdef STASH
		DP(("gtr%d: INTR: received %d bytes in %d frames\n",
			gtr->gif_unit, found, i));
#endif
	}

	/* free xmit bufs */
	if ((i = xrfree(gtr->xr)) > 0) {
		while (i > 0) {
			struct mbuf *m;
			IF_DEQUEUE(&gtr->gif.if_snd, m);
			if (!m) {
				QDBGMSG(("%d xmit mbufs lost\n", i));
				break;
			}
			QDBGUP(gtrqs.sndq_freed,1);
			QDBGDOWN(gtr->sndhiwat,1);
			m_freem(m);
			i--;
		}
		QDBGUP(gtrints.xmit,1);
	}

	/* allocate mbufs for recv */
	if ((found > 0) || !poolpeek(gtr->sml) ||
	     !poolpeek(gtr->med) || !poolpeek(gtr->big)) {
		gtr_hang(gtr);
	}
	if (gtr->hc->state&FSTATE_NORBUF) {
		gtr_poolov(gtr);
	}
	splx(s);
}

static void
gtr_read_llc2(struct mbuf *m0)
{
	TR_MAC *mh;
	LLC1 *llcp;
	u_short sid;
	LLI_CB *cbp;
	TR_ICB *icbp;
	mblk_t *s0, *s1;
	struct mbuf *m1;
	struct ifnet *ifp;
	struct ifheader *ifh;
	int s2, s, ftype, off;
	int len, hlen, rlen, llen, mlen;
	u_int cur_seq;

	if (m0 == 0)
		return;

	s2 = spl1();
	s = splimp();
	s0 = 0;
	ifh = mtod(m0, struct ifheader *);
	ifp = ifh->ifh_ifp;
	M_ADJ(m0, ifh->ifh_hdrlen);
	for (len = 0, m1 = m0; m1 != 0; m1 = m1->m_next)
		len += m1->m_len;
	if (len > (ifp->if_mtu+sizeof(TR_MAC_HDR)))
		goto llcret_2;

	mh = mtod(m0, TR_MAC*);
	if ((mh->mac_fc & TR_FC_TYPE) != TR_FC_TYPE_LLC)
		goto llcret_1;

	if (mh->mac_sa[0] & TR_MACADR_RII) {
		TR_RII *rii = (TR_RII *)(mh+1);
		rlen = (rii->rii & TR_RIF_LENGTH) >> 8;
		mh->mac_sa[0] &= ~TR_MACADR_RII;
	} else {
		rlen = 0;
	}
	hlen = sizeof(TR_MAC) + rlen;
	llcp = (LLC1 *)(((char*)mh)+hlen);
	if ((sid = tr_addr2sid(ifp,(char*)&mh->mac_sa[0],llcp->llc1_dsap)) == 0)
		goto llcret_1;
	if (rlen != 0)
		mh->mac_sa[0] |= TR_MACADR_RII;

	switch (llcp->llc1_cont&LLC_FORMAT_MASK) {
	case LLC_UFORMAT:			/* UFORMAT */
		switch (llcp->llc1_cont&~LLC_UPF_MASK) {
		case LLC_CMD_UI:		/* TYPE 1 */
			ftype = TR_UI_FRAME; break;
		case LLC_CMD_XID:		/* TYPE 1 */
			if (llcp->llc1_ssap&LLC_SSAP_RESP) {
				if (llcp->llc1_cont&LLC_UPF_MASK)
					ftype = TR_XID_RESP_F1;
				else
					ftype = TR_XID_RESP_F0;
			} else {
				if (llcp->llc1_cont&LLC_UPF_MASK)
					ftype = TR_XID_CMD_P1;
				else
					ftype = TR_XID_CMD_P0;
			}
			break;
		case LLC_CMD_TEST:		/* TYPE 1 */
			if (llcp->llc1_cont&LLC_UPF_MASK)
				ftype = TR_TEST_RESP_F1;
			else
				ftype = TR_TEST_RESP_F0;
			break;
		default:
		case LLC_CMD_SABME:		/* TYPE 2 */
		case LLC_CMD_DISC:		/* TYPE 2 */
		case LLC_CMD_UARESP:		/* TYPE 2 */
		case LLC_CMD_DMRESP:		/* TYPE 2 */
		case LLC_CMD_FRMRRESP:		/* TYPE 2 */
			ftype = TR_NON_MAC_FRAME;/* for the time being */
			break;
		}
		llen = sizeof(LLC1);
		break;
	case LLC_SFORMAT:			/* SFORMAT - TYPE 2 */
		ftype = TR_NON_MAC_FRAME;/* for the time being */
		switch (llcp->llc1_cont&LLC_SCMD_MASK) {
		default:
		DP(("%s%d: unknown SFORMAT = %x",
			ifp->if_name,ifp->if_unit, (int)llcp->llc1_cont));
		case LLC_CMD_RR:
		case LLC_CMD_REJ:
		case LLC_CMD_RNR:
			break;
		}
		llen = sizeof(LLC2);
		break;
	default:			/* IFORMAT - TYPE 2 */
		ASSERT((llcp->llc1_cont&LLC_SFORMAT) == 0);
		ftype = TR_I_FRAME;
		llen = sizeof(LLC2);
		cur_seq = llcp->llc1_cont;
		cur_seq &= 0xfe;
		QDBGUP(gtrifx.iiq_tot,1);
		break;
	}

	mlen = min(((sizeof(LLI_CB)+sizeof(TR_ICB)+len)), VCL_MAX);
	if ((s0 = allocb(mlen,BPRI_MED)) == 0) {
		printf("gtr_read_llc2: no mblk size=%d\n",mlen);
		goto llcret_2;
	}
	s0->b_datap->db_type = M_PROTO;
	s0->b_wptr = s0->b_rptr + sizeof(LLI_CB);
	cbp = (LLI_CB *)s0->b_rptr;
	cbp->port = ifp->if_unit;
	cbp->command = LLI_RCV_PENDING;
	cbp->param_blk = s0->b_wptr;

	icbp = (TR_ICB*)s0->b_wptr;
	s0->b_wptr += sizeof(*icbp);
	icbp->lan_header_len = hlen + rlen;
	icbp->llc_header_len = llen;
	icbp->frame_len = len;
	icbp->frame_type = ftype;
	icbp->station_id_r = sid;

	mlen -= (sizeof(LLI_CB)+sizeof(TR_ICB));
	len -= mlen;
	for(off = 0, s1 = s0;;) {
		off =  m_datacopy(m0, off, mlen, (char*)s1->b_wptr);
		s1->b_wptr += mlen;
		len -= mlen;
		if (len <= 0)
			break;
		mlen = min(len, VCL_MAX);
		if ((s1->b_cont = allocb(mlen,BPRI_MED)) == 0) {
			printf("%s%d: no mblk size=%d\n",
				ifp->if_name,ifp->if_unit,mlen);
			goto llcret_2;
		}
		s1 = s1->b_cont;
	}

	if (ftype != TR_I_FRAME) {
		tr_lli_input(s0);
		goto llcret;
	}

{
	int tmp;
	LCB *lcb;
	IIQ *iiq;

	if (ifp == 0) {
		printf("ifp = 0\n");
		goto llcret_2;
	}
	if ((lcb = tr_find_WR_q(ifp, sid)) == 0) {
		printf("%s%d:read_llc2: no lcb, sid=%x\n", ifp->if_name, ifp->if_unit, sid);
		tr_lli_input(s0);
		goto llcret;
	}

/* gtr_hexdump(&s0->b_rptr[sizeof(LLI_CB)+sizeof(TR_ICB)+14+6], 12); */
	if (lcb->iiq_seq == cur_seq) {
		tr_lli_input(s0);
		lcb->iiq_seq = ((cur_seq+2)&0xfe);
	} else {
		for (iiq = 0, tmp = 0; tmp < NIIQ; tmp++) {
			if (lcb->iiq[tmp].frm == 0) {
				iiq = &lcb->iiq[tmp];
				break;
			}
		}
		if (!iiq) {
			printf("IIQ full\n");
			goto llcret_2;
		}
		lcb->iiq_noob++;
		iiq->seq = cur_seq;
		iiq->frm = s0;
		QDBGUP(gtrifx.iiq_oobs,1);
	}

	/* pending oobs */
	while (lcb->iiq_noob > 0) {
		for (iiq = 0, tmp = 0; tmp < NIIQ; tmp++) {
			if (lcb->iiq[tmp].seq == lcb->iiq_seq) {
				iiq = &lcb->iiq[tmp];
				break;
			}
		}
		if (iiq == 0) {
			goto llcret;
		}

		QDBGUP(gtrifx.iiq_rcvrd,1);
		tr_lli_input(iiq->frm);
		iiq->frm = 0;
		iiq->seq = 0xff;
		lcb->iiq_noob--;
		lcb->iiq_seq = ((lcb->iiq_seq+2)&0xfe);
	}
}

llcret:
	m_freem(m0);
	splx(s);
	splx(s2);
	return;

llcret_2:
	ifp->if_ierrors++;
llcret_1:
	if (s0)
		freemsg(s0);
	ifp->if_iqdrops++;
	m_freem(m0);
	splx(s);
	splx(s2);
}

/*
 * read frame off the board.
 */
#define DRAINFRM	0
#define IPFRM		1
#define LLCFRM		2
static int
gtr_recv(struct gtr_info *gtr, SLOT *sl)
{
	int i;
	TR_MAC *mh;
	LLC1 *llcp;
	TR_RII *rii;
	u_short *sri;
	int rpadlen = sizeof(TR_RII), rilen = 0;
	register char *cp, *dp;
	register int hlen, frm, ft;
	u_short	port = 0;
	int snoopflag = 0;
	struct ifqueue *ifq = 0;
	struct mbuf *m0, *mt, *m;
	int bds = (int)sl->sl_bds;
	int hpad = (int)sl->sl_sts;
	int length = (int)sl->sl_len;
	int len = 0;

#ifdef DEBUG_FW
	trdbginfo_ix = (trdbginfo_ix + 1) & 0xFF;
	tr_dbg_info[trdbginfo_ix].tlen = length;
	tr_dbg_info[trdbginfo_ix].nbufs = bds;
	tr_dbg_info[trdbginfo_ix].b1len = sl->busd[0].bd_len;
	tr_dbg_info[trdbginfo_ix].b2len = sl->busd[1].bd_len;
#endif
	/* Whether of not the packet is good, buffers MUST be recycled. */
	ASSERT(bds <= MAX_BDS);
	ASSERT(hpad == 0 || hpad == 2);
	for (m0 = 0, i = 0; i < bds; i++) {
		BUSD *bus = &sl->busd[i];
		switch (bus->bd_addr&BP_MASK) {
		case BP_SML: m = poolget(gtr,gtr->sml,bus->bd_addr);
			     ASSERT(bus->bd_len <= GTR_SML_SIZE);
			     QDBGUP(gtrqs.smlq_freed,1);
			     if (!m) {
                                DP(("gtr%u: recv: m = 0\n", gtr->gif_unit));                                
				return(0);
                             }
			     break;
		case BP_MED: m = poolget(gtr,gtr->med,bus->bd_addr);
			     ASSERT(bus->bd_len <= GTR_MED_SIZE);
			     QDBGUP(gtrqs.medq_freed,1);
#ifdef WBCACHE
/* This works around wild pointer somewhere in the system that is
 * polluting the cache. - paranoid
 */
			     if (!m) {
                                DP(("gtr%u: recv: m = 0\n", gtr->gif_unit));                                
				return(0);
                             }
			     dki_dcache_inval(mtod(m,caddr_t), GTR_MED_SIZE);
#endif
			     break;
		case BP_BIG: m = poolget(gtr,gtr->big,bus->bd_addr);
			     ASSERT(bus->bd_len <= GTR_BIG_SIZE);
			     QDBGUP(gtrqs.bigq_freed,1);
#ifdef WBCACHE
/* This works around wild pointer somewhere in the system that is
 * polluting the cache. - paranoid
 */
			     if (!m) {
                                DP(("gtr%u: recv: m = 0\n", gtr->gif_unit));                                
				return(0);
                             }
			     dki_dcache_inval(mtod(m,caddr_t), GTR_BIG_SIZE);
#endif
			     break;
		default:
			DP(("gtr%u: BP_MASK=%x\n",
				gtr->gif_unit, (bus->bd_addr&BP_MASK)));
			return(length);
		}

		ASSERT(m != 0);
		ASSERT(IS_KSEG0(m) && ((ulong)(m) % MSIZE) == 0);
		/*
		 * fw gets in weird mode when running against the IP19.
		 * Claims it has 2 buffer (bds == 2), but total
		 * length is that of 1st bd_len only, and second
		 * bd_addr in array is bogus, causing poolget() to
		 * return NULL and thereby crashing the system since
		 * the above ASSERT is turned off in shipped code. So,
		 * if m==0 but length is satisfied, ignore the rest of
		 * the buffers. -- IK 1/20/94.
		 */
		if (!m) {
			if (len == length) {
#ifdef DEBUG_FW
printf ("gtr0 fw bug: bds=%d, i=%d, sl_len=%d, bd_len=%d,%d,%d,%d\n",
	 bds, i, length, sl->busd[0].bd_len, sl->busd[1].bd_len, 
	 sl->busd[2].bd_len, sl->busd[3].bd_len);
				fwbug_hit++;
#endif
    				break;
			} else {
				/* drop what we have so far */
#ifdef DEBUG_FW
				even_worse++;
#endif
				gtr->gif.if_ierrors++;
				if (m0)
					m_freem(m0);
				return (len);
			}

		}
#ifdef WBCACHE
/* This works around wild pointer somewhere in the system that is
 * polluting the cache. - paranoid
 */
	        dki_dcache_inval((caddr_t)m, MLEN);
#endif
		len += bus->bd_len;
		if (m0 == 0) {
			m0 = m;
			mt = m;
			m->m_len = gtr->rcvpad + hpad + bus->bd_len;
		} else {
			mt->m_next = m;
			mt = m;
			m->m_len = bus->bd_len;
		}
		ASSERT(m0 != 0);
		ASSERT(mt != 0);
#ifndef WBCACHE
		/* XXX: be conservative on cache for now */
		vflush(mtod(m, char*), m->m_len);
#endif
	}
	if (!m0) {
		DP(("gtr%u: recv: m0 = 0\n", gtr->gif_unit));
		return(0);
	}

	/*
	 * Even though iface is not up buffer recycl was necessary.
	 * That is why ALIVE test is here after buf recycle.
	 */
	if ((gtr->gif.if_flags&(IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING))
		goto drop;

	/* Count packet/bytes regardless of error */
	gtr->gif.if_ipackets++;
	gtr->gif.if_ibytes += length;

	/* Checking small packet is done by board firmware */

	/* get hdr ptrs. SRI PAD must be applied uncoditionally */
	mh = (TR_MAC*)(mtod(m0, char*) + gtr->rcvpad);
	rii = (TR_RII*)(((char*)mh)+sizeof(*mh));
	if (mh->mac_sa[0]&TR_MACADR_RII) {
		sri = (u_short*)rii;
		rilen = (*sri & TR_RIF_LENGTH) >> 8;
		if (rilen <= 2) {
			sri = 0;
			rilen = 0;
		}
	} else {
		sri = 0;
		rilen = 0;
	}
	llcp = (LLC1 *)(((char*)rii)+sizeof(*rii));

	/* Quick parse for frame type */
	if ((mh->mac_fc & TR_FC_TYPE) == TR_FC_TYPE_LLC) {
		if (llcp->llc1_dsap == TR_IP_SAP) {
			hlen=sizeof(*mh)+sizeof(*rii)+sizeof(LLC1)+sizeof(SNAP);
			frm = IPFRM;
		} else {
			hlen = sizeof(*mh)+sizeof(*rii);
			frm = LLCFRM;
		}
	} else {
		hlen = sizeof(*mh)+sizeof(*rii);
		frm = DRAINFRM;
	}

	/* purge my own frame */
	if ((mh->mac_sa[5] != gtr->gif_enaddr[5]) ||
	    (mh->mac_sa[4] != gtr->gif_enaddr[4]) ||
	    (mh->mac_sa[3] != gtr->gif_enaddr[3]) ||
	    (mh->mac_sa[2] != gtr->gif_enaddr[2]) ||
	    (mh->mac_sa[1] != gtr->gif_enaddr[1]) ||
	    ((mh->mac_sa[0]&~TR_MACADR_RII) != gtr->gif_enaddr[0]))
		goto chksnoop;
	m_freem(m0); goto read_ret;
	/* NOT REACHED... */

chksnoop:
	if (snoopflag && !RAWIF_SNOOPING(&gtr->gtr_rawif)) {
		DP(("gtr%d: DROP ftype=%d, err=%x\n",
				gtr->gif_unit,frm,snoopflag));
		goto drop;
	}

	/* Move in a frame is done already */
	IF_INITHEADER(mtod(m0, char*), &gtr->gif, TRPAD+hlen);

	/* Do snoop iff applicable */
	if (RAWIF_SNOOPING(&gtr->gtr_rawif)) {
		snoop_input(&gtr->gtr_rawif,snoopflag,(char*)mh,m0,length-hlen);
		if (snoopflag&~SN_PROMISC)
			goto read_ret;
	}

	/* Final frame type adjustment */
	if (frm == IPFRM) {
		int roff = rpadlen-rilen;
		SNAP *snap=(SNAP*)(((char*)mh)+hlen-sizeof(SNAP));
		port = (snap->snap_etype[0]<<8)+(snap->snap_etype[1]);
#ifdef HOSTLLC
        {
                dlsap_family_t *dlp = dlsap_find(llcp->llc1_dsap,DL_LLC_ENCAP);
                if (dlp)
                        IF_INITHEADER(mtod(m0, char*), &gtr->gif, TRPAD+roff);
                if (dlp && (*dlp->dl_infunc)(dlp, &gtr->gif, m0, mh))
                        goto read_ret;
        }
#endif
	} else if (frm == LLCFRM) {
		int roff = rpadlen-rilen;
		if (roff > 0) {
			char tbuf[128];	/* paranoid */
			char *tbp = mtod(m0, char*)+TRPAD+roff;
			bcopy((char*)mh, tbuf, (sizeof(TR_MAC)+rilen));
			bcopy(tbuf, tbp, (sizeof(TR_MAC)+rilen));
			mh = (TR_MAC*)tbp;
		}
		IF_INITHEADER(mtod(m0, char*), &gtr->gif, TRPAD+roff);
#ifdef HOSTLLC
	{
		dlsap_family_t *dlp = dlsap_find(llcp->llc1_dsap,DL_LLC_ENCAP);
		if (dlp && (*dlp->dl_infunc)(dlp, &gtr->gif, m0, mh))
			goto read_ret;
		port = TR_PORT_LLC;
	}
#else
		gtr_read_llc2(m0);
		goto read_ret;
#endif
	} else {
		port = TR_PORT_MAC;
	}

	/* check mcast/bcast */
	if (TOKEN_ISGROUP(mh->mac_da)) {
		gtr->gif.if_imcasts++;
		if (TOKEN_ISBROAD(mh->mac_da))
			m0->m_flags |= M_BCAST;
		else
			m0->m_flags |= M_MCAST;
	}

	switch (port) {
	case ETHERTYPE_IP:
		schednetisr(NETISR_IP);
		ifq = &ipintrq;
       		if (IF_QFULL(ifq)) {
			IF_DROP(ifq);
			gtr->gif.if_iqdrops++;
			goto drop;
		}
       		IF_ENQUEUE(ifq, m0);
		m0 = 0;
		goto read_ret;
	case ETHERTYPE_ARP:
		if (sri) {
			*sri &= ~TR_RIF_SRB_NBR;
			*sri ^= TR_RIF_DIRECTION;
		}
		srarpinput(&gtr->gtr_ac, m0, sri);
		goto read_ret;
	default:
#ifdef _IRIX4
		if (snif_input(m0, mh))
			goto read_ret;
#else
		/* TBD */
#endif
		if (RAWIF_DRAINING(&gtr->gtr_rawif)) {
	        	drain_input(&gtr->gtr_rawif, port, (caddr_t)mh, m0);
			goto read_ret;
		}
 		gtr->gif.if_noproto++;
		goto drop;
        }
drop:
	if ((snoopflag&~SN_PROMISC) == 0) {
		if (RAWIF_SNOOPING(&gtr->gtr_rawif))
			snoop_drop(&gtr->gtr_rawif,snoopflag,(caddr_t)mh,length);
		if (RAWIF_DRAINING(&gtr->gtr_rawif))
			drain_drop(&gtr->gtr_rawif, port);
	}
	if (m0)
	        m_freem(m0);

read_ret:
	if ((snoopflag&~SN_PROMISC) != 0)
		gtr->gif.if_ierrors++;
	return(length);
}

/*
 * Token-Ring output.
 */
static int
if_gtroutput(struct ifnet *ifp, struct mbuf *m0, struct sockaddr *dst)
{
	char *map;
	XLOB xlob;	/* sandwitch to enforce alignment */
	TR_MAC *mh;
	LLC1 *llcp;
	SNAP *snap;
	int s, etyp;
	struct mbuf *m1;
	struct gtr_info *gtr;
	char maddr[TR_MAC_ADDR_SIZE];
	u_short ftype;
	extern int looutput(struct ifnet *, struct mbuf *, struct sockaddr *);

	int hlen = 0;
	int error = 0;
	int sri_len = 0;
	u_short *sri = 0;
	struct mbuf *mloop = 0;

	/* 1: check and get info struct */
	ASSERT((ifp->if_unit >= 0) && (ifp->if_unit < GTR_MAXBD));
	gtr = &gtr_info[ifp->if_unit];

	/* 2: make sure board has been initialized properly */
	if (gtr->bd_state != GTR_STATE_OK || iff_dead(ifp->if_flags)) {
		DP(("gtr%d: xmit: board not operational\n", gtr->gif_unit));
		m_freem(m0); error = ENETDOWN; goto output_ret;
	}

	switch (dst->sa_family) {
	case AF_INET: {
		int trail;	/* use trailer = always false in TR */
		struct in_addr *idst = &((struct sockaddr_in *)dst)->sin_addr;
		if (!arpresolve(&gtr->gtr_ac,m0,idst,(u_char*)&maddr[0],&trail))
			return(0);      /* arp will hang onto m0 */

		/* At this point, ARP resolved,
		 * which means, SRI is also learned
		 * whether it is over bridge or not.
		 */
		if (TOKEN_ISBROAD(maddr)) {
			if ((mloop = m_copy(m0, 0, M_COPYALL)) == 0) {
				m_freem(m0);
				error = ENOBUFS; goto output_ret;
			}
			map = (gtr_batype==1)?gtr_baddr_c:gtr_baddr_f;
		} else {
			map = maddr;
		}
		etyp = ETHERTYPE_IP;

		/* prepare SRCROUTE for resolved IP packets */
		if ((gtr->gif.if_flags & IFF_ALLCAST) == 0) {
			sri_len = 0;
			break;
		}
		if (TOKEN_ISGROUP(map)) {
			/* This is IP broadcat */
			sri = &gtr->nullsri;
			sri_len = 2;
		} else if ((sri = srilook(map, 1)) != 0) {
			/* This IP unicast for remote ring */
			sri_len = ((*sri)&TR_RIF_LENGTH)>>8;
		} else {
			/* This is IP unicast for local ring */
			sri_len = 0;
		}
	} break;

#ifdef HOSTLLC /* { */
	case AF_SDL: {
		struct sockaddr_sdl *sdl;
	
		sdl = (struct sockaddr_sdl *)dst;
		map = (char *)sdl->ssdl_addr;
		sri = (u_short *)&map[TR_MAC_ADDR_SIZE];
		etyp = (int)sdl->ssdl_ethersap;
		sri_len = sdl->ssdl_addr_len - TR_MAC_ADDR_SIZE;

		if (sdl->ssdl_addr_len < TR_MAC_ADDR_SIZE) {
			error = EAFNOSUPPORT;
			goto output_ret;
		}

		if (sri_len > 0) {
			/* If SRI is provided,
			 * then just honor it with sanity check.
			 */
			if ((gtr->gif.if_flags & IFF_ALLCAST) == 0) {
				error = EAFNOSUPPORT;
				goto output_ret;
			}
			if ((sri_len > TR_MAC_RIF_SIZE) ||
			    (sri_len != (((*sri)&TR_RIF_LENGTH)>>8))) {
				error = EINVAL;
				goto output_ret;
			}
		} else if (TOKEN_ISBROAD(map)) {
			/* If not, just provide null-sri for broadcast only.
			 */
			map = (gtr_batype == 1) ? gtr_baddr_c : gtr_baddr_f;
			sri = &(gtr->nullsri);
			sri_len = 2;
		}

		/* 4: Allocate MAC/SRI/LLC/SNAP header space */
		if ((m1 = m_get(M_DONTWAIT, MT_DATA)) == 0) {
			error = ENOBUFS; goto output_ret;
		}
		m1->m_next = m0;
		m0 = m1;
		m0->m_len = sizeof(*mh)+sri_len;

		/* 5: setup MAC header */
		mh = mtod(m0, TR_MAC*);
		mh->mac_ac = TR_AC_FRAME;
		mh->mac_fc = TR_FC_TYPE_LLC;
		bcopy(gtr->gif_enaddr, mh->mac_sa, TR_MAC_ADDR_SIZE);
		bcopy(map, mh->mac_da, TR_MAC_ADDR_SIZE);

		/* 6: insert SRCROUTE info into xmit header */
		if (sri_len > 0) {
			char *tsri = ((char*)mh)+sizeof(*mh);
			bcopy((char*)sri,tsri,sri_len);
			mh->mac_sa[0] |= TR_MACADR_RII;
		}

		xlob.sid = gtr->ip_sid;
		xlob.rsap = TR_IP_SAP;
		xlob.hlen_corel = sizeof(*mh)+sri_len;
		ftype = TR_FTYPE_DIRECT;

		s = splimp();
		error = gtrsend(gtr, m0, ftype, &xlob);
		splx(s);
		goto output_ret;
	}
#endif /* HOSTLLC } */

	case AF_UNSPEC:

# define EP ((struct ether_header *)&dst->sa_data[0])
		etyp = EP->ether_type; map = (char*)&EP->ether_dhost[0];
		if (TOKEN_ISBROAD(map)) {
			map = (gtr_batype==1)?gtr_baddr_c:gtr_baddr_f;
		}
# undef EP
		/* prepare SRCROUTE for ARP packets */
		if ((gtr->gif.if_flags & IFF_ALLCAST) == 0) {
			sri_len = 0;
			break;
		}
		if (TOKEN_ISGROUP(map)) {
			/* This is ARP request */
			sri = &gtr->nullsri;
			sri_len = 2;
		} else if ((sri = srilook(map, 1)) != 0) {
			/* This ARP response to remote ring */
			sri_len = ((*sri)&TR_RIF_LENGTH)>>8;
		} else {
			/* This is ARP response to local ring */
			sri_len = 0;
		}
		break;

	case AF_RAW:
		s = splimp();
		/* NO NEED to prepare SRCROUTE, it is RAW packet */
		xlob.sid = gtr->ip_sid;
		xlob.rsap = TR_IP_SAP;
		xlob.hlen_corel = 0;	/* don't care */
		error = gtrsend(gtr, m0, (u_short)TR_FTYPE_DIRECT, &xlob);
		splx(s);
		if (error)
			goto output_ret2;
		goto output_ret;

	default:
		DP(("gtr%d: can not handle address family 0x%x\n",
				gtr->gif_unit, dst->sa_family));
		m_freem(m0);
		error = EAFNOSUPPORT; goto output_ret;
	}

	/* 3: prepare for SRCROUTE is done already */

	/* 4: Allocate MAC/SRI/LLC/SNAP header space */
	/*    For simplicity just allocate new mbuf for header */
	if ((m1 = m_get(M_DONTWAIT, MT_DATA)) == 0) {
		m_freem(m0);
		error = ENOBUFS; goto output_ret;
	}
	m1->m_next = m0;
	m1->m_len = LLCOVHD+sri_len;
	m0 = m1;

	/* 5: setup MAC header */
	mh = mtod(m0, TR_MAC*);
	mh->mac_ac = TR_AC_FRAME;
	mh->mac_fc = TR_FC_TYPE_LLC;
	bcopy(map, mh->mac_da, TR_MAC_ADDR_SIZE);
	bcopy(gtr->gif_enaddr, mh->mac_sa, TR_MAC_ADDR_SIZE);

	/* 6: setup SRCROUTE info */
	if (sri_len > 0) {
		char *tsri = ((char*)mh)+sizeof(*mh);
		bcopy((char*)sri,tsri,sri_len);
		mh->mac_sa[0] |= TR_MACADR_RII;
	}

	/* 7: setup LLC header */
	if ((gtr->flags & GTR_ONCHIPLLC) && (gtr->ip_sid != 0)) {
		xlob.sid = gtr->ip_sid;
		xlob.rsap = TR_IP_SAP;
		xlob.hlen_corel = sizeof(*mh)+sri_len;
		m1->m_len -= sizeof(LLC1);
		snap = (SNAP*)(((char*)mh)+sizeof(*mh)+sri_len);
		ftype = TR_FTYPE_UI;
	} else {
		llcp = (LLC1*)(((char*)mh)+sizeof(*mh)+sri_len);
		llcp->llc1_dsap = TR_IP_SAP;
		llcp->llc1_ssap = TR_IP_SAP;
		llcp->llc1_cont = RFC1042_CONT;
		snap = (SNAP*)(((char*)llcp)+sizeof(*llcp));
		ftype = TR_FTYPE_DIRECT;
	}

	/* 8: setup SNAP */
	snap->snap_org[0] = RFC1042_K2;
	snap->snap_org[1] = RFC1042_K2;
	snap->snap_org[2] = RFC1042_K2;
	snap->snap_etype[0] = etyp>>8;
	snap->snap_etype[1] = etyp&0xff;

	/* 9: send out packet */
	if (mloop) {
		looutput(&loif, mloop, dst);
		mloop = 0;
	}
/* gtr_hexdump(mtod(m0, char*), 40); */
	s = splimp();
	error = gtrsend(gtr, m0, ftype, &xlob);
	splx(s);
	if (error)
		goto output_ret2;

output_ret:
	/* log error generated other than gtrsend */
	if (error)
		ifp->if_oerrors++;
output_ret2:
	if (mloop)
		m_freem(mloop);
	return(error);
}

/*
 * send a word to the board thru the GIO bus slave interface.
 */
static void
gtr_nibble(struct gtr_info *gtr, u_long w)
{
	int i;

	for (i = 0; i < 8; i++) {
		/* send a nibble and delay */
		*(gtr->gio) = (((w * GTR_GIO_FLAG0) & GTR_GIO_FLAG_MASK)
			| GTR_GIO_H2B_INTR);
		w >>= 4;
		DELAYBUS(DELAY_NIBBLE);
	}
}

/*
 * report PUD/BUD/INIT errors.
 */
static char *bud_rc[] = {
	"Initial test error",
	"Adapter software checksum error",
	"Adapter RAM error First 128 Kbyte",
	"Instruction test error",
	"Context/interrupt test error",
	"Protocol handler hardware error",
	"System interface register error",
	"BUD time out",		/* non-official */
	"Unkonwn error 0x8",
	"Unkonwn error 0x9",
	"Unkonwn error 0xa",
	"Unkonwn error 0xb",
	"Unkonwn error 0xc",
	"Unkonwn error 0xd",
	"Unkonwn error 0xe",
	"Unkonwn error 0xf"
};
static char *init_rc[] = {
	"Unkown error",
	"invalid initialization block",
	"invalid options",
	"invalid receive burst size",
	"invalid transmit burst count",
	"invalid DMA abort threshold",
	"invalid SCB",
	"invalid SSB",
	"DIO parity error",
	"DMA timeout",
	"DMA parity error",
	"DMA bus error",
	"DMA data error",
	"adapter check",
	"SCB init error",
	"SSB init error"
};

/* Specific Open Phase errors */
static char *OpenPhases[] = {
	" ",
	"Lobe media test",
	"Physical insertion",
	"Address verification",
	"Participation in ring poll",
	"Request Initialization",
	" ",
	" ",
	" ",
	" ",
	" ",
	" ",
	" ",
	" ",
	" ",
	" "
};

static char *OpenErrs[] = {
	" ",
	"Function failure",
	"Signal loss",
	" ",
	" ",
	"Timeout",
	"Ring failure",
	"Ring beaconing",
	"Duplicate node address",
	"Request initialization",
	"Remove received",
	" ",
	" ",
	" ",
	" ",
	" "
};

static int
errmsg(struct gtr_info *gtr)
{
	GTR_HC *hc = gtr->hc;
	u_long	rc;
	int	un = gtr->gif_unit;
	u_long	sts = hc->state;

	if ((sts&FSTATE_SICK) == 0)
		return(0);

	rc = sts&FSTATE_RC;
	if (sts&FSTATE_PUD) {
		printf("gtr%d: EPROM checksum error(%x)\n", un,sts);
	} else if (sts&FSTATE_BUD) {
		printf("gtr%d: BUD: %s(%x)\n",un,bud_rc[rc],sts);
	} else if (sts&FSTATE_INIT) {
		printf("gtr%d: INIT: %s(%x)\n",un,init_rc[rc],sts);
	} else if (sts&FSTATE_OPEN) {
		printf("gtr%d: OPEN: open failed(%x)",un,sts);
	} else if (sts&FSTATE_RUN) {
		printf("gtr%d: RUN: died(%x)\n",un,sts);
	} else if (sts&FSTATE_CHECK) {
		if (sts&TR_DIO_PARITY)
			printf("gtr%d: ADAPTER DIO parity error\n",un);
		if (sts&TR_DMA_READ_ABORT)
			printf("gtr%d: ADAPTER DMA read error\n",un);
		if (sts&TR_DMA_WRITE_ABORT)
			printf("gtr%d: ADAPTER DMA write error\n",un);
		if (sts&TR_ILLEGAL_OPCODE)
			printf("gtr%d: ADAPTER checksum error\n", un);
		if (sts&TR_PARITY_ERRORS)
			printf("gtr%d: ADAPTER internal bus error\n", un);
		if (sts&TR_RING_UNDERRUN)
			printf("gtr%d: ADAPTER DMA underrun\n",un);
		if (sts&TR_INVALID_INT)
			printf("gtr%d: ADAPTER invalid interrupt\n",un);
		if (sts&TR_INVALID_ERR_INT)
			printf("gtr%d: ADAPTER invalid error interrupt\n",un);
		if (sts&TR_INVALID_XOP)
			printf("gtr%d: ADAPTER invalid xop\n",un);
	} else if (sts&FSTATE_OERR) {
		if (sts & TR_OPEN_NODE_ADDR_ERR)
			printf("gtr%d: OPEN: invalid node address\n",un);
		if (sts & TR_OPEN_LIST_SIZE_ERR)
			printf("gtr%d: OPEN: invalid list size\n",un);
		if (sts & TR_OPEN_BUF_SIZE_ERR)
			printf("gtr%d: OPEN: invalid buffer size\n",un);
		if (sts & TR_OPEN_XMIT_BUFCNT_ERR)
			printf("gtr%d: OPEN: too few recv buffers\n", un);
		if (sts & TR_LLC_ERR)
			printf("gtr%d: OPEN: invalid LLC options\n",un);
		if (sts & TR_OPEN_ERR) {
			u_long phase = ((sts & TR_OPEN_PHASE) >> 4);
			printf("gtr%d: OPEN: %s phase: %s\n",
			    un,OpenPhases[phase], OpenErrs[rc]);
		}
	} else if (sts&FSTATE_MERR) {
		printf("gtr%d: invalid MOD options(%x)\n",un,sts);
	} else if (sts&FSTATE_CERR) {
		printf("gtr%d: CLOSE failed(%x)\n",un,sts);
	} else {
		printf("gtr%d: Unknown error(%x)\n", un,sts);
	}
#if DBG_LVL >= DBG_LVL_DRV
	gtr_hexdump((char*)&gtr->hc->sign, 64);
#endif
	return(-1);
}

/*
 * make it empty initially for hang.
 */
static void
ringinit(register struct ringhdr *rh, int siz)
{
	ASSERT(IS_KSEG1(rh));
	rh->len = siz;
	rh->ridx = 0;
	rh->ridxs = 0;
	rh->widx = 0;
}

/*
 * xmit ring access funcs.
 */
static int
xrfree(register RING *rp)
{
	int num = 0;

	ASSERT(IS_KSEG1(rp));
	while (rp->rh_ridx != rp->rh_widx) {
		register SLOT *p = &rp->sl[rp->rh_ridx];
		ASSERT(IS_KSEG1(p));
		if (p->sl_len)
			break;
		rp->rh_ridx = (rp->rh_ridx+1)&(rp->rh_len-1);
		num++;
	}
	return(num);
}

static SLOT *
xrnxt(register RING *rp)
{
	register int idx;
	register SLOT *p = &rp->sl[rp->rh_widx];

	ASSERT(IS_KSEG1(rp));
	ASSERT(IS_KSEG1(p));
	idx = ((rp->rh_widx+1)&(rp->rh_len-1));
	if (idx != rp->rh_ridx) {
		if (p->sl_len == 0) {
			rp->rh_widx = idx;
			return(p);
		}
	}
	DP(("xrnxt: XRP=%x, WIDX=%d, RIDX=%d, SL_LEN=%d\n",
		(u_long)rp,rp->rh_widx,rp->rh_ridx, p->sl_len));
	return(0);
}

/*
 * recv ring access funcs.
 */
static SLOT *
rrnxt(register RING *rp)
{
	register SLOT *p = &rp->sl[rp->rh_ridx];

	ASSERT(IS_KSEG1(rp));
	ASSERT(IS_KSEG1(p));
	if (p->sl_len) {
		rp->rh_ridx = (rp->rh_ridx+1)&(rp->rh_len-1);
		return(p);
	}
	return(0);
}

static void
rrfree(register SLOT *p)
{
	ASSERT(IS_KSEG1(p));
	p->sl_len = 0;
	p->sl_sts = 0;		/* complete status */
	p->sl_bds = 0;		/* total num of BDs */
}

/*
 * Buffer pool access funcs.
 */
static struct mbuf*
poolget(struct gtr_info *gtr, register BPOOL *bp, u_long tag)
{
	register int i, j;
	struct mbuf *mb;

	ASSERT(IS_KSEG1(bp));
	ASSERT(tag != 0);
	i = bp->rh_ridx;
	if (i == bp->rh_widx) {
		/* empty */
		return(0);
	}
	if (bp->bp_pbuf[i] == (tag&~BP_MASK)) {
		/* match */
		mb = (struct mbuf*)bp->bp_hbuf[i];
		bp->bp_pbuf[i] = 0;
		bp->bp_hbuf[i] = 0;
		bp->rh_ridx = (i+1)&(bp->rh_len-1);
		return(mb);
	}

	DP(("ROUND ROBIN broken for %s ridx=%d, hb=%x pb=%x\n",
		(bp == gtr->sml) ? "sml" :
		(bp == gtr->med) ? "med" :
		"big",
		i, bp->bp_hbuf[i], bp->bp_pbuf[i]));
	return(0);

#ifdef STASH
	/* ROUND ROBIN broken. try to fix and if not fixable
	 * reset the board and restart.
	 */
	for (i = 1; i < MAX_POOL; i++) {
		j = (bp->rh_ridx+i)&(bp->rh_len-1);
		if (bp->bp_pbuf[j] != (tag&~BP_MASK))
			continue;

		DP(("gtr%d: indone: %d expected=%d type=%d\n",
			gtr->gif_unit,j,bp->rh_ridx,(tag&BP_MASK)));
		while (j != bp->rh_ridx) {
			mb = (struct mbuf*)bp->bp_hbuf[bp->rh_ridx];
			bp->bp_pbuf[bp->rh_ridx] = 0;
			bp->bp_hbuf[bp->rh_ridx] = 0;
			bp->rh_ridx = (bp->rh_ridx+1)&(bp->rh_len-1);
			if (mb)
				m_freem(mb);
		}
		mb = (struct mbuf*)bp->bp_hbuf[j];
		bp->bp_pbuf[j] = 0;
		bp->bp_hbuf[j] = 0;
		bp->rh_ridx = j;
		return(mb);
	}
	DP(("gtr%d: indone mismatch not fixable\n", gtr->gif_unit));
	gtr->bd_state = GTR_STATE_SICK;
	return(0);
#endif
}

static int
poolput(BPOOL *bp, struct mbuf *mb)
{
	register int idx;

	ASSERT(IS_KSEG1(bp));
	idx = ((bp->rh_widx+1)&(bp->rh_len-1));
	if (idx == bp->rh_ridx) {
		/* full */
		return(-1);
	}
	ASSERT(bp->bp_pbuf[bp->rh_widx] == 0);
	ASSERT(bp->bp_hbuf[bp->rh_widx] == 0);
	bp->bp_pbuf[bp->rh_widx] = svirtophys(mtod(mb,u_long));
	bp->bp_hbuf[bp->rh_widx] = (u_long)mb;
	bp->rh_widx = idx;
	return(0);
}

static int
poolpeek(BPOOL *bp)
{
	register int idx;

	ASSERT(IS_KSEG1(bp));
	idx = ((bp->rh_widx+1)&(bp->rh_len-1));
	if (idx != bp->rh_ridx)
		return(0);	/* not full */
	return(-1);		/* full */
}

/*
 * fill up the board stocks of empty buffers
 * called in splimp already.
 */
static void
gtr_hang(struct gtr_info *gtr)
{
	caddr_t p;
	struct mbuf *m;
	register int i;
	u_long arg[BP_MASK];

	ASSERT(gtr->sml->rh_len == MAX_POOL);
	arg[BP_SML] = 0;
	for (i = 0; i < (GTR_MAX_SML/1); i++) {

		if ((m = m_get(M_DONTWAIT,MT_DATA)) == 0)
			break;
		/*
		 * min wvinval is 128 anyway.
		 * m and mtod is in same cache block anyway.
		 */
#ifdef WBCACHE
		dki_dcache_wb(m, MSIZE);
#endif
		if (poolput(gtr->sml, m)) {
			m_freem(m);
			break;
		}
		arg[BP_SML]++;
	}

	ASSERT(gtr->med->rh_len == MAX_POOL);
	arg[BP_MED] = 0;
	for (i = 0; i < (GTR_MAX_MED/1); i++) {
		if ((m = m_vget(M_DONTWAIT, GTR_MED_SIZE, MT_DATA)) == 0)
			break;
		p = mtod(m,caddr_t);
#ifdef WBCACHE
		dki_dcache_inval(p, GTR_MED_SIZE);
		dki_dcache_wb(m, MSIZE);	/* paranoid */
#endif
		if (poolput(gtr->med, m)) {
			m_freem(m);
			break;
		}
		arg[BP_MED]++;
	}

	ASSERT(gtr->big->rh_len == MAX_POOL);
	arg[BP_BIG] = 0;
	for (i = 0; i < (GTR_MAX_BIG/1); i++) {

		if ((m = m_vget(M_DONTWAIT, GTR_BIG_SIZE, MT_DATA)) == 0)
			break;
		p = mtod(m,caddr_t);
#ifdef WBCACHE
		dki_dcache_inval(p, GTR_BIG_SIZE);
		dki_dcache_wb(m, MSIZE);	/* paranoid */
#endif
		if (poolput(gtr->big, m)) {
			m_freem(m);
			break;
		}
		arg[BP_BIG]++;
	}

	QDBGUP(gtrqs.smlq, arg[BP_SML]);
	QDBGUP(gtrqs.medq, arg[BP_MED]);
	QDBGUP(gtrqs.bigq, arg[BP_BIG]);

	if ((gtr->firstsml) && (arg[BP_SML] == 0))
		gtr->firstsml = 0;
	if (gtr->firstsml == 0) {
		QDBGHW(arg[BP_SML],gtrqs.smlq_hiwat);
	}
	if ((gtr->firstmed) && (arg[BP_MED] == 0))
		gtr->firstmed = 0;
	if (gtr->firstmed == 0) {
		QDBGHW(arg[BP_MED],gtrqs.medq_hiwat);
	}
	if ((gtr->firstbig) && (arg[BP_BIG] == 0))
		gtr->firstbig = 0;
	if (gtr->firstbig == 0) {
		QDBGHW(arg[BP_BIG],gtrqs.bigq_hiwat);
	}

	if ((arg[BP_SML] > 0) || (arg[BP_MED] > 0) || (arg[BP_BIG] > 0)) {
		QDBGUP(gtrints.buf,1);
		i=gtr_xcmd(gtr,XCMD_BUF,(char*)arg,sizeof(arg), 0,0,0);
		if (i != 0) {
			/* should recover the mbufs, BACK OUT??? */
			QDBGMSG(("HANG failed for %d, %d, %d\n",
				arg[BP_SML], arg[BP_MED], arg[BP_BIG]));
		}
	}
}

/*
 * ReArm pool info.
 *
 * XXX - this doesn't work very well. To work properly, we should
 * update the receive ring indices as well as the bpool indices,
 * but that requires a simultaneous change in the firmware - think
 * about backward compatibility issues.
 */
static void
gtr_poolov(struct gtr_info *gtr)
{
	BPOOL *bp;
	register int i, n, nxt, limit, mask;
	struct hc_pool hcp;

	printf("fixing bpool:\n");
	i = gtr_xcmd(gtr,XCMD_GPOOL,0,0,(char*)&hcp,sizeof(hcp),1);
	if (i != 0) {
		printf("GETPOOL failed\n");
		gtr->bd_state = GTR_STATE_SICK;
		return;
	}

	bp = gtr->sml;
	ASSERT(bp->rh_len == MAX_POOL);
	n = 0; nxt = bp->rh_ridx; limit = bp->rh_widx; mask = bp->rh_len-1;
	for (i = nxt; i != limit;) {
		if (bp->bp_hbuf[i] == 0) {
			printf("sml pool broken at %d, hb=%x pb=%x\n",
				i, bp->bp_hbuf[i], bp->bp_pbuf[i]);
			gtr->bd_state = GTR_STATE_SICK;
			return;
		}
		if (bp->bp_pbuf[i] == 0) {
			printf("sml pool skipped at %d, hb=%x pb=%x\n",
				i, bp->bp_hbuf[i], bp->bp_pbuf[i]);
			nxt = i;
		} else {
			n++;
		}
		i = (i+1)&mask;
	}
	if (hcp.nsml != n) {
		printf("  nsml %d -> %d\n", hcp.nsml, n);
		hcp.nsml = n;
	}
	if ((nxt = (hcp.smls+(nxt*4))) != hcp.nxtsml) {
		printf("  nxtsml %x -> %x\n",hcp.nxtsml,nxt);
		hcp.nxtsml = nxt;
	}

	bp = gtr->med;
	ASSERT(bp->rh_len == MAX_POOL);
	n = 0; nxt = bp->rh_ridx; limit = bp->rh_widx; mask = bp->rh_len-1;
	for (i = nxt; i != limit;) {
		if (bp->bp_hbuf[i] == 0) {
			printf("med pool broken at %d, hb=%x pb=%x\n",
				i, bp->bp_hbuf[i], bp->bp_pbuf[i]);
			gtr->bd_state = GTR_STATE_SICK;
			return;
		}
		if (bp->bp_pbuf[i] == 0) {
			printf("med pool skipped at %d, hb=%x pb=%x\n",
				i, bp->bp_hbuf[i], bp->bp_pbuf[i]);
			nxt = i;
		} else {
			n++;
		}
		i = (i+1)&mask;
	}
	if (hcp.nmed != n) {
		printf("  nmed %d -> %d\n", hcp.nmed, n);
		hcp.nmed = n;
	}
	if ((nxt = (hcp.meds+(nxt*4))) != hcp.nxtmed) {
		printf("  nxtmed %x -> %x\n",hcp.nxtmed,nxt);
		hcp.nxtmed = nxt;
	}

	bp = gtr->big;
	ASSERT(bp->rh_len == MAX_POOL);
	n = 0; nxt = bp->rh_ridx; limit = bp->rh_widx; mask = bp->rh_len-1;
	for (i = nxt; i != limit;) {
		if (bp->bp_hbuf[i] == 0) {
			printf("big pool broken at %d, hb=%x pb=%x\n",
				i, bp->bp_hbuf[i], bp->bp_pbuf[i]);
			gtr->bd_state = GTR_STATE_SICK;
			return;
		}
		if (bp->bp_pbuf[i] == 0) {
			printf("big pool skipped at %d, hb=%x pb=%x\n",
				i, bp->bp_hbuf[i], bp->bp_pbuf[i]);
			nxt = i;
		} else {
			n++;
		}
		i = (i+1)&mask;
	}
	if (hcp.nbig != n) {
		printf("  nbig %d -> %d\n", hcp.nbig, n);
		hcp.nbig = n;
	}
	if ((nxt = (hcp.bigs+(nxt*4))) != hcp.nxtbig) {
		printf("  nxtbig %x -> %x\n",hcp.nxtbig,nxt);
		hcp.nxtbig = nxt;
	}


	i = gtr_xcmd(gtr,XCMD_SPOOL,(char*)&hcp,sizeof(hcp), 0,0,1);
	if (i != 0) {
		printf("SETPOOL failed\n");
		gtr->bd_state = GTR_STATE_SICK;
		return;
	}
}


/*
 * recover recv bufs.
 */
static int
gtr_recovermb(struct gtr_info *gtr, BPOOL *p)
{
	struct mbuf *mb;
	register int i, nf;

	for (i = 0, nf = 0; i < MAX_POOL; i++) {
		if (p->bp_pbuf[p->rh_ridx] == 0)
			break;
		if ((mb = poolget(gtr,p,p->bp_pbuf[p->rh_ridx])) == 0)
			break;
		m_freem(mb);
		nf++;
	}
	return(nf);
}

/* reset the board */
#define ZAPBD(g)			\
	*((g)->gio) = GTR_GIO_RESET;	\
	(g)->bd_state = GTR_STATE_RESET;\
	DELAYBUS(GTR_GIO_RESET_DELAY);	\
	*((g)->gio) = 0;

static void
gtr_initring(struct gtr_info *gtr, int buffers_only)
{
	if (gtr->flags&GTR_EDTINIT) {
		int i;
		BPOOL *p;
		struct mbuf *mb;

		/* recover xmit queue */
		for (;;) {
			struct mbuf *m;
			IF_DEQUEUE(&gtr->gif.if_snd, m);
			if (!m)
				break;
			m_freem(m);
		}

		/* recover recv bufs, just reuse them */
		i = gtr_recovermb(gtr, gtr->big);
		QDBGUP(gtrqs.bigq_freed,i);

		i = gtr_recovermb(gtr, gtr->med);
		QDBGUP(gtrqs.medq_freed,i);

		i = gtr_recovermb(gtr, gtr->sml);
		QDBGUP(gtrqs.smlq_freed,i);

		if (!buffers_only) {
			bzero(gtr->dma1, NBPC);
			bzero(gtr->dma2, NBPC);
		}
	} else {
		gtr->flags |= gtr_cksum;
		gtr->flags |= gtr_low_latency;
#if defined(IP20) || defined(IP22)
		setgioconfig(gtr->gif_unit ? GIO_SLOT_1 : GIO_SLOT_0,
			     GIO64_ARB_EXP0_RT | GIO64_ARB_EXP0_MST);
#endif
		gtr->flags |= GTR_EDTINIT;
	}
	/* Initialize rings. */
	ringinit(&(gtr->rr->rh), RH_MAX);
	ringinit(&(gtr->xr->rh), RH_MAX);
	ringinit(&(gtr->sml->rh), MAX_POOL);
	ringinit(&(gtr->med->rh), MAX_POOL);
	ringinit(&(gtr->big->rh), MAX_POOL);
	gtr->gif.if_snd.ifq_maxlen = RH_MAX;
}

/*
 * signal an address and op number to the board.
 *
 *	return 0 for success.
 */
static int
gtr_nibplay(struct gtr_info *gtr, int dlay, u_long op)
{
	u_long sts;
	u_long nibsum;
	GTR_HC *hc = gtr->hc;
	u_long hca = K1_TO_PHYS(hc);

#ifdef ONCHIPLLC
	if (gtr->flags&GTR_ONCHIPLLC) {
		op |= GTR_ONCHIPLLC;
	}
#endif
	nibsum = (GTR_GIO_DL_PREAMBLE^op^hca);
	hc->state = FSTATE_SANITY;
	do {
		gtr_nibble(gtr, GTR_GIO_DL_PREAMBLE);
		gtr_nibble(gtr, op);
		gtr_nibble(gtr, hca);
		gtr_nibble(gtr, nibsum);
		DELAYBUS(START_CHECK);
		if ((sts = hc->state) != FSTATE_SANITY)
			break;
	} while (0 <= (dlay -= (DELAY_NIBBLE*4*8+START_CHECK)));

	if ((sts == FSTATE_SANITY) || (sts&FSTATE_SICK)) {
		printf("gtr%d: serial handshake failed. FSTATE=0x%x\n",
				gtr->gif_unit, sts);
		return(errmsg(gtr));
	}
	if (hc->sign != GTR_SIGN) {
		printf("gtr%d: bad SIGNATURE=%x\n",gtr->gif_unit,(int)hc->sign);
		return(-1);
	}
	return(0);
}

/*
 * Check version number.
 */
static int
gtr_chkvers(struct gtr_info *gtr)
{
	u_long nsts,msts;
	inventory_t *pinv;
	GTR_HC *hc = gtr->hc;

	if (hc->sign != GTR_SIGN) {
		printf("gtr%d: BAD version(%x). expected=%x\n", gtr->gif_unit,
			hc->sign, GTR_SIGN);
		return(EIO);
	}

	DP(("gtr%d: correct SIGN\n",gtr->gif_unit));
	DP(("gtr%d: VERS=0x%x\n",gtr->gif_unit,(int)hc->vers));

	msts = gtr->vers&GTR_4MBPS_VERS ? 0 : 1;
	nsts = hc->vers&GTR_4MBPS_VERS ? 0 : 1;

	gtr->vers = hc->vers;
	if (nsts) {
		gtr->gif_mtu = min(max(1,gtr_mtu), TRMTU_16M);
	} else {
		gtr->gif_mtu = min(max(1,gtr_mtu), TRMTU_4M);
	}
	pinv = find_inventory(0, INV_NETWORK, INV_NET_TOKEN, INV_TOKEN_GTR,
						gtr->gif_unit, msts);
	if (pinv) {
#ifdef _IRIX4
		pinv->state = nsts;
#else
		pinv->inv_state = nsts;
#endif
	}
#ifdef ONCHIPLLC
	if ((gtr->vers&GTR_LLC_VERS) == 0) {
		gtr->flags &= ~GTR_ONCHIPLLC;
		if (showconfig)
			printf("executing CAF\n");
	} else {
		if (showconfig)
			printf("executing LLC223\n");
	}
#endif
	gtr_setnullsri(gtr);
	return(0);
}

/*
 * RESET:
 *
 * reset the board
 *	- Atomic operation.
 */
static int
gtr_reset(struct gtr_info *gtr, int alltheway)
{
	u_long nsts,msts;
	int sts = 0;
	GTR_HC *hc = gtr->hc;

	DP(("gtr%d: RESET!\n", gtr->gif_unit));
	ZAPBD(gtr);

	/* Initialize buf rings */
	gtr_initring(gtr,0);

	/* Init flags */
	gtr->first_xmit = 1;
	gtr->gif.if_timer = 0;			/* turn off watchdog */
	gtr->xid = 1;				/* start xid from 1 */
	gtr->cmd_Flags &= ~CMD_BUSY;		/* clear cmd busy flag */
	gtr->ip_sid = 0;
#ifdef ONCHIPLLC
	gtr->send_seq = 0;
#endif
	/* Initiate handshake. */
	hc->cmd = XCMD_NOP;			/* sync for over sampling */
	hc->cmd_ack = 0;
	hc->cmd_id = gtr->xid;
	if (gtr_nibplay(gtr,DELAY_RESET,(u_long)XCMD_NOP)) {
		return(EIO);
	}

	if (gtr_chkvers(gtr)) {
		DP(("gtr%d: version check failed\n", gtr->gif_unit));
		return(EIO);
	}

	if (alltheway) {
		gtr->reset_tid = cputimeout((void (*)())gtr_init,
					(caddr_t)gtr,TR_INCR_TIMEOUT);
	}
	return(0);
}

/*
 * Check iff board is ready to accept cmd.
 *	- this routine should be called within splimp().
 */
static int
gtr_ready(struct gtr_info *gtr)
{
	u_long sts;
	GTR_HC *hc = gtr->hc;
	int dlay = (3+11)*1000*1000;

	gtr->bd_state = GTR_STATE_INIT;
	do {
		DELAYBUS(1000);		/* check BUD every 1 msec */
		sts = hc->state;
		if ((sts&FSTATE_SICK) || (sts == FSTATE_INIT))
			break;
	} while (0 <= (dlay -= 1000));

	if (sts == FSTATE_INIT) {
		return(0);
	}
	if (dlay < 0) {
		printf("gtr%d: BUD/INIT time out", gtr->gif_unit);
		return(EIO);
	}
	return(errmsg(gtr));
}

/*
 * init: the board
 *
 *	- gtr_init is called by timer intr or within splimp already.
 */
static int
gtr_init(struct gtr_info *gtr)
{
	int s, s1;
	int sts;
	RING *r;
	int dlay;
	u_long op;
	GTR_HC *hc = gtr->hc;
	struct hc_init *hip = &hc->arg.hc_init;

	if (gtr->bd_state >= GTR_STATE_INIT && gtr->bd_state <= GTR_STATE_OK) {
		DP(("gtr%d: Multiple init going on\n",gtr->gif_unit));
		return(EIO);
	}

	/* Lower to splimp */
	s1 = spl1();

	untimeout(gtr->reset_tid);
	gtr->reset_tid = 0;
	gtr->gif.if_opackets = 0;
	gtr->gif.if_obytes = 0;
	gtr->gif.if_oerrors = 0;

	/* Check BUD/INIT - try upto 3+11 sec */
	if ((sts = gtr_ready(gtr)) != 0) {
		DP(("gtr%d: INIT failed\n",gtr->gif_unit));
		goto gir;
	}
	DP(("gtr%d: INIT done\n",gtr->gif_unit));

	if (!gtr->addrset) {
		sts = gtr_xcmd(gtr,XCMD_GADDR,0,0,
				gtr->gtr_bia,sizeof(gtr->gtr_bia),1);
		if (sts != 0) {
			printf("gtr%d: read BIA failed\n",gtr->gif_unit);
			goto gir;
		}
		DP(("gtr%d: GETADDR: %s\n",
				gtr->gif_unit,ether_sprintf(gtr->gtr_bia)));
		bcopy(gtr->gtr_bia,gtr->gif_enaddr,sizeof(gtr->gif_enaddr));
	}

	DP(("gtr%d: SETADDR: %s\n",
			gtr->gif_unit,ether_sprintf(gtr->gif_enaddr)));
	sts = gtr_xcmd(gtr,XCMD_SADDR, (char*)&gtr->gif_enaddr[0],
			sizeof(gtr->gif_enaddr),0,0,1);
	if (sts != 0) {
		printf("gtr%d: set address failed\n",gtr->gif_unit);
		goto gir;
	}

	if (showconfig) {
		printf("gtr%d: tokenring address %s\n",
			gtr->gif_unit, ether_sprintf(gtr->gif_enaddr));
	}

	/*
	 * Do Open, Receive, Transmit, and Mod.
	 */
	s = splimp();
	hip->flags = gtr->flags;
	hip->rcvpad = gtr->rcvpad;
	hip->ctptr = svirtophys(gtr->stat);		/* counters */

	r = gtr->xr;
	hip->h2b_ring = svirtophys(&r->sl[0]);		/* first x-slot */
	hip->h2b_ring_last=svirtophys(&r->sl[RH_MAX-1]);/* last x-slot */

	r = gtr->rr;
	hip->b2h_ring = svirtophys(&r->sl[0]);		/* first r-slot */
	hip->b2h_ring_last=svirtophys(&r->sl[RH_MAX-1]);/* last r-slot */

	hip->b2h_sml = svirtophys(&gtr->sml->bp_pbuf[0]);/* small rcv bufs */
	hip->b2h_med = svirtophys(&gtr->med->bp_pbuf[0]);/* medium rcv bufs */
	hip->b2h_big = svirtophys(&gtr->big->bp_pbuf[0]);/* large rcv bufs */

#ifdef ONCHIPLLC
	hip->ssb = svirtophys(gtr->ssb);
	hip->ssb_parm = svirtophys(&gtr->ssb_parm[0]);
#endif
	gtr->bd_state = GTR_STATE_OPEN;
	splx(s);
	sts = gtr_xcmd(gtr, XCMD_OPEN, 0,0,0,0,1);
	if (sts != 0) {
		(void)errmsg(gtr);
		goto gir;
	}

{
	u_long hcs;
	u_long os = -1;
#ifdef STASH
	int delay = DELAY_OPEN;
#else
	int delay = DELAY_RESET*5;
#endif

	s = splimp();
	while (delay > 0) {
		splx(s);
		DELAYBUS(DELAY_OP);
		s = splimp();
		delay -= DELAY_OP;
		hcs = hc->state;

		if ((hcs&FSTATE_RUN) == FSTATE_RUN) {
			DP(("gtr%d: RING INSERT SUCCEEDED\n", gtr->gif_unit));
			break;
		}

		if (hcs != os) {
			DP(("gtr%d: state=0x%x\n",gtr->gif_unit, hcs));
			os = hcs;
		}

		if (hcs & FSTATE_SICK) {
			(void)errmsg(gtr);
			splx(s); goto gir;
		}
	}
	if ((hcs&FSTATE_RUN) != FSTATE_RUN) {
		printf("gtr%d: RING INSERT failed\n", gtr->gif_unit);
		gtr->bd_state = GTR_STATE_SICK;
		splx(s); goto gir;
	}
}

#ifdef ONCHIPLLC
	if (gtr->flags&GTR_ONCHIPLLC) {
		/* should open IP SAP */
		(void)gtr_open_ip_sap(gtr);
	}
#endif
	/* let's do hang */
	gtr->firstsml = 1;
	gtr->firstmed = 1;
	gtr->firstbig = 1;
	gtr_hang(gtr);
	gtr->bd_state = GTR_STATE_OK;
	splx(s);

gir:	splx(s1);
	wakeup((caddr_t)&gtr->bd_state);
	return(0);
}

/*
 * gtr_xcmd:
 *	Issue a command to the GIOTR board
 *		generic routine used all over the driver to issue
 *		a variety of commands to the board.
 */
static int
gtr_xcmd(struct gtr_info *gtr, u_long cmd,
	caddr_t src, u_long slen, caddr_t res, u_long rlen, u_long wcom)
{
	int sts;
	u_long tdly;
	int lcmd_id, s1;
	int s = splimp();
	GTR_HC *hc = gtr->hc;
	int delay = DELAY_RESET;

	QDBGUP(gtrqs.cmdq,1);
	QDBGUP(gtr->cmdhiwat,1);
	QDBGHW(gtr->cmdhiwat,gtrqs.cmdq_hiwat);

	tdly = delay;
	lcmd_id = gtr->xid;
	while (!CMD_OK(gtr,lcmd_id) && tdly > 0) { /* wait for completion */
		SYIELD;
		DELAYBUS(NORM_CHECK);
		EYIELD;
		delay -= NORM_CHECK;
#if defined(ONCHIPLLC) && !defined(IP12)
		(void)dcmd_hndl(gtr);
#endif
		lcmd_id = gtr->xid;
		if (gtr->hc->cmd_ack > lcmd_id) {
			printf("bd missed intr ack(%x) != lxid(%x)\n",
				gtr->hc->cmd_ack, lcmd_id);
			break;
		}
	}

	/* set parameters */
	if (slen > 0)
		bcopy((char*)src, hc->arg.cmd_data, slen);

	CMDBD(gtr,cmd);

	if ((!rlen) && (!wcom)) {
		sts = 0;
		goto xcmd_ret;
	}

	lcmd_id = gtr->xid;
	while (delay > 0) { /* wait for completion */
		if (CMD_OK(gtr,lcmd_id)) {
			sts = hc->cmd_sts;
			if (sts) {
				IWMSG(("send: cmd sts=%x\n", sts));
				sts = EIO;
				goto xcmd_ret;
			}
			if (rlen > 0) {
				bcopy((char*)&hc->res.cmd_res[0],
						(char*)res,rlen);
			}
			goto xcmd_ret;
		}
		SYIELD;
		DELAYBUS(DELAY_NIBBLE);
		EYIELD;
		delay -= DELAY_NIBBLE;
#if defined(ONCHIPLLC) && !defined(IP12)
		(void)dcmd_hndl(gtr);
#endif
	}

	sts = EIO;
	IWMSG(("gtr%d: cmd %x failed(%x)\n", gtr->gif_unit, cmd, sts));

xcmd_ret:
	QDBGUP(gtrqs.cmdq_freed,1);
	QDBGDOWN(gtr->cmdhiwat,1);
	splx(s);
	return(sts);
}

#ifdef ONCHIPLLC
/*
 * gtr_xllc:
 *
 *	- convert s0 to m0 and transmit.
 */
static int
gtr_xllc(struct gtr_info *gtr, register TR_XMIT_PARMS *xp, u_short corel)
{
	TR_MAC *mh;
	XLOB xlob;	/* sandwitch for enforce alignment */
	TR_RII *rii;
	mblk_t *s0, *s1;
	char *mcp, *scp;
	struct mbuf *m0, *m1, *mt;
	int hlen, slen, mlen, clen, tlen;
	int alen = 0;
	int error = 0;

	ASSERT(xp != 0);
	s0 = (mblk_t*)xp->xmit_data;
	ASSERT(s0 != 0);

	if (xp->frame_type != TR_FTYPE_I_FRAME) {
		/* Non-IP/IFRAME has full 18byte RII pad in it */
		mh = (TR_MAC*)s0->b_rptr;
		hlen = sizeof(*mh);
		if (mh->mac_sa[0]&TR_MACADR_RII) {
			u_long off, rlen;

			rii = (TR_RII*)(mh+1);
			rlen = (rii->rii&TR_RIF_LENGTH) >> 8;
			off = sizeof(*rii)-rlen;
			hlen += rlen;
			if (off > 0) {
				bcopy((char*)mh,(char*)mh+off,hlen);
				s0->b_rptr += off;
				xp->frame_size -= off;
			}
			if (rlen == 0) {
				mh = (TR_MAC*)s0->b_rptr;
				mh->mac_sa[0] &= ~TR_MACADR_RII;
			}
		}
	}

	/* copy s0 to m0 */
	tlen = xp->frame_size;
	if (tlen <= 0) {
		IWMSG(("xllc: 0 byte xp failed\n"));
		error = EINVAL;
		goto xllc_ret;
	}
	for (s1 = s0, alen = 0; s1 != 0; s1 = s1->b_cont) {
		alen += s1->b_wptr-s1->b_rptr;
	}
	if (tlen != alen) {
		IWMSG(("xllc: len %d fixed to %d(ft=%d)\n",
				tlen, alen, xp->frame_type));
		tlen = alen;
	}

	for (s1 = s0, m0 = 0, mlen = 0; s1 != 0; s1 = s1->b_cont) {
		scp = (char *)s1->b_rptr;
		if ((slen = s1->b_wptr-s1->b_rptr) == 0)
			continue;
sss:
		if (mlen <= 0) {
			mlen = min(tlen, VCL_MAX);
			if (mlen <= 0) {
				/* let go */
				break;
			}
			mt = m_vget(M_DONTWAIT, mlen, MT_DATA);
			if (mt == 0) {
				if (m0)
					m_freem(m0);
				error = ENOBUFS;
				IWMSG(("xllc: %d byte mbuf failed\n",mlen));
				goto xllc_ret;
			}
			tlen -= mlen;
			if (!m0)
				m0 = mt;
			else
				m1->m_next = mt;
			m1 = mt;
			m1->m_len = mlen;
			mcp = mtod(m1, char*);
		}

		clen = min(mlen, slen);
		bcopy(scp, mcp, clen);
		mcp += clen;
		scp += clen;
		mlen -= clen;
		slen -= clen;
		if (slen > 0) {
			goto sss;
		}
	}

	/* send out */
	xlob.sid = xp->station_id;
	xlob.rsap = xp->remote_sap;
	xlob.hlen_corel = (xp->frame_type==TR_FTYPE_I_FRAME)?corel:hlen;
	error = gtrsend(gtr, m0, xp->frame_type, &xlob);
	goto xllc_ret2;	/* skip double oerror bumpups */

xllc_ret:
	if (error)
		gtr->gif.if_oerrors++;
xllc_ret2:
	/* release s0 - Iframe mblk will be released when ACK recvd */
	return(error);
}
#endif

/*
 * gtrsend:
 *
 *	- accepts only exact amount of RII
 *	- should CONSUME mbuf
 *	- should be called in splimp.
 *	- should count oerror.
 */
static int
gtrsend(struct gtr_info *gtr, struct mbuf *m0, u_short ftype, XLOB *xlob)
{
	char *cp;
	SLOT *sl;
	TR_MAC *mh;
	struct mbuf *n0;
	int i, bds, len, mlen, oddlist;
	int error = 0;

	ASSERT(m0 != 0);
	if (IF_QFULL(&gtr->gif.if_snd)) {
		gtr->gif.if_odrops++;
		error = ENOBUFS;
		IWMSG(("gtrsend%d: QUEUE full\n", gtr->gif_unit));
		goto snderr;
	}
	
	/*
	 * Count the number of buffers to go out in the packet and
	 * total length of the packet. Even if the packet has
	 * odd buf addr, counting total lenght should go on for
	 * use in rebuild phase.
	 */
massage:
	for (n0=m0, len=0, bds=0, oddlist=0; n0 != 0; n0=n0->m_next) {
		if ((mlen = n0->m_len) == 0)
			continue;
		len += mlen;
		if (oddlist)
			continue;	/* If odd, then just get tot len */
		cp = mtod(n0, caddr_t);
#ifdef STASH
		if (((u_long)cp & 0x3) ||
		   (n0->m_next && ((((u_long)cp+mlen)&0x3) != 0))) {
#endif
		if ((u_long)cp & 0x1) {
			oddlist = 1;
			continue;
		}

		if (btoct(cp) != btoct(cp+mlen-1)) {
			bds++;
		}
		bds++;
	}

	if (len > (gtr->gif_mtu+LLCOVHD+SRCROVHD)) {
		IWMSG(("gtrsend%d: too large(%d) packet flushed\n",
				gtr->gif_unit, (int)len));
		error = EINVAL;
		goto snderr;
	}

	if (oddlist != 0 || bds >= MAX_BDS) {
		/* rebuild data chain and retest
		 * XXX: inefficient but this case is RARE, iff it happens
		 *      data length is normally short.
		 * ">=" above guarantees one extra busd free for xlob for
		 * ONCHIPLLC case.
		 */
		int maxlen;
		int off = 0;
		struct mbuf *tmh = 0;
		struct mbuf *tmt = 0;
		
		while (len > 0) {
			maxlen = min(len, VCL_MAX);
			len -= maxlen;
			if ((n0 = m_vget(M_DONTWAIT,maxlen,MT_DATA)) == 0) {
				DP(("gtrsend%d: mget failed\n", gtr->gif_unit));
				if (tmh != 0)
					m_freem(tmh);
				error = ENOBUFS;
				IWMSG(("gtrsend: m_get failed\n"));
				goto snderr;
			}
			if (tmh == 0)
				tmh = n0;
			else
				tmt->m_next = n0;
			tmt = n0;
			off += m_datacopy(m0,off,maxlen,mtod(tmt,caddr_t));
		}
		m_freem(m0);
		m0 = tmh;
		goto massage;
	}

	/* Now m0 and mh is final */
	if ((m0 == 0) || (len == 0)) {
		printf("m0=%x, len=%x\n", m0, len);
		error = EIO;
		goto snderr;
	}
	mh = mtod(m0, TR_MAC*);

	/*
	 * self-snoop must be done prior to dma bus desc. build because
	 * bus-desc. may change mbuf len and off.
	 */
	if (RAWIF_SNOOPING(&gtr->gtr_rawif)
	    && snoop_match(&gtr->gtr_rawif,(caddr_t)mh,m0->m_len)) {
		    ether_selfsnoop((struct etherif*)gtr,
				(struct ether_header *)mh,
				    m0, sizeof(*mh), min(len,VCL_MAX));
	}

	if ((sl = xrnxt(gtr->xr)) == 0) {
		IWMSG(("gtrsend%d: No slot\n", gtr->gif_unit));
		error = ENOBUFS;
		goto snderr;
	}

	i = 0; n0 = m0;
	do {
		int singlepage;

		singlepage = 1;
		mlen = n0->m_len;
		if (mlen) {
			cp = mtod(n0, caddr_t);
			if (btoct(cp) != btoct(cp + mlen - 1)) {
				singlepage = 0;
				mlen = NBPP - poff(cp);
				n0->m_len -= mlen;
				n0->m_off += mlen;
			}

			if (IS_KSEG2(cp))
				sl->busd[i].bd_addr= kvtophys(cp);
			else
				sl->busd[i].bd_addr= svirtophys(cp);
			sl->busd[i].bd_len = mlen;
			i++;
#ifdef WBCACHE
			/* write-back data for the DMA to find it */
			dki_dcache_wb(cp, mlen);
#endif
		}
		if (singlepage)
			n0 = n0->m_next;
	} while (n0);

	ASSERT(IS_KSEG1(sl));
#ifdef ONCHIPLLC
	bcopy((char*)xlob, (char*)&sl->busd[i], sizeof(*xlob));
	sl->sl_sts = (u_char)ftype;
	sl->busd[i].bd_addr = gtr->send_seq;
	gtr->send_seq += 1;
#endif
	sl->sl_bds = bds;
	sl->sl_len = len;

	gtr->gif.if_opackets++;
	gtr->gif.if_obytes += len;

	/* set end of frame bits in last frame status field */
#ifdef STASH
	/* scheduling xmit intr does not really help for slow tokenring */
	if ((lbolt-gtr->xmit_hz) > 1)
		gtr->first_xmit |= 1;
	gtr->xmit_hz = lbolt;
	gtr->first_xmit++;
	if ((gtr->first_xmit%2) == 0) {
#endif
		error = gtr_xcmd(gtr, XCMD_XMIT,0,0,0,0,0);
		if (error) {
			/* Don't have to recover xl and sene_seq.
			 * Current xmit will be simple delayed a bit.
			 */
			error = EIO;
			goto snderr;
		}
#ifdef STASH
	}
#endif
	QDBGUP(gtrqs.sndq,1);
	QDBGUP(gtr->sndhiwat,1);
	QDBGHW(gtr->sndhiwat,gtrqs.sndq_hiwat);
	IF_ENQUEUE(&gtr->gif.if_snd, m0);
	return(0);

snderr:
	if (m0)
		m_freem(m0);
	gtr->gif.if_oerrors++;
	return(error);
}

static int
gtr_close(struct gtr_info *gtr, int restart)
{
	int s = splimp();

	DP(("gtr%d: close\n", gtr->gif_unit));
	gtr->ip_sid = 0;
	gtr->first_xmit = 1;
	if ((gtr->bd_state == GTR_STATE_OK) ||
	    (gtr->bd_state == GTR_STATE_SICK)) {
		gtr->bd_state = GTR_STATE_CLOSE;
		gtr->gif.if_flags &= ~(IFF_UP|IFF_RUNNING);
		if (gtr_reset(gtr, restart)) {
			splx(s);
			DP(("gtr%d: close FAILED\n", gtr->gif_unit));
			return(EIO);
		}
	}
	splx(s);
	DP(("gtr%d: close SUCCEEDED\n", gtr->gif_unit));
	return(0);
}

static int
gtr_insert(struct gtr_info *gtr)
{
	int s, error = 0;

	if (gtr->bd_state == GTR_STATE_OK)
		return 0;

	s = splimp();

	switch (gtr->bd_state) {
	case GTR_STATE_CLOSE:
	case GTR_STATE_SICK:
		(void)gtr_reset(gtr,0);
		/* FALL THRU... */
	case GTR_STATE_RESET:
	case GTR_STATE_INIT:
		gtr_init(gtr);
		/* FALL THRU... */
	case GTR_STATE_OPEN:
	case GTR_STATE_OPENSAP:
		break;
	default:
		DP(("gtr BD STATE = %x\n", gtr->bd_state));
		error = EIO;
		goto isret;
	}

isret:	splx(s);
	return(error);
}

static void
if_gtrwdog(int unit)
{
	struct gtr_info *gtr = &gtr_info[unit];
	int s = splimp();

	if (gtr->bd_state == GTR_STATE_SICK) {
		printf("gtr%d: died and restarted(%x)\n", unit);
		(void)gtr_reset(gtr,1);
	}

	gtr->gif.if_timer = DOG;
	splx(s);
	return;
}

#if DBG_LVL > 0
	char *gtr_spool;
	char *gtr_mpool;
	char *gtr_bpool;
	char *gtr_rring;
	char *gtr_xring;
#endif

static struct trlliops gtrlliops =
	{ gtr_valid_unit, gtr_get_sa, (int (*)())gtr_lli_output };

/*
 * Probe for a single board and attach if present.
 */
void
if_gtredtinit(struct edt *edtp)
{
	int s;
	struct ifnet *ifp;
	struct gtr_info *gtr;
	static int unit = 0;

	/* check unit number */
	if (GTR_MAXBD <= unit) {
		printf("gtr%d: bad EDT entry\n", unit);
		return;
	}
	s = splimp();
	gtr = &gtr_info[unit];
	if (0 != gtr->gio) {
		printf("gtr%d: duplicate EDT entry\n", unit);
		goto gtredt_ret;
	}

	/* Probe */
	if (badaddr((volatile unsigned long*)edtp->e_base, 4) ||
	   ((*(u_long*)edtp->e_base & GIO_SID_MASK) != GIO_ID_GTR)) {
		printf("gtr%d: missing\n", unit);
		goto gtredt_ret;
	}
	if (showconfig)
		printf("gtr%d: present\n", unit);

	/* setup PRIVATE data. */
	/* Allocate a multicast filter table with an initial size of 10. */
	if ((!mfnew(&gtr->mfilter, 10)) || (gtr->mfilter.mf_base == 0)) {
		printf("gtr%d: no memory for frame filter\n", unit);
		goto gtredt_ret;
	}

	/* get buffers for non-page crossing DMA structures
	 * These pages are referenced only via non-cached address space
	 * a.k.a K1 segment in order to avoid cache coherency problem.
	 */
	gtr->dma1 = (u_char*)kvpalloc(1,VM_DIRECT|VM_STALE|VM_UNCACHED,0);
	gtr->dma2 = (u_char*)kvpalloc(1,VM_DIRECT|VM_STALE|VM_UNCACHED,0);
	if ((gtr->dma1 == 0) || (gtr->dma2 == 0)) {
		printf("gtr%d: no memory for dma\n", unit);
		goto gtredt_ret;
	}

	bzero(gtr->dma1, NBPC);
	bzero(gtr->dma2, NBPC);
#ifdef WBCACHE
	dki_dcache_wbinval(gtr->dma1, NBPC);
	dki_dcache_wbinval(gtr->dma2, NBPC);
#endif
	gtr->dma1 = (u_char*)K0_TO_K1(gtr->dma1);
	gtr->dma2 = (u_char*)K0_TO_K1(gtr->dma2);
	bzero(gtr->dma1, NBPC);
	bzero(gtr->dma2, NBPC);

	gtr->hc = (GTR_HC*)&(gtr->dma1[0*512]);		/* cmd area - 128 */
	gtr->stat = (GTR_CNTS*)&(gtr->dma1[1*512]);	/* stat counts - 240 */
#ifdef ONCHIPLLC
	gtr->ssb = (SSB*)&(gtr->dma1[2*512]);		/* ssb - 8+4 */
	gtr->ssb_parm = &(gtr->dma1[2*512+8+4]);	/* ssb - 64 */
#endif
	gtr->xr = (RING*)&(gtr->dma1[3*512]);		/* xmit ring - 2192 */

	gtr->sml = (BPOOL*)&(gtr->dma2[0*sizeof(BPOOL)]);/* small bufs - 528 */
	gtr->med = (BPOOL*)&(gtr->dma2[1*sizeof(BPOOL)]);/* medium bufs - 528 */
	gtr->big = (BPOOL*)&(gtr->dma2[2*sizeof(BPOOL)]);/* large bufs - 528 */
	gtr->rr = (RING*)&(gtr->dma2[3*sizeof(BPOOL)]);	 /* recv ring - 2192 */

	gtr->gio = (volatile unsigned long*)edtp->e_base;

	ASSERT(IS_KSEG1(gtr->hc));
	ASSERT(IS_KSEG1(gtr->stat));
	ASSERT(IS_KSEG1(gtr->xr));
	ASSERT(IS_KSEG1(gtr->sml));
	ASSERT(IS_KSEG1(gtr->med));
	ASSERT(IS_KSEG1(gtr->big));
	ASSERT(IS_KSEG1(gtr->rr));
	ASSERT(IS_KSEG1(gtr->gio));
#if DBG_LVL > 0
	gtr_spool = (char*)gtr->sml;
	gtr_mpool = (char*)gtr->med;
	gtr_bpool = (char*)gtr->big;
	gtr_rring = (char*)gtr->rr;
	gtr_xring = (char*)gtr->xr;
#endif

	gtr->rcvpad = TRPAD;
#ifdef ONCHIPLLC
	gtr->flags = GTR_ONCHIPLLC;
#else
	gtr->flags = 0;
#endif
	gtr->addrset = 0;
	gtr->naddrset = 0;
	gtr->xmit_hz = 0;
	bzero(&gtr->tr_fcn_addr,sizeof(gtr->tr_fcn_addr));
	bzero(&gtr->tr_grp_addr,sizeof(gtr->tr_grp_addr));
	setgiovector(GIO_INTERRUPT_1, unit+GIO_SLOT_0, (void (*)())if_gtrintr, (int)gtr);

	/* set initial default invent */
	add_to_inventory(INV_NETWORK, INV_NET_TOKEN, INV_TOKEN_GTR, unit, 1);

	/* setup Broadcast address */
	if (gtr_batype != 1 && gtr_batype != 2) {
		if (showconfig)
		    printf("gtr%d: broadcast address set to 0xC000FFFFFFFF\n",
				ifp->if_unit);
		gtr_batype = 1;
	}

	/* prepare for ATTACH */
	ifp = &gtr->gif;
	ifp->if_unit = unit;
	ifp->if_name = "gtr";
	ifp->if_type = IFT_ISO88025;
	ifp->if_mtu = min(max(1,gtr_mtu), TRMTU_16M);
	ifp->if_flags = IFF_BROADCAST|IFF_MULTICAST|IFF_NOTRAILERS|IFF_ALLCAST;
	ifp->if_output = if_gtroutput;
	ifp->if_ioctl = (int (*)(struct ifnet*, int, void*))if_gtrioctl;
	ifp->if_watchdog = if_gtrwdog;
	gtr_setnullsri(gtr);

	/* start the board ASAP */ 
	if (gtr_reset(gtr, 1) != 0) {
		printf("gtr%d: reset failed\n", unit);
		goto gtredt_ret;
	}

	/* ATTACH interface */
	if_attach(ifp);
	rawif_attach(&gtr->gtr_rawif, ifp,
		     (caddr_t)&gtr->gif_enaddr[0],
		     (caddr_t)&etherbroadcastaddr[0],
		     sizeof(gtr->gif_enaddr), sizeof(TR_MAC)+sizeof(TR_RII),
		     structoff(tr_mac, mac_sa[0]),
		     structoff(tr_mac, mac_da[0]));
	(void)tr_lli_register(&gtrlliops);

	unit++;

gtredt_ret:;
	splx(s);
}

/*
 * LLC BANDAGE:
 */
static int
gtr_valid_unit(int unit)
{
	if (unit >= 0 && unit < GTR_MAXBD) {
		return(1);
	}
	return(0);
}

static char *
gtr_get_sa(int unit)
{
	struct gtr_info *gtr;

	if (unit >= 0 && unit < GTR_MAXBD) {
		gtr = &gtr_info[unit];
		return((char*)&gtr->gif_enaddr[0]);
	}
	return((char*)0);
}

#ifdef ONCHIPLLC
/************************************************************************/
/*									*/
/*	ONCHIPLLC BANDAGE:						*/
/*									*/
/************************************************************************/

#define BDINT(b,h) { u_short *sp = (u_short*)&b; \
	sp[0] = htons((u_long)h>>16); sp[1] = htons((u_long)h&0xffff); }
#define HTINT(h,b) { u_short *sp = (u_short*)&b; u_long *tp = (u_long*)&h;\
	*tp=(u_long)(ntohs(sp[0])<<16)+(ulong)ntohs(sp[1]); }

#ifdef IFRMDBG
IFRAMES gtrifr0[0xff];
mblk_t	gtrmb0[0xff];
struct	datab gtrdp0[0xff];
char	gtrmd0[0xff][128];
u_long	gtrnxtcorel = 0xfe;
u_long	gtr_scorel = 1;

static void
stash_ifrm(IFRAMES *ifrm, u_long corel)
{
	char *cp;
	mblk_t *mb;

	if (ifrm == 0 || corel >= 255) {
		printf("Iframe Internel error, ptr=%x, corel=%x\n",ifrm,corel);
		return;
	}

	if (gtr_scorel && (corel != gtrnxtcorel)) {
		printf("Iframe corel out of sync. exp=%x, actual=%x\n",
			gtrnxtcorel, corel);
		gtr_scorel = 0;
	} else if (!gtr_scorel && (corel == gtrnxtcorel)) {
		printf("Iframe corel back int sync at %x\n", corel);
		gtr_scorel = 1;
	}

	cp = (char*)&gtrifr0[corel];
	bcopy((char*)ifrm,cp,sizeof(*ifrm));

	if ((mb=ifrm->xmit_parms.xmit_data)==0) {
		printf("No Iframe mblk for corel=%d\n",(int)corel);
		return;
	}
	cp = (char*)&gtrmb0[corel];
	bcopy((char*)mb, cp, sizeof(*mb));

	if (mb->b_datap == 0) {
		printf("No Iframe datap for corel=%d\n",(int)corel);
		return;
	}
	cp = (char*)&gtrdp0[corel];
	bcopy((char*)mb->b_datap, cp, sizeof(struct datab));

	if (mb->b_rptr == 0) {
		printf("No rptr for corel=%d\n",(int)corel);
		return;
	}

	cp = &gtrmd0[corel][0];
	bcopy(mb->b_rptr,cp, 32);
	if (cp[0]==0 && cp[1]==0 && cp[2]==0 && cp[3]==0) {
		printf("Already BAD Iframe for corel=%d\n", (int)corel);
		gtr_hexdump((char*)mb, sizeof(*mb));
		gtr_hexdump((char*)mb->b_datap, 16);
		printf("bad data:\n");
		gtr_hexdump(mb->b_rptr, 32);
	}
}

static void
chkrdy_ifrm(IFRAMES *ifrm)
{
	IFRAMES *oifr, *nifr;
	TR_XMIT_PARMS *oxp, *nxp;
	mblk_t *omb, *nmb;
	char *odp, *ndp;
	struct datab *obp, *nbp;
	u_long corel = (u_long)ifrm->correlator;

	nifr = ifrm;
	nxp = &nifr->xmit_parms;
	nmb = (mblk_t*)nxp->xmit_data;
	nbp = nmb->b_datap;
	ndp = nmb->b_rptr;

	if (ifrm->state == TR_IFRM_WAIT_XMIT_STAT) {
		printf("Waited Iframe ready, corel=%x\n", corel);
	} else if (ifrm->state == TR_IFRM_WAIT_ACK) {
		printf("Rexmit Iframe, corel=%x\n", corel);
	} else if (ifrm->state != TR_IFRM_WAIT_IFRM_RSP) {
		printf("Invalid Iframe ready, corel=%x, stat=%x\n",
			corel, (int)(ifrm->state));
		return;
	}

	oifr = &gtrifr0[corel];
	oxp = &oifr->xmit_parms;
	omb = &gtrmb0[corel];
	obp =  &gtrdp0[corel];
	odp = &gtrmd0[corel][0];

	if (!nxp) {
		printf("Bad XP ptr: oxp = %x, nxp = %x\n", oxp, nxp);
	} if (bcmp((char*)oxp, (char*)nxp, sizeof(*nxp))) {
		printf("Iframe XP altered\n");
		printf("nxp:\n");
		gtr_hexdump((char*)oxp, sizeof(*oxp));
		printf("oxp:\n");
		gtr_hexdump((char*)nxp, sizeof(*nxp));
	}

	if (!nmb) {
		printf("Bad MBLK ptr: omb = %x, nmb = %x\n", omb, nmb);
	} else if (bcmp((char*)omb, (char*)nmb, sizeof(*nmb))) {
		printf("Iframe MBLK altered\n");
		printf("omb:\n");
		gtr_hexdump((char*)omb, sizeof(*omb));
		printf("nmb:\n");
		gtr_hexdump((char*)nmb, sizeof(*nmb));
	}

	if (!nbp) {
		printf("Bad DATA BLOCK ptr: obp = %x, nbp = %x\n", obp, nbp);
	} else if (bcmp(obp, nbp, 16)) {
		printf("Iframe DATA altered\n");
		printf("obp:\n");
		gtr_hexdump((char*)obp, 16);
		printf("nbp:\n");
		gtr_hexdump((char*)nbp, 16);
	}

	if (!ndp) {
		printf("Bad DATA ptr: odp = %x, ndp = %x\n", odp, ndp);
	} else if (bcmp(odp, ndp, 32)) {
		printf("Iframe DATA altered\n");
		printf("odp:\n");
		gtr_hexdump((char*)odp, 32);
		printf("ndp:\n");
		gtr_hexdump((char*)ndp, 32);
	}
}
#endif /* IFRMDBG */

/************************************************************************/
/*	IFRAME OUTPUT:							*/
/************************************************************************/

/*
 * get scb:
 */
static int
gtr_getscb(struct gtr_info *gtr, int cmd)
{
	int s1;
	register int err = 0;

	while (gtr->dcmd != 0) {
		SYIELD;
		DELAY(1000);
		EYIELD;
		if (++err >= 10) {	/* wait for 10 msec */
			return(EBUSY);
		}
	}
	gtr->dcmd = cmd;	/* XXX: possible race condition */
	return(0);
}

/*
 * gtr_dcmd: Direct CMD - serialized.
 */
static int
gtr_dcmd(struct gtr_info *gtr, caddr_t src, u_long slen,
			caddr_t res, u_long rlen, u_long delay)
{
	int err;
	int limit = 300;	/* max = 3 sec */
	SCB *scb = (SCB *)src;

#ifdef STASH
	if (gtr->bd_state != GTR_STATE_OK)
		return(ENETDOWN);
#endif
	while ((err=gtr_getscb(gtr, (int)scb->scb_cmd)) != 0) {
		if (dcmd_hndl(gtr) == 0)
			break;
		if (--limit == 0) {
			printf("gtr%d: override dcmd(0x%x) with 0x%x\n",
				gtr->gif_unit,gtr->dcmd, (int)scb->scb_cmd);
			break;
		}
	}

	err = gtr_xcmd(gtr,XCMD_DIRECT,src,slen,res,rlen,1);
	return(err);
}

/*
 * transmit I-frame
 */
int
gtr_req_iframe(struct gtr_info *gtr, LCB *lcb, IFRAMES *ifrp)
{
	SCB scb;
	int error;
	u_short sid;

	ASSERT(lcb != 0);
	ASSERT(ifrp != 0);

	if (lcb->ifr_state == IFR_STATE_WAIT) {
		/* C16 issued HOLD, so we can not even issue req. */
		IWMSG(("Iframe REQ not issued due to C16 busy\n"));
		return(EBUSY);
	}

	sid = lcb->stn_id;
	if (lcb->ifr_pend != 0) {
		/* req cmd is pending already */
		return(EBUSY);
	}
	if ((ifrp->correlator&0x1) == 0) {
		IWMSG(("corel valid already=%x", (u_int)ifrp->correlator));
	}
	lcb->ifr_pend = 1;

	QDBGUP(gtrifx.wcorel,1);
	ifrp->state = TR_IFRM_WAIT_COREL;

	scb.scb_cmd = TR_CMD_XMIT_IFRAME;
	scb.scb_parm1 = sid;
	scb.scb_parm2 = 0;

	if ((error=gtr_dcmd(gtr,(char*)&scb,sizeof(scb),0,0,DELAY_OP)) != 0) {
		IWMSG(("Iframe(%x) REQ not issued due to %s\n",
		    ifrp->correlator, (error==CMDSTS_IWAIT)?"IWAIT":"NO SCB"));
		lcb->ifr_pend = 0;
		QDBGDOWN(gtrifx.wcorel,1);
		ifrp->state = TR_IFRM_PENDING;
		return(EIO);
	}
#ifdef IWAITDBG
	if (babble) {
		printf("Iframe(%x) request to board\n", ifrp->correlator);
		babble--;
	}
#endif
	return(0);
}

/*
 * transmit I-frame
 * - takes only brand new request.
 */
gtr_xmit_iframe(struct gtr_info *gtr, TR_XMIT_PARMS *xp)
{
	SCB scb;
	LCB *lcb;
	int newlast;
	IFRAMES *ifrp;
	int s = splimp();
	int error = 0;
	u_short sid = xp->station_id;

	DP(("gtr%d: xmit_iframe, sid=%x, mblk=%x\n",
				gtr->gif_unit, sid, xp->xmit_data));

#ifdef STASH
{
	int alen;
	mblk_t *s0 = (mblk_t*)xp->xmit_data;
	for (alen = 0; s0 != 0; s0 = s0->b_cont) {
		alen += s0->b_wptr-s0->b_rptr;
	}
	if (alen != xp->frame_size) {
		printf("Jay sir, len %d fixed to %d(ft=%d)\n",
				xp->frame_size, alen, xp->frame_type);
	}
}
#endif

	lcb = tr_find_WR_q(&gtr->gif, sid);
	if (lcb == 0) {
		printf("gtr%d: xmit_iframe: No LCB for sid=%x\n",
				gtr->gif_unit, sid);
		error = EIO; goto xif_ret;
	}
#ifdef QDBG
	if (gtrifx.lcb != (u_long)lcb) {
		gtrifx.lcb = (u_long)lcb;
	}
#endif

	/* look for a place to store the iframe information in the LCB */
	if (lcb->ifr_cnt >= (NIFRAMES-1)) {
		/* full , but left 1 as DMZ */
		IWMSG(("Xiframe Q is full\n"));
		error = ENOBUFS; goto xif_ret;
	}
	if ((newlast = lcb->ifr_last + 1) >= NIFRAMES)
		newlast = 0;

	/* save IFRAME info */
	ifrp = &lcb->iframes[lcb->ifr_last];
	if (ifrp->state != TR_IFRM_NULL) {
		IWMSG(("invalid new state = %d\n",(u_int)ifrp->state));
	}

	ifrp->correlator = (lcb->ifr_corel|0x1);
	bcopy((char*)xp,(char*)&ifrp->xmit_parms,sizeof(*xp));

	lcb->ifr_cnt += 1;
	lcb->ifr_last = newlast;

	lcb->ifr_corel += 2;
	lcb->ifr_corel &= 0xfe;

	ifrp->state = TR_IFRM_PENDING;

	if (lcb->ifr_pend != 0) {
		/* ether cmd pending already, do nothing */
		IWMSG(("Holding Iframe ec=%x\n", lcb->ifr_corel));
		QDBGUP(gtrifx.pend,1);
		/* it is no error */
		goto xif_ret;
	}

	/* cmd is not peding, so get the first pending iframe */
	ifrp=tr_find_iframe_q(&gtr->gif,sid,(u_short)TR_IFRM_PENDING);
	if (ifrp == 0) {
		error = EIO;
		goto xif_ret;
	}

	/* since the iframe is queued, should return success
	 * even if current request failed.
	 */
	error = 0;	/* paranoid */
	(void)gtr_req_iframe(gtr, lcb, ifrp);

xif_ret:
	splx(s);
	return(error);
}

/*
 * llc reset
 */
static int
gtr_llc_reset(struct gtr_info *gtr, u_short sid)
{
	SCB scb;

	LDP(("gtr_llc_reset:\n"));
	scb.scb_cmd = TR_CMD_LLC_RESET;
	scb.scb_parm1 = sid;
	scb.scb_parm2 = 0;
	return(gtr_dcmd(gtr,(char*)&scb,sizeof(scb),0,0,DELAY_OP));
}

/*
 * open sap:
 */
static int
gtr_open_sap(struct gtr_info *gtr, struct tr_open_sap_block *osb, char *gsaps)
{
	int argsiz;
	char arg[64];
	SCB *scb = (SCB*)&arg[0];
	struct tr_open_sap_block *tosb;

	LDP(("gtr_open_sap:\n"));
	bzero((caddr_t)&arg[0], 64);
	scb->scb_cmd = TR_CMD_OPEN_SAP;

	tosb = (struct tr_open_sap_block *)&arg[sizeof(*scb)];
	bcopy((char*)osb, (char*)tosb, sizeof(*osb));
	/* don't worry about gsap ptr */

	if (gsaps != 0) {
		argsiz = sizeof(*scb)+sizeof(*osb)+8;
		bcopy(gsaps,&arg[sizeof(*scb)+sizeof(*osb)], 8); /* XXX */
	} else {
		argsiz = sizeof(*scb)+sizeof(*osb);
		osb->gsap_list_ptr_hi = 0;
		osb->gsap_list_ptr_lo = 0;
	}
	return(gtr_dcmd(gtr,arg,argsiz,0,0,DELAY_OP));
}

/*
 * open IP SAP
 */
static int
gtr_open_ip_sap(struct gtr_info *gtr)
{
	int sts, iptid;
	struct tr_open_sap_block osb;

	LDP(("gtr_open_ip_sap:\n"));

	if (gtr->ip_sid != 0) {
		LDP(("gtr_open_ip_sap: already opened(%x)\n", gtr->ip_sid));
	}

	/* set open IP sap */
	bzero((caddr_t)&osb, sizeof(osb));
	osb.sap_value = TR_IP_SAP;
	osb.sap_options = TR_OPEN_IP_SAP_OPTIONS;

	if ((sts = gtr_open_sap(gtr, &osb, 0)) != 0) {
		LDP(("open_ip_sap failed\n"));
		return(sts);
	}

#ifdef STASH
	sts = DELAY_OPEN;
	while ((gtr->ip_sid == 0) && (sts > 0)) {
		DELAYBUS(DELAY_OP);
		sts -= DELAY_OP;
	}
	LDP(("open_ip_sap: ip_sid = %x\n", gtr->ip_sid));
	if (gtr->ip_sid == 0)
		return(EIO);
#endif
	return(0);
}

/*
 * close sap:
 */
static int
gtr_close_sap(struct gtr_info *gtr, u_short sid)
{
	SCB scb;

	LDP(("gtr_close_sap:\n"));
	scb.scb_cmd = TR_CMD_CLOSE_SAP;
	scb.scb_parm1 = sid;
	scb.scb_parm2 = 0;
	return(gtr_dcmd(gtr,(char*)&scb,sizeof(scb),0,0,DELAY_OP));
}

/*
 * open station:
 */
static int
gtr_open_station(struct gtr_info *gtr,
	struct tr_open_station_block *ostn, char* rnodea)
{
	SCB *scb;
	int argsiz;
	char arg[64];
	char *rnode;
	struct tr_open_station_block *tostn;

	LDP(("gtr_open_station:\n"));
	scb = (SCB*)&arg[0];
	scb->scb_cmd = TR_CMD_OPEN_STATION;
	/* don't worry about parm_ptr */

	tostn = (struct tr_open_station_block*)&arg[sizeof(*scb)];
	bcopy((char*)ostn, (char*)tostn, sizeof(*ostn));
	/* don't worry about rnode addr ptr */

	rnode = &arg[sizeof(*scb)+sizeof(*tostn)];
	bcopy(rnodea, rnode, TR_MAC_ADDR_SIZE);

	argsiz = sizeof(*scb)+sizeof(*tostn)+TR_MAC_ADDR_SIZE;
	return(gtr_dcmd(gtr,arg,argsiz,0,0,DELAY_OP));
}

/*
 * close station:
 */
static int
gtr_close_station(struct gtr_info *gtr, u_short sid)
{
	SCB scb;

	LDP(("gtr_close_station:\n"));
	scb.scb_cmd = TR_CMD_CLOSE_STATION;
	scb.scb_parm1 = sid;
	scb.scb_parm2 = 0;
	return(gtr_dcmd(gtr,(char*)&scb,sizeof(scb),0,0,DELAY_OP));
}

/*
 * connect station:
 */
static int
gtr_connect_station(struct gtr_info *gtr,
		struct tr_connect_station_block *csb, TR_RII *ri)
{
	int argsiz;
	TR_RII *tri;
	char arg[64];
	SCB *scb = (SCB*)&arg[0];
	struct tr_connect_station_block *tcsb;

	LDP(("gtr_connect_station:\n"));
	scb->scb_cmd = TR_CMD_CONNECT_STATION;
	/* don't worry about param blk ptr */

	tcsb = (struct tr_connect_station_block *)&arg[sizeof(*scb)];
	bcopy((char*)csb,(char*)tcsb, sizeof(*csb));
	/* don't worry about param blk ptr */

	argsiz = sizeof(*scb)+sizeof(*csb);
	if (ri != 0) {
		tri = (TR_RII*)&arg[argsiz];
		argsiz += sizeof(*ri);
		bcopy((char*)ri,(char*)tri, sizeof(*ri));
	} else {
		tcsb->routing_info_high = 0;
		tcsb->routing_info_low = 0;
	}

	return(gtr_dcmd(gtr,arg,argsiz,0,0,DELAY_OP));
}

/*
 * open station:
 */
static int
gtr_llc_stat(struct gtr_info *gtr)
{
	SCB *scb;
	int argsiz;
	char arg[64];
	TR_LLC_STATISTICS_BLOCK *sb;

	LDP(("gtr_llc_stat:\n"));
	scb = (SCB*)&arg[0];
	scb->scb_cmd = TR_CMD_LLC_STATS;
	/* don't worry about parm_ptr */

	sb = (TR_LLC_STATISTICS_BLOCK*)&arg[sizeof(*scb)];
	sb->station_id = ((LCB*)(gtrifx.lcb))->stn_id;
	sb->buffer_length = sizeof(TR_LLC_STAT_DATA);
	sb->buffer_addr_high = 0;
	sb->buffer_addr_low = 0;
	sb->actual_buff_length = sizeof(TR_LLC_STAT_DATA);
	sb->options = 0;

	argsiz = sizeof(*scb)+sizeof(*sb);
	return(gtr_dcmd(gtr,arg,argsiz,0,0,DELAY_OP));
}

/*----------------------------------------------------------------------*
 * 	gtr_lli_output							*
 *									*
 *	 	Interface to token-ring driver from LLI module		*
 *		and the rest of the SNA protocol stack			*
 *									*
 *		returns -1 if error occurs				*
 *		or 0 if successful					*
 *----------------------------------------------------------------------*/
static int
gtr_lli_output(u_short unit, SCB *scb)
{
	int s;
	char *addr;
	caddr_t	pblk;
	mblk_t *em_ptr;
	struct gtr_info *gtr;
	TR_XMIT_PARMS *xp;
    	TR_OPEN_STATION_BLOCK *opn_blk;
	u_short	error = 0;

	/* ensure unit number is valid */
	LP(("gtr_lli_output: req = %d\n", scb->scb_cmd));
	if ((unit < 0) || (unit >= GTR_MAXBD)) {
		LP(("gtr_lli_output: bad unit = %d\n", unit));
		return(0);
	}

	gtr = &gtr_info[unit];
	if (gtr->bd_state != GTR_STATE_OK || iff_dead(gtr->gif.if_flags)) {
		LP(("gtr_lli_output: board is not operational\n"));
		return(0);
	}
	if ((gtr->flags&GTR_ONCHIPLLC) == 0) {
		printf("gtr%d: LLC ucode is not running\n", gtr->gif_unit);
		return(EINVAL);
	}

	switch (scb->scb_cmd) {
	case TR_CMD_TRANSMIT:
		/* Here, the scb will point to a transmit parameter
		 * block (on the stack). In turn, the transmit parameter
		 * block will point to xmit_data, which will be an mblk.
		 */

		HTINT(xp, scb->scb_parm1);
		if (xp->frame_type == TR_FTYPE_I_FRAME) {
#ifdef IFRMDBG
			char *cp = xp->xmit_data->b_rptr;
			gtrnxtcorel += 2;
			gtrnxtcorel &= 0xff;
			if (cp[0]==0 && cp[1]==0 && cp[2]==0 && cp[3]==0) {
				printf("lli_output: BAD Iframe\n");
				gtr_hexdump(cp, 32);
			}
#endif
			QDBGUP(gtrifx.tot,1);
			error = gtr_xmit_iframe(gtr, xp);
#ifdef QDBG
			if (error) {
				QDBGUP(gtrifx.reject,1);
			}
#endif
			break;
		}
		LP(("gtr_lli_output(xmit): remote sap = %x, stn id = %x\n",
				xp->remote_sap, xp->station_id));
		s = splimp();
		error = gtr_xllc(gtr, xp, 0);
		splx(s);
		break;

	case TR_CMD_LLC_RESET:
		if ((error = gtr_llc_reset(gtr, scb->scb_parm1)) != 0)
			break;
		if (scb->scb_parm1 == 0)
			error = gtr_open_ip_sap(gtr);
		break;

	case TR_CMD_OPEN_SAP: {
		char *gsaps;
		TR_OPEN_SAP_BLOCK *osb;

		if (iff_dead(gtr->gif.if_flags)) {
			error = EINVAL;
			break;
		}
		HTINT(pblk, scb->scb_parm1);
		osb = (TR_OPEN_SAP_BLOCK*)pblk;
		HTINT(gsaps, osb->gsap_list_ptr_hi);
#ifdef QQQ
		osb->maxout = NIFRAMES;	/* XXX: OUTMAX congestion avoidance */
#endif
		error = gtr_open_sap(gtr, osb, gsaps);
		} break;

	case TR_CMD_CLOSE_SAP:
		error = gtr_close_sap(gtr, scb->scb_parm1);
		break;

	case TR_CMD_OPEN_STATION:
		HTINT(pblk, scb->scb_parm1);
		opn_blk = (TR_OPEN_STATION_BLOCK *)pblk;
		HTINT(addr, opn_blk->rnode_addr_high);
		error = gtr_open_station(gtr, opn_blk, addr);
		break;

	case TR_CMD_CLOSE_STATION:
		error = gtr_close_station(gtr, scb->scb_parm1);
		break;

	case TR_CMD_CONNECT_STATION: {
		TR_RII *ri;
		TR_CONNECT_STN_BLK *csb;

		HTINT(csb, scb->scb_parm1);
		HTINT(ri, csb->routing_info_high);
		error = gtr_connect_station(gtr, csb, ri);
		} break;

	case TR_CMD_LLC_STATS:
		break;

	default:
		return(0);
		break;
	}

	if (error) {
		LP(("gtr_lli_output: error = %x \n", error));
		return(0);
	}
	LP(("gtr_lli_output: successful completion\n"));
	return(1);
}


/************************************************************************/
/*	IFRAME INPUT:							*/
/************************************************************************/
static void
gtr_lli(struct mbuf *m0)
{
	int s1, s;
	mblk_t *s0;
	LLI_CB *cbp;

	ASSERT(m0->m_next == 0);
	if (m0 == 0)
		return;

	/* XXX: llc ctl frame size < 4K */
	s1 = spl1();
	s = splimp();
	if ((s0 = allocb(m0->m_len,BPRI_MED)) == 0) {
		printf("gtr_lli: no mblk size=%d\n",m0->m_len);
		goto lliret_1;
	}

	bcopy(mtod(m0, char*), s0->b_wptr, m0->m_len);
	s0->b_datap->db_type = M_PROTO;
	s0->b_wptr += m0->m_len;
	cbp = (LLI_CB*)s0->b_rptr;
	cbp->param_blk = (cbp+1);
	tr_lli_input(s0);

lliret_1:
	m_freem(m0);
	splx(s);
	splx(s1);
}

static void
gtr_xmit_pend(struct gtr_info *gtr, u_short sid)
{
	int sts;
	IFRAMES *ifrp;
	LCB *lcb = tr_find_WR_q(&gtr->gif, sid);

	if (lcb == 0)
		return;

	/* get the first pending iframe to be requested */
	ifrp=tr_find_iframe_q(&gtr->gif,sid,(u_short)TR_IFRM_PENDING);
	if (ifrp == 0) {
		return;
	}

	IWMSG(("Requesting held iframe(%x)\n", ifrp->correlator));
	sts = gtr_req_iframe(gtr, lcb, ifrp);
	if (sts) {
		IWMSG(("Request held iframe(%x) failed\n",ifrp->correlator));
		QDBGUP(gtrifx.rfail,1);
	} else {
		QDBGUP(gtrifx.rtry,1);
	}
}

static int
gtr_irearm(struct gtr_info *gtr)
{
#ifdef IREARM
	int s, s1;
	IFRAMES *ifrp;
	LCB *lcb;
	u_short sid = gtr->ir_sid;

	gtr->ir_tid = 0;
	if (sid == 0)
		return(0);
	if ((lcb = (LCB *)gtrifx.lcb) == 0) {
		return(0);
	}
	if (lcb->ifr_state == IFR_STATE_OK) {
		return(0);
	}

	s1 = spl1();
	s = splimp();
	ifrp = tr_find_iframe_q(&gtr->gif,sid,(u_short)TR_IFRM_WAIT_ACK);
	if (ifrp == 0) {
		IWMSG(("Rearm Iframe not found\n"));
		goto irret;
	}
	
	QDBGUP(gtrifx.rxmit,1);	/* total rexmited */
	if (gtr_xllc(gtr, &ifrp->xmit_parms, (u_short)ifrp->correlator)) {
		IWMSG(("Rearm Iframe(%x) failed\n", ifrp->correlator));
		QDBGUP(gtrifx.rfail,1);		/* total IFR req failed */
	} else {
		IWMSG(("Rearm Iframe(%x) succeeded\n", ifrp->correlator));
		QDBGUP(gtrifx.rtry,1);	/* total IFR req succeeded */
	}

irret:
	splx(s);
	splx(s1);
#endif
	return(0);
}

/*
 * Handle Iframe xmit status.
 */
static void
gtr_ifr_stat(struct mbuf *m0)
{
	int *bp;
	SSB *ssb;
	LCB *lcb;
	int i, j;
	u_short sid;
	u_short count;
	u_short corel;
	IFRAMES *ifrp;
	struct gtr_info *gtr;
#ifdef ISTRINTR
	int s1 = spl1();
	int s = splimp();
#endif
	bp = mtod(m0, int*);
	gtr = (struct gtr_info *)(bp[0]);
	ssb = (SSB*)(&bp[1]);
	sid = ssb->ssb_parm1;

	if ((lcb = tr_find_WR_q(&gtr->gif, sid)) == 0) {
#if defined(IFRMDBG) || defined(IWAITDBG)
		printf("lcb not found for sid=%x\n",(u_long)sid);
#endif
		goto ifs_ret;
	}
	corel = (ssb->ssb_parm2 >> 8);

	if (ssb->ssb_stat & TR_XMIT_I_FRAME_ACK) {
		/* When this bit is set, only sid and count has meaning.
		 * correlator has no meaning */
		count = ssb->ssb_stat & 0xff;
		lcb->ifr_cnt -= (u_int)count;
		ASSERT((int)lcb->ifr_cnt >= 0);
		QDBGUP(gtrifx.freed,1);

		for (i = lcb->ifr_first; count > 0; count--) {
			ifrp = &lcb->iframes[i];

			if (ifrp->state != TR_IFRM_WAIT_ACK) {
				printf("Unexpected Iframe ACK: corel=%x\n",
					(int)(ifrp->correlator));
#ifdef IFRMDBG
	if (ifrp->xmit_parms.xmit_data == 0)
		printf("Iframe ACK has no xmit_data\n");
	else
		gtr_hexdump(((mblk_t*)ifrp->xmit_parms.xmit_data)->b_rptr,
				(int)32);
#endif
				break;
			}
			if (i == lcb->ifr_last) {
				printf("too many Iframe ACKs\n");
				break;
			}

#ifdef IWAITDBG
		if (babble)
			printf("got ACK for %x\n",(u_long)ifrp->correlator);
#endif
			ifrp->correlator = 0xff;
			ifrp->state = TR_IFRM_NULL;
			if (ifrp->xmit_parms.xmit_data) {
				DP(("freeing iframe info, mblk=%x, index=%x\n",
					ifrp->xmit_parms.xmit_data, i));
	 			freemsg(ifrp->xmit_parms.xmit_data);
				ifrp->xmit_parms.xmit_data = NULL;
			}

			if (++i >= NIFRAMES)
				i = 0;
		}
		lcb->ifr_first = i;
	}

	/* See if ready to take iframe req again */
	if (ssb->ssb_stat & TR_XMIT_I_READY) {
		/* ready to xmit Iframe; find iframe-q, and xmit. */
#ifdef IWAITDBG
		if (babble) {
			printf("gtr: iframe ready, corel=%x\n", corel);
			babble--;
		}
#endif
		lcb->ifr_state = IFR_STATE_OK;
#ifdef QDBG
		gtrifx.stat = IFR_STATE_OK;
#endif
		for (i = 0, j = lcb->ifr_first; i < NIFRAMES; i++) {
			if (j == lcb->ifr_last)
				break;
			if (lcb->iframes[j].correlator == corel) {
				TR_XMIT_PARMS *xp = &lcb->iframes[j].xmit_parms;
#ifdef IFRMDBG
				chkrdy_ifrm(&lcb->iframes[j]);
#endif
				if (lcb->iframes[j].state==TR_IFRM_WAIT_ACK) {
					QDBGUP(gtrifx.rxmit,1);
				} else {
					lcb->iframes[j].state=TR_IFRM_WAIT_ACK;
				}
				if (gtr_xllc(gtr, xp, corel)) {
					IWMSG(("send failed for corel=%x\n",
						corel));
					goto ifs_ret;
				}
				QDBGUP(gtrifx.wack,1);   /* total IFR xmited */
				goto ifs_ret;
			}
			if (++j >= NIFRAMES)
	    			j = 0;
		}
		printf("can't find Iframe: corel=%x\n", (int)corel);
		ASSERT(0);
	}

ifs_ret:
	m_freem(m0);

	/* request pending iframe iff exist */
	if ((lcb != 0) && (lcb->ifr_pend==0) && (lcb->ifr_cnt > 0) &&
	    (lcb->ifr_state == IFR_STATE_OK)) {
		gtr_xmit_pend(gtr, sid);
	}
#ifdef ISTRINTR
	splx(s);
	splx(s1);
#endif
}

/*
 * Handle Iframe request.
 */
static void
gtr_ifr_intr(struct mbuf *m0)
{
	int *bp;
	SSB *ssb;
	LCB *lcb;
	int i, j;
	u_short sid;
	LLI_CB *cbp;
	u_short corel;
	struct mbuf *m1;
	struct gtr_info *gtr;
	IFRAMES *ifrp = 0;
#ifdef ISTRINTR
	int s1 = spl1();
	int s = splimp();
#endif

	bp = mtod(m0, int*);
	gtr = (struct gtr_info *)(bp[0]);
	ssb = (SSB*)(&bp[1]);
	sid = ssb->ssb_parm1;
	corel = ssb->ssb_parm2 >> 8;

	if ((lcb = tr_find_WR_q(&gtr->gif, sid)) == 0) {
		IWMSG(("ifr_intr: NO LCB!!!\n"));
		goto ifr_ret;
	}

	switch(ssb->ssb_stat) {
	case TR_IFRAME_BUILD:
		lcb->ifr_pend = 0;
		/* See if it is rexmit request */
		for (i = 0, j = lcb->ifr_first; i < NIFRAMES; i++) {
			if (j == lcb->ifr_last)
				break;
			if (lcb->iframes[j].correlator == corel) {
				ifrp = &lcb->iframes[j];
				break;
			}
			if (++j >= NIFRAMES)
				j = 0;
		}
		if (!ifrp) {
			/* This is new corel assignment */
			ifrp = tr_find_iframe_q(&gtr->gif,sid,
					(u_short)TR_IFRM_WAIT_COREL);
			if (ifrp == 0) {
				IWMSG(("NO BUILD IFRP: sts=%x corel=%x\n",
						(u_long)ssb->ssb_stat,
						(u_long)ssb->ssb_parm2));
				goto ifr_ret;
			}
if ((ifrp->correlator&0xfe) != (corel&0xfe)) {
	IWMSG(("corel=%x, expected=%x\n", corel, ifrp->correlator));
	for (i = 0, j = lcb->ifr_first; i < NIFRAMES; i++) {
		if (j == lcb->ifr_last)
			break;
		if ((lcb->iframes[j].correlator&0xfe) == (corel&0xfe)) {
			ifrp = &lcb->iframes[j];
			IWMSG(("FOUND: corel=%x, expected=%x\n",
				corel, ifrp->correlator));
			break;
		}
		if (++j >= NIFRAMES)
			j = 0;
	}
}
		}

		/* board says it's okay to send iframe request now. */
		if ((ifrp->correlator&0x1) == 0) {
			QDBGUP(gtrifx.rxmit,1);	/* total rexmited */
			IWMSG(("Retransmit Iframe, corel=%x\n",corel));
		}
#ifdef IWAITDBG
		else if (babble) {
			printf("Build Iframe, corel=%x\n",corel);
			babble--;
		}
#endif
	        ifrp->correlator = corel;
#ifdef IFRMDBG
		stash_ifrm(ifrp, (u_long)(ifrp->correlator));
#endif
		if (gtr_xllc(gtr, &ifrp->xmit_parms,
					(u_short)ifrp->correlator)) {
			IWMSG(("Queue Iframe failed\n"));
			QDBGUP(gtrifx.rfail,1);	/* total IFR req failed */
			break;
		}

	    	ifrp->state = TR_IFRM_WAIT_ACK;
		QDBGUP(gtrifx.wack,1);   /* total IFR xmited */

		/* request pending iframe iff exist */
		gtr_xmit_pend(gtr, sid);
		break;

	case TR_IFRAME_WAIT: {
		/* board says to wait before sending iframe request. */
#ifdef IWAITDBG
		babble = 100;
		printf("Wait Iframe, corel=%x\n", corel);
#endif
		lcb->ifr_pend = 0;
		lcb->ifr_state = IFR_STATE_WAIT;
#ifdef QDBG
		gtrifx.stat = TR_IFRM_WAIT_READY;
#endif
		ifrp = tr_find_iframe_q(&gtr->gif,sid,
					(u_short)TR_IFRM_WAIT_COREL);
		if (ifrp == 0) {
			printf("NO WAIT IFRP: sts=%x corel=%x\n",
					(u_long)ssb->ssb_stat,
					(u_long)ssb->ssb_parm2);
			break;
		}

if ((ifrp->correlator&0xfe) != (corel&0xfe)) {
	IWMSG(("WAIT corel=%x, expected=%x\n", corel, ifrp->correlator));
	for (i = 0, j = lcb->ifr_first; i < NIFRAMES; i++) {
		if (j == lcb->ifr_last)
			break;
		if ((lcb->iframes[j].correlator&0xfe) == (corel&0xfe)) {
			ifrp = &lcb->iframes[j];
			IWMSG(("WAIT FOUND: corel=%x, expected=%x\n",
				corel, ifrp->correlator));
			break;
		}
		if (++j >= NIFRAMES)
			j = 0;
	}
}

		if (ifrp) {
			/* board says it's okay to send iframe request now. */
			if ((ifrp->correlator&0x1) == 0) {
				QDBGUP(gtrifx.rxmit,1);	/* total rexmited */
				IWMSG(("Retransmit Iframe, corel=%x\n",corel));
			}
			ifrp->correlator = corel;
			ifrp->state = TR_IFRM_WAIT_READY;
#ifdef IFRMDBG
			stash_ifrm(ifrp, (u_long)(ifrp->correlator));
#endif
		}
		if (gtr->ir_tid == 0) {
			gtr->ir_sid = sid;
			gtr->ir_tid = cputimeout((void (*)())gtr_irearm,
					(caddr_t)gtr,3*TR_INCR_TIMEOUT);
		}
		} break;

	case TR_IFRAME_MAX_CMDS:
		printf("Iframe request rejected: more than 127 outstanding iframe requests\n");
		goto xmit_iframe_err;

	case TR_IFRAME_INV_STN:
		printf("Iframe request rejected: Invalid station id\n");
		goto xmit_iframe_err;

	case TR_IFRAME_PROT_ERR:
		printf("Iframe request rejected: Protocol error\n");
xmit_iframe_err:
		m1 = m_vget(M_DONTWAIT, (sizeof(*cbp)+sizeof(*ssb)),MT_DATA);
		if (m1 == 0) {
			printf("gtr%d: no mbuf for Iframe PROT ERR\n",
					gtr->gif_unit);
			break;
		}

		cbp = mtod(m1, LLI_CB*);
		cbp->port = gtr->gif_unit;
		cbp->command = LLI_CMD_RSP;
		cbp->param_blk = (cbp+1);
		bcopy((char*)ssb, (char*)cbp->param_blk, sizeof(*ssb));
		m1->m_len = (sizeof(*cbp)+sizeof(*ssb));
		gtr_lli(m1);
		break;
	default:
		printf("Iframe: unknown ERROR returned = %x\n", ssb->ssb_stat);
		break;
	}

ifr_ret:
	m_freem(m0);
#ifdef ISTRINTR
	splx(s);
	splx(s1);
#endif
}

/*
 * Process a command status interrupt.
 */
static void
gtr_process_cmd(struct gtr_info *gtr, SSB *ssb, TR_ICB *icb)
{
	LLI_CB *cbp;
	struct mbuf *m0;
	u_long cmd = ssb->ssb_command;

	LDP(("dcmd process for cmd=%x\n",cmd));
	switch (cmd) {
	    case TR_SSB_COMMAND_REJ:
		printf("gtr%d: cmd %x rejected, status = %x\n", gtr->gif_unit,
				ssb->ssb_parm1, ssb->ssb_parm0);
		gtr->dcmd = 0;		/* free dcmd queue */
		break;

	    case TR_CMD_OPEN_SAP:
		if (ssb->ssb_stat != 0) {
			printf("gtr%d: OPEN_SAP failed, status=%x, sid=%x\n",
				gtr->gif_unit, ssb->ssb_stat, ssb->ssb_stn_id);
		}

		/* XXX: first SAP opened must be IPSAP */
		if (gtr->ip_sid == 0) {
			if (ssb->ssb_stat == 0) {
				gtr->ip_sid = ssb->ssb_parm1;
				LDP(("IP_SID = %x\n", (int)gtr->ip_sid));
			}
			return;
		}

		DP(("gtr%d: OPEN_SAP succeeded, sid = %x\n",
				gtr->gif_unit, ssb->ssb_stn_id));
		m0 = m_vget(M_DONTWAIT,(sizeof(*cbp)+sizeof(*ssb)),MT_DATA);
		if (m0 == 0) {
			printf("gtr%d: no mbuf for OPen SAP\n", gtr->gif_unit);
			break;
		}

		cbp = mtod(m0, LLI_CB*);
		cbp->port = gtr->gif_unit;
		cbp->command = LLI_CMD_RSP;
		cbp->param_blk = (cbp+1);
		bcopy((char*)ssb, (char*)cbp->param_blk, sizeof(*ssb));
		m0->m_len = (sizeof(*cbp)+sizeof(*ssb));
		gtr_lli(m0);
		break;

	    case TR_CMD_LLC_RESET:
		DP(("gtr%d: LLC RESET complete, status = %x\n",
				gtr->gif_unit, ssb->ssb_stat));
		break;

	    case TR_CMD_OPEN_STATION:
	    case TR_CMD_CLOSE_SAP:
	    case TR_CMD_CLOSE_STATION:
	    case TR_CMD_CONNECT_STATION:
		DP(("gtr%d: cmd = %x,status = %x, sap = %x\n",
			gtr->gif_unit, ssb->ssb_command,
			ssb->ssb_stat, ssb->ssb_parm1));
		m0 = m_vget(M_DONTWAIT,(sizeof(*cbp)+sizeof(*ssb)),MT_DATA);
		if (m0 == 0) {
			printf("gtr%d: no mbuf for OPEN/CLOSE/CONNECT stat\n",						gtr->gif_unit);
			break;
		}

		cbp = mtod(m0, LLI_CB*);
		cbp->port = gtr->gif_unit;
		cbp->command = LLI_CMD_RSP;
		cbp->param_blk = (cbp+1);
		bcopy((char*)ssb, (char*)cbp->param_blk, sizeof(*ssb));
		m0->m_len = (sizeof(*cbp)+sizeof(*ssb));
		gtr_lli(m0);
		break;

	    case TR_CMD_XMIT_IFRAME: {
		int *bp;
		m0 = m_vget(M_DONTWAIT,(sizeof(int)+sizeof(*ssb)),MT_DATA);
		if (!m0) {
			printf("gtr%d: no mbuf for XMIT_IFRAME stat\n",
					gtr->gif_unit);
			break;
		}
		bp = mtod(m0, int*);
		bp[0] = (int)gtr;
		bcopy((char*)ssb, (char*)(&bp[1]), sizeof(*ssb));
		m0->m_len = sizeof(int)+sizeof(*ssb);
#ifdef IWAITDBG
		if (babble) {
			printf("got ifr intr(%x)\n",(u_long)ssb->ssb_stat);
			babble--;
		}
#endif
#ifdef ISTRINTR
		gtr_ifr_intr(m0);
#else
		gtr_ifr_intr(m0);
#endif
		} break;

		case TR_CMD_LLC_STATS: {
			/* icb is more than 16 bytes, so it is o.K. */
			bcopy((char*)icb, (char*)&gtr->llcstat,
					sizeof(gtr->llcstat));
		} break;
	}
}

static void
gtr_dcmdintr(struct gtr_info *gtr, u_long sifsts, SSB *ssb, TR_ICB *icb)
{
	LDP(("dcmd intr: sifsts=%x\n",sifsts));
	switch(sifsts&TR_STAT_INT_CODE) {
	case TR_INT_COMMAND_STAT:		/* OK */
		gtr_process_cmd(gtr, ssb, icb);
		break;

	case TR_INT_XMIT_STAT: {
		int *bp;
		struct mbuf *m0;

		if ((ssb->ssb_stat&(TR_XMIT_I_FRAME_ACK|TR_XMIT_I_READY))==0) {
			IWMSG(("unknown XMIT intr: %x\n", ssb->ssb_stat));
			return;
		}
		m0 = m_vget(M_DONTWAIT,(sizeof(int)+sizeof(*ssb)),MT_DATA);
		if (!m0) {
			printf("gtr%d: no mbuf for xmit stat\n",gtr->gif_unit);
			return;
		}
		bp = mtod(m0, int*);
		bp[0] = (int)gtr;
		bcopy((char*)ssb, (char*)(&bp[1]), sizeof(*ssb));
		m0->m_len = sizeof(int) + sizeof(*ssb);
#ifdef IWAITDBG
		if (babble) {
			printf("got ifr stat(%x)\n",(u_long)ssb->ssb_stat);
			babble--;
		}
#endif
#ifdef ISTRINTR
		gtr_ifr_stat(m0);
#else
		gtr_ifr_stat(m0);
#endif
		} break;

	case TR_INT_LLC_STAT: {		/* OK */
		LLI_CB *cbp;
		struct mbuf *m0;

		LP(("gtr%d: LLC_STATUS: sid=%x code=%x\n",
			gtr->gif_unit, icb->station_id_s, icb->llc_stat_code));

		if ((m0=m_vget(M_DONTWAIT,(sizeof(*cbp)+sizeof(*icb)),MT_DATA))==0) {
			DP(("no buf for llc status\n"));
			return;
		}

		cbp = mtod(m0, LLI_CB*);
		cbp->port = gtr->gif_unit;
		cbp->command = LLI_LLC_STATUS;
		cbp->param_blk = (cbp+1);
		bcopy((char*)icb, (char*)cbp->param_blk, sizeof(*icb));
		m0->m_len = (sizeof(*cbp)+sizeof(*icb));
		gtr_lli(m0);
}
		break;


	case TR_INT_ADAP_CHECK:		/* OK */
	case TR_INT_RING_STAT:		/* OK */
	case TR_INT_RCV_PENDING:	/* OK */
	case TR_INT_RCV_STAT:		/* OK */
		/* Non-direct cmd status */

	case TR_INT_SCB_CLEAR:		/* OK */
	case TR_INT_TIMER:		/* OK */
		/* Unused features */
		break;
	default:
		printf("gtr%d: spurious direct interrupt = %x\n",
				gtr->gif_unit, sifsts);
		break;
	}
}
#else /* ONCHIPLLC */
static int
gtr_lli_output(u_short unit, SCB *scb)
{
	return(0);	/* for now */
}
#endif
#endif /* IP12 || IP20 || IP22 */
