
/*
 * net64.h -
 *
 * File of bogus routines made to cram 64-bit addresses into 32-bit
 * quantities _temporarily_.
 *
 * BSD networking code has a nasty way of overlaying fixed-size structures
 * like the IP header with structures that contain pointers.  Since we can't
 * overlay 32-bit fields with pointers in the 64-bit kernel, we need to
 * rewrite the reassembly code for IP and TCP.  These macros provide a
 * band-aid(TM) solution.  They'll need to be replaced to take advantage of
 * the whole 64-bit address space.
 *
 *	--Steve Whitney, 10/93
 */

#include <sys/mips_addrspace.h>


#if _K64U64

#if IP20 || IP22 || IP26 || IP28
/* MC based systems have small physical addresses, so we can use bit 31 for
 * mapped/unmapped, and leave the bottom bits unchanged so unaligned transfers
 * work ok.
 */
#define NET64_SHIFTBITS	0
#else
#define NET64_SHIFTBITS	3
#endif

/*
 * NET64_SHIFTBITS is the number of bit of alignment we can count on.  We
 * get (31 + NET64_SHIFTBITS) bits of address space for network operations.
 * Since all structures we use these macros on are doubleword aligned, the
 * current value of NET64_SHIFTBITS is 3, giving us 16 GB of usable address
 * space.
 */

#define NET64_IN_RANGE(_x)						\
	((IS_KSEG0(_x) || IS_KSEG2(_x))					\
			&& ((K0_TO_PHYS(_x) >> NET64_SHIFTBITS) < 0x7fffffffL) \
			&& (!(K0_TO_PHYS(_x) & ((1 << NET64_SHIFTBITS) - 1))))

#define NET64_TOINT(_x)							\
	(__uint32_t)((K0_TO_PHYS(_x) >> NET64_SHIFTBITS)  |		\
		((0x4000000000000000L & (__psint_t)(_x)) >> 31))

#define	NET64_TOPTR(_x)						\
	(void *)(((long)(_x) & 0x80000000 ? K2BASE : K0BASE) |		\
			((0x7fffffff & ((long)_x)) << NET64_SHIFTBITS))
#if NET64_SHIFTBITS == 3
#define	NET64_ALIGN_CHECK	0x7
#endif	/* NET64_SHIFTBITS == 3 */
#if NET64_SHIFTBITS == 2
#define	NET64_ALIGN_CHECK	0x3
#endif	/* NET64_SHIFTBITS == 2 */
#if NET64_SHIFTBITS == 0
#define	NET64_ALIGN_CHECK	0x0
#endif	/* NET64_SHIFTBITS == 0 */
#define	NET64_ALIGN_MASK	(~NET64_ALIGN_CHECK)

#else	/* !_K64U64 */

#define	NET64_IN_RANGE(_x)	1
#define NET64_TOINT(_x)		(__uint32_t)(_x)
#define NET64_TOPTR(_x)		(void *)(_x)
#endif
