/*
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 *	@(#)if_ether.c	7.6 (Berkeley) 12/7/87 plus MULTICAST 1.0
 *	plus portions of 7.12 (Berkeley) 6/28/90
 */

/*
 * Ethernet address resolution protocol.
 * TODO:
 *	run at splnet (add ARP protocol intr.)
 *	link entries onto hash chains, keep free list
 *	add "inuse/lock" bit (or ref. count) along with valid bit
 */

#include "tcp-param.h"
#include "sys/param.h"
#include "sys/systm.h"
#include "sys/sema.h"
#include "sys/hashing.h"

#include "sys/mbuf.h"
#include "sys/socket.h"
#include "sys/errno.h"
#include "sys/immu.h"
#include "sys/pda.h"
#include "sys/debug.h"
#include "sys/sema.h"
#include "sys/cmn_err.h"

#define hz HZ
#include "net/soioctl.h"
#include "net/if.h"

#include "in.h"
#include "in_systm.h"
#include "in_var.h"
#include "ip.h"
#include "if_ether.h"

/* SRCROUTE */
#include "net/if_types.h"
#include "sys/kmem.h"

/*
 * Forward Referenced Procedures
 */
int in_arp_match(struct hashbucket *h, caddr_t key, caddr_t arg1,caddr_t arg2);

/*
 * External Procedures
 */
extern char *inet_ntoa(struct in_addr);
extern struct ifaddr *ifa_ifwithnet(struct sockaddr *);
extern void arptfree(struct arptab *at);
extern int looutput(struct ifnet *, struct mbuf *, struct sockaddr *);

/*
 * Module global variables
 */
#ifdef MP
#define	ARPTAB_BSIZ	16   /* bucket size */
#else
#define	ARPTAB_BSIZ	8    /* bucket size (smaller for dekstop machines) */
#endif

#define	ARPTAB_NB	41		/* number of buckets */

#define	ARPTAB_SIZE	(ARPTAB_BSIZ * ARPTAB_NB)
struct	arptab arptab[ARPTAB_SIZE];
int	arptab_size = ARPTAB_SIZE;	/* for arp command */

#ifdef _MP_NETLOCKS
lock_t	arp_lock;
#define ARP_LOCK(s)		s = mp_mutex_spinlock(&arp_lock)
#define ARP_UNLOCK(s)		mp_mutex_spinunlock(&arp_lock, s)
#define ARPLOCK_ISLOCKED()	(spinlock_islocked(&arp_lock))
#else
#define ARP_LOCK(s)		s = splimp()
#define ARP_UNLOCK(s)		splx(s)
#define ARPLOCK_ISLOCKED()	(1)
#endif /* _MP_NETLOCKS */

#define	ARPTAB_HASH(a) \
	((u_long)(a) % ARPTAB_NB)

#define	ARPTAB_LOOK(at,addr) { \
	register n; \
	ASSERT(ARPLOCK_ISLOCKED()); \
	at = &arptab[ARPTAB_HASH(addr) * ARPTAB_BSIZ]; \
	for (n = 0 ; n < ARPTAB_BSIZ ; n++,at++) \
		if (at->at_iaddr.s_addr == addr) \
			break; \
	if (n >= ARPTAB_BSIZ) \
		at = 0; \
}

/*
 * Source Route table:
 * This table is dynamically allocated, since it is used only by token
 * ring drivers.
 */
struct	sritab	*sritab;

#define	SRITAB_HASH(a)	((u_long)((((u_short*)a)[1]<<16)+((u_short*)a)[2])%ARPTAB_NB)

#define	SRITAB_LOOK(st,addr,arpflg) { \
	register u_int n; \
	st = &sritab[SRITAB_HASH(addr) * ARPTAB_BSIZ]; \
	for (n = 0 ; n < ARPTAB_BSIZ ; n++,st++) { \
		if ((arpflg && (st->st_flags & SRIF_NONARP)) || \
		    (!arpflg && (st->st_flags & SRIF_ARP))) \
			continue; \
		if (!bcmp(st->st_addr,addr,sizeof(st->st_addr))) \
			break; \
	} \
	if (n >= ARPTAB_BSIZ) \
		st = 0; \
}

static void in_arpinput(struct arpcom*, struct mbuf*, u_short*);
extern struct sritab* srifind(char*, u_char);
static int sriset(char*, u_short*, u_char, u_char);
static struct sritab* srinew(char*, u_char);
static void sritfree(char*, u_char);
extern void sritdelete(struct sritab*);

/* timer values */
#define	ARPT_AGE    1			/* aging timer 1 sec */
#define	ARPT_HOLD   (1/ARPT_AGE)	/* hold mbufs 1 sec */
#define	ARPT_KILLI  ((3*60)/ARPT_AGE)	/* hold incomplete entry 3 min */
#define	ARPT_KILLC  (20/ARPT_AGE)	/* hold complete entry 20 min */
#if ARPT_KILLI > 255
? error ? /* ARPT_KILLI must fit in a byte */
#endif

/*
 * The ether_copy() macro casts elements of these arrays to short, so
 * we must enforce proper alignment to avoid read address error exceptions.
 */
union ebcast_align ebcast_align = { 0xff,0xff,0xff,0xff,0xff,0xff };

/* MULTICAST */
ushort ether_ipmulticast_min[3] = { 0x0100, 0x5e00, 0x0000 };
ushort ether_ipmulticast_max[3] = { 0x0100, 0x5e7f, 0xffff };

extern struct ifnet loif;
extern int useloopback;		/* use loopback interface for local traffic */

#define	SCREAM_TIME (15/ARPT_AGE)
static struct arpcom *scream_arpcom = 0;
static struct in_addr scream_iaddr;
static u_char scream_sha[6];
static int scream_timer;

/*
 * Timeout routine.  Age arp_tab entries.
 */
void
arptimer(__psint_t all)
{
	register struct arptab *at;
	register i;
	struct arpcom *ac;
	struct in_addr inaddr;

	/* lower to imp, but no lower to protect the values in the loop.
	 *	Since splimp() refuses to go lower, pass thru level 1.
	 */
	int s1 = spl1();
	int s2;

	ARP_LOCK(s2);

	timeout( arptimer, (caddr_t)((all+1) % (60/ARPT_AGE)), ARPT_AGE*hz);

	/* if duplicated IP address detected, defend aggressively */
	if (ac = scream_arpcom) {
		int s;

		inaddr = scream_iaddr;
		ARP_UNLOCK(s2);
		IFNET_LOCK(&ac->ac_if, s);
		arpwhohas(ac, &inaddr);
		IFNET_UNLOCK(&ac->ac_if, s);
		ARP_LOCK(s2);
		if (++scream_timer >= SCREAM_TIME*2)
			scream_arpcom = 0;
	}

	at = &arptab[0];
	for (i = 0; i < ARPTAB_SIZE; i++, at++) {
		if (at->at_flags == 0 || (at->at_flags & ATF_PERM))
			continue;

		/* age complete entries once a minute */
		if (at->at_flags & ATF_COM) {
			if (all == 0 && ++at->at_timer > ARPT_KILLC)
				arptfree(at);
			continue;
		}

		/* release mbufs after a short time so NFS will not hang */
		if (at->at_hold != 0 && at->at_timer >= ARPT_HOLD) {
			m_freem(at->at_hold);
			at->at_hold = 0;
		}

		if (++at->at_timer >= ARPT_KILLI)
			arptfree(at);
		if (sritab)
		    ++(sritab[i].st_timer);
	}
	ARP_UNLOCK(s2);
	splx(s1);
}

/*
 * Broadcast an ARP packet, asking who has addr on interface ac.
 */
void
arpwhohas(register struct arpcom *ac, struct in_addr *addr)
{
	register struct mbuf *m;
	register struct ether_header *eh;
	register struct ether_arp *ea;
	struct sockaddr sa;

	ASSERT(IFNET_ISLOCKED(&ac->ac_if));
	if ((m = m_get(M_DONTWAIT, MT_DATA)) == NULL)
		return;
	m->m_len = sizeof *ea;
	m->m_off = MMAXOFF - m->m_len;
	ea = mtod(m, struct ether_arp *);
	eh = (struct ether_header *)sa.sa_data;
	bzero((caddr_t)ea, sizeof (*ea));
	ether_copy(etherbroadcastaddr, eh->ether_dhost);
	eh->ether_type = ETHERTYPE_ARP;		/* if_output will swap */
	if (ac->ac_if.if_type == IFT_ISO88025)
		ea->arp_hrd = htons(ARPHRD_802);
	else
		ea->arp_hrd = htons(ARPHRD_ETHER);
	ea->arp_pro = htons(ETHERTYPE_IP);
	ea->arp_hln = sizeof(ea->arp_sha);	/* hardware address length */
	ea->arp_pln = sizeof(ea->arp_spa);	/* protocol address length */
	ea->arp_op = htons(ARPOP_REQUEST);
	ether_copy(ac->ac_enaddr, ea->arp_sha);
	IP_copy(&ac->ac_ipaddr, ea->arp_spa);
	IP_copy(addr, ea->arp_tpa);
	sa.sa_family = AF_UNSPEC;
	(*ac->ac_if.if_output)(&ac->ac_if, m, &sa);
}

/*
 * Resolve an IP address into an ethernet address.  If success, 
 * desten is filled in.  If there is no entry in arptab,
 * set one up and broadcast a request for the IP address.
 * Hold onto this mbuf and resend it once the address
 * is finally resolved.  A return value of 1 indicates
 * that desten has been filled in and the packet should be sent
 * normally; a 0 return indicates that the packet has been
 * taken over here, either now or for later transmission.
 *
 * We do some (conservative) locking here at splimp, since
 * arptab is also altered from input interrupt service (ecintr/ilintr
 * calls arpinput when ETHERTYPE_ARP packets come in).
 */
int
ip_arpresolve(struct arpcom *ac,
	struct mbuf *m,
	struct in_addr *destip,
	u_char *desten)
{
	struct arptab *at;
	struct sockaddr_in sin;
	register struct in_ifaddr *ia;
	int s;

	if (IN_MULTICAST(ntohl(destip->s_addr))) {
	        if (ac->ac_if.if_type == IFT_ISO88025){
			TOKEN_MAP_IP_MULTICAST(destip, desten);
		}
		else {
			ETHER_MAP_IP_MULTICAST(destip, desten);
		}
		return(1);
	} else if (m->m_flags & M_BCAST) {
		ether_copy(etherbroadcastaddr, desten);
		return (1);
	}

	/* if for us, use software loopback driver if up */
	ia = (struct in_ifaddr *)hash_lookup(&hashinfo_inaddr,
		in_arp_match,
		(caddr_t)destip,
		(caddr_t)&ac->ac_if,
		(caddr_t)0);
	if (ia) {
		/*
		 * This test used to be
		 *	if (loif.if_flags & IFF_UP)
		 * It allowed local traffic to be forced
		 * through the hardware by configuring the loopback down.
		 * However, it causes problems during network configuration
		 * for boards that can't receive packets they send.
		 * It is now necessary to clear "useloopback"
		 * to force traffic out to the hardware.
		 */
		if (useloopback) {
			sin.sin_family = AF_INET;
			sin.sin_addr = *destip;
			(void) looutput(&loif, m, (struct sockaddr *)&sin);
			/*
			 * The packet has already been sent and freed.
			 */
			return (0);
		} else {
			ether_copy(ac->ac_enaddr, desten);
			return (1);
		}
	}
	ARP_LOCK(s);
	ARPTAB_LOOK(at, destip->s_addr);
	if (at == 0) {			/* not found */
		at = arptnew(destip);
		ASSERT(at != 0);
		at->at_hold = m;
		ARP_UNLOCK(s);
		arpwhohas(ac, destip);
		return (0);
	}
	at->at_timer = 0;		/* restart the timer */
	if (at->at_flags & ATF_COM) {	/* entry IS complete */
		ether_copy(at->at_enaddr, desten);
		ARP_UNLOCK(s);
		return (1);
	}
	/*
	 * There is an arptab entry, but no ethernet address
	 * response yet.  Replace the held mbuf with this
	 * latest one.
	 */
	if (at->at_hold)
		m_freem(at->at_hold);
	at->at_hold = m;
	ARP_UNLOCK(s);
	arpwhohas(ac, destip);		/* ask again */
	return (0);
}

/*
 * Called from 10 Mb/s Ethernet interrupt handlers
 * when ether packet type ETHERTYPE_ARP
 * is received.  Common length and type checks are done here,
 * then the protocol-specific routine is called.
 */
void
arpinput(struct arpcom *ac, struct mbuf *m)
{
	register struct arphdr *ar;

	M_SKIPIFHEADER(m);
	if (m->m_len < sizeof(struct arphdr))
		goto out;
	ar = mtod(m, struct arphdr *);
	if (ntohs(ar->ar_hrd) != ARPHRD_ETHER
	    && ntohs(ar->ar_hrd) != ARPHRD_802)
		goto out;
	if (m->m_len < sizeof(struct arphdr) + 2 * ar->ar_hln + 2 * ar->ar_pln)
		goto out;

	switch (ntohs(ar->ar_pro)) {

	case ETHERTYPE_IP:
		in_arpinput(ac, m, 0);
		return;

	default:
		break;
	}
out:
	m_freem(m);
}

/* SRCROUTE */
void
srarpinput(struct arpcom *ac, struct mbuf *m, u_short *sri)
{
	register struct arphdr *ar;

	M_SKIPIFHEADER(m);

	if (m->m_len < sizeof(struct arphdr))
		goto out;
	ar = mtod(m, struct arphdr *);

	/* O.K. Let's be a little more tolarent */
	if (ntohs(ar->ar_hrd) != ARPHRD_ETHER
	    && ntohs(ar->ar_hrd) != ARPHRD_802)
		goto out;

	if (m->m_len < sizeof(struct arphdr) + 2 * ar->ar_hln + 2 * ar->ar_pln)
		goto out;

	switch (ntohs(ar->ar_pro)) {
	case ETHERTYPE_IP:
		in_arpinput(ac, m, sri);
		return;
	default:
		break;
	}
out:
	m_freem(m);
}

/* ARGSUSED */
int
in_arp_match(struct hashbucket *h, caddr_t key, caddr_t arg1, caddr_t arg2)
{
	struct in_ifaddr *ia = (struct in_ifaddr *)h;
	struct in_addr *addr = (struct in_addr *)key;
	struct ifnet *ifp = (struct ifnet *)arg1;
	int match = 0;

	if (h->flags & HTF_INET) {
		if ((ia->ia_ifp == ifp) &&
		 (addr->s_addr == IA_SIN(ia)->sin_addr.s_addr)) {
			match = 1;
		}
	}
	return match;
}

/*
 * ARP for Internet protocols on 10 Mb/s Ethernet.
 * Algorithm is that given in RFC 826.
 * In addition, a sanity check is performed on the sender
 * protocol address, to catch impersonators.
 */
static void
in_arpinput(register struct arpcom *ac, struct mbuf *m, u_short *sri)
{
	register struct ether_arp *ea;
	struct ether_header *eh;
	register struct arptab *at;  /* same as "merge" flag */
	register struct in_ifaddr *ia;
	struct sockaddr_in sin;
	struct sockaddr sa;
	struct in_addr isaddr, itaddr, myaddr;
	int op, s;

	ASSERT(IFNET_ISLOCKED(&ac->ac_if));
	myaddr = ac->ac_ipaddr;
	ea = mtod(m, struct ether_arp *);
	op = ntohs(ea->arp_op);
	IP_copy(ea->arp_spa, &isaddr);
	IP_copy(ea->arp_tpa, &itaddr);

	ia = (struct in_ifaddr *)hash_lookup(&hashinfo_inaddr,
		in_arp_match, (caddr_t)&itaddr,
		(caddr_t)&ac->ac_if, (caddr_t)0);

	if (ia == 0) {
		/* failed so now check if it matches source address */

		ia = (struct in_ifaddr *)hash_lookup(&hashinfo_inaddr,
			in_arp_match, (caddr_t)&isaddr,
			(caddr_t)&ac->ac_if, (caddr_t)0);
	
		if (ia == 0) { /* double failure so find first matching i/f */

			ia = (struct in_ifaddr *)hash_enum(&hashinfo_inaddr,
					ifnet_enummatch,
					HTF_INET, (caddr_t)0,
					(caddr_t)&ac->ac_if, (caddr_t)0);

			if (ia == 0) { /* big dog-doo now so quit */
				goto out;
			}
		}
	}
	myaddr = IA_SIN(ia)->sin_addr;

	if (!bcmp((caddr_t)ea->arp_sha, (caddr_t)ac->ac_enaddr,
	  sizeof (ea->arp_sha)))
		goto out;	/* it's from me, ignore it. */
	if (!bcmp((caddr_t)ea->arp_sha, (caddr_t)etherbroadcastaddr,
	    sizeof (ea->arp_sha))) {
		printf(
		    "arp: ether address is broadcast for IP address %s!\n",
		    inet_ntoa(isaddr));
		goto out;
	}
	if (isaddr.s_addr == myaddr.s_addr) {
		ARP_LOCK(s);
		if (0 != scream_arpcom) {
			if (scream_timer >= SCREAM_TIME
			    && scream_arpcom == ac) {
				scream_timer = 0;
				ARP_UNLOCK(s);
				cmn_err(CE_ALERT,
					"arp: host with MAC address %s"
					" is still using my IP address %s!\n",
					ether_sprintf(&scream_sha[0]),
					inet_ntoa(scream_iaddr));
			} else {
				ARP_UNLOCK(s);
			}
		} else {
			ether_copy(ea->arp_sha, &scream_sha[0]);
			scream_iaddr.s_addr = myaddr.s_addr;
			scream_arpcom = ac;
			scream_timer = 0;
			ARP_UNLOCK(s);
			cmn_err(CE_ALERT, "arp: host with MAC address %s"
				" is using my IP address %s!\n",
				ether_sprintf(&scream_sha[0]),
				inet_ntoa(scream_iaddr));
		}
		itaddr = myaddr;
		if (op == ARPOP_REQUEST)
			goto reply;
		goto out;
	}
	ARP_LOCK(s);
	ARPTAB_LOOK(at, isaddr.s_addr);
	if (at) {
		ether_copy(ea->arp_sha, at->at_enaddr);
		at->at_flags |= ATF_COM;

		/* SRCROUTE */
		if (sri != 0)
			(void)sriset((char*)&ea->arp_sha[0], sri, 1,0);

		if (at->at_hold) {
			struct mbuf *om;

			sin.sin_family = AF_INET;
			sin.sin_addr = isaddr;
			om = at->at_hold;
			at->at_hold = 0;
			ARP_UNLOCK(s);
			(*ac->ac_if.if_output)(&ac->ac_if, 
			    om, (struct sockaddr *)&sin);
			goto reply;
		}
	}
	if (at == 0 && itaddr.s_addr == myaddr.s_addr) {
		/* ensure we have a table entry */
		if (at = arptnew(&isaddr)) {
			ether_copy(ea->arp_sha, at->at_enaddr);
			at->at_flags |= ATF_COM;

			/* SRCROUTE */
			if (sri != 0)
				(void)sriset((char*)&ea->arp_sha[0], sri, 1,1);
		}
	}
	ARP_UNLOCK(s);
reply:
	/*
	 * Reply if this is an IP request,
	 */
	if (op != ARPOP_REQUEST)
	    goto out;

	if (itaddr.s_addr == myaddr.s_addr) {
		/* I am the target */
		ether_copy(ea->arp_sha, ea->arp_tha);
		ether_copy(ac->ac_enaddr, ea->arp_sha);
	} else {
		ARP_LOCK(s);
		ARPTAB_LOOK(at, itaddr.s_addr);
		if (at == NULL || (at->at_flags & ATF_PUBL) == 0) {
			ARP_UNLOCK(s);
			goto out;
		}
		ether_copy(ea->arp_sha, ea->arp_tha);
		ether_copy(at->at_enaddr, ea->arp_sha);
		ARP_UNLOCK(s);
	}

	IP_copy(ea->arp_spa, ea->arp_tpa);
	IP_copy(&itaddr, ea->arp_spa);
	ea->arp_op = htons(ARPOP_REPLY); 
	eh = (struct ether_header *)sa.sa_data;
	ether_copy(ea->arp_tha, eh->ether_dhost);
	eh->ether_type = ETHERTYPE_ARP;
	sa.sa_family = AF_UNSPEC;
	(*ac->ac_if.if_output)(&ac->ac_if, m, &sa);
	return;
out:
	m_freem(m);
	return;
}

/*
 * Free an arptab entry.
 */
void
arptfree(register struct arptab *at)
{
	int s = splimp();
	ASSERT(ARPLOCK_ISLOCKED());
	sritfree((char*)&at->at_enaddr[0], 1);
	if (at->at_hold)
		m_freem(at->at_hold);
	at->at_hold = 0;
	at->at_timer = at->at_flags = 0;
	at->at_iaddr.s_addr = 0;
	splx(s);
}

/*
 * Enter a new address in arptab, pushing out the oldest entry 
 * from the bucket if there is no room.
 * This always succeeds since no bucket can be completely filled
 * with permanent entries (except from arpioctl when testing whether
 * another permanent entry will fit).
 * MUST BE CALLED AT SPLIMP.
 */
struct arptab *
arptnew(struct in_addr *addr)
{
	register n;
	int oldest = -1;
	register struct arptab *at, *ato = NULL;
	static int first = 1;

	ASSERT(ARPLOCK_ISLOCKED());
	if (first) {
		first = 0;
		timeout(arptimer, (caddr_t)0, hz);
	}
	at = &arptab[ARPTAB_HASH(addr->s_addr) * ARPTAB_BSIZ];
	for (n = 0; n < ARPTAB_BSIZ; n++,at++) {
		if (at->at_flags == 0)
			goto out;	 /* found an empty entry */
		if (at->at_flags & ATF_PERM)
			continue;
		if ((int) at->at_timer > oldest) {
			oldest = at->at_timer;
			ato = at;
		}
	}
	if (ato == NULL)
		return (NULL);
	at = ato;
	arptfree(at);
out:
	at->at_iaddr = *addr;
	at->at_flags = ATF_INUSE;
	return (at);
}

arpioctl(int cmd, caddr_t data)
{
	register struct arpreq *ar = (struct arpreq *)data;
	register struct arptab *at;
	register struct sockaddr_in *sin;
	int s;

	if (ar->arp_pa.sa_family != AF_INET ||
	    ar->arp_ha.sa_family != AF_UNSPEC)
		return (EAFNOSUPPORT);

	sin = (struct sockaddr_in *)&ar->arp_pa;
	ARP_LOCK(s);
	ARPTAB_LOOK(at, sin->sin_addr.s_addr);

	if (at == NULL) { /* not found */
		if (cmd != SIOCSARP) {
			ARP_UNLOCK(s);
			return (ENXIO);
		}
		if (ifa_ifwithnet(&ar->arp_pa) == NULL) {
			ARP_UNLOCK(s);
			return (ENETUNREACH);
		}
	}
	switch (cmd) {

	case SIOCSARP:		/* set entry */
	case SIOCSARPX:	{	/* Ext set entry */
		u_char newarp = 0;
		if (at == NULL) {
			at = arptnew(&sin->sin_addr);
			if (at == NULL) {
				ARP_UNLOCK(s);
				return (EADDRNOTAVAIL);
			}
			if (ar->arp_flags & ATF_PERM) {
			/* never make all entries in a bucket permanent */
				register struct arptab *tat;

				/*
				 * fix for 366431; avoid panic by ensuring
				 * that we don't return two pointers to
				 * the same bucket entry.
				 * We set the perm. flag so that the second
				 * arptnew() won't succeed.
				 */
				at->at_flags = (ATF_PERM|ATF_INUSE);
				/* try to re-allocate */
				tat = arptnew(&sin->sin_addr);
				if (tat == NULL) {
					arptfree(at);
					ARP_UNLOCK(s);
					return (EADDRNOTAVAIL);
				}
				arptfree(tat);
			}
			newarp = 1;
		}
		bcopy((caddr_t)ar->arp_ha.sa_data, (caddr_t)at->at_enaddr,
		    sizeof(at->at_enaddr));

		/* SRCROUTE */
		if (cmd == SIOCSARPX) {
			struct sritab *st;
			struct arpreqx *arx = (struct arpreqx *)ar;

			if (sriset((char*)&at->at_enaddr[0],arx->arp_sri,1,
			    newarp)) {
				arptfree(at);
				ARP_UNLOCK(s);
				return(EADDRNOTAVAIL);
			}
			if ((st = srifind((char*)&at->at_enaddr[0],1)) == 0) {
				arptfree(at);
				ARP_UNLOCK(s);
				return(EADDRNOTAVAIL);
			}
			st->st_flags |= SRIF_PERM;
		}
		/* SRCROUTE */
		at->at_flags = ATF_COM | ATF_INUSE |
			(ar->arp_flags & (ATF_PERM|ATF_PUBL));
		at->at_timer = 0;
		}
		break;

	/* SRCROUTE */
	case SIOCDARPX:		/* Ext delete entry */
		sritfree((char*)&at->at_enaddr[0], 1);
		/* FALL THRU... */

	case SIOCDARP:		/* delete entry */
		arptfree(at);
		break;

	case SIOCGARP:		/* get entry */
	case SIOCGARPX:		/* Ext get entry */
		bcopy((caddr_t)at->at_enaddr, (caddr_t)ar->arp_ha.sa_data,
		    sizeof(at->at_enaddr));
		ar->arp_flags = at->at_flags;

		/* SRCROUTE */
		if (cmd == SIOCGARPX) {
			struct sritab *st =srifind((char*)&at->at_enaddr[0],1);
			struct arpreqx *arx = (struct arpreqx *)ar;

			arx->arp_sri[0] = 0;
			if (!st)
				break;
			bcopy((caddr_t)&st->st_ri[0],(caddr_t)&arx->arp_sri[0],
				sizeof(st->st_ri));
		}
		break;
	}
	ARP_UNLOCK(s);
	return (0);
}

/*
 * Convert Ethernet address to printable (loggable) representation.
 */
char *
ether_sprintf(register u_char *ap)
{
	register i;
	static char etherbuf[18];
	register char *cp = etherbuf;
	static char digits[] = "0123456789abcdef";

	for (i = 0; i < 6; i++) {
		*cp++ = digits[*ap >> 4];
		*cp++ = digits[*ap++ & 0xf];
		*cp++ = ':';
	}
	*--cp = 0;
	return (etherbuf);
}

/*
 * Convert an Internet multicast socket address into an ethernet address.
 */
ether_cvtmulti(register struct sockaddr *sa, int *all)
{
    switch (sa->sa_family) {
	case AF_UNSPEC:
	    break;

	case AF_INET:
	    if (((struct sockaddr_in *)sa)->sin_addr.s_addr == INADDR_ANY) {
		/*
		 * An IP address of INADDR_ANY means listen to all
		 * of the Ethernet multicast addresses used for IP.
		 * (This is for the sake of IP multicast routers.)
		 */
		*all = 1;
		return 0;
	    } else {
		struct in_addr ia;
		ia = ((struct sockaddr_in *)sa)->sin_addr;
		ETHER_MAP_IP_MULTICAST(&ia, sa->sa_data);
		sa->sa_family = AF_UNSPEC;
	    }
	    break;

	case AF_RAW:
	    sa->sa_family = AF_UNSPEC;
	    break;

	default:
	    return EAFNOSUPPORT;
    }

    /*
     * Verify that we have a valid Ethernet multicast addresses.
     */
    if ((sa->sa_data[0] & 1) != 1) {
	    return EINVAL;
    }
    *all = 0;
    return(0);
}

/* SRCROUTE */
/*
 * This routine should be called within splimp.
 * Called by delete ioctl and when corresponding arp entry is deleted.
 */
static void
sritfree(char *addr, u_char arpflg)
{
	struct sritab *st;

	if (!sritab) 
	    return;

	SRITAB_LOOK(st,addr,arpflg);
	if (st)
		if (st->st_ref_cnt > 0)
			st->st_ref_cnt--;
		else
			bzero((char*)st, sizeof(*st));
}

/*
 * flushing sri/arp table is expensive.
 * Arptimo should compensate to minimize sri stealing.
 */
void
sritdelete(struct sritab *st)
{
	/* XXX: should call sritfree(), but for speed.... */
	bzero((char*)st, sizeof(*st));
}

/*
 * Allocate new src route table.
 * Entries used by arp and none-arp are mutually excluded.
 */
static struct sritab *
srinew(char *addr, u_char arpflg)
{
	register n;
	int oldest = -1;
	register struct sritab *st, *sto = NULL;

	st = &sritab[SRITAB_HASH(addr)*ARPTAB_BSIZ];
	for (n = 0; n < ARPTAB_BSIZ; n++,st++) {
		if ((st->st_flags&SRIF_VALID) == 0)
			goto out;	 /* found an empty entry */
		if (st->st_flags&SRIF_PERM)
			continue;
		if ((arpflg && (st->st_flags & SRIF_NONARP)) ||
		    (!arpflg && (st->st_flags & SRIF_ARP)))
			continue;
		if (st->st_ref_cnt)     /* this is shared by many */
			continue;
		if ((int) st->st_timer > oldest) {
			oldest = st->st_timer;
			sto = st;
		}
	}
	if (sto == NULL)
		return (NULL);
	st = sto;

	/* If stealing a sri, then arptab also must be freed */
	sritdelete(st);
out:
	bcopy(addr, (char*)&st->st_addr[0], sizeof(st->st_addr));
	st->st_flags = (arpflg == 0) ? SRIF_NONARP : SRIF_ARP;
	return(st);
}

/*
 * This routine is call by 802 drivers to fill route info into MAC header.
 */
struct sritab *
srifind(char *addr, u_char arpflg)
{
	struct sritab *st;

	if (!sritab) {
		sritab = kmem_zalloc(ARPTAB_SIZE * sizeof (struct sritab), KM_NOSLEEP);
		if (!sritab)
		    return(NULL);
	}

	SRITAB_LOOK(st,addr,arpflg);
	if (!st)
		return(NULL);
	if ((st->st_flags & SRIF_VALID) == 0)
		return(NULL);
	st->st_timer = 0;		/* reset the timer */
	return(st);
}

u_short *
srilook(char *addr, u_char arpflg)
{
	struct sritab *st;

	st = srifind(addr, arpflg);
	if (!st)
		return(NULL);
	return(&st->st_ri[0]);
}

/*
 * This routine is called only by in_arpinput and set ioctl.
 */
static int
sriset(char*addr, u_short *ri, u_char arpflg, u_char newarp)
{
	u_long len;
	u_short *ori;
	struct sritab *st;
	register u_long n;

	if (!sritab) {
		sritab = kmem_zalloc(ARPTAB_SIZE * sizeof (struct sritab), KM_NOSLEEP);
		if (!sritab)
		    return(-1);
	}

	st = &sritab[SRITAB_HASH(addr) * ARPTAB_BSIZ];
	for (n = 0 ; n < ARPTAB_BSIZ ; n++,st++) {
		if ((arpflg && (st->st_flags & SRIF_NONARP)) ||
		   (!arpflg && (st->st_flags & SRIF_ARP)))
			continue;
		if (!bcmp(st->st_addr,addr,sizeof(st->st_addr)))
			break;
		}
	if (n >= ARPTAB_BSIZ) {
		st = srinew(addr, arpflg);
		if (!st)
			return(-1);
	} else {
		if (newarp) {
			st->st_ref_cnt++;
		}
	}
	/* Take latest no matter what because the shortest but old route
	 * maybe dead by now.
	 */
	ori = &st->st_ri[0];
#ifdef STASH
	if ((st->st_flags & SRIF_VALID) &&
	   ((ori[0]&SRI_LENMASK) <= (ri[0]&SRI_LENMASK)))
		return(0);
#endif
	len = ((ri[0]&SRI_LENMASK) >> 8);
	bcopy((char*)ri, (char*)ori, len);
	if (arpflg)
		st->st_flags |= (SRIF_VALID|SRIF_ARP);
	else
		st->st_flags |= (SRIF_VALID|SRIF_NONARP);
	return(0);
}
