#ifndef __NETINET_IN_H__
#define __NETINET_IN_H__
/*
 * Copyright (c) 1982, 1986, 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution is only permitted until one year after the first shipment
 * of 4.4BSD by the Regents.  Otherwise, redistribution and use in source and
 * binary forms are permitted provided that: (1) source distributions retain
 * this entire copyright notice and comment, and (2) distributions including
 * binaries display the following acknowledgement:  This product includes
 * software developed by the University of California, Berkeley and its
 * contributors'' in the documentation or other materials provided with the
 * distribution and in all advertising materials mentioning features or use
 * of this software.  Neither the name of the University nor the names of
 * its contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *	@(#)in.h	7.10 (Berkeley) 6/28/90 plus MULTICAST
 */

#include <sgidefs.h>
#include <sys/endian.h>
#include <sys/bsd_types.h>

/*
 * Constants and structures defined by the internet system,
 * Per RFC 790, September 1981.
 */

/*
 * Protocols
 */
#define	IPPROTO_IP		0		/* dummy for IP */
#define	IPPROTO_ICMP		1		/* control message protocol */
#define	IPPROTO_IGMP		2		/* group mgmt protocol */
#define	IPPROTO_GGP		3		/* gateway^2 (deprecated) */
#define IPPROTO_IPIP		4		/* (multicast) encapsulation */
#define IPPROTO_ENCAP IPPROTO_IPIP		/* old name */
#define IPPROTO_ST		5		/* STream protocol */
#define	IPPROTO_TCP		6		/* tcp */
#define	IPPROTO_EGP		8		/* exterior gateway protocol */
#define	IPPROTO_PUP		12		/* pup */
#define	IPPROTO_UDP		17		/* user datagram protocol */
#define	IPPROTO_IDP		22		/* xns idp */
#define	IPPROTO_TP		29 		/* tp-4 w/ class negotiation */
#define	IPPROTO_XTP		36		/* XTP */
#define IPPROTO_RSVP		46		/* resource reservation proto */
#define IPPROTO_HELLO		63	/* depricated "hello" routing protocol */
#define IPPROTO_ND		77	/* depricated net disk proto */
#define	IPPROTO_EON		80		/* ISO clnp */
#define IPPROTO_OSPF		89		/* Open SPF routing */
#define IPPROTO_SWIPE		94		/* swIPe protocol */

#define	IPPROTO_RAW		255		/* raw IP packet */
#define	IPPROTO_MAX		256


/*
 * Local port number conventions:
 * Ports < IPPORT_RESERVED are reserved for
 * privileged processes (e.g. root).
 * Ports > 5000 used to be reserved for non-root servers,
 * but this convention can no longer be used with large scale systems.
 * All well-known ports should be registered with the Internet Assigned
 * Numbers Authority.
 */
#define	IPPORT_RESERVED		1024
#define IPPORT_MAXPORT		65535

/*
 * Internet address (a structure for historical reasons)
 */
struct in_addr {
	__uint32_t s_addr;
};

/*
 * Definitions of bits in internet address integers.
 * On subnets, the decomposition of addresses to host and net parts
 * is done according to subnet mask, not the masks here.
 */
#define	IN_CLASSA(i)		(((__int32_t)(i) & 0x80000000) == 0)
#define	IN_CLASSA_NET		0xff000000
#define	IN_CLASSA_NSHIFT	24
#define	IN_CLASSA_HOST		0x00ffffff
#define	IN_CLASSA_MAX		128

#define	IN_CLASSB(i)		(((__int32_t)(i) & 0xc0000000) == 0x80000000)
#define	IN_CLASSB_NET		0xffff0000
#define	IN_CLASSB_NSHIFT	16
#define	IN_CLASSB_HOST		0x0000ffff
#define	IN_CLASSB_MAX		65536

#define	IN_CLASSC(i)		(((__int32_t)(i) & 0xe0000000) == 0xc0000000)
#define	IN_CLASSC_NET		0xffffff00
#define	IN_CLASSC_NSHIFT	8
#define	IN_CLASSC_HOST		0x000000ff

#define	IN_CLASSD(i)		(((__int32_t)(i) & 0xf0000000) == 0xe0000000)
#define	IN_CLASSD_NET		0xf0000000	/* These ones aren't really */
#define	IN_CLASSD_NSHIFT	28		/* net and host fields, but */
#define	IN_CLASSD_HOST		0x0fffffff	/* routing needn't know.    */
#define	IN_MULTICAST(i)		IN_CLASSD(i)

#define	IN_EXPERIMENTAL(i)	(((__int32_t)(i) & 0xf0000000) == 0xf0000000)
#define	IN_BADCLASS(i)		(((__int32_t)(i) & 0xf0000000) == 0xf0000000)

#define	INADDR_ANY		(__uint32_t)0x00000000
#define	INADDR_BROADCAST	(__uint32_t)0xffffffff	/* must be masked */
#define	INADDR_LOOPBACK		(__uint32_t)0x7F000001

/* Multicast */
#define	INADDR_UNSPEC_GROUP	(__uint32_t)0xe0000000	/* 224.0.0.0   */
#define	INADDR_ALLHOSTS_GROUP	(__uint32_t)0xe0000001	/* 224.0.0.1   */
#define INADDR_ALLROUTERS_GROUP (__uint32_t)0xe0000002  /* 224.0.0.2   */
#define	INADDR_MAX_LOCAL_GROUP 	(__uint32_t)0xe00000ff	/* 224.0.0.255 */

#ifndef _KERNEL
#define	INADDR_NONE		0xffffffff		/* -1 return */
#endif

#define	IN_LOOPBACKNET		127			/* official! */

/*
 * Socket address, internet style.
 */
struct sockaddr_in {
#undef _HAVE_SIN_LEN			/* marker for some places to change */
#ifdef _HAVE_SIN_LEN			/* when we have sin_len */
	u_char	sin_len;
	u_char	sin_family;
#else
	short	sin_family;
#endif
	u_short	sin_port;
	struct	in_addr sin_addr;
	char	sin_zero[8];
};
#ifndef _HAVE_SIN_LEN
struct sockaddr_in_new {
	u_char	sin_len;
	u_char	sin_family;
	u_short	sin_port;
	struct	in_addr sin_addr;
	char	sin_zero[8];
};
#endif

/*
 * Structure used to describe IP options.
 * Used to store options internally, to pass them to a process,
 * or to restore options retrieved earlier.
 * The ip_dst is used for the first-hop gateway when using a source route
 * (this gets put into the header proper).
 */
struct ip_opts {
	struct	in_addr ip_dst;		/* first hop, 0 w/o src rt */
	char	ip_opts[40];		/* actually variable in size */
};

/*
 * Options for use with [gs]etsockopt at the IP level.
 * First word of comment is data type; bool is stored in int.
 */

#define	IP_OPTIONS	    1	/* buf/ip_opts; set/get IP per-packet options */
#define	IP_HDRINCL	    2	/* int; header is included with data (raw) */
#define	IP_TOS		    3	/* int; IP type of service and precedence */
#define	IP_TTL		    4	/* int; IP time to live */
#define	IP_RECVOPTS	    5 	/* bool; receive all IP options w/datagram */
#define	IP_RECVRETOPTS	    6 	/* bool; receive IP options for response */
#define	IP_RECVDSTADDR	    7 	/* bool; receive IP dst addr w/datagram */
#define	IP_RETOPTS	    8 	/* ip_opts; set/get IP per-packet options */
#define	IP_MULTICAST_IF	   20	/* in_addr; set/get IP multicast interface */
#define	IP_MULTICAST_TTL   21	/* u_char;  set/get IP multicast timetolive */
#define	IP_MULTICAST_LOOP  22	/* u_char;  set/get IP multicast loopback */
#define	IP_ADD_MEMBERSHIP  23	/* ip_mreq; add  an IP group membership */
#define	IP_DROP_MEMBERSHIP 24	/* ip_mreq; drop an IP group membership */
#define IP_MULTICAST_VIF   25	/* int; set/get IP mcast vir. interface */
#define IP_RSVP_VIF_ON	   26	/* int: set rsvp per-vif socket */
#define IP_RSVP_VIF_OFF	   27	/* int: unset rsvp per-vif socket */
#define IP_RSVP_ON         28	/* void; enable rsvp in kernel */
#define	IP_SENDSRCADDR     36 	/* in_addr; send from specified address */

/* End of [gs]etsockopt options */

#define	IP_DEFAULT_MULTICAST_TTL   1	/* normally limit m'casts to 1 hop  */
#define	IP_DEFAULT_MULTICAST_LOOP  1	/* normally hear sends if a member  */
#define	IP_MAX_MEMBERSHIPS         20	/* per socket; must fit in one mbuf */

/*
 * Argument structure for IP_ADD_MEMBERSHIP and IP_DROP_MEMBERSHIP.
 */
struct ip_mreq {
	struct in_addr	imr_multiaddr;	/* IP multicast address of group */
	struct in_addr	imr_interface;	/* local IP address of interface */
};

#ifdef _KERNEL
extern	struct domain inetdomain;
extern	struct protosw inetsw[];
struct ifnet;				/* forward reference */

extern	int		in_broadcast(struct in_addr, struct ifnet *);
extern	int		in_canforward(struct in_addr);
extern	int		in_localaddr(struct in_addr);
extern	__uint32_t	in_netof(struct in_addr);
extern	__uint32_t	inet_addr(char *);
#endif /* _KERNEL */
#endif /* __NETINET_IN_H__ */
