/*
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution is only permitted until one year after the first shipment
 * of 4.4BSD by the Regents.  Otherwise, redistribution and use in source and
 * binary forms are permitted provided that: (1) source distributions retain
 * this entire copyright notice and comment, and (2) distributions including
 * binaries display the following acknowledgement:  This product includes
 * software developed by the University of California, Berkeley and its
 * contributors'' in the documentation or other materials provided with the
 * distribution and in all advertising materials mentioning features or use
 * of this software.  Neither the name of the University nor the names of
 * its contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *	@(#)in_pcb.c	7.13 (Berkeley) 6/28/90  plus MULTICAST 1.0
 */

#include "tcp-param.h"
#include "sys/param.h"
#include "sys/errno.h"
#include "sys/sema.h"
#include "sys/hashing.h"

#include "sys/mbuf.h"
#include "sys/protosw.h"
#include "sys/socket.h"
#include "sys/socketvar.h"
#include "sys/user.h"
#include "sys/proc.h"
#include "net/soioctl.h"
#include "net/if.h"
#include "net/route.h"
#include "sys/sat.h"
#include "sys/debug.h"
#include "sys/sysmacros.h"
#include "sys/capability.h"
#include "sys/cmn_err.h"

#include "in.h"
#include "in_systm.h"
#include "ip.h"
#include "in_pcb.h"
#include "in_var.h"
#include "ip_var.h"
#include "sys/tcpipstats.h"
#include "sys/kmem.h"
#include "sys/atomic_ops.h"

/*
 * External Procedures
 */
extern struct in_ifaddr *ifatoia(struct ifaddr *);

extern struct ifaddr *ifa_ifwithaddr(struct sockaddr *);
extern struct ifaddr *ifa_ifwithdstaddr(struct sockaddr *);
extern struct ifaddr *ifa_ifwithnet(struct sockaddr *);
extern int ifnet_enummatch(struct hashbucket *, caddr_t, caddr_t, caddr_t);

/*
 * Local macro definitions
 */
#define	SATOSIN(sa)	((struct sockaddr_in *)(sa))
#define SINTOSA(sin)	((struct sockaddr *)(sin))

/*
 * External Global Variables
 */
extern struct inpcb *tcp_last_inpcb;
extern u_char inetctlerrmap[];
extern struct hashinfo hashinfo_inaddr;
extern int in_interfaces;	/* number of non-loopback network interfaces */
extern struct ifnet *ifnet;	/* head of network interface list */
extern int in_ifaddr_count;     /* Number of address'es configured */

/*
 * Module Global Variables
 */
struct	in_addr zeroin_addr;
zone_t *inpcb_zone;

/*
 * Following two defines are taken from bsd/sys/mbuf.h to avoid having to
 * include lots of extra header files
 * XXX
 */
#define MBSTAT_INC(x, c)	\
	atomicAddInt(&(((struct kna *)(private.knaptr))->mbstat.x),c)

#define MBSTAT_DEC(x, c)	\
	atomicAddInt(&(((struct kna *)(private.knaptr))->mbstat.x),-c)

/*
 * Flags to in_pcblookupx
 */
enum LookupPolicy {
	LOOKUP_NO_OP,		/* No operation */
	ALREADY_LOCKED,		/* Already locked */
	GIVE_LOCK_ALWAYS,	/* Always return hash locked */
	DONT_HOLD		/* Don't hold inp on return */
};

/*
 * Flags to in_pcbassign
 */
enum AssignFlags {
	ASSIGN_NO_OP,		/* No operation */
	USE_SECURITY,		/* Use security protocols for assignment */
	DONT_REHASH		/* Don't rehash on changed address */
};

/*
 * Forward Referenced Procedures
 */
u_short in_pcbnextport(struct inpcb *inp);

int in_pcbassign(struct inpcb *, struct in_addr,
		u_short, struct in_addr, u_short,
		int, int, struct inaddrpair *, enum AssignFlags);
int in_pcbsetaddrx(struct inpcb *, struct sockaddr_in *, struct in_addr,
		struct inaddrpair *);

#ifdef _MP_NETLOCKS
struct inpcb *in_pcblookupx(struct inpcb *, struct in_addr, u_short,
	struct in_addr, u_short, int, int, enum LookupPolicy);
#endif /* _MP_NETLOCKS */

u_short in_pcbhashfun(struct inpcb *, struct in_addr,
			u_short, struct in_addr, u_short);

u_short in_pcbcltshashfun(struct inpcb *, struct in_addr,
			u_short, struct in_addr, u_short);

#if defined(_MP_NETLOCKS)
#define IN_PCBLOOKUPX(hd, fa, fp, la, lp, w, lk, a) \
		in_pcblookupx(hd, fa, fp, la, lp, w, lk, a)
#else
#define IN_PCBLOOKUPX(hd, fa, fp, la, lp, w, lk, a) \
		in_pcblookup(hd, fa, fp, la, lp, w)
#endif	/* _MP_NETLOCKS */

#define INPCB_STATINC(head, stat) PCBSTAT(stat, (head)->inp_stats)
#define INPCB_STATDEC(head, stat) PCBSTAT_DEC(stat, (head)->inp_stats)
#define INPCB_STATSET(head, stat, val) PCBSTAT_SET(stat, (head)->inp_stats,val)

/*
 * The use of the two-level cache is intended as a hint to the lookup routines
 * and the protocol provider.  There is no write-locking of the caches.
 *
 * We currently use hash bucket 0 to contain pcbs that are in the
 * process of being created.  Theoretically, we could remove
 * hash bucket 0 from the lists that we use when we're setting
 * up notifications on all sockets.
 */
#define LEADING_BUCKETS		(1)		/* Hash0 */
#define TRAILING_BUCKETS	(0)
#define BUCKET_0		(0)
#define BUCKET_INVAL		((u_short)-1)

#define SPECIAL_BUCKETS		(LEADING_BUCKETS+TRAILING_BUCKETS)

/*
 * The following constant MUST be power of 2
 * The dynamic sizing algorithm is 'INPCB_NUMHASHPERMB' base hash bucket
 * entries for every megabyte of main memory up to the limit of
 * the minimum or maximum for either the UDP or TCP tables.
 */
#define INPCB_NUMHASHPERMB      4
#define INPCB_CACHE           1

int
in_pcb_hashtablesize()
{
	return ((int)((ctob(physmem) / (1024*1024)) * INPCB_NUMHASHPERMB));
}

/* 
 * Initialize the head of the in_pcb lookup structure
 */
void
in_pcbinitcb(struct inpcb *head, int tablesz, u_short flags, u_short stats)
{
	struct in_pcbhead *h;
	int i;
	static need_zone = 1;

	if (need_zone) {
		inpcb_zone = kmem_zone_init(sizeof(*head), "inpcb struct");
		need_zone = 0;
	}

	head->inp_next = head->inp_prev = head->inp_head = head;
	/*
	 * We try to compute a table size based on the requested number.
	 *
	 * It makes our lives easier (and more efficient) if we end
	 * up with a table size based on a power of 2.  This avoids having
	 * to do modulus operations in the hash functions.
	 */
	for (i = 1; i < tablesz; i <<= 1) {
	}
	head->inp_tablesz = i + SPECIAL_BUCKETS;
	head->inp_table = kern_calloc(head->inp_tablesz, sizeof(*h));
	head->inp_dstats = kern_calloc(head->inp_tablesz, sizeof(u_int));

#ifdef INPCB_CACHE
	head->inp_cache= kern_calloc(head->inp_tablesz,sizeof(struct inpcb *));
	bzero (head->inp_cache, head->inp_tablesz * sizeof(struct inpcb *));
#else
	head->inp_cache = (struct inpcb **)0;
#endif

	head->inp_stats = (stats < NUM_PCBSTAT) /* validate */
		? stats   /* use one supplied */
		: UDP_PCBSTAT; /* use datagram if bogus */

	head->inp_hashflags = flags;
	head->inp_last = 0;
	head->inp_refcnt = 1;

	head->inp_hashfun = (flags & INPFLAGS_CLTS)
		? in_pcbcltshashfun	/* Connectionless hash function */
		: in_pcbhashfun;	/* Connection oriented hash function */

	head->inp_wildport = IPPORT_RESERVED;

	INHEAD_INITMRLOCK(head);

	h = &head->inp_table[0];
	for (i = 0; i < head->inp_tablesz; i++) {
		h->hinp_next = h->hinp_prev = (struct inpcb *)h;
		h->hinp_waste = head;
		h->hinp_refcnt = 1;
		h++;
	}
	return;
}

int
in_pcballoc(struct socket *so, struct inpcb *head)
{
	register struct inpcb *inp, *tinp;

	inp = (struct inpcb *)kmem_zone_zalloc(inpcb_zone, KM_NOSLEEP);
	if (inp == NULL)
		return (ENOBUFS);
	MBSTAT_INC(m_mtypes[MT_PCB], 1);
	inp->inp_head = head;
	inp->inp_socket = so;
	inp->inp_hashval = BUCKET_0;
	inp->inp_refcnt = 1;
	INHEAD_MRWLOCK(head);
	tinp = _TBLELEMTAIL(inp);
	insque(inp, tinp);
	inp->inp_hhead = &head->inp_table[inp->inp_hashval];
	head->inp_dstats[inp->inp_hashval]++;
	INPCB_STATINC(head, inpcb_curpcbs);
	so->so_pcb = (caddr_t)inp;
	INHEAD_MRUNLOCK(head);
	return (0);
}

int
in_pcbbind(register struct inpcb *inp, struct mbuf *nam)
{
	register struct socket *so = inp->inp_socket;
	register struct sockaddr_in *sin = 0;
	u_short lport = 0;
	int wild = 0, reuseport = (so->so_options & SO_REUSEPORT);
	int r = 0;

	INPCB_STATINC(inp->inp_head, inpcb_bind);
	ASSERT(SOCKET_ISLOCKED(so));

	if (in_ifaddr_count <= 0) { /* no addr's configured NOT even LB */
		return EADDRNOTAVAIL;
	}

	if (inp->inp_lport || inp->inp_laddr.s_addr != INADDR_ANY) {
		return (EINVAL);
	}

	if ((so->so_options & (SO_REUSEADDR|SO_REUSEPORT)) == 0 &&
	    ((so->so_proto->pr_flags & PR_CONNREQUIRED) == 0 ||
	     (so->so_options & SO_ACCEPTCONN) == 0))
		wild = INPLOOKUP_WILDCARD;
	if (nam) {
		sin = mtod(nam, struct sockaddr_in *);
		if (nam->m_len != sizeof (*sin)) 
			return (EINVAL);
		/* 
		 * Since bind() NEVER checks sin_family, programs often fail
		 * to initialize it properly. Fix it here for audit. Yech.
		 */
		sin->sin_family = AF_INET;
		lport = sin->sin_port;
		if (IN_MULTICAST(ntohl(sin->sin_addr.s_addr))) {
			/* treat SO_REUSEADDR as SO_REUSEPORT for multicast */
			if (so->so_options & SO_REUSEADDR)
				reuseport = SO_REUSEADDR|SO_REUSEPORT;
		} else if (sin->sin_addr.s_addr != INADDR_ANY) {
			sin->sin_port = 0;		/* yech... */
			if (ifa_ifwithaddr((struct sockaddr *)sin) == 0) {
				sin->sin_port = lport;
				return (EADDRNOTAVAIL);
			}
		}
		if (lport) {
			/* GROSS */
			if (ntohs(lport) < IPPORT_RESERVED &&
			    !_CAP_CRABLE(curprocp->p_cred,CAP_PRIV_PORT)) {
				return (EACCES);
			}
			r = in_pcbassign(inp, zeroin_addr, 0,
			    sin->sin_addr, lport, reuseport, 
				INPLOOKUP_BIND|INPLOOKUP_ALL|wild,
				0, USE_SECURITY);
			if (r == 0) {
				/* 
				 * Audit use of privilege: *u.u_ap is 
				 * file descriptor.
				 * XXX:casey
				 *	Using the U-area here is pretty bogus.
				 *	Checking the pointer is a bandage
				 *	for a problem found in diskless booting
				 */
				if (ntohs(lport) < IPPORT_RESERVED && 
				    !(IS_KTHREAD(curprocp)) && u.u_ap) {
					_SAT_BSDIPC_RESVPORT(*u.u_ap, so, 
					ntohs(lport), 0);
				}
			}
			return r;
		}
	}
	if (sin) {
		inp->inp_laddr = sin->sin_addr;
	}
	ASSERT(lport == 0);
	return in_pcbassign(inp, zeroin_addr, 0, inp->inp_laddr,
		lport, reuseport, INPLOOKUP_BIND|INPLOOKUP_ALL|wild, 
		0, ASSIGN_NO_OP);
}

/*
 * This routine does the work in moving pcbs from one hash chain
 * to another an assigning port numbers to pcbs as necessary.
 */
int
in_pcbassign(struct inpcb *inp,
	struct in_addr faddr, u_short fport,
	struct in_addr laddr, u_short lport,
	int reuseport,
	int wild,
	struct inaddrpair *iap,
	enum AssignFlags flags)
{
	struct inpcb *t = 0;
	struct inpcb *head = inp->inp_head;
	u_short ohashval = inp->inp_hashval;
	u_short nhashval;
	u_short olport = lport;
	int didbind = 0;

	/*
	 * Try to find a unique value for the port in the pcb table.
	 */
	do {
		if (t) {
			INHEAD_MRUNLOCK(head);
#ifdef _MP_NETLOCKS
			if (t != inp) {
				SOCKET_MPLOCK(t->inp_socket);
				if (!INPCB_RELE(t))
					SOCKET_MPUNLOCK(t->inp_socket);
			} else {
				INPCB_RELE(t);
			}
#else
			INPCB_RELE(t);
#endif
		}

		/*
		 * If no port was given, try to get the next port.  Rehash
		 * to determine where we're going.
		 */
		if (!olport)
			lport = in_pcbnextport(inp);
		nhashval = (*head->inp_hashfun)(inp->inp_head,
					faddr, fport, laddr, lport);
	} while ((t = IN_PCBLOOKUPX(head,
		    faddr, fport, laddr, lport, wild,
			MR_UPDATE, GIVE_LOCK_ALWAYS)) && !olport);

	if (t) {
		if (flags != USE_SECURITY ||
				(reuseport & t->inp_socket->so_options) == 0) {
			INHEAD_MRUNLOCK(head);
#ifdef _MP_NETLOCKS
			if (t != inp) {
				SOCKET_MPLOCK(t->inp_socket);
				if (!INPCB_RELE(t))
					SOCKET_MPUNLOCK(t->inp_socket);
			} else {
				INPCB_RELE(t);
			}
#else
			INPCB_RELE(t);
#endif
			return EADDRINUSE;
		}
		/* clean t up down below */
	}

	if (!inp->inp_lport && iap) {
		/*
		 * Bind PCB now.  This code is needed for applications that
		 * have not previously bound and then call the UDP connect
		 * code passing in the inaddrpair structure.
		 * The local port will never get put into the PCB otherwise.
		 */
		inp->inp_lport = lport;
		didbind = 1;
	}

	/*
	 * Assign addresses.  If iap is provided, we insert the
	 * information into it.  Otherwise, we set up the inp
	 * and rehash, if necessary.
	 */

	if (iap) {
		iap->iap_faddr = faddr;
		iap->iap_fport = fport;
		iap->iap_laddr = laddr;
		iap->iap_lport = lport;
		if (didbind) {
			goto fixhash;
		}
		goto done;
	}

	inp->inp_faddr = faddr;
	inp->inp_fport = fport;
	inp->inp_laddr = laddr;
	inp->inp_lport = lport;

fixhash:
	/*
	 * Reset hashval for the inp
	 */
	inp->inp_hashval = nhashval;

	/*
	 * If the two hashvals are different, move from one list to the other.
	 */
	if (ohashval != nhashval) {
		struct inpcb *tinp = _TBLELEMTAIL(inp);
#ifdef INPCB_CACHE
		struct inpcb **cache = &_TBLCACHE(inp);
#endif
		if (ohashval != BUCKET_INVAL) {
			remque(inp);
			head->inp_dstats[ohashval]--;
#ifdef INPCB_CACHE
			if (inp == *cache) {
				*cache = 0;
			}
#endif
		}
		insque(inp, tinp);
		inp->inp_hhead = &head->inp_table[inp->inp_hashval];
		head->inp_dstats[inp->inp_hashval]++;
	}

done:
	INHEAD_MRUNLOCK(head);
	if (t) {
		if (t != inp) {
			/* release here now that list is unlocked */
			SOCKET_MPLOCK(t->inp_socket);
			if (!INPCB_RELE(t))
				SOCKET_MPUNLOCK(t->inp_socket);
		} else {
			INPCB_RELE(t);
		}
	}
	return (0);
}

/*
 * Connect from a socket to a specified address.
 * Both address and port must be specified in argument sin.
 * If don't have a local address for this socket yet,
 * then pick one.
 */
int
in_pcbconnect(register struct inpcb *inp, struct mbuf *nam)
{
	register struct sockaddr_in *sin = mtod(nam, struct sockaddr_in *);

	INPCB_STATINC(inp->inp_head, inpcb_connect);
	if (nam->m_len != sizeof (*sin))
		return (EINVAL);
	return (in_pcbsetaddrx(inp, sin,inp->inp_laddr,(struct inaddrpair *)0));
}

int
in_pcbconnectbyaddr(struct inpcb *inp, struct mbuf *nam, struct in_addr laddr)
{
	register struct sockaddr_in *sin = mtod(nam, struct sockaddr_in *);

	INPCB_STATINC(inp->inp_head, inpcb_connectbyaddr);
	if (nam->m_len != sizeof (*sin))
		return (EINVAL);
	return (in_pcbsetaddrx(inp, sin, laddr, (struct inaddrpair *)0));
}

int
in_pcbconnectbyaddrsin(struct inpcb *inp, struct sockaddr_in *sin,
	struct in_addr laddr)
{
	INPCB_STATINC(inp->inp_head, inpcb_connectbyaddr);
	return (in_pcbsetaddrx(inp, sin, laddr, (struct inaddrpair *)0));
}

int
in_pcbsetaddr(struct inpcb *inp,struct sockaddr_in *sin,struct inaddrpair *iap)
{
	INPCB_STATINC(inp->inp_head, inpcb_setaddr);
	return in_pcbsetaddrx(inp, sin, inp->inp_laddr, iap);
}

int
in_pcbsetaddrx(struct inpcb *inp,
	struct sockaddr_in *sin,
	struct in_addr laddr,
	struct inaddrpair *iap)
{
	register struct in_ifaddr *ia, *ia_xx;
	register struct sockaddr_in *ifaddr;
	int error;
	u_short fport;
	NETSPL_MPDECL(s)

	INPCB_STATINC(inp->inp_head, inpcb_setaddrx);
	ASSERT(SOCKET_ISLOCKED(inp->inp_socket));

	if (sin->sin_family != AF_INET)
		return (EAFNOSUPPORT);
	if (sin->sin_port == 0)
		return (EADDRNOTAVAIL);

	if (in_ifaddr_count > 0) {

		if (ia_xx = (struct in_ifaddr *)(ifnet->in_ifaddr)) {
			/*
			 * If the destination address is INADDR_ANY,
			 * use the first interface's primary local address.
			 * If the supplied address is INADDR_BROADCAST,
			 * and the primary interface supports broadcast,
			 * choose the broadcast address for that interface.
			 */
			if (sin->sin_addr.s_addr == INADDR_ANY) {
				    sin->sin_addr = ia_xx->ia_addr.sin_addr;
			} else {
				if (sin->sin_addr.s_addr ==
				     (u_long)INADDR_BROADCAST &&
				   (ia_xx->ia_ifp->if_flags & IFF_BROADCAST)) {

					sin->sin_addr =
						ia_xx->ia_broadaddr.sin_addr;
				}
			}
		}
  	}
	if (inp->inp_laddr.s_addr == INADDR_ANY) {
		register struct route *ro;

		ia = (struct in_ifaddr *)0;
		/* 
		 * If route is known or can be allocated now,
		 * our src addr is taken from the i/f, else punt.
		 */
		ROUTE_MPLOCK(s);
		ro = &inp->inp_route;
		if (ro->ro_rt &&
		    (SATOSIN(&ro->ro_dst)->sin_addr.s_addr !=
			sin->sin_addr.s_addr || 
		    inp->inp_socket->so_options & SO_DONTROUTE)) {
			RTFREE(ro->ro_rt);
			ro->ro_rt = (struct rtentry *)0;
		}
		if ((inp->inp_socket->so_options & SO_DONTROUTE) == 0 && /*XX*/
		    (ro->ro_rt == (struct rtentry *)0 ||
		    ro->ro_rt->rt_ifp == (struct ifnet *)0)) {
			/*
			 * No route yet, so try to acquire one
			 */
			ro->ro_dst.sa_family = AF_INET;
#ifdef _HAVE_SA_LEN
			ro->ro_dst.sa_len = sizeof(struct sockaddr_in);
#endif
			((struct sockaddr_in *) &ro->ro_dst)->sin_addr =
				sin->sin_addr;
			rtalloc(ro);
		}
		ROUTE_MPUNLOCK(s);
		/*
		 * If we found a route, use the address corresponding to the
		 * outgoing interface unless it is the loopback (in case a
		 * route to our address on another net goes to loopback).
		 */
		if (ro->ro_rt && (ro->ro_rt->rt_srcifa ||
			!(ro->ro_rt->rt_ifp->if_flags & IFF_LOOPBACK))) {

			ia = (ro->ro_rt->rt_srcifa)
				? ifatoia(ro->ro_rt->rt_srcifa)
				: ifatoia(ro->ro_rt->rt_ifa);

			if (ia && (ia->ia_addrflags & IADDR_ALIAS)) {
				/*
				 * If the address is an alias AND has the
				 * same network interface as the routing
				 * entry we use the primary address of the
				 * network interface instead.
				 */
				if (ia->ia_ifp == ro->ro_rt->rt_ifp) {
					ia = (struct in_ifaddr *)
						ia->ia_ifp->in_ifaddr;
					if (ia == 0) {
						goto findit;
					}
				}
				/* fast exit case */
				goto gotit;
			}
		}
findit:
		if (ia == 0) {
			fport = sin->sin_port;
			sin->sin_port = 0;

			/* check if address associated with ppp interface */
			if (ia = ifatoia(ifa_ifwithdstaddr(SINTOSA(sin)))) {
				/* restore port number */
				sin->sin_port = fport;
				goto gotit;
			}
			/*
			 * Now do the more expensive check to find the
			 * best matching network number of all of the
			 * address'es we know about.
			 * NOTE: With huge numbers of IP alias'es this is a
			 * very expensive operation, to be avoided if possible.
			 */
			if (ia = ifatoia(ifa_ifwithnet(SINTOSA(sin)))) {
				/* restore port number */
				sin->sin_port = fport;
				goto gotit;
			}

			/* restore port number in case we fast exit */
			sin->sin_port = fport;

			if (in_interfaces > 0) { /* use primary if not any */
				ia = (struct in_ifaddr *)(ifnet->in_ifaddr);
				error = (ia) ? 0 : EADDRNOTAVAIL;
			} else {
				error = EADDRNOTAVAIL;
			}
			if (error) { /* failed so quit */
				return error;
			}
		}
gotit:
		/*
		 * If the destination address is multicast and an outgoing
		 * interface has been set as a multicast option, use the
		 * address of that interface as our source address.
		 */
		if (IN_MULTICAST(ntohl(sin->sin_addr.s_addr)) &&
				inp->inp_moptions != NULL) {
			struct ip_moptions *imo;
			struct ifnet *ifp;

			imo = mtod(inp->inp_moptions, struct ip_moptions *);
			if (imo->imo_multicast_ifp != NULL) {
				ifp = imo->imo_multicast_ifp;

				/*
				 * Find interface address record and
				 * return its primary address.
				 */
				ia = (struct in_ifaddr *)hash_enum(
					&hashinfo_inaddr, ifnet_enummatch,
					HTF_INET, (caddr_t)0,
					(caddr_t)ifp, (caddr_t)0);

				if (ia == 0)
					return (EADDRNOTAVAIL);
			}
		}
		ifaddr = (struct sockaddr_in *)&ia->ia_addr;
	}

	laddr = laddr.s_addr ? inp->inp_laddr : ifaddr->sin_addr;

	if (inp->inp_lport == 0) {
		int error;
		/*
		 * It might be possible to just pass INPLOOKUP_WILD
		 * into in_pcbassign() below, but that *might* cause spurious
		 * conflicts.  Just get a local port the old-fashioned way.
		 */
		if (error = in_pcbbind(inp, (struct mbuf *)0)) {
			return error;
		}
	}
	return in_pcbassign(inp, sin->sin_addr, sin->sin_port, laddr,
		inp->inp_lport, 0, INPLOOKUP_ALL, iap, ASSIGN_NO_OP);
}

void
in_pcbdisconnect(struct inpcb *inp)
{
	ASSERT(SOCKET_ISLOCKED(inp->inp_socket));
	INHEAD_MRWLOCK(inp->inp_head);
	inp->inp_faddr.s_addr = INADDR_ANY;
	inp->inp_fport = 0;
	if (inp->inp_socket->so_state & SS_NOFDREF)
		in_pcbdetach(inp);
	INHEAD_MRUNLOCK(inp->inp_head);
	return;
}

/*
 * Removes a PCB from the active list, clearing various bits of cached state.
 * Normally called from in_pcbdetach(), but may be called directly by TCP to
 * ensure that segments do not arrive for connections in the middle of being
 * torn down.
 */
void
in_pcbunlink(struct inpcb *inp)
{
	struct inpcb *head = inp->inp_head;
#ifdef INPCB_CACHE
	struct inpcb **cache = &_TBLCACHE(inp);
#endif

	ASSERT(SOCKET_ISLOCKED(inp->inp_socket));

	INHEAD_MRWLOCK(head);
	if (inp->inp_prev == 0) {
		/*
		 * Unlink was already done, so no need to do it again.
		 */
		INHEAD_MRUNLOCK(head);
		return;
	}

	if (inp->inp_hashval != BUCKET_INVAL) {
		head->inp_dstats[inp->inp_hashval]--;
		remque(inp);
	}
	inp->inp_hashval = BUCKET_INVAL;
	inp->inp_next = inp->inp_prev = 0;
#ifdef INPCB_CACHE
	if (inp == *cache) {
		*cache = 0;
	}
#endif
	if (inp == head->inp_last) {
		head->inp_last = 0;
	}

	INPCB_STATDEC(head, inpcb_curpcbs);

	INHEAD_MRUNLOCK(head);
	return;
}

int
in_pcbdetach(struct inpcb *inp)
{
	ASSERT(SOCKET_ISLOCKED(inp->inp_socket));

	/*
	 * Take this PCB off the list of active PCBs, if necessary.
	 */
	in_pcbunlink(inp);

	if (inp->inp_socket->so_state & SS_CLOSED) {
		return 0;
	}

	/*
	 * Since we've detached once, we don't want to detach again
	 */
	inp->inp_socket->so_state |= SS_CLOSED;

	return(INPCB_RELE(inp));
}

/*
 * Decrement the ref count of the PCB and free if last reference
 */
int
inpcb_rele(struct inpcb *inp)
{
	NETSPL_MPDECL(s)
	int rv;

#ifdef _MP_NETLOCKS
 	if (atomicAddInt(&inp->inp_refcnt, -1)) {
		return (0);
	}
#else
	inp->inp_refcnt--;
	if (inp->inp_refcnt) {
		return (0);
	}
#endif
	rv = 0;
	ASSERT(SOCKET_ISLOCKED(inp->inp_socket));
	ASSERT(inp->inp_next == 0 && inp->inp_prev == 0);
	ASSERT(inp->inp_hashval == BUCKET_INVAL);

if (inp->inp_next || inp->inp_prev) {
	cmn_err(CE_PANIC,"freeing linked PCB: 0x%x", inp);
}
	inp->inp_socket->so_pcb = 0;
	if (sofree(inp->inp_socket)) {
		inp->inp_socket = 0;
		rv = 1;
	}
	if (inp->inp_options)
		(void)m_free(inp->inp_options);
	ROUTE_MPLOCK(s);
	if (inp->inp_route.ro_rt)
		rtfree(inp->inp_route.ro_rt);
	ROUTE_MPUNLOCK(s);
	ip_freemoptions(inp->inp_moptions);
	(void) kmem_zone_free(inpcb_zone, inp);
	MBSTAT_DEC(m_mtypes[MT_PCB], 1);
	return(rv);
}

void
in_setsockaddr(register struct inpcb *inp, struct mbuf *nam)
{
	register struct sockaddr_in *sin;

	ASSERT(SOCKET_ISLOCKED(inp->inp_socket));	
	nam->m_len = sizeof (*sin);
	sin = mtod(nam, struct sockaddr_in *);
	bzero((caddr_t)sin, sizeof (*sin));
	sin->sin_family = AF_INET;
	sin->sin_port = inp->inp_lport;
	sin->sin_addr = inp->inp_laddr;
	return;
}

void
in_setpeeraddr(struct inpcb *inp, struct mbuf *nam)
{
	register struct sockaddr_in *sin;
	
	ASSERT(SOCKET_ISLOCKED(inp->inp_socket));	
	nam->m_len = sizeof (*sin);
	sin = mtod(nam, struct sockaddr_in *);
	bzero((caddr_t)sin, sizeof (*sin));
	sin->sin_family = AF_INET;
	sin->sin_port = inp->inp_fport;
	sin->sin_addr = inp->inp_faddr;
	return;
}

/*
 * Pass some notification to all connections of a protocol
 * associated with address dst.  The local address and/or port numbers
 * may be specified to limit the search.  The "usual action" will be
 * taken, depending on the ctlinput cmd.  The caller must filter any
 * cmds that are uninteresting (e.g., no error in the map).
 * Call the protocol specific routine (if any) to report
 * any errors for each matching socket.
 *
 * Must be called at splnet.
 */

void
in_pcbnotify(
	struct inpcb *head,
	struct sockaddr *dst,
	u_short fport,
	struct in_addr laddr,
	u_short lport,
	int cmd,
	void (*notify)(struct inpcb *, int, void *),
	void *data)
{
	register struct inpcb *inp, *inpnxt;
	struct in_addr faddr;
	int errno;
	void in_rtchange(struct inpcb *, int, void *);
	int hash;
	struct in_pcbhead *hinp;
	int havelock = 0;
	int start = 1, end = head->inp_tablesz;

	if ((unsigned)cmd > PRC_NCMDS || dst->sa_family != AF_INET)
		return;
	faddr = ((struct sockaddr_in *)dst)->sin_addr;
	if (faddr.s_addr == INADDR_ANY)
		return;

	/*
	 * Redirects go to all references to the destination,
	 * and use in_rtchange to invalidate the route cache.
	 * Dead host indications: notify all references to the destination.
	 * Otherwise, if we have knowledge of the local port and address,
	 * deliver only to that socket.
	 */
	if (PRC_IS_REDIRECT(cmd) || cmd == PRC_HOSTDEAD) {
		fport = 0;
		lport = 0;
		laddr.s_addr = 0;
		if (cmd != PRC_HOSTDEAD)
			notify = in_rtchange;
	}

	/*
	 * See if we can optimize the delivery of this message; if we have
	 * the required values for the hash function, we can just scan one
	 * bucket.
	 */
	if (head->inp_hashflags & INPFLAGS_CLTS) {
		if (lport) {
			start = (*head->inp_hashfun)(head, 
				faddr, fport, laddr, lport);
			end = start + 1;
		}
	} else {
		/* This ignores TIME-WAITers for TCP, but who cares? */
		if (faddr.s_addr && laddr.s_addr && lport && fport) {
			start = (*head->inp_hashfun)(head, 
				faddr, fport, laddr, lport);
			end = start + 1;
		}
	}
	errno = inetctlerrmap[cmd];
	for (hash = start; hash < end; hash++) {
resync:
	INHEAD_MRRLOCK(head);
	havelock = 1;
	hinp = &head->inp_table[hash];
	inp = hinp->hinp_next;
	if (inp != (struct inpcb *)hinp) {
	INPCB_HOLD(inp);
	for (; (inp != (struct inpcb *)hinp) &&
	       (inp->inp_hhead == hinp); inp = inpnxt) {
		ASSERT(INHEAD_ISMRLOCKED(head));
		inpnxt = inp->inp_next;
		INPCB_HOLD(inpnxt);
		if (inp->inp_faddr.s_addr != faddr.s_addr ||
		    inp->inp_socket == 0 ||
		    (lport && inp->inp_lport != lport) ||
		    (laddr.s_addr && inp->inp_laddr.s_addr != laddr.s_addr) ||
		    (fport && inp->inp_fport != fport)) {
			INHEAD_MRUNLOCK(head);
#ifdef _MP_NETLOCKS
			SOCKET_MPLOCK(inp->inp_socket);
			if (!INPCB_RELE(inp))
				SOCKET_MPUNLOCK(inp->inp_socket);
#else
			INPCB_RELE(inp);
#endif
			INHEAD_MRRLOCK(head);
			if (inpnxt->inp_next == 0) {
				INHEAD_MRUNLOCK(head);
				SOCKET_MPLOCK(inpnxt->inp_socket);
				if (!INPCB_RELE(inpnxt)) {
					SOCKET_MPUNLOCK(inpnxt->inp_socket);
				}
				goto resync;
			}
			continue;
		}
		INHEAD_MRUNLOCK(head);
		SOCKET_MPLOCK(inp->inp_socket);
		if (notify)
			(*notify)(inp, errno, data);
		if (!INPCB_RELE(inp))
			SOCKET_MPUNLOCK(inp->inp_socket);
		INHEAD_MRRLOCK(head);
		if (inpnxt->inp_next == 0) {
			INHEAD_MRUNLOCK(head);
			SOCKET_MPLOCK(inpnxt->inp_socket);
			if (!INPCB_RELE(inpnxt))
				SOCKET_MPUNLOCK(inpnxt->inp_socket);
			goto resync;
		}
	}
	if (inp != (struct inpcb *)hinp) {
		INHEAD_MRUNLOCK(head);
		havelock = 0;
		SOCKET_MPLOCK(inp->inp_socket);
		if (!INPCB_RELE(inp)) {
			SOCKET_MPUNLOCK(inp->inp_socket);
		}
	}
	} /* if */
	if (havelock) {
		INHEAD_MRUNLOCK(head);
	}
	}	/* end for */
}

/*
 * Check for alternatives when higher level complains
 * about service problems.  For now, invalidate cached
 * routing information.  If the route was created dynamically
 * (by a redirect), time to try a default gateway again.
 */
void
in_losing(struct inpcb *inp)
{
	register struct rtentry *rt;
	struct rt_addrinfo info;
	NETSPL_MPDECL(s)

	ASSERT(SOCKET_ISLOCKED(inp->inp_socket));
	if ((rt = inp->inp_route.ro_rt)) {
		ROUTE_MPLOCK(s);
		inp->inp_route.ro_rt = 0;
		bzero((caddr_t)&info, sizeof(info));
		info.rti_info[RTAX_DST] =
			(struct sockaddr *)&inp->inp_route.ro_dst;
		info.rti_info[RTAX_GATEWAY] = rt->rt_gateway;
#ifdef _HAVE_SIN_LEN
		info.rti_info[RTAX_NETMASK] = rt_mask(rt);
#else
		info.rti_info[RTAX_NETMASK] = (struct sockaddr*)rt_mask(rt);
#endif
		rt_missmsg(RTM_LOSING, &info, rt->rt_flags, 0);
		if (rt->rt_flags & RTF_DYNAMIC)
			(void) rtrequest(RTM_DELETE, rt_key(rt),
				rt->rt_gateway, rt_mask(rt), rt->rt_flags, 
				(struct rtentry **)0);
		else 
		/*
		 * A new route can be allocated
		 * the next time output is attempted.
		 */
			rtfree(rt);
		ROUTE_MPUNLOCK(s);
	}
	return;
}

/*
 * After a routing change, flush old routing
 * and allocate a (hopefully) better one.
 */
/*ARGSUSED*/
void
in_rtchange(register struct inpcb *inp, int flag, void *data)
{
	NETSPL_MPDECL(s)

	ASSERT(SOCKET_ISLOCKED(inp->inp_socket));
	ROUTE_MPLOCK(s);
	if (inp->inp_route.ro_rt) {
		rtfree(inp->inp_route.ro_rt);
		inp->inp_route.ro_rt = 0;
		/*
		 * A new route can be allocated the next time
		 * output is attempted.
		 */
	}
	ROUTE_MPUNLOCK(s);
	return;
}

struct inpcb *
in_pcblookup(struct inpcb *head,
	struct in_addr faddr, u_short fport,
	struct in_addr laddr, u_short lport,
	int flags)
#if defined(_MP_NETLOCKS)
{
	return IN_PCBLOOKUPX(head, faddr, fport, laddr, lport,
				flags, MR_ACCESS, LOOKUP_NO_OP);
}

struct inpcb *
in_pcblookupx(struct inpcb *head,
	struct in_addr faddr, u_short fport,
	struct in_addr laddr, u_short lport,
	int flags, int type, enum LookupPolicy policy)
#endif /* _MP_NETLOCKS */
{
	register struct inpcb *inp = 0, *match = 0;
	struct in_pcbhead *hinp;
	int matchwild = 3, wildcard;
	int all = (flags & INPLOOKUP_ALL);
	u_short hashval, smhashval, start, end;
	int havelock = 0;

	flags &= ~INPLOOKUP_ALL;

	INPCB_STATINC(head, inpcb_lookup);
	hashval = (*head->inp_hashfun)(head, faddr, fport, laddr, lport);

	/*
	 * If we are CLTS, we jump over the exact match and cache
	 * code and go directly to the wildcard setup.
	 *
	 * The idea here is that we've set up to hash by local port number
	 * only (see in_pcbcltshashfun, below).  This means that we only
	 * need to check the hash bucket corresponding to the local port
	 * number instead of the entire hash.  We still need to look for
	 * wildcard matches so we can't really take advantage of the cache
	 * since we may have a "better" match.  The point of hashing
	 * by local port numbers is that "most" clts inpcbs are going
	 * to be wildcarded.  We try to minimize the amount of the total
	 * inpcb list that we have to search while ensuring "completeness"
	 * in the search.
	 */
	if (head->inp_hashflags & INPFLAGS_CLTS) {
		start = hashval;
		end = hashval + 1;
		INPCB_STATINC(head, inpcb_wildcard);
		goto lookwild;
	} else {
		if (flags & INPLOOKUP_LISTEN) { /* listen/wildcard lookup */
			INPCB_STATINC(head, inpcb_listen);
			/* only search first bucket */
			start = 0;
			end = 1;
			goto lookwild;
		} else if (flags & INPLOOKUP_WILDCARD) { /* wildcard lookup */
			INPCB_STATINC(head, inpcb_wildcard);
		} else if (flags & INPLOOKUP_BIND) {
			start = 0;
			end = (head->inp_tablesz - SPECIAL_BUCKETS) / 2;
			goto lookwild;
		} else { /* normal lookup */
			INPCB_STATINC(head, inpcb_regular);
		}
	}

	/* Try quick exact match lookup first from the inpcb cache */
	INHEAD_MRLOCK(head, type);
again:
#ifdef INPCB_CACHE
	inp = head->inp_cache[hashval];
	if (inp && IP_ADDRS_EQUAL(inp, laddr, lport, faddr, fport)) {
		head->inp_last = inp;
		INPCB_STATSET(head, inpcb_last, inp);
		INPCB_HOLD(inp);

		INPCB_STATINC(head, inpcb_found_cache);
		goto foundit;
	}
#endif
	/*
	 * Compute the hash function and search the
	 * hash chain associated with this bucket for a match.
	 */
	hinp = &head->inp_table[hashval];
	for (inp = hinp->hinp_next; inp != (struct inpcb *)hinp; 
		inp = inp->inp_next) {
		if (!IP_ADDRS_EQUAL(inp, laddr, lport, faddr, fport))
			continue;
#ifdef INPCB_CACHE
		head->inp_cache[hashval] = inp;
#endif
		head->inp_last = inp;
		INPCB_STATSET(head, inpcb_last, inp);
		INPCB_HOLD(inp);
		INPCB_STATINC(head, inpcb_found_chain);
		goto foundit;
	}

	inp = 0;
foundit:
	if (inp || !(flags & INPLOOKUP_WILDCARD)) {
		if (inp) { /* success */
			INPCB_STATINC(head, inpcb_success);
		} else { /* failed*/
			if (all) {
				all = 0;
				hashval += 
				 ((head->inp_tablesz - SPECIAL_BUCKETS) / 2);
				goto again;
			}
			INPCB_STATINC(head, inpcb_failed);
		}
#ifdef _MP_NETLOCKS
		if (policy != GIVE_LOCK_ALWAYS)
			INHEAD_MRUNLOCK(head);
#endif
		return inp;
	}

	havelock = 1;
	
	/*
	 * Arrive here to search the entire hash table (all buckets)
	 * when a wildcard match is required.  This is used for COTS only.
	 */
	start = 0;
	end = (head->inp_tablesz - SPECIAL_BUCKETS) / 2;

lookwild:
	if (!(flags & INPLOOKUP_LISTEN)) {
		INPCB_STATINC(head, inpcb_searched_all);
	}

	if (havelock == 0) {
		INHEAD_MRLOCK(head, type);
	}
lookwild2:
	/*
	 * Wildcard lookup, we search all the hash
	 * buckets and associated chains.
	 */
	for (hashval = start, hinp = &head->inp_table[start];
			hashval < end;
			hashval++, hinp++) {
#ifdef _MP_NETLOCKS
	    for (; hashval < head->inp_tablesz && (hashval < end) &&
		hinp == (struct in_pcbhead *)hinp->hinp_next;
				hashval++, hinp++) {
	    }
	    if ((hashval >= head->inp_tablesz) || (hashval >= end)) {
		continue;
	    }
#endif
	    for (inp = hinp->hinp_next; inp != (struct inpcb *)hinp; 
		inp = inp->inp_next) {
		if ((flags & INPLOOKUP_LISTEN) && 
		    !(inp->inp_hashflags & INPFLAGS_LISTEN)) {
			continue;
		}
	    	if (inp->inp_lport != lport)
			continue;
		wildcard = 0;
		if (inp->inp_laddr.s_addr != INADDR_ANY) {
		    	if (laddr.s_addr == INADDR_ANY)
				wildcard++;
		    	else if (inp->inp_laddr.s_addr != laddr.s_addr)
			    	continue;
		} else {
		    	if (laddr.s_addr != INADDR_ANY)
				wildcard++;
		}
		if (inp->inp_faddr.s_addr != INADDR_ANY) {
		    	if (faddr.s_addr == INADDR_ANY)
				wildcard++;
		    	else if (inp->inp_faddr.s_addr != faddr.s_addr ||
			     	inp->inp_fport != fport)
				continue;
		} else {
		    	if (faddr.s_addr != INADDR_ANY)
				wildcard++;
		}
		if (wildcard && (flags & INPLOOKUP_WILDCARD) == 0)
		    	continue;
		if (wildcard < matchwild) {
		    	match = inp;
			smhashval = hashval;
		    	matchwild = wildcard;
		    	if (matchwild == 0)
				break;
		}
	    }

	    if (!matchwild) {
		break; /* out of for loop */
	    }
	    continue;	/* top of for loop */
	}

	if (match) {
#ifdef INPCB_CACHE
		head->inp_cache[smhashval] = match;
#endif
		head->inp_last = match;
		INPCB_HOLD(match);
		INPCB_STATSET(head, inpcb_last, match);
	} else {
		if (all) {
			all = 0;
			start = end;
			end = head->inp_tablesz;
			goto lookwild2;
		}
	}

#ifdef _MP_NETLOCKS
	if (policy != GIVE_LOCK_ALWAYS)
		INHEAD_MRUNLOCK(head);
#endif
	if (match) { /* success */
		INPCB_STATINC(head, inpcb_success);
	} else { /* failed */
		INPCB_STATINC(head, inpcb_failed);
	}
	return (match);
}
/*
 * This function returns a "hint" indicating what the next lport
 * value might be by incrementing through the list of possible
 * port numbers.
 *
 * The caller should verify the port during a bind operation by
 * looking up the completed inp address using a locking form of
 * in_pcblookup.
 */
u_short
in_pcbnextport(struct inpcb *inp)
{
	struct inpcb *head = inp->inp_head;
	u_short lport;
#ifdef _MP_NETLOCKS
	extern int atomicIncPort(int *, int, int);
#endif
	/*
	 * This used to wrap at IPPORT_USERRESERVED (5000)
	 * but that would cause an infinite loop with 4000
	 * anonymous ports. The limit needs to be MAXPORT/2
	 * in order to work around a bug in the boot PROMs:
	 * they do not handle ports with the sign bit set.
	 */
#ifdef _MP_NETLOCKS
	/* avoid lock on head via atomic op */
	lport = (u_short)atomicIncPort(&head->inp_wildport, 
		IPPORT_MAXPORT / 2, IPPORT_RESERVED);
	lport = htons(lport);
#else
	if (head->inp_lport++ < IPPORT_RESERVED ||
	    head->inp_lport > IPPORT_MAXPORT/2) {
		head->inp_lport = IPPORT_RESERVED;
	}
	lport = htons(head->inp_lport);
#endif
	return lport;
}

/*
 * NOTE:
 * These hash functions are based on the fact that the number of
 * hash buckets will be (a power of 2) + 1.
 *
 * Check out the procedure in_pcbinitcb()
 * This is the TCP hash function.
 * Note that we "never" hash into bucket 0.
 */
u_short
in_pcbhashfun(struct inpcb *head,
		struct in_addr faddr, u_short fport,
		struct in_addr laddr, u_short lport)
{
	__uint32_t value;
	
	value = faddr.s_addr + fport + laddr.s_addr + lport;
	value &= (((head->inp_tablesz - SPECIAL_BUCKETS) / 2) - 1);
	value += LEADING_BUCKETS;

	ASSERT(value <= head->inp_tablesz);

	return (u_short)value;
}

/*
 * This is the connectionless (UDP) hash function.
 * NOTE: We "never" hash into bucket 0.
 */
/* ARGSUSED */
u_short
in_pcbcltshashfun(struct inpcb *head,
		struct in_addr faddr, u_short fport,
		struct in_addr laddr, u_short lport)
{
	__uint32_t value;

	if (lport == 0) {
		value = 0;	
	} else {
		value = lport;
		value &= ((head->inp_tablesz - SPECIAL_BUCKETS) - 1);
		value += LEADING_BUCKETS;
	}
	return ((u_short)value);
}

/*
 * Support for listening TCP endpoints
 * We'd like to have these all hashed into the same bucket so that
 * we don't ever have to do wildcard lookups in the normal TCP case.
 * Only an incoming initial SYN can match a wildcard PCB.  Outgoing 
 * connections or PCBs created by an accept are all fully specified,
 * so they don't need a wildcard lookup.
 * We put the listening endpoints into bucket zero since nothing else
 * uses it.
 */
void
in_pcblisten(struct inpcb *inp)
{
	struct inpcb *tinp;
#ifdef INPCB_CACHE
	struct inpcb **cache = &_TBLCACHE(inp);
#endif

	INHEAD_MRWLOCK(inp->inp_head);
	if (inp->inp_hashval != BUCKET_INVAL) {
		inp->inp_head->inp_dstats[inp->inp_hashval]--;
		remque(inp);
#ifdef INPCB_CACHE
		if (inp == *cache) {
			*cache = 0;
		}
#endif
	}

	inp->inp_hashval = BUCKET_0;
	tinp = _TBLELEMTAIL(inp);
	insque(inp, tinp);

	/* A lookup might have cached this PCB; make sure we clear it. */
	if (inp == inp->inp_head->inp_last) {
		inp->inp_head->inp_last = 0;
	}
	inp->inp_hhead = &inp->inp_head->inp_table[inp->inp_hashval];
	inp->inp_head->inp_dstats[inp->inp_hashval]++;
	inp->inp_hashflags = INPFLAGS_LISTEN;

	INHEAD_MRUNLOCK(inp->inp_head);
	return;
}

void
in_pcbunlisten(struct inpcb *inp)
{
	inp->inp_hashflags &= ~INPFLAGS_LISTEN;
	/*
	 * The PCB will get rehashed when in_pcbconnect() is called
	 */
	return;
}

/*
 * Move a PCB in a hash table -- COTS only.
 */
void
in_pcbmove(struct inpcb *inp)
{
	struct inpcb *head;
#ifdef INPCB_CACHE
	struct inpcb **cache = &_TBLCACHE(inp);
#endif

	ASSERT(inp);
	head = inp->inp_head;
	ASSERT(head);
	INHEAD_MRWLOCK(head);
	remque(inp);
	head->inp_dstats[inp->inp_hashval]--;
	if (inp == head->inp_last) {
		head->inp_last = 0;
	}
#ifdef INPCB_CACHE
	if (inp == *cache) {
		*cache = 0;
	}
#endif

	ASSERT(inp->inp_hashval != BUCKET_INVAL);

	inp->inp_hashflags |= INPFLAGS_MOVED;
	inp->inp_hashval += (head->inp_tablesz - SPECIAL_BUCKETS) / 2;

	insque(inp, _TBLELEMTAIL(inp));
	inp->inp_hhead = &head->inp_table[inp->inp_hashval];
	head->inp_dstats[inp->inp_hashval]++;
	INHEAD_MRUNLOCK(head);
	return;
}

