/*
 * Copyright (c) 1982, 1986, 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)in_pcb.h	8.1 (Berkeley) 6/10/93
 */

#include "sys/sema.h"
struct in_addr;
struct mbuf;
struct route;
struct socket;
struct inpcbstat;

struct inaddrpair {
	struct	in_addr iap_faddr;	/* foreign host table entry */
	u_short	iap_fport;		/* foreign port */
	struct	in_addr	iap_laddr;	/* local host table entry */
	u_short	iap_lport;		/* local port */
};

/*
 * To save space, each hash list uses a small head structure consisting
 * only of threading pointers.
 */
struct in_pcbhead {
	/* fields must line up with PCB below */
	struct	inpcb *hinp_next;
	struct	inpcb *hinp_prev;
	struct  inpcb *hinp_waste;
	int	hinp_refcnt;
};

/*
 * Common structure pcb for internet protocol implementation.
 * Here are stored pointers to local and foreign host table
 * entries, local and foreign socket numbers, and pointers
 * up (to a socket structure) and down (to a protocol-specific)
 * control block.
 */
struct inpcb {
	struct	inpcb *inp_next,*inp_prev;
					/* pointers to other pcb's */
	struct	inpcb *inp_head;	/* pointer back to chain of inpcb's
					   for this protocol */

	int	inp_refcnt;		/* reference count */
	struct	in_pcbhead *inp_hhead;	/* pointer to hash head */
	struct	inaddrpair inp_iap;	/* foreign and local addr pair */
	/* needed by both heads and individual PCBs */
	u_short		inp_hashflags;		/* hashing flags */
	union {
	   /* The real PCB */		
	   struct {
		struct	socket *u1_socket;/* back pointer to socket */
		caddr_t	u1_ppcb;	  /* pointer to per-protocol pcb */
		struct	route u1_route;  /* placeholder for routing entry */
		int	u1_flags;	  /* generic IP/datagram flags */
		struct	ip u1_ip;	/* header prototype; should have more */
		struct	mbuf *u1_options;	/* IP options */
		struct	mbuf *u1_moptions;	/* IP multicast options */
		u_short	u1_hashval;	/* Hash value */
	   } pcb_u1;
	   /* The hashing stuff */
	   struct {
		/*
		 * Pcb_u2 is currently
#ifdef _K64U64
		 *	24 bytes smaller
#else
		 *	36 bytes smaller
#endif	
		 * than pcb_u1.
		 */
		mrsplock_t	*u2_mrlock;	/* MP locking for list */
		u_short		(*u2_hashfun)(struct inpcb *, struct in_addr,
			u_short,struct in_addr,u_short);

		int		u2_wildport;	/* wildcard port */
	    	int		u2_tablesz;	/* size of table */

		struct in_pcbhead    *u2_table; /* hash tbl inpcb's ptrs, etc*/
		struct inpcb	**u2_cache;	/* cached pcb address */

		struct inpcb	*u2_last;	/* last pcb per protocol */
		u_int		*u2_dstats;	/* distribution statistics */
		u_short		u2_stats;	/* statistics */
	   } pcb_u2;
	} inp_u;
};

#define	inp_faddr	inp_iap.iap_faddr
#define	inp_fport	inp_iap.iap_fport
#define	inp_laddr	inp_iap.iap_laddr
#define	inp_lport	inp_iap.iap_lport

#define	inp_socket	inp_u.pcb_u1.u1_socket
#define	inp_ppcb	inp_u.pcb_u1.u1_ppcb
#define	inp_route	inp_u.pcb_u1.u1_route
#define	inp_options	inp_u.pcb_u1.u1_options

#define	inp_moptions	inp_u.pcb_u1.u1_moptions
#define	inp_ip		inp_u.pcb_u1.u1_ip
#define	inp_ip_ttl	inp_u.pcb_u1.u1_ip.ip_ttl
#define	inp_ip_tos	inp_u.pcb_u1.u1_ip.ip_tos
#define	inp_hashval	inp_u.pcb_u1.u1_hashval
#define inp_flags	inp_u.pcb_u1.u1_flags

#define	inp_mrlock	inp_u.pcb_u2.u2_mrlock
#define	inp_hashfun	inp_u.pcb_u2.u2_hashfun
#define	inp_tablesz	inp_u.pcb_u2.u2_tablesz
#define	inp_table	inp_u.pcb_u2.u2_table
#define	inp_cache	inp_u.pcb_u2.u2_cache
#define inp_last	inp_u.pcb_u2.u2_last
#define inp_stats	inp_u.pcb_u2.u2_stats
#define inp_dstats	inp_u.pcb_u2.u2_dstats
#define inp_wildport	inp_u.pcb_u2.u2_wildport

/* flags in inp_flags: */
#define	INP_RECVOPTS		0x01	/* receive incoming IP options */
#define	INP_RECVRETOPTS		0x02	/* receive IP options for reply */
#define	INP_RECVDSTADDR		0x04	/* receive IP dst address */
#define	INP_CONTROLOPTS		(INP_RECVOPTS|INP_RECVRETOPTS|INP_RECVDSTADDR)
#define	INP_HDRINCL		0x08	/* user supplies entire IP header */

/* flags for in_pcblookup operation */
#define	INPLOOKUP_WILDCARD	0x1
#define	INPLOOKUP_SETLOCAL	0x2
#define	INPLOOKUP_LISTEN	0x4
#define	INPLOOKUP_ALL		0x8
#define	INPLOOKUP_BIND		0x10

/* flags in inp_hashflags */
#define INPFLAGS_CLTS		0x0001		/* Connectionless service */
#define INPFLAGS_LISTEN		0x0002		/* Listening endpoint */
#define INPFLAGS_MOVED		0x0004		/* PCB has moved */

#define IP_ADDRS_EQUAL(inp, laddr, lport, faddr, fport) \
			(((inp)->inp_lport == (lport)) && \
			((inp)->inp_fport == (fport)) && \
			((inp)->inp_faddr.s_addr == (faddr).s_addr) && \
			((inp)->inp_laddr.s_addr == (laddr).s_addr))

#define	sotoinpcb(so)	((struct inpcb *)(so)->so_pcb)

#ifdef _KERNEL

#define _TBLELEM(inp)	(&(inp)->inp_head->inp_table[(inp)->inp_hashval])
#define _TBLCACHE(inp)	((inp)->inp_head->inp_cache[(inp)->inp_hashval])
#define _TBLELEMHEAD(inp)	_TBLELEM(inp)->hinp_next
#define _TBLELEMTAIL(inp)	_TBLELEM(inp)->hinp_prev

#ifdef _MP_NETLOCKS
#include "sys/atomic_ops.h"
#define INPCB_HOLD(inp)			atomicAddInt(&(inp)->inp_refcnt, 1)

#define INHEAD_SPINCOUNT	10
#define INHEAD_INITMRLOCK(inp) { \
	(inp)->inp_mrlock = (mrsplock_t *)kern_calloc(1, sizeof(mrsplock_t)); \
	mrinit_splock((inp)->inp_mrlock, "inpmrlk", INHEAD_SPINCOUNT); \
}
#define INHEAD_MRLOCK(inp, type)	mrsplock((inp)->inp_mrlock, (type))
#define INHEAD_MRRLOCK(inp)		mrsplock((inp)->inp_mrlock, MR_ACCESS)
#define INHEAD_MRWLOCK(inp)		mrsplock((inp)->inp_mrlock, MR_UPDATE)
#define INHEAD_MRUNLOCK(inp)		mrspunlock((inp)->inp_mrlock)
#define INHEAD_ISMRLOCKED(inp)		ismrsplocked((inp)->inp_mrlock, \
					MR_ACCESS|MR_UPDATE)
#define INPCB_RELE(inp)			inpcb_rele(inp)

#else
#define INHEAD_INITMRLOCK(inp) 
#define INHEAD_MRLOCK(inp, type)	
#define INHEAD_MRRLOCK(inp)	
#define INHEAD_MRWLOCK(inp)
#define INHEAD_MRUNLOCK(inp)	
#define	INHEAD_ISMRLOCKED(inp)	(1)
#define INPCB_HOLD(inp) { \
		(inp)->inp_refcnt++; \
}
#define INPCB_RELE(inp)			inpcb_rele(inp)

#endif /* _MP_NETLOCKS */
extern void	in_pcbinitcb(struct inpcb *, int, u_short, u_short);
extern void	in_losing(struct inpcb *);
extern int	in_pcballoc(struct socket *, struct inpcb *);
extern int	in_pcbbind(struct inpcb *, struct mbuf *);
extern int	in_pcbconnect(struct inpcb *, struct mbuf *);
extern int	in_pcbconnectbyaddr(struct inpcb *, struct mbuf *, 
			struct in_addr);
extern int	in_pcbconnectbyaddrsin(struct inpcb *, struct sockaddr_in *, 
			struct in_addr);
extern int	in_pcbdetach(struct inpcb *);
extern void	in_pcbdisconnect(struct inpcb *);
extern void	in_pcbnotify(struct inpcb *, struct sockaddr *, u_short,
			     struct in_addr, u_short, int,
			     void (*)(struct inpcb *, int, void *),
			     void *);
extern int	in_pcbsetaddr(struct inpcb *, struct sockaddr_in *,
				struct inaddrpair *);

extern void	in_setpeeraddr(struct inpcb *, struct mbuf *);
extern void	in_setsockaddr(struct inpcb *, struct mbuf *);
extern int	inpcb_rele(struct inpcb *);

extern struct inpcb *in_pcblookup(struct inpcb *, struct in_addr, u_short,
				  struct in_addr, u_short, int);
extern void	in_pcbunlink(struct inpcb *);
/* support for TCP listening endpoints */
extern void	in_pcblisten(struct inpcb *);
extern void	in_pcbunlisten(struct inpcb *);
/* support for TCP TIME-WAIT endpoints */
extern void	in_pcbmove(struct inpcb *);
#endif /* _KERNEL */
