/*
 * Copyright 1996 Silicon Graphics, Inc.  All rights reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#ident	"$Revision: 1.2 $"

#include "sys/asm.h"
#include "sys/regdef.h"
#if IP28
#include "ml/ml.h"
#endif

/*
 * Implementation of atomicIncPort
 * Used to avoid locking overhead on MP systems
 *
 * a0 -- address of 32-bit integer to increment
 * a1 -- wrap value
 * a2 -- new value
 *	  When *a0 becomes greater than a1, it is set to a2
 * v0 -- return value; previous contents of *a0
 */
LEAF(atomicIncPort)
	.set 	noreorder

#if IP28
	CACHE_BARRIER_AT(0,a0)          # top factored for sc
#endif
	# Load the old value
1:
	ll	t2,0(a0)
	addiu	t1,t2,1
	ble	t1,a1,norm
	nop
	move	t1, a2
norm:		
	# Try to set the new one
#if IP28
	AUTO_CACHE_BARRIERS_DISABLE
#endif
	sc	t1,0(a0)	
#if IP28
	AUTO_CACHE_BARRIERS_ENABLE
#endif
	beq	t1,zero,1b
	# set return value in branch-delay slot
	move	v0,t2	# BDSLOT
	# Return to caller
	j	ra
	nop
	.set	reorder
END(atomicIncPort)
