/**************************************************************************
 *									  *
 * 		 Copyright (C) 1995, Silicon Graphics, Inc.		  *
 *									  *
 *  These coded instructions, statements, and computer programs  contain  *
 *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
 *  are protected by Federal copyright law.  They  may  not be disclosed  *
 *  to  third  parties  or copied or duplicated in any form, in whole or  *
 *  in part, without the prior written consent of Silicon Graphics, Inc.  *
 *									  *
 **************************************************************************/

#ident	"$Revision: 1.1 $"
#include <sys/types.h> 
#include <sys/mbuf.h>
#include <sys/tcp-param.h>
#include <net/if.h>
#include <net/route.h>
#include <sys/buf.h>
#include <sys/cipso.h>
#include <sys/debug.h>
#include "sys/cmn_err.h"
#include <sys/errno.h>
#include <sys/mac_label.h>
#include <sys/proc.h>
#include <sys/protosw.h>
#include <sys/sat.h>
#include <sys/so_dac.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/systm.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/user.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip_secopts.h>
#include <netinet/ip.h>
#include <netinet/in_pcb.h>
#include <net/ksoioctl.h>
#include <sys/xlate.h>
#include <sys/capability.h>
#include <sys/var.h>
#include <sys/kmem.h>
#include "sys/sesmgr.h"

extern int sesmgr_enabled;

/*
 * Simply print a message to the effect that network session management is enabled.
 */
void
sesmgr_confignote(void)
{
	cmn_err(CE_CONT, "Network Session Manager Enabled.\n");
}

void
sesmgr_init()
{
	sesmgr_enabled = 1;

#ifdef DEBUG
	printf("Session Manager is enabled\n");
#endif

}

/*ARGSUSED*/
int
sesmgr_usrreq(
        struct socket *so,
        int req,
        struct mbuf *m,
        struct mbuf *nam,       /* XXX sometimes really an "int" */
        struct mbuf *rights)
{
	return 0;
}

int
sesmgr_ip_input (
	struct mbuf *        m,
	struct ifnet *       ifp) 
{
	uid_t	uid;
	mac_label * lbl;

	/* Drop everything until interface labeled */
	if (!ifp->if_sec)
		return 0;

	/* As test, always accept packet */
	return 1;

/*NOTREACHED*/
	lbl = ip_recvlabel(m, ifp, &uid);
	if (lbl == NULL)
		return 0;
	else
		return 1;
}

struct mbuf *
sesmgr_ip_output (
	struct ifnet * ifp,
	struct mbuf * m,
	int * error)
{
	struct mbuf * lbl;

	lbl = ip_xmitlabel(ifp, m, ifp->if_sec->ifs_label_max,
		ifp->if_sec->ifs_uid, error);

	return lbl;
}

struct mac_label *
sesmgr_get_label(struct socket * so) 
{
	struct sesmgr_cipso * sd = (struct sesmgr_cipso *)so->so_sesmgr_data;
	ASSERT(sd != NULL);
	return sd->sm_label;
}

void
sesmgr_set_label(struct socket * so, struct mac_label * label) 
{
	struct sesmgr_cipso * sd = (struct sesmgr_cipso *)so->so_sesmgr_data;
	ASSERT(sd != NULL);
	sd->sm_label = label;
}

struct mac_label *
sesmgr_get_sendlabel(struct socket * so) 
{
	struct sesmgr_cipso * sd = (struct sesmgr_cipso *)so->so_sesmgr_data;
	ASSERT(sd != NULL);
	return sd->sm_sendlabel;
}

void
sesmgr_set_sendlabel(struct socket * so, struct mac_label * label) 
{
	struct sesmgr_cipso * sd = (struct sesmgr_cipso *)so->so_sesmgr_data;
	ASSERT(sd != NULL);
	sd->sm_sendlabel = label;
}

struct soacl *
sesmgr_get_soacl(struct socket * so)
{
	struct sesmgr_cipso * sd = (struct sesmgr_cipso *)so->so_sesmgr_data;
	ASSERT(sd != NULL);
	return sd->sm_soacl;
}

void
sesmgr_set_soacl(struct socket * so, struct soacl * soacl)
{
	struct sesmgr_cipso * sd = (struct sesmgr_cipso *)so->so_sesmgr_data;
	ASSERT(sd != NULL);
	sd->sm_soacl = soacl;
}

uid_t
sesmgr_get_uid(struct socket * so)
{
	struct sesmgr_cipso * sd = (struct sesmgr_cipso *)so->so_sesmgr_data;
	ASSERT(sd != NULL);
	return sd->sm_uid;
}

uid_t
sesmgr_set_uid(struct socket * so, uid_t uid)
{
	struct sesmgr_cipso * sd = (struct sesmgr_cipso *)so->so_sesmgr_data;
	ASSERT(sd != NULL);
	sd->sm_uid = uid;
	return sd->sm_uid;
}

uid_t
sesmgr_get_rcvuid(struct socket * so)
{
	struct sesmgr_cipso * sd = (struct sesmgr_cipso *)so->so_sesmgr_data;
	ASSERT(sd != NULL);
	return sd->sm_rcvuid;
}

uid_t
sesmgr_set_rcvuid(struct socket * so, uid_t uid)
{
	struct sesmgr_cipso * sd = (struct sesmgr_cipso *)so->so_sesmgr_data;
	ASSERT(sd != NULL);
	sd->sm_rcvuid = uid;
	return sd->sm_rcvuid;
}
