/*
 * Copyright 1988 Silicon Graphics, Inc.  All rights reserved.
 *
 * Initialize and miscellany for Irix BSD networking.
 */
#ident "$Revision: 3.23 $"

#include "tcp-param.h"
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/systm.h>
#include <sys/sema.h>
#include <sys/hashing.h>

#include <sys/capability.h>
#include <sys/sat.h>

#include <sys/sysmp.h>
#include <sys/mbuf.h>
#include <sys/mac_label.h>
#include <net/if.h>
#include <net/route.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/in_pcb.h>
#include <netinet/tcp.h>
#include <netinet/tcp_fsm.h>
#include <netinet/tcp_seq.h>
#include <netinet/tcp_timer.h>
#ifdef _MP_NETLOCKS
#include <sys/sockd.h>
#include <net/netisr.h>
#endif

/*
 * Initialize 4.3BSD-derived networking.
 *
 * This function must be called after the drivers have attached themselves.
 * Therefore, it is not simply a module-initialization function that lboot
 * knows about.
 *
 * Notice that mbufs must be initialized before the drivers are attached,
 * and so each driver should call the mbuf initialization function.
 */
void
bsd_init(void)
{
	int s;
	static int initialized = 0;
	extern int domaininit(void);
	extern int in_init(void);
	extern int mflipinit(void);

	s = splimp();
	if (!initialized) {
		loattach();		/* start the loop back interface */
		ifinit();		/* generally initialize interfaces */
		in_init();		/* initialize internet globals */
		domaininit();		/* initialize the domain tables */
		mflipinit();		/* check version of os.a */
		initialized = 1;
		spl0();
#ifdef _MP_NETLOCKS
		netisr_init();
		sockd_init();
		tpisockd_init();
#endif
	}
	splx(s);
}

__int32_t	hostid;		/* XXX move this... */

/* ARGSUSED */
int
gethostid(void *uap, rval_t *rvp)
{

	rvp->r_val1 = hostid;
	return 0;
}

sethostid(sysarg_t *uap)
{

	if (!_CAP_ABLE(CAP_SYSINFO_MGT))
		return EPERM;
	hostid = *uap;
	_SAT_HOSTID_SET(*uap,0);
	return 0;
}
