/*
 * Copyright 1995, Silicon Graphics, Inc.
 * ALL RIGHTS RESERVED
 *
 * UNPUBLISHED -- Rights reserved under the copyright laws of the United
 * States.   Use of a copyright notice is precautionary only and does not
 * imply publication or disclosure.
 *
 * U.S. GOVERNMENT RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in FAR 52.227.19(c)(2) or subparagraph (c)(1)(ii) of the Rights
 * in Technical Data and Computer Software clause at DFARS 252.227-7013 and/or
 * in similar or successor clauses in the FAR, or the DOD or NASA FAR
 * Supplement.  Contractor/manufacturer is Silicon Graphics, Inc.,
 * 2011 N. Shoreline Blvd. Mountain View, CA 94039-7311.
 *
 * THE CONTENT OF THIS WORK CONTAINS CONFIDENTIAL AND PROPRIETARY
 * INFORMATION OF SILICON GRAPHICS, INC. ANY DUPLICATION, MODIFICATION,
 * DISTRIBUTION, OR DISCLOSURE IN ANY FORM, IN WHOLE, OR IN PART, IS STRICTLY
 * PROHIBITED WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SILICON
 * GRAPHICS, INC.
 */

/*
 * IOC3 10/100 Mbps Fast Ethernet Device Driver definitions
 */

#ident "$Header: /proj/banyan/isms/irix/kern/bsd/sys/RCS/if_ef.h,v 1.3 1995/09/18 20:32:58 dwong Exp $"

#ifndef __IF_EF_H__
#define __IF_EF_H__

#include <sys/mii.h>
/*
 * IOC3 10/100 Mbit Fast Ethernet device driver
 */

/*
 * RX buffer
 */
#define	EFRDATASZ	(2048 - sizeof (struct ioc3_erxbuf) \
			 - sizeof(struct etherbufhead))
struct efrxbuf {
	struct ioc3_erxbuf	rb_rxb;
	struct etherbufhead	rb_ebh;
	u_char			rb_data[EFRDATASZ];
};

/*
 * TX descriptor
 */
#define	TXDATASZ	88
struct eftxd {
	__uint32_t	cmd;
	__uint32_t	bufcnt;
	__uint64_t	p1;
	__uint64_t	p2;
	struct ether_header	eh;
	u_char		data[TXDATASZ];
};

/*
 * Multicast logical address filter.
 */
struct lafilter {
	u_char	laf_vec[8];
};

/*
 * Principle (per-interface) structure.
 */
struct ef_info {
	struct etherif	ei_eif;		/* common ethernet interface */
	volatile ioc3_eregs_t	*ei_regs;	/* ioc3 ether registers */
	int		ei_phyunit;

	int		ei_ntxd;		/* # tx descriptors */
	volatile struct eftxd	*ei_txd;	/* tx descriptor ring */
	int		ei_txhead;		/* next txd to post */
	int		ei_txtail;		/* next txd to reclaim */
	struct mbuf	**ei_txm;		/* tx mbufs pending */
	int		ei_linkdown;		/* T/F eisr link failure */

	int		ei_nrbuf;		/* # rbufs to keep posted */
	volatile __uint64_t	*ei_rxd;	/* rx descriptor ring */
	int		ei_rxhead;		/* next rxd to read */
	int		ei_rxtail;		/* next rxd to post */
	struct efrxbuf	*ei_rb;			/* quick ptr to head rbuf */
	struct mbuf	**ei_rxm;		/* posted rbufs */

	struct lafilter	ei_laf;			/* logical address filter */
	int		ei_nmulti;		/* # multicast enabled */

	int		ei_erbar;		/* store into erbar reg */
	int		ei_ssram_bits;		/* emcr parity/bufsiz bits */

	int		ei_remfaulting;		/* T/F phy remote fault */
	int		ei_jabbering;		/* T/F phy jabber fault */

	int		ei_defers;		/* # cumulative defers */
	int		ei_runt;		/* # runt packets */
	int		ei_crc;			/* # crc errors */
	int		ei_frame;		/* # framing errors */
	int		ei_code;		/* # coding error */
	int		ei_giant;		/* # giant pkt errors */
	int		ei_invpreamb;		/* # invalid preamble */
	int		ei_drop;		/* # dropped input packets */
	int		ei_ssramoflo;		/* # ssram overflow */
	int		ei_ssramuflo;		/* # ssram underflow */
	int		ei_memerr;		/* # rx/tx memory error */
	int		ei_parityerr;		/* # rx/tx parity error */
	int		ei_rtry;		/* # tx retry errors */
	int		ei_exdefer;		/* # excessive defers */
	int		ei_lcol;		/* # late collisions */
	int		ei_txgiant;		/* # tx giant packets */
	volatile __uint32_t	*ei_ssram;	/* diag access to ssram */
};

#define	ei_ac		ei_eif.eif_arpcom	/* common arp stuff */
#define	ei_if		ei_ac.ac_if		/* ifnet */
#define	ei_unit		ei_if.if_unit		/* logical unit # */
#define	ei_curaddr	ei_ac.ac_enaddr		/* current ether addr */
#define	ei_addr		ei_eif.eif_addr		/* i/f official addr */
#define	ei_rawif	ei_eif.eif_rawif	/* raw domain interface */
#define	eiftoei(eif)	((struct ef_info*)(eif)->eif_private)

#define	TXDSZ	(sizeof (struct eftxd))
#define	RXDSZ	(sizeof (__uint64_t))

#define	NTXDSML	128			/* # txd desktop configuration */
#define	NTXDBIG	512			/* # txd server configuration */
#define	NRXD	512			/* # rxd */

#define	GIANT_PACKET_SIZE	1536

#endif /* __IF_EF_H__ */
