/*
 * GIO bus Token Ring board
 * Copyright 1992 Silicon Graphics, Inc.  All rights reserved.
 */

#ifndef __IF_GTR_H__
#define __IF_GTR_H__

#ident "$Revision: 1.6 $"

/* special IOCTLs */
#define SIOC_GTR_STO	_IOW('P', 1, struct gtr_dwn)
#define SIOC_GTR_EXEC	_IOW('P', 2, struct gtr_dwn)
#define SIOC_GTR_FET	_IOWR('P', 3, struct gtr_dwn)
#define SIOC_GTR_VERS	_IOR('P', 4, u_long)
#define SIOC_GTR_SIGNAL	_IOW('P', 5, u_long)

#define	GTR_MAXBD	2		/* allow this many boards */

#define GTR_GIO_RESET		0x0100	/* set to reset board (latch) */
#define GTR_GIO_FLAG0		0x0200
#define	GTR_GIO_FLAG1		0x0400
#define	GTR_GIO_FLAG2		0x0800
#define	GTR_GIO_FLAG3		0x1000
#define	GTR_GIO_FLAG_MASK	0x1e00
#define	GTR_GIO_H2B_INTR	0x2000	/* write to create intr on board */

#define	GTR_GIO_B2H_CLR		GTR_GIO_FLAG3
#define GTR_GIO_RESET_DELAY	200	/* usec delay for reset */




/* Following section should match if_gtr_s.h */

#define	GTR_BD_MHZ		16	/* approximate CPU speed */

/* initial 32 bits of message specifying the host communications area */
#define	GTR_GIO_DL_PREAMBLE	0x7ffffffe

/* shape of the board memory as far as the host knows */
/*XXX*/
#define	GTR_EPROM		0
#define	GTR_SRAM		0x400000
#define	GTR_EPROM_SIZE		(128*1024)

/* use SRAM starting here for downloading firmware */
#define	GTR_DOWN		(GTR_SRAM + GTR_EPROM_SIZE)

#define	GTR_FIRM		0
#define	GTR_FIRM_SIZE		(16*1024)

/* the MAC address lives here */
#define	GTR_MAC_SIZE		16
#define	GTR_MAC_ADDR		(GTR_EPROM_SIZE - GTR_MAC_SIZE)

/* the ring speed user defined lives here. */
#define	GTR_SPD_SIZE		16
#define	GTR_SPD_ADDR		(GTR_MAC_ADDR - GTR_SPD_SIZE)

/* the host driver is stored here */
#define	GTR_DRV			(GTR_FIRM + GTR_FIRM_SIZE)
#define	GTR_DRV_SIZE		(GTR_EPROM_SIZE - GTR_FIRM_SIZE - GTR_MAC_SIZE)

/* the EEPROM programming code is put here */
#define	GTR_PROG_SIZE		(1024*2)
#define	GTR_PROG		(GTR_DOWN-GTR_PROG_SIZE)

/* host-to-board commands: values for hc.cmd */
enum gtr_cmd {
	XCMD_NOP=0,		/* 00  do nothing, could be used as poke */
	XCMD_OPEN,		/* 01  insert to the ring */
	XCMD_DIRECT,		/* 02  run direct cmd */
	XCMD_CLOSE,		/* 03  deinsert from the ring */
	XCMD_STO,		/* 04  download to the board */
	XCMD_EXEC,		/* 05  execute downloaded code */
	XCMD_FET,		/* 06  upload from the board */
	XCMD_COUNT,		/* 07  fetch statistic counts */
	XCMD_BUF,		/* 08  add recv buffers */
	XCMD_PROMISC,		/* 09  promiscuous mode */
	XCMD_MADDR,		/* 0A  set multicast address */
	XCMD_FADDR,		/* 0B  set functional address */
	XCMD_GADDR,		/* 0C  get address */
	XCMD_SADDR,		/* 0D  set address */
	XCMD_XMIT,		/* 0E  xmit */
	XCMD_DIAG,		/* 0F  run extensive diagnostics */
	XCMD_GPOOL,		/* 10  Get Recv buf Pool info */
	XCMD_SPOOL		/* 11  Set Recv buf pool info */
};
#define	XCMD_MAX	18

typedef struct gtr_cnts {
/*
 * Adapter stat counts.
 */
	/* From RING STATUS interrupts */
	u_long	signal_loss;		/* # of signal losses */
	u_long	hard_error;		/* # of hard errors */
	u_long	soft_error;		/* # of soft errors */
	u_long	lobe_wire_fault;	/* # of lobe wire fault errors */
	u_long	auto_removal_err;	/* # of auto removal errors */
	u_long	remove_rcv;		/* # of remove receive errors */
	u_long	ring_recovery;		/* # of ring recovery errors */
	u_long	beacons;		/* # of beacon frames received */

	/* from error log */
	u_long	line_err;		/* FCS (CRC) error, etc. */
	u_long	burst_err;		/* no transition for 5 half bits */
	u_long	ari_fci_err;		/* upstream neighbor can't set ari */
	u_long	lost_frame_err;		/* strip whole frame from ring failed */
	u_long	rcv_cong_err;		/* no space to copy in frame */
	u_long	frame_copied_err;	/* line hit or duplicate address */
	u_long	token_err;		/* token error (monitor node only) */
	u_long	dma_bus_errs;		/* DMA bus errors (up to init limit) */
	u_long	dma_parity_errs;	/* DMA parity errs (up to init limit)*/

	/* operational */
	u_long	single_station;		/* # of single station errors */
	u_long	cmd_rejects;		/* # of cmds rejected by tms380 */

	/*  frame xmit err */
	u_long	xmit_incmplt;		/* incomplete xmit frame */
	u_long  xmit_suspend;		/* suspended xmit commands */
	u_long  xmit_list;		/* xmit list error */
	u_long  xmit_listcnt;		/* data cnt list error */
	u_long  xmit_frmsiz;		/* frame size error */
	u_long  xmit_oddaddr;		/* odd fwd ptr */
	u_long  xmit_framgment;		/* bad start bit */
	u_long  xmit_priority;		/* bad access pri */
	u_long  xmit_class;		/* bad MAC class */
	u_long  xmit_format;		/* bad FC error */

	/*  frame recv err */
	u_long	rcv_incmplt;		/* incomplete receives */
	u_long	rcv_suspend;		/* suspended rcv commands */

/*
 * Firmware stat counts
 */
	u_long	misfrm;			/* R frm dropped due to no input buf */
	u_long	xmtabt;			/* X frm aborted due to no xmitlist */
	u_long	frame_ct;		/* R total frame count */
	u_long	byte_ct;		/* R bytes seen by board */

	u_long	e;			/* received E bit */
	u_long	a;			/* received A bit */
	u_long	c;			/* received C bit */

	u_long	bd_hiwat;		/* R BUSD hiwat */
	u_long	rq_hiwat;		/* R max lists in use */
	u_long	xq_hiwat;		/* X max lists in use */

	u_long	xmit_frm_ct;		/* X frame count */
	u_long	frm_cpy_ct;		/* R copied frame count */

	/* new */
	u_long	x_callouts;		/* R callouts */
	u_long	r_callouts;		/* X callouts */

	u_long	x_cpy_ct;		/* X copied frame count */
	u_long	x_byte_ct;		/* X bytes sent by board */

	u_long	x_bd_hiwat;		/* X BUSD hiwat */
	u_long	x_maxsize;		/* X max xmit frame size */
	u_long	r_maxsize;		/* R max recv frame size */
	u_long	r_maxlists;		/* R max lists per frame */
	u_long	x_maxlists;		/* X max lists per frame */
	u_long	x_maxfpc;		/* X max frames per callout */
	u_long	r_maxfpc;		/* R max frames per callout */

	u_long	isema;			/* Iframe semaphore */
	u_long	bad_xseq;		/* bad xmit sequence number */

	u_long	rsv_2;
	u_long	rsv_3;
	u_long	rsv_4;
	u_long	rsv_5;
} GTR_CNTS;

/*
 * Host Communication area
 */
typedef struct gtr_hc {
    /* host-written/board-read portion of the communcations area
     */
    u_long	cmd;			/* one of the xcmd enum */
    u_long	cmd_id;			/* ID used by board to ack cmd */

    union {
	char	cmd_data[64-8];		/* minimize future version problems */

#define	GTR_DWN_LEN	8
	struct gtr_dwn {		/* XCMD_STO, XCMD_EXEC, XCMD_FET */
	    u_long	baddr;		/* board address */
	    u_long	cnt;		/* number of values */
	    union {
	    	u_long	l[GTR_DWN_LEN];
		char	c[GTR_DWN_LEN*4];
	    } val;
	} dwn;

	u_long	    exec;		/* goto downloaded code, exec addr */

	struct hc_init {		/* initialization parameters */
#define	GTR_IN_CKSUM_SIZE 4096
#define	GTR_ONCHIPLLC 	0x80000000	/* onchip llc */
#define	GTR_INCKSUM 	0x40000000	/* board IP cksums received frames */
#define	GTR_OUTCKSUM 	0x20000000	/* board IP cksums xmitted frames */
#define	GTR_PROMISC 	0x10000000	/* Promiscuous mode */
#define	GTR_ALLMULTI 	0x08000000	/* all multi */
#define	GTR_MULTI 	0x04000000	/* all multi */
#define	GTR_XTHRU 	0x02000000	/* break RoundRobin for xmit */
#define	GTR_RTHRU 	0x01000000	/* break RoundRobin for recv */
#define	GTR_EDTINIT 	0x00800000	/* only used by host */
	    u_long  flags;
	    u_long  rcvpad;		/* recv pad bytes */
	    u_long  ctptr;		/* counters for XCMD_FET_CT */

	    u_long  h2b_ring;		/* first xmit ring slot */
	    u_long  h2b_ring_last;	/* last xmit ring slot */

	    u_long  b2h_ring;		/* first recv ring slot */
	    u_long  b2h_ring_last;	/* last recv ring slot */

	    u_long    b2h_sml;		/* small recv bufs */
	    u_long    b2h_med;		/* medium recv bufs */
	    u_long    b2h_big;		/* large recv bufs */

	    u_long  ssb;		/* Direct LLC ssb ptr for ONCHIPLLC */
	    u_long  ssb_parm;		/* Direct LLC ssb parm ptr */
	} hc_init;

	struct hc_mult {		/* XCMD_MAC_MODE */
	    u_long  mode;
#define MMODE_HLEN   (256/8)
	    u_char  hash[MMODE_HLEN];	/* multicast hash bits */
	} hc_mult;

	struct hc_pool {
		u_long	nsml;		/* # of valid sml bufs */
		u_long	nmed;		/* # of valid med bufs */
		u_long	nbig;		/* # of valid big bufs */
		u_long	nxtsml;		/* ridx for sml buf */
		u_long	nxtmed;		/* ridx for med buf */
		u_long	nxtbig;		/* ridx for big buf */
		u_long	smls;		/* sml buf base */
		u_long	meds;		/* med buf base */
		u_long	bigs;		/* big buf base */
	} hc_pool;
    } arg;



/* host-read/board-written part of the communications area */
#define GTR_SIGN	0xfeedbab1
    volatile unsigned long sign;	/* board signature */
    /*			((((yy-92)*13 +  m)*32 +  d)*24 + hr)*60 + minute */
#define GTR_MIN_VERS	((((92-92)*13 +  7)*32 + 22)*24 +  0)*60 + 0
#define GTR_VERS_MASK	0x00ffffff
#define GTR_CKSUM_VERS	0x80000000	/* bad checksum */
#define GTR_LLC_VERS	0x40000000	/* have onchip llc */
#define GTR_4MBPS_VERS	0x20000000	/* it is on 4mbps ring */
    volatile unsigned long vers;	/* board/eprom version */

#define	FSTATE_SANITY	0x00000000	/* sanity */
#define	FSTATE_PUD	0x80000000	/* Power Up diagnostics */
#define	FSTATE_BUD	0x40000000	/* BUD */
#define	FSTATE_INIT	0x20000000	/* INIT */
#define	FSTATE_OPEN	0x10000000	/* OPEN */
#define	FSTATE_RUN	0x08000000	/* RUN */
#define	FSTATE_CLOSE	0x04000000	/* CLOSED */
#define	FSTATE_CHECK	0x02000000	/* ADAPTOR CHECK OCCURED */
#define	FSTATE_OERR	0x01000000	/* Open failed */
#define	FSTATE_MERR	0x00800000	/* MOD Open param failed */
#define	FSTATE_CERR	0x00400000	/* CLOSE failed */
#define	FSTATE_XERR	0x00200000	/* XMIT suspended */
#define	FSTATE_RERR	0x00100000	/* RECV suspended */
#define	FSTATE_NORBUF	0x00080000	/* no RECV buf - operation */

#define	FSTATE_SICK	0x00010000	/* SICK */
#define	FSTATE_RC	0x0000000f	/* State Reason Code */
#define	FSTATE_RC_SUM	0x00000001	/*  eprom check sum failed */
    volatile unsigned long state;	/* init state */

    volatile unsigned long inum;	/* ++ on board-to-host interrupt */
    volatile unsigned long cmd_ack;	/* ID of last completed command */

#define	CMDSTS_SUCCESS	0		/* success */
#define	CMDSTS_INVAL	1		/* invalid cmd */
#define	CMDSTS_FAIL	2		/* cmd failed */
#define CMDSTS_SCBTIMO	3		/* SCB timeout */
#define CMDSTS_CHKSUM	4		/* bad check sum */
#define CMDSTS_IWAIT	5		/* Iframe Wait */
    volatile unsigned long cmd_sts;	/* cmd completionstatus */

    volatile union {
	char		cmd_res[64-24];	/* minimize future version problems */
	struct gtr_dwn	fet;		/* XCMD_FET */
    } res;
} GTR_HC;


/*
 * XMIT/RECV ring
 */
#define	GTR_PAD_IN_DMA	24		/* input padding added by board */
#define	MAX_BDS		8		/* max no of scatter/gather */
#define	RH_MAX		32

typedef struct	busd {
	u_long bd_len;			/* length */
	u_long bd_addr;			/* physical address */
} BUSD;

typedef struct slot {
	u_short	sl_len;			/* total length in a slot */
/* XMIT */
#define SLX_FMASK	0xe0		/* 3bit ONCHIPLLC frame type */

/* RECV */
#define	SLE_PAD1	1		/* 1 byte pad applied */
#define	SLE_PAD2	2		/* 2 byte pad applied */
#define	SLE_PAD3	3		/* 3 byte pad applied */
#define	SLE_LIST	4		/* odd pointer error */
	u_char	sl_sts;			/* complete status/pad bytes */
	u_char	sl_bds;			/* total num of BDs */
	struct busd busd[MAX_BDS];
} SLOT;

/*
 * If ONCHIPLLC, real MAXBD becomes 7 for xmit, and sl_bds+1th bd_len
 * becomes last 32bit of xmit list for SID,RSAP, and hlen.
 */
struct ringhdr {
	u_long	len;			/* len of ring */
	u_long	ridx;			/* read index */
	u_long	ridxs;			/* read index shdow */
	u_long	widx;			/* write index */
};
#define	rh_len		rh.len
#define	rh_ridx		rh.ridx
#define	rh_ridxs	rh.ridxs
#define	rh_widx		rh.widx

typedef struct ring {
	struct ringhdr	rh;
	struct slot	sl[RH_MAX];
} RING;

/*
 * Buffer Pool
 */
#define GTR_MAX_BIG	32		/* clusters in pool */
#define GTR_MAX_MED	64		/* medium sized mbufs in pool */
#define GTR_MAX_SML	64		/* little mbufs in pool */
#define GTR_BIG_SIZE	MCLBYTES	/* size of big mbufs */
#define GTR_MED_SIZE (((1514+GTR_PAD_IN_DMA+MSIZE-1)/MSIZE)*MSIZE-32)
#define GTR_SML_SIZE	MLEN

#define BP_MASK		0x3		/* buf type indicator */
#define	BP_SML		0		/* Small sized buf id. */
#define	BP_MED		1		/* Med sized buf id. */
#define	BP_BIG		2		/* Large sized buf id. */

#define MAX_POOL	(RH_MAX*2)
typedef struct bpool {
	struct ringhdr rh;
	u_long bp_pbuf[MAX_POOL];
	u_long bp_hbuf[MAX_POOL];
} BPOOL;

typedef struct xlob {
	u_short	sid;
	u_char	rsap;
	u_char	hlen_corel;
} XLOB;

#endif /* __IF_GTR_H__ */
