/* Point-to-Point Protocol
 */

#ident "$Revision: 1.16 $"

/* packet codes */
#define PPP_CODE_CR	1		/* Configure-Request */
#define PPP_CODE_CA	2		/* Configure-Ack */
#define PPP_CODE_CN	3		/* Configure-Nak */
#define PPP_CODE_CONFJ	4		/* Configure-Reject */
#define PPP_CODE_TR	5		/* Terminate-Request */
#define PPP_CODE_TA	6		/* Terminate-Ack */
#define PPP_CODE_CJ	7		/* Code-Reject */
#define PPP_CODE_PJ	8		/* Protocol-Reject */
#define PPP_CODE_EREQ	9		/* Echo-Request */
#define PPP_CODE_EREP	10		/* Echo-Reply */
#define PPP_CODE_DIS	11		/* Discard-Request */
/* #define PPP_CODE_??	12 */
/* #define PPP_CODE_??	13 */
#define PPP_CODE_RREQ	14		/* (compression) reset request */
#define PPP_CODE_RACK	15		/* (compression) reset ACK */


#define MAX_PPP_DEVS 64			/* max total interfaces */
#define MAX_PPP_LINKS 4			/* max devices/interface */

#define MIN_VJ_SLOT 3
#define DEF_VJ_SLOT 16
#define MAX_VJ_SLOT 254

#define PPP_DEF_MRU	1500		/* default MRU */
#define PPP_MP_MAX_OVHD 5
#define PPP_MAX_MTU	(10*1024)
#define PPP_MIN_MTU	64

#define PPP_FLAG	0x7e		/* frame flag */
#define PPP_ADDR	0xff
#define PPP_CNTL	0x03

#define PPP_ESC		0x7d		/* Control Escape */
#define PPP_ESC_BIT	0x20		/* toggle to escape control chars */

#define PPP_SET_ACCM(accm,c) ((accm)[(u_char)(c)/SFM] |= (1<<((c)%SFM)))

#define PPP_ASYNC_MAP(c,accm) (0!=(((accm)[(u_char)(c)/SFM]	\
				    >>(((u_char)c)%SFM)) & 1))
#define PPP_ACCM_DEF	0xffffffff

/* protocol numbers */
#define	PPP_IP		0x0021		/* IP */
#define PPP_IPXCP	0x002b		/* IPX */
#define PPP_VJC_COMP	0x002d		/* Van Jacobson Compressed TCP/IP */
#define	PPP_VJC_UNCOMP	0x002f		/* Van Jacobson Uncompressed TCP/IP */
#define PPP_BP		0x0031		/* bridging */
#define PPP_MP		0x003d		/* multilink */
#define PPP_CP		0x00fd		/* compressed packets */
#define PPP_CP_MULTI	0x00fb		/* compression below multi-link */

#define	PPP_IPCP	0x8021		/* IP Control Protocol */
#define PPP_IPXP	0x802b		/* IPX */
#define PPP_BCP		0x8031		/* bridging */
#define PPP_CCP		0x80fd		/* Compression Control Protocol */
#define PPP_CCP_MULTI	0x80fb		/* compression below multi-link */

#define PPP_LCP		0xc021		/* Link Control Protocol */
#define PPP_PAP		0xc023		/* Password Authentication Protocol */
#define	PPP_LQR		0xc025		/* Link Quality Report */
#define PPP_CHAP	0xc223		/* Challenge Handshake Auth Proto */

#define MIN_BSD_COMP_PROTO  0		/* compress these protocols with */
#define MAX_BSD_COMP_PROTO  0x3fff	/*	BSD Compress */

/* Endpoint Discriminator types */
#define LCP_CO_ENDPOINT_NULL 0
#define LCP_CO_ENDPOINT_LOC 1
#define LCP_CO_ENDPOINT_IP  2
#define LCP_CO_ENDPOINT_MAC 3
#define LCP_CO_ENDPOINT_MAGIC 4
#define LCP_CO_ENDPOINT_PHONE 5
#define LCP_MAX_CO_ENDPOINT LCP_CO_ENDPOINT_PHONE


#ifdef _IRIX4
#define __uint32_t u_int
#define __psunsigned_t u_long
#endif
#define PPP_FM  __uint32_t		/* type of filter bit mask */
#define SFM (sizeof(PPP_FM)*8)		/* size of filter bit mask in bits */

/* port activity filter
 * 1=>do not worry about this packet
 */
struct afilter {
    PPP_FM	port[0x10000/sizeof(PPP_FM)];
    PPP_FM	icmp[256/sizeof(PPP_FM)];
};
#define IGNORING_PORT(f,p) ((f) != 0 && (((f)->port[(p)/SFM]>>((p)%SFM)) & 1))
#define IGNORING_ICMP(f,t) ((f) != 0 && (((f)->icmp[(t)/SFM]>>((t)%SFM)) & 1))
#define SET_PORT(f,p)	    ((f)->port[(p)/SFM] |= ((PPP_FM)1<<((p)%SFM)))
#define SET_ICMP(f,t)	    ((f)->icmp[(t)/SFM] |= ((PPP_FM)1<<((t)%SFM)))


/* PPP "Predictor" compression */
struct pred_db {
	u_short	hash;
	u_short mru;
	char	debug;
	u_char  unit;
	u_char	tbl[65536];
};
#define PRED_OVHD (4+2)			/* Predictor 1 overhead/packet */
#ifdef _KERNEL
extern struct pred_db *pf_pred1_init(struct pred_db*,int,int);
extern int pf_pred1_comp(struct pred_db*, u_char*,int,struct mbuf*,int,int);
extern mblk_t* pf_pred1_decomp(struct pred_db*, mblk_t*);
#endif


/* PPP "BSD Compress" */
struct bsd_db {
    int	    totlen;			/* length of this structure */
    u_int   hsize;			/* size of the hash table */
    u_char  hshift;			/* used in hash function */
    u_char  n_bits;			/* current bits/code */
    u_char  debug;
    u_char  unit;
    u_short mru;
    u_short seqno;			/* # of last byte of packet */
    u_int   maxmaxcode;			/* largest valid code */
    u_int   max_ent;			/* largest code in use */
    u_int   in_count;			/* uncompressed bytes */
    u_int   bytes_out;			/* compressed bytes */
    u_int   ratio;			/* recent compression ratio */
    u_int   checkpoint;			/* when to next check the ratio */
    int	    clear_count;		/* times dictionary cleared */
    int	    incomp_count;		/* incompressible packets */
    int	    decomp_count;		/* packets decompressed */
    int	    overshoot;			/* excess decompression buf */
    int	    undershoot;			/* insufficient decomp. buf */
    u_short *lens;			/* array of lengths of codes */
    struct bsd_dict {
	union {				/* hash value */
	    __uint32_t	fcode;
	    struct {
#ifdef BSD_LITTLE_ENDIAN
		u_short prefix;		/* preceding code */
		u_char  suffix;		/* last character of new code */
		u_char	pad;
#else
		u_char	pad;
		u_char  suffix;		/* last character of new code */
		u_short prefix;		/* preceding code */
#endif
	    } hs;
	} f;
	u_short codem1;			/* output of hash table -1 */
	u_short cptr;			/* map code to hash table entry */
    } dict[1];
};
#define BSD_OVHD (2+2)			/* BSD compress overhead/packet */
#define MIN_BSD_BITS	9
#define MAX_BSD_BITS	15		/* implementation limit */
#define BSD_VERS	1		/* when shifted */
#ifdef _KERNEL
extern struct bsd_db *pf_bsd_init(struct bsd_db*, int, int, int);
extern int pf_bsd_comp(struct bsd_db*,u_char*,int,struct mbuf*,int);
extern mblk_t* pf_bsd_decomp(struct bsd_db*, mblk_t*);
extern void pf_bsd_incomp(struct bsd_db*, mblk_t*, u_int);
#endif


/* All int's in the kernel-user interface are 32-bits long to avoid
 * translation problems, so that a 32-bit daemon can be used on
 * all systems.
 */
#define IF_PPP_VERSION 4		/* change when the interface does */
struct ppp_arg {
    __uint32_t	dev_index;
    __uint32_t	version;		/* version of kernel modules */
    union {
	u_char	ok;
	struct {
	    __uint32_t	mtru;
	    __uint32_t	qmax;
	    __uint32_t  bigxmit;
	    u_char	telnettos;
	} mtu;
	struct ppp_arg_lcp {
#	    define  PPP_ACCM_LEN (256/SFM)
	    PPP_FM  tx_accm[PPP_ACCM_LEN];
	    PPP_FM  rx_accm;
	    u_char  bits;
#		define SIOC_PPP_LCP_ACOMP   0x01    /* compress A field */
#		define SIOC_PPP_LCP_PCOMP   0x02    /* compress protocol */
#		define SIOC_PPP_LCP_SYNC    0x04    /* SYNC link */
	    u_char  s_d_pad;
	} lcp;
	struct {
	    u_char  tx_comp;
	    u_char  tx_compslot;
	    u_char  tx_slots;
	    u_char  rx_slots;
	    u_char  link_num;
	} vj;
	struct {
	    u_char  proto;
#		define	SIOC_PPP_CCP_PRED1  0x01    /* Predictor type 1 */
#		define	SIOC_PPP_CCP_BSD    0x02    /* BSD compress */
#		define	SIOC_PPP_CCP_PROTOS (SIOC_PPP_CCP_PRED1	\
					     | SIOC_PPP_CCP_BSD)
	    u_char  debug;
	    u_char  bits;
	    u_short mru;
	    u_char  unit;
	} ccp;
	struct {
	    u_char  id;
	} ccp_arm;
	struct {
	    __uint32_t	mtu;		/* per-link MTU */
	    __uint32_t	frag_size;	/* 0 or max frag size */
	    __uint32_t	reasm_window;	/* limit on unassembled fragments */
	    u_char	on;		/* 1=turn on MP */
	    u_char	send_ssn;	/* 1=send short sequence numbers */
	    u_char	recv_ssn;	/* 1=receive short sequence numbers */
	    u_char	must_use_mp;	/* 1=must use MP headers */
	    u_char	mp_nulls;	/* 1=ok to use MP null fragmentss */
	    u_char	debug;
	} mp;
    } v;
};


/* these IOCTLs do not currently work via sockets. */
#ifdef _KERNEL
#define SIOC_PPP_AFILTER _IOWR('P',1, __uint32_t)	/* address filter */
#else
#ifdef _K64U64
		/* this interface is only for 32-bit mode programs */
#else
#endif
#define SIOC_PPP_AFILTER _IOWR('P',1, struct afilter*)	/* address filter */
#endif
#define SIOC_PPP_NOICMP	_IOWR('P',2, struct ppp_arg)	/* suppress ICMP */
#define SIOC_PPP_MTU	_IOWR('P',3, struct ppp_arg)	/* set MTU */
#define SIOC_PPP_LCP	_IOWR('P',4, struct ppp_arg)	/* set LCP parameters */
#define SIOC_PPP_VJ	_IOWR('P',5, struct ppp_arg)	/* set VJ comp */
#define SIOC_PPP_RX_OK	_IOWR('P',6, struct ppp_arg)	/* allow link IP rx */
#define SIOC_PPP_TX_OK	_IOWR('P',7, struct ppp_arg)	/* allow link IP tx */
#define SIOC_PPP_1_RX	_IOWR('P',8, struct ppp_arg)	/* allow 1 rx frame */
#define SIOC_PPP_CCP_RX	_IOWR('P',9, struct ppp_arg)	/* set rx CCP */
#define SIOC_PPP_CCP_TX	_IOWR('P',10,struct ppp_arg)	/* set tx CCP */
#define SIOC_PPP_CCP_ARM_CA _IOWR('P',12,struct ppp_arg)
#   define PPP_CCP_DISARM_ID 0
#define SIOC_PPP_IP_TX_OK _IOWR('P',13,struct ppp_arg)	/* enable IP TX */
#define SIOC_PPP_MP	_IOWR('P',14,struct ppp_arg)	/* IETF MP on or off */

#ifdef _KERNEL
extern int pf_sioc_rx_init(struct iocblk*, struct ppp_arg*,
			   struct pred_db**, struct bsd_db**, int, int);
#endif


#define BEEP		(HZ*5)		/* poke daemon this often */

/* shape of a PPP packet */
struct ppp_buf {
    __uint32_t	type;
    short	dev_index;
    u_char	bits;
    u_char	frame;			/* last of PPP_BUF_HEAD bytes */
    u_char	addr;
    u_char	cntl;
    u_short	proto;
    union {
	u_char	    info[PPP_MAX_MTU];
	__uint32_t  l_info[(PPP_MAX_MTU+3)/4];
	struct	{			/* ordinary packet */
	    u_char	code;
	    u_char	id;
	    u_short	len;
	} pkt;
	struct mp_l {			/* MP header with long sequence # */
	    u_int	mp_bits:2;
	    int		rsvd:6;
	    u_int	sn:24;
	} mp_l;
	struct mp_s {			/* MP header with short sequence # */
	    u_int	mp_bits:2;
	    int		rsvd:2;
	    u_int	sn:12;
	} mp_s;
#	define MP_B_BIT	0x02		/* 1st MP fragment */
#	define MP_E_BIT 0x01		/* last MP fragment */
#	define MP_BIT_SHIFT 6
#	define MP_LSN_MASK 0xffffff
#	define MP_SSN_MASK 0xfff
#	define MP_MAX_SN (0x40000000)	/* wrap seq # at this */
#	define MP_TRIM_SN (MP_MAX_SN/2)	/* trim by this at wrap */
    } un;
};
#define PPP_BUF_HEAD 8			/* bytes before link-layer packet */
#define PPP_BUF_HEAD_FRAME (PPP_BUF_HEAD-1) /* bytes before framing byte */
#define PPP_BUF_HEAD_PROTO (PPP_BUF_HEAD+2) /* bytes before before protocol */
#define PPP_BUF_HEAD_INFO (PPP_BUF_HEAD+4)  /* bytes before good stuff */
#define PPP_BUF_TAIL 3			/* bytes after info, FCS and frame */
#define PPP_BUF_MAX sizeof(struct ppp_buf)
#define PPP_BUF_MIN (PPP_BUF_MAX-PPP_MAX_MTU)

#define PPP_BUF_ALIGN(p) (0 == ((__psunsigned_t)(p) % sizeof(__uint32_t)))

/* bytes before protocol in MP packet */
#define MB_BUF_HEAD_L  (PPP_BUF_HEAD_INFO+4)
#define MB_BUF_HEAD_S  (PPP_BUF_HEAD_INFO+2)
#define MP_BUF_HEAD(ssn) ((ssn) ? MB_BUF_HEAD_S : MB_BUF_HEAD_L)

/* values for ppp_buf.type */
enum ppp_buf_type {
    BEEP_ACTIVE	=1,			/* looks like long activity */
    BEEP_WEAK	=2,			/* the streams module says "Hi" */
    BEEP_DEAD	=3,			/* the modem is dead */
    BEEP_CP_BAD	=4,			/* compression has gone off */
    BEEP_FRAME	=5,			/* a PPP frame */
    BEEP_FCS	=6,			/* complaint about bad FCS */
    BEEP_ABORT	=7,			/* complaint about aborted frame */
    BEEP_TINY	=8,			/* complaint about truncated frame */
    BEEP_BABBLE	=9,			/* complaint about bloated frame */
    BEEP_S_D_PAD=10,			/* bad self-describing-padding */
    BEEP_MP	=11,			/* MP sequence number */
    BEEP_MP_ERROR=12			/* bad MP header */
};

/* bits in ppp_buf.addr for BEEP_ACTIVE and BEEP_WEAK */
#define BEEP_ADDR_BUSY	1
#define BEEP_ADDR_IDLE	2


#define PPP_FCS_INIT	0xffff		/* Initial FCS value */
#define PPP_FCS_GOOD	0xf0b8		/* Good final FCS value */

/* add to the FCS */
#define PPP_FCS(fcs,c) ((((u_short)(fcs))>>8)	\
			^ ppp_fcstab[((fcs) ^ (c)) & 0xff])
extern u_short ppp_fcstab[256];

#define PPP_MAX_SBUF MIN(PPP_BUF_MAX, MAXBSIZE)
