#ifndef __SYS_MBUF_H__
#define __SYS_MBUF_H__

/*
 * Copyright (C) 1994 Silicon Graphics, Inc. All rights reserved.
 *
 * Copyright (c) 1982, 1986, 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 *	@(#)mbuf.h	7.10 (Berkeley) 2/8/88
 */

#include <sys/immu.h>

/*
 * Constants related to memory allocator.
 */
#if NBPP == 4096
#if _MIPS_SZLONG == 32
#define	MSIZE	    128			/* size of an mbuf */
#else
#define	MSIZE	    256			/* size of an mbuf */
#endif /* _MIPS_SZLONG == 32 */
#elif NBPP == 16384
#define	MSIZE	    256			/* size of an mbuf */
#elif !defined(_KERNEL)
#define MSIZE	128
#else
? ? ?	The size of a mbuf is dependent on the page size.
#endif
#define	MBUFCL_SIZE	2048		/* Size in bytes of cluster mbuf */

#define VCL_MAX	    CLBYTES		/* largest variable cluster */
#define	MCLBYTES    CLBYTES

#if _MIPS_SZPTR == 32
#define	MMINOFF	    16			/* mbuf header length */
#elif _MIPS_SZPTR == 64
#define	MMINOFF	    32			/* mbuf header length */
#else
? ? ?   We must know the size of pointers to compute the header size.
#endif

#define	MTAIL	    0
#define	MMAXOFF	    (MSIZE-MTAIL)	/* offset where data ends */
#define	MLEN	    (MSIZE-MMINOFF-MTAIL)   /* mbuf data length */

/*
 * This structure is used to define the sizes of various constants that are
 * shared between kernel and administrative utilities.
 */
struct mbufconst {
	int		m_msize;	/* MSIZE */
	int		m_mbufclsize;	/* MBUFCL_SIZE */
	int		m_mclbytes;	/* MCLBYTES */
	int		m_mminoff;	/* MMINOFF */
	int		m_mtail;	/* MTAIL */
	int		m_maxoff;	/* MMAXOFF */
	int		m_mlen;		/* MLEN */
};

/*
 * Macros for type conversion
 */

/* address in mbuf to mbuf head */
#define	datatom(x)	((struct mbuf *)((__psint_t)(x)-1 & ~(MSIZE-1)))
#define sotom(x)	datatom(x)
#define rttom(x)	datatom(x)
#define rcbtom(x)	datatom(x)
#define inpcbtom(x)	datatom(x)
#define tcptmplttom(x)	datatom(x)
#define unptom(x)	datatom(x)
#define mholdtom(x)	datatom(x)

/* mbuf head, to typed data */
#define	mtod(x,t)	((t)((__psint_t)(x) + (x)->m_off))

struct mbuf {
	struct	mbuf *m_next;		/* next buffer in chain */
	__psint_t m_off;		/* offset of data */
	struct	mbuf *m_act;		/* link in higher-level mbuf list */
	u_short	m_len;			/* amount of data in this mbuf */
	u_char	m_flags;		/* attributes */
	u_char	m_type;			/* mbuf type (MT_FREE == free) */
#if (_MIPS_SZPTR == 64)
	__int32_t m_filler;		/* Make sure mu_dat's 8-byte aligned. */
#endif /* _MIPS_SZPTR == 64 */
	union {
		/*
		 * cluster mbuf can have it's underlying memory anywhere.
		 * The 'freefunc' is called with the given arguments when
		 * the cluster mbuf is done being used.
		 */
		struct {
#ifdef DEBUG
#define	_FUNCPARAM	struct mbuf *, inst_t *
#else
#define	_FUNCPARAM	struct mbuf *
#endif
			void	(*mu_freefunc)(_FUNCPARAM);
			long	mu_farg;
			void	(*mu_dupfunc)(_FUNCPARAM);
			long	mu_darg;
#undef _FUNCPARAM
			u_short mu_nodesize; /* cluster buffer size in bytes */
		} m_us;
		u_char	mu_dat[MLEN];	/* data storage */
	} m_u;
};
#define	m_dat		m_u.mu_dat		/* compatability */
#define	m_freefunc	m_u.m_us.mu_freefunc	/* lazy */
#define	m_farg		m_u.m_us.mu_farg
#define	m_dupfunc	m_u.m_us.mu_dupfunc
#define	m_darg		m_u.m_us.mu_darg
#define	m_cluster	m_flags			/* compatibility */
#define m_nodesize	m_u.m_us.mu_nodesize

/*
 * mbuf types
 */
#define	MT_FREE		0	/* should be on free list */
#define	MT_DATA		1	/* dynamic (data) allocation */
#define	MT_HEADER	2	/* packet header */
#define	MT_SOCKET	3	/* socket structure */
#define	MT_PCB		4	/* protocol control block */
#define	MT_RTABLE	5	/* routing tables */
#define	MT_HTABLE	6	/* IMP host tables */
#define	MT_ATABLE	7	/* address resolution tables */
#define	MT_SONAME	8	/* socket name */
#define	MT_SAT		9	/* security audit trail */
#define	MT_SOOPTS	10	/* socket options */
#define	MT_FTABLE	11	/* fragment reassembly header */
#define	MT_RIGHTS	12	/* access rights */
#define	MT_IFADDR	13	/* interface address */

#define MT_CPI		14	/* cluster page info nodes in use (DEBUG) */
#define MT_CLDATA	15	/* number cluster size mbufs in use (DEBUG) */
#define MT_CLPAGE	16	/* number page size mbufs in use (DEBUG) */

#define	MT_DN_DRBUF	14	/* 4DDN driver buffer allocation (OBSOLETE) */
#define	MT_DN_BLK	15	/* 4DDN block allocation (OBSOLETE) */
#define	MT_DN_BD	16	/* 4DDN buffer to mbuf (OBSOLETE) */

#define MT_MCAST_RINFO	16	/* number multicast route info struct in use */
#define MT_IPMOPTS	17	/* internet multicast options */
#define MT_MRTABLE	18	/* multicast routing tables */

#define MT_SOACL	19	/* access control list for socket */
#ifndef XTP
#define MT_MAX		20	/* 'largest' type + 1 */
#else
#define	MT_X_CTRL	20	/* xtp_sendcontrol packet */
#define	MT_X_DIAG	21	/* xtp_senddiag packet */
#define	MT_X_MKPKT	22	/* xtp_outqcopy header/data */
#define	MT_X_MKPKT_T	23	/* xtp_outqcopy trailer */
#define	MT_X_OQ		24	/* xtp_outqappend data */
#define	MT_X_CTXT	25	/* newcontext */
#define	MT_X_NAME	26	/* xsd_matchaddr accept */
#define MT_MAX		27
#endif

/*
 * mbuf flags attributes (MUST be bit assigned, 8-bit wide field)
 */
#define MCL_CLUSTER	0x01	/* undistinguished cluster (compatability) */
#define M_CLUSTER	0x01	/* undistinguished cluster mbuf */

#define M_CKSUMMED	0x02	/* link-level handles checksums */
#define M_BCAST		0x04	/* send/received as link-level broadcast */
#define M_MCAST		0x08	/* send/received as link-level multicast */
#define M_SHARED	0x10	/* shared page (page flipped) cluster mbuf */
#define M_DIRTY		0x20	/* dirty data cache page cluster mbuf */
#define M_COPYFLAGS	(M_CKSUMMED|M_MCAST|M_BCAST)

/*
 * wait flags
 */
#define	M_DONTWAIT	0
#define	M_WAIT		1

/*
 * Length to m_copy to copy all
 */
#define	M_COPYALL	1000000000

/*
 * m_pullup() will pull up additional length if convenient;
 * should be enough to hold headers of second-level and higher protocols.
 */
#define	MPULL_EXTRA	32

/*
 * 4.3BSD compatiblity macros
 */
#define	MFREE(m, n)	((n) = m_free(m))
#define	MGET(m, i, t)	((m) = m_get(i, t))
#define	MGETHDR(m, i, t) ((m) = m_get(i, t))

#define	M_HASCL(m)	(((m)->m_flags & M_CLUSTER) != 0)
#define M_TRAILINGSPACE(m) (M_HASCL(m) ? 0 : MLEN - ((m)->m_off + (m)->m_len))

#define MBUF_IS_CLUSTER(m)	((m)->m_flags & M_CLUSTER)
#define MBUF_IS_DIRTY(m)	((m)->m_flags & M_DIRTY)
#define MBUF_IS_SHARED(m)	((m)->m_flags & M_SHARED)

/*
 * Similar to MCLALLOC() and MCLGET(), allocating a buffer of CLBYTES.
 * Unlike the 4.3BSD macros, it allocates regular mbuf as well as the page.
 */
#define MCL_GET(m, i)	(m = m_vget(M_DONTWAIT, CLBYTES, MT_DATA))

/*
 * Release the CLBYTES mbuf allocated byte MCL_GET()
 */
#define MCL_FREE(m)	m_freem(m)

#ifdef DEBUG
/* assert that an mbuf type is reasonable */
#define GOODMT(t) {ASSERT((t) != MT_FREE); ASSERT((t) < MT_MAX);}

/* assert that an mbuf pointer is 'good' (may be K2 in large mem configs) */
#define GOODMP(m) ASSERT((IS_KSEG0(m) || IS_KSEG2(m)) && ((__psint_t)(m) % MSIZE) == 0)

#else
#define GOODMT(t)
#define GOODMP(m)
#endif

#ifdef	_KERNEL
extern void	mbinit(void);	/* initialize mbuf stuff; OBSOLETE now no-op */

extern void	m_adj(struct mbuf*, int);	/* trim an mbuf */
extern void	m_cat(struct mbuf*, struct mbuf *);	/* concatenate */
extern struct mbuf *	m_copy(struct mbuf*, int, int);
extern int	m_datacopy(struct mbuf*, int, int, caddr_t); /* mbuf to buf */
extern caddr_t	m_flip(struct mbuf*, caddr_t, int);  /* flip to user */
extern void	m_flip_finished(caddr_t *, int); /* free pages saved by m_flip */
extern struct mbuf *m_shget(int, int, caddr_t, int, int);
extern void	m_shget_finished(void);

extern struct mbuf *m_free(struct mbuf*);   /* free an mbuf/cluster */
extern void	    m_freem(struct mbuf*);  /* free a chain */

extern struct mbuf *m_get(int, int);	/* alloc an mbuf */
extern struct mbuf *m_getclr(int, int);	/* alloc and zero-fill an mbuf */
extern struct mbuf *m_vget(int, int, int);  /* alloc variable-sized cluster */

extern int	m_length(struct mbuf*);  /* data length of chain */
extern struct mbuf *m_pullup(struct mbuf*, int);    /* condense an mbuf */

extern void *mcb_get(int, int, int); /* alloc control block by type */
extern void mcb_free(void *, int, int); /* free  control block by type */

#endif /* _KERNEL */

#endif /* !__SYS_MBUF_H__ */
