#ifndef __SYS_TCPIPSTATS_H__
#define __SYS_TCPIPSTATS_H__

#ident	"$Revision: 1.7 $"
/*
 * System wide TCP/IP network statistics.
 * These include those for IP, UDP, TCP and Multicast IP packets
 * The various structures were copied from there respective header files
 * to allow them to be easily incorporated into the system wide statistics
 * maintained for each cpu. This allows us to avoid having to acquire a
 * spin lock on MP machines just to update the statistic count.
 */
struct	ipstat {			/* IP statistics */
	u_int	ips_total;		/* total packets received */
	u_int	ips_badsum;		/* checksum bad */
	u_int	ips_tooshort;		/* packet too short */
	u_int	ips_toosmall;		/* not enough data */
	u_int	ips_badhlen;		/* ip header length < data size */
	u_int	ips_badlen;		/* ip length < ip header length */
	u_int	ips_fragments;		/* fragments received */
	u_int	ips_fragdropped;	/* frags dropped (dups, out of space)*/
	u_int	ips_fragtimeout;	/* fragments timed out */
	u_int	ips_forward;		/* packets forwarded */
	u_int	ips_cantforward;	/* packets rcvd for unreachable dest */
	u_int	ips_redirectsent;	/* packets forwarded on same net */
	u_int	ips_noproto;		/* unknown or unsupported protocol */
	u_int	ips_delivered;		/* datagrams delivered to upper level*/
	u_int	ips_localout;		/* total ip packets generated here */
	u_int	ips_odropped;		/* lost packets due to nobufs, etc. */
	u_int	ips_reassembled;	/* total packets reassembled ok */
	u_int	ips_fragmented;		/* datagrams sucessfully fragmented */
	u_int	ips_ofragments;		/* output fragments created */
	u_int	ips_cantfrag;		/* don't fragment flag was set, etc. */
	u_int	ips_badoptions;		/* error in option processing */
	u_int	ips_noroute;		/* packtes discarded due to no route */
	u_int	ips_badvers;		/* ip version != 4 */
	u_int	ips_rawout;		/* total raw ip packets generated */
};

struct	udpstat {			/* UDP statistics */
					/* input statistics */
	u_int	udps_ipackets;		/* total input packets */
	u_int	udps_hdrops;		/* packet shorter than header */
	u_int	udps_badsum;		/* checksum error */
	u_int	udps_badlen;		/* data length larger than packet */
	u_int	udps_noport;		/* no socket on port */
	u_int	udps_noportbcast;	/* of above, arrived as broadcast */
	u_int	udps_fullsock;		/* not delivered, input socket full */
					/* output statistics */
	u_int	udps_opackets;		/* total output packets */
	u_int	udpps_pcbcachemiss;	/* input packets missing pcb cache */
};

/*
 * TCP statistics.
 * Many of these should be kept per connection,
 * but that's inconvenient at the moment.
 */
struct	tcpstat {			/* TCP statistics */
	u_int	tcps_connattempt;	/* connections initiated */
	u_int	tcps_accepts;		/* connections accepted */
	u_int	tcps_connects;		/* connections established */
	u_int	tcps_drops;		/* connections dropped */
	u_int	tcps_conndrops;		/* embryonic connections dropped */
	u_int	tcps_closed;		/* conn. closed (includes drops) */
	u_int	tcps_segstimed;		/* segs where we tried to get rtt */
	u_int	tcps_rttupdated;	/* times we succeeded */
	u_int	tcps_delack;		/* delayed acks sent */
	u_int	tcps_timeoutdrop;	/* conn. dropped in rxmt timeout */
	u_int	tcps_rexmttimeo;	/* retransmit timeouts */
	u_int	tcps_persisttimeo;	/* persist timeouts */
	u_int	tcps_keeptimeo;		/* keepalive timeouts */
	u_int	tcps_keepprobe;		/* keepalive probes sent */
	u_int	tcps_keepdrops;		/* connections dropped in keepalive */

	u_int	tcps_sndtotal;		/* total packets sent */
	u_int	tcps_sndpack;		/* data packets sent */
	u_int	tcps_sndbyte;		/* data bytes sent */
	u_int	tcps_sndrexmitpack;	/* data packets retransmitted */
	u_int	tcps_sndrexmitbyte;	/* data bytes retransmitted */
	u_int	tcps_sndacks;		/* ack-only packets sent */
	u_int	tcps_sndprobe;		/* window probes sent */
	u_int	tcps_sndurg;		/* packets sent with URG only */
	u_int	tcps_sndwinup;		/* window update-only packets sent */
	u_int	tcps_sndctrl;		/* control (SYN|FIN|RST) packets sent*/
	u_int	tcps_sndrst;		/* packets with RST sent */

	u_int	tcps_rcvtotal;		/* total packets received */
	u_int	tcps_rcvpack;		/* packets received in sequence */
	u_int	tcps_rcvbyte;		/* bytes received in sequence */
	u_int	tcps_rcvbadsum;		/* packets received with ccksum errs */
	u_int	tcps_rcvbadoff;		/* packets received with bad offset */
	u_int	tcps_rcvshort;		/* packets received too short */
	u_int	tcps_rcvduppack;	/* duplicate-only packets received */
	u_int	tcps_rcvdupbyte;	/* duplicate-only bytes received */
	u_int	tcps_rcvpartduppack;	/* packets with some duplicate data */
	u_int	tcps_rcvpartdupbyte;	/* dup. bytes in part-dup. packets */
	u_int	tcps_rcvoopack;		/* out-of-order packets received */
	u_int	tcps_rcvoobyte;		/* out-of-order bytes received */
	u_int	tcps_rcvpackafterwin;	/* packets with data after window */
	u_int	tcps_rcvbyteafterwin;	/* bytes rcvd after window */
	u_int	tcps_rcvafterclose;	/* packets rcvd after "close" */
	u_int	tcps_rcvwinprobe;	/* rcvd window probe packets */
	u_int	tcps_rcvdupack;		/* rcvd duplicate acks */
	u_int	tcps_rcvacktoomuch;	/* rcvd acks for unsent data */
	u_int	tcps_rcvackpack;	/* rcvd ack packets */
	u_int	tcps_rcvackbyte;	/* bytes acked by rcvd acks */
	u_int	tcps_rcvwinupd;		/* rcvd window update packets */
	u_int	tcpps_pcbcachemiss;	/* input packets missing pcb cache */
	u_int	tcps_predack;		/* XXX: # hdr predict ok for acks */
	u_int	tcps_preddat;		/* XXX: # hdr predict ok for data */
	u_int	tcps_pawsdrop;		/* segments dropped due to PAWS */
};

/*
 * Internet Control Message Protocol statistics.
 */

#define	ICMP_MAXTYPE		18
/*
 * Also defined in ip_icmp.h but duplicated here to avoid massive header
 * file includes to get all the rest required symbols and constants.
 */

struct icmpstat {
	u_int	icps_dropped;		/* packets deliberately dropped */

		/* ICMP statistics related to packets generated */
	u_int	icps_error;		/* # of calls to icmp_error */
	u_int	icps_oldshort;		/* no error 'cuz old ip too short */
	u_int	icps_oldicmp;		/* no error 'cuz old was icmp */
	u_int	icps_outhist[ICMP_MAXTYPE + 1];

		/* statistics related to input messages processed */
 	u_int	icps_badcode;		/* icmp_code out of range */
	u_int	icps_tooshort;		/* packet < ICMP_MINLEN */
	u_int	icps_checksum;		/* bad checksum */
	u_int	icps_badlen;		/* calculated bound mismatch */
	u_int	icps_reflect;		/* number of responses */
	u_int	icps_inhist[ICMP_MAXTYPE + 1];
};

struct igmpstat {			/* IGMP statistics */
	u_int	igps_rcv_total;		/* total IGMP messages received    */
	u_int	igps_rcv_tooshort;	/* received with too few bytes     */
	u_int	igps_rcv_badsum;	/* received with bad checksum      */
	u_int	igps_rcv_queries;	/* received membership queries     */
	u_int	igps_rcv_badqueries;	/* received invalid queries        */
	u_int	igps_rcv_reports;	/* received membership reports     */
	u_int	igps_rcv_badreports;	/* received invalid reports        */
	u_int	igps_rcv_ourreports;	/* received reports for our groups */
	u_int	igps_snd_reports;	/* sent membership reports         */
};

#define MT_MAX		20	/* maximum number of mbuf types from mbuf.h */
/*
 * mbuf statistics
 */
struct mbstat {			/* MBUF related statistics */
	int	m_mbufs;	/* mbufs obtained from page pool */
	int	m_clusters;	/* clusters obtained from page pool */
	int	m_spare;	/* spare field */
	int	m_clfree;	/* free clusters */
	int	m_drops;	/* times failed to find space */
	int	m_wait;		/* times waited for space */
	int	m_drain;	/* times drained protocols for space */
	int	m_mtypes[MT_MAX];	/* type specific mbuf allocations */
};

/*
 * Per-protocol inpcb statistics
 */
struct inpcbstat {			/* INPCB related statistics */
	u_int	inpcb_notify;		/* Calls to "really" notify */
	u_int	inpcb_lookup;		/* Calls to lookup in the inpcb table */
	u_int	inpcb_regular;		/* Lookup: regular lookup requested */
	u_int	inpcb_wildcard;		/* Lookup: wildcard lookup requested */

	u_int	inpcb_found_cache; 	/* Lookup: Found in head cache */
	u_int	inpcb_found_chain; 	/* Lookup: Found on hash chain */
	u_int	inpcb_searched_all; 	/* Lookup: Searched all hash buckets */
	u_int	inpcb_success;		/* Lookup: Success */
	u_int	inpcb_failed;		/* Lookup: Failed to find a match */
	u_int	inpcb_bind;		/* # of bind calls */
	u_int	inpcb_listen;		/* # of listen lookups */
	u_int	inpcb_connect;		/* # of connect calls */
	u_int	inpcb_connectbyaddr;	/* # of connectbyaddr calls */
	u_int	inpcb_setaddr;		/* # of setaddr */
	u_int	inpcb_setaddrx;		/* # of setaddrx */
	u_int	inpcb_curpcbs;		/* current # of pcbs */
	u_int	inpcb_maxpcbs;		/* max # of pcbs */
	u_int	inpcb_maxbucklen;	/* maximum bucket length */
	struct inpcb *inpcb_last;	/* last PCB looked up on CPU */
};

#define UDP_PCBSTAT		0
#define TCP_PCBSTAT		1
#define NUM_PCBSTAT		2

/*
 * Kernel network activity statistics for the TCP/IP protocol family
 */
struct kna {
	struct ipstat		ipstat;
	struct udpstat		udpstat;
	struct tcpstat		tcpstat;
	struct icmpstat		icmpstat;
	struct igmpstat		igmpstat;
	struct mbstat		mbstat;
	struct inpcbstat	pcbstat[NUM_PCBSTAT];
};

#ifdef _KERNEL

#include <sys/pda.h>	/* for private macro */
#include <sys/immu.h>	/* for PDAPAGE, inside private macro */

#define IPSTAT(x)		((struct kna *)(private.knaptr))->ipstat.x++
#define IPSTAT_DEC(x, c)	((struct kna *)(private.knaptr))->ipstat.x -=c

#define UDPSTAT(x)		((struct kna *)(private.knaptr))->udpstat.x++

#define TCPSTAT(x)		((struct kna *)(private.knaptr))->tcpstat.x++
#define TCPSTAT_ADD(x, c)	((struct kna *)(private.knaptr))->tcpstat.x +=c

#define ICMPSTAT(x)		((struct kna *)(private.knaptr))->icmpstat.x++
#define ICMPSTAT_INHIST(x)	((struct kna *)(private.knaptr))->icmpstat.icps_inhist[x]++
#define ICMPSTAT_OUTHIST(x)	((struct kna *)(private.knaptr))->icmpstat.icps_outhist[x]++

#define IGMPSTAT(x)		((struct kna *)(private.knaptr))->igmpstat.x++

#define PCBSTAT(x,y)		((struct kna *)(private.knaptr))->pcbstat[y].x++
#define PCBSTAT_SET(x,y,c)	((struct kna *)(private.knaptr))->pcbstat[y].x = c
#define PCBSTAT_DEC(x,y)	((struct kna *)(private.knaptr))->pcbstat[y].x--
#define PCBSTAT_ADD(x,y,c)	((struct kna *)(private.knaptr))->pcbstat[y].x += c

#endif /* _KERNEL */
#endif /* __SYS_TCPIPSTATS_H__ */
