/*
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *	based on @(#)uipc_syscalls.c	(Berkeley) tahoe
 */


#include <tcp-param.h>
#include <sys/param.h>
#include <sys/buf.h>
#include <sys/conf.h>
#include <sys/debug.h>
#include <sys/domain.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/mbuf.h>
#include <sys/proc.h>
#include <sys/protosw.h>
#include <sys/sema.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/uio.h>
#include <sys/user.h>
#include <sys/vnode.h>
#include <sys/xlate.h>
#include <sys/systm.h>
#include <sys/sat.h>

extern struct domain unixdomain;
extern struct domain *domains;
#ifdef _MP_NETLOCKS
#ifdef _SCA_DEBUG
extern mutex_t unp_lock;
#define UNP_INITLOCK()	mutex_init(&unp_lock, MUTEX_DEFAULT, "unplock")
#else
extern lock_t unp_lock;
#define UNP_INITLOCK()	spinlock_init(&unp_lock, "unplock")
#endif
#else
#define UNP_INITLOCK()
#endif

int
uds_add(void)
{
	unixdomain.dom_next = domains;
	domains = &unixdomain;
	UNP_INITLOCK();
	return(1);
}

/*
 * System call interface to the socketpair abstraction.
 */
struct socketpaira {
	sysarg_t	domain;
	sysarg_t	type;
	sysarg_t	protocol;
	sysarg_t	rsv;
};


int
socketpair(struct socketpaira *uap)
{
	int error;
	struct file *fp1, *fp2;
	struct socket *so1, *so2;
	int sv[2];
	register struct protosw *prp;
	register struct vnode *vp;

	if (useracc((void *)uap->rsv, 2 * sizeof (int), B_READ) == 0) {
		_SAT_BSDIPC_CREATE_PAIR(0,(struct socket *)NULL,uap->domain,
			uap->protocol, 0,(struct socket *)NULL,EFAULT);
		return EFAULT;
	}

	/*
	 * useracc acquired shdlock (SH_ACC), but we release it now,
	 * therefore the copyout below could fail if a shared process
	 * makes the address unavailable.
	 */
	shdunlock(curprocp);

	if (uap->protocol) {
		error = EPROTONOSUPPORT;
		goto free;
	}

	if (uap->domain != AF_UNIX) {
		error = EPROTONOSUPPORT;
		goto free;
	}

	prp = pffindtype(uap->domain, uap->type);
	if (prp == 0) {
		error = EPROTONOSUPPORT;
		goto free;
	}

	if (prp->pr_type != uap->type) {
		error = EPROTOTYPE;
		goto free;
	}

	error = socreate(uap->domain, &so1, uap->type, uap->protocol);
	if (error)
		goto free;
	error = socreate(uap->domain, &so2, uap->type, uap->protocol);
	if (error)
		goto free1;

	error = makesockvp(&fp1, &sv[0], (void *)so1);
	if (error) {
		goto free2;
	}

	error = makesockvp(&fp2, &sv[1], (void *)so2);
	if (error)
		goto free3;

	error = soconnect2(so1, so2);
	if (error)
		goto free4;
	if (uap->type == SOCK_DGRAM) {
		/*
		 * Datagram socket connection is asymmetric.
		 */
		 error = soconnect2(so2, so1);
		 if (error)
			goto free4;
	}
	fready(fp1);
	fready(fp2);

	(void) copyout(sv, (void *)uap->rsv, sizeof sv);
	_SAT_BSDIPC_CREATE_PAIR(sv[0],so1,uap->domain,uap->protocol,
		sv[1],so2,0);
	return 0;

free4:
	fp2->f_count = 0;
	setf(sv[1], NULLFP);
	vp = fp2->f_vnode;
	unfalloc(fp2);
	dropsockvp(vp);
	VN_RELE(vp);
free3:
	fp1->f_count = 0;
	setf(sv[0], NULLFP);
	vp = fp1->f_vnode;
	unfalloc(fp1);
	dropsockvp(vp);
	VN_RELE(vp);
free2:
	(void)soclose(so2);
free1:
	(void)soclose(so1);
free:
	_SAT_BSDIPC_CREATE_PAIR(0,(struct socket *)NULL,uap->domain,
		uap->protocol, 0,(struct socket *)NULL,error);
	return error;
}
