/*	Copyright (c) 1990, 1991 UNIX System Laboratories, Inc.	*/
/*	Copyright (c) 1984, 1986, 1987, 1988, 1989, 1990 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF     	*/
/*	UNIX System Laboratories, Inc.                     	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident "$Revision: 3.49 $"

#include <sys/param.h>
#include <sys/types.h>
#include <sys/conf.h>
#include <sys/idbgentry.h>
#include <sys/stream.h>
#include <sys/strmp.h>
#include <sys/strmdep.h>
#include <sys/strstat.h>
#include <sys/strsubr.h>
#include <sys/errno.h>
#include <sys/cmn_err.h>
#include <sys/debug.h>
#include <sys/systm.h>
#include <sys/idbgactlog.h>

/*
 * This file contains all the STREAMS utility routines that may
 * be used by modules and drivers.
 */

extern char strbcwait;

#ifdef STREAMS_MEM_TRACE

#define	ALLOCEDBY	1
#define	DUPEDBY		2
#define	FREEDBY		3

static void str_mem_trace(struct str_mem_trace_log *, int, void *, void *);

#define STR_MEM_TRACE(b,t,a1,a2) str_mem_trace(&((b)->tracelog),t,a1,a2)

#else

#define STR_MEM_TRACE(b,t,a1,a2)

#endif /* STREAMS_MEM_TRACE */

/*
 *	Pointer types:
 *			(struct mdbblock *)	mdbfreelist
 *			(struct msgb *)		msgfreelist
 *			(struct mbinfo *)	m_prev, m_next
 */

/*
 * Resource lock and/or appropriate spl held across the following primitives.
 */

#ifdef DEBUG

#define	FREEMDBBLOCK( mdbp )	{ \
	register struct msgb *blockp; \
	blockp = (struct msgb *) mdbfreelist; \
	while (blockp) { \
		ASSERT(blockp != (struct msgb *)(mdbp)); \
		blockp = blockp->b_next; \
	} \
	((struct msgb *)(mdbp))->b_next = (struct msgb *) mdbfreelist; \
	mdbfreelist = (mdbp); \
	STR_MEM_TRACE(((struct msgb *)(mdbp)), FREEDBY, \
		      (inst_t *)__return_address, 0); \
	strst.mdbblock.use--; \
	_DELETE_MDB_INUSE(mdbp); \
}

#define	FREEMSGBLOCK( mesgp )	{ \
	register struct msgb *blockp; \
	blockp = msgfreelist; \
	while (blockp) { \
		ASSERT(blockp != (mesgp)); \
		blockp = blockp->b_next; \
	} \
	(mesgp)->b_next = (struct msgb *) msgfreelist; \
	msgfreelist = mesgp; \
	STR_MEM_TRACE((mesgp), FREEDBY, (inst_t *)__return_address, 0); \
	strst.msgblock.use--; \
	_DELETE_MSG_INUSE((struct mbinfo *)mesgp); \
}

#else /* !DEBUG */

#define	FREEMDBBLOCK( mdbp )	{ \
	((struct msgb *)(mdbp))->b_next = (struct msgb *) mdbfreelist; \
	mdbfreelist = (mdbp); \
	strst.mdbblock.use--; \
	_DELETE_MDB_INUSE(mdbp); \
}

#define	FREEMSGBLOCK( mesgp )	{ \
	(mesgp)->b_next = (struct msgb *) msgfreelist; \
	msgfreelist = mesgp; \
	strst.msgblock.use--; \
	_DELETE_MSG_INUSE((struct mbinfo *)mesgp); \
}

#endif /* DEBUG */

#define ALLOCMDBBLOCK(mdbp)	{ \
	mdbp = mdbfreelist; \
	STR_MEM_TRACE(((struct msgb *)(mdbp)), ALLOCEDBY, \
		      (inst_t *)__return_address, 0); \
	mdbfreelist = (struct mdbblock *)(((struct msgb *)mdbp)->b_next );\
	BUMPUP(strst.mdbblock);	\
	_INSERT_MDB_INUSE(mdbp); \
}


#define ALLOCMSGBLOCK(mesgp)		{ \
	mesgp = msgfreelist; \
	STR_MEM_TRACE((mesgp), ALLOCEDBY, (inst_t *)__return_address, 0); \
	msgfreelist = mesgp->b_next; \
	BUMPUP(strst.msgblock);	\
	_INSERT_MSG_INUSE((struct mbinfo *)mesgp); \
}

static void *str_mem_alloc(int, unsigned char *);
static int str_mem_free(struct datab *);

/* Status flags for triplets */
/* _MDBBLK:	Both message and data headers are valid.*/
/* _NOT_M :	Invalid message header.			*/
/* _NOT_D :	Invalid data header.			*/
/* _XCONN :	Valid headers, but the message header
		points to a different data header	*/ 

#define		_MDBBLK		0
#define		_NOT_M		1
#define		_NOT_D		2
#define		_XCONN		3


/* Convenient macros for accessing fields */

#define	m_bdatap	m_mblock.b_datap
#define m_bnext		m_mblock.b_next
#define m_bprev		m_mblock.b_prev
#define	m_bcont		m_mblock.b_cont
#define m_brptr		m_mblock.b_rptr
#define	m_bwptr		m_mblock.b_wptr
#define	m_bband		m_mblock.b_band
#define	m_bflag		m_mblock.b_flag

#define	d_dbfreep	d_dblock.db_freep
#define d_dbfrtnp	d_dblock.db_frtnp
#define	d_dbbase	d_dblock.db_base
#define	d_dblim		d_dblock.db_lim
#define	d_dbref		d_dblock.db_ref
#define	d_dbtype	d_dblock.db_type
#define d_dbiswhat	d_dblock.db_iswhat
#define	d_dbsize	d_dblock.db_size
#define	d_dbztype	d_dblock.db_ztype
#define	d_dbmsgaddr	d_dblock.db_msgaddr
#define d_dbmbufref	d_dblock.db_mbufref

#define	 _DUPB( newbp, oldbp) { \
	register int sl; \
	LOCK_STR_RESOURCE(sl); \
	ASSERT((oldbp)); \
	if (!msgfreelist){ \
		UNLOCK_STR_RESOURCE(sl); \
		newbp	= dupb(oldbp); \
	} \
	else { \
		ALLOCMSGBLOCK( newbp ); \
		saveaddr(&(((struct mbinfo *)newbp)->m_func)); \
		((newbp)->b_datap = (oldbp)->b_datap)->db_ref++; \
		UNLOCK_STR_RESOURCE(sl); \
		(newbp)->b_next = NULL; \
		(newbp)->b_prev = NULL; \
		(newbp)->b_cont = NULL; \
		(newbp)->b_rptr = (oldbp)->b_rptr; \
		(newbp)->b_wptr = (oldbp)->b_wptr; \
		(newbp)->b_flag = (oldbp)->b_flag; \
		(newbp)->b_band = (oldbp)->b_band; \
		STR_MEM_TRACE(newbp, DUPEDBY, \
			      (inst_t *)__return_address, oldbp); \
		STR_MEM_TRACE(oldbp, DUPEDBY, \
			      (inst_t *)__return_address, newbp); \
	} \
}


/*
 * Caller is expected to be running at splstr() minimum.
 */
#define	 _FREEB(msgp) { \
	register int sl; \
	register struct datab *dbp; \
	LOCK_STR_RESOURCE(sl); \
	dbp  = msgp->b_datap; \
	ASSERT( dbp->db_ref > 0 ); \
	if (dbp->db_msgaddr != (caddr_t)msgp) { \
		FREEMSGBLOCK(msgp); \
	} \
	else { \
		if (((struct mdbblock *)msgp)->datblk.d_dbiswhat == _NOT_D) { \
			FREEMDBBLOCK((struct mdbblock *)msgp); \
		} \
		else ((struct mdbblock *)msgp)->datblk.d_dbiswhat = _NOT_M; \
	} \
	if ( --(dbp->db_ref) == 0 ) { \
		ASSERT(dbp->db_mbufref == 0); \
		STR_MEM_TRACE(dbp, FREEDBY, (inst_t *)__return_address, 0); \
		if ((char *)dbp + sizeof(struct dbinfo) != (char *)(dbp->db_base)) { \
			if (dbp->db_frtnp == NULL)  \
				Strcount -= str_mem_free(dbp); \
			else if (dbp->db_frtnp->free_func) { \
					UNLOCK_STR_RESOURCE(sl); \
				 	(*dbp->db_frtnp->free_func)(dbp->db_frtnp->free_arg); \
					LOCK_STR_RESOURCE(sl); \
			} \
		} \
		if (dbp->db_iswhat == _NOT_M) { \
			msgp = (struct msgb *)((char *)dbp - sizeof(struct mbinfo)); \
			FREEMDBBLOCK((struct mdbblock *)msgp); \
		} \
		else \
			dbp->db_iswhat = _NOT_D; \
	} \
	UNLOCK_STR_RESOURCE(sl); \
}

# ifdef DEBUG

/* routines for managing in-use lists : only in DEBUG mode */

/* Resource lock held by caller */
/* ARGSUSED */
void
delete_msg_inuse(register struct mbinfo *msgptr)
{
#if !_K64U64
	if (msgptr->m_prev) { 
		msgptr->m_prev->m_next = msgptr->m_next; 
		if (msgptr->m_next) 
			msgptr->m_next->m_prev = msgptr->m_prev; 
	} 
	else { 
		Strinfo[DYN_MSGBLOCK].sd_head = (void *)(msgptr->m_next); 
		if (msgptr->m_next) 
			msgptr->m_next->m_prev = NULL;
	} 
#endif
}

/* Resource lock held by caller */
/* ARGSUSED */
void
delete_mdb_inuse(struct mdbblock *mdbptr)
{
#if !_K64U64
	register struct mbinfo *msgptr = &(mdbptr->msgblk);
	register struct dbinfo *datptr = &(mdbptr->datblk);

	if (msgptr->m_prev) { 
		msgptr->m_prev->m_next = msgptr->m_next; 
		if (msgptr->m_next) 
			msgptr->m_next->m_prev = msgptr->m_prev; 
	} 
	else { 
		Strinfo[DYN_MSGBLOCK].sd_head = (void *)(msgptr->m_next); 
		if (msgptr->m_next) 
			msgptr->m_next->m_prev = NULL;
	} 
	if (datptr->d_prev) { 
		datptr->d_prev->d_next = datptr->d_next; 
		if (datptr->d_next) 
			datptr->d_next->d_prev = datptr->d_prev; 
	} 
	else { 
		Strinfo[DYN_MDBBLOCK].sd_head = (void *)(datptr->d_next); 
		if (datptr->d_next) 
			datptr->d_next->d_prev = NULL;
	} 
#endif
}

/* Resource lock held by caller */
/* ARGSUSED */
void
insert_msg_inuse(register struct mbinfo *msgptr)
{
#ifdef XXXrs_BOGUS
#if !_K64U64
	register int s = splstr();

	msgptr->m_next = (struct mbinfo *)(Strinfo[DYN_MSGBLOCK].sd_head);
	Strinfo[DYN_MSGBLOCK].sd_head = (void *)(msgptr);
	if (msgptr->m_next)
		msgptr->m_next->m_prev = (struct mbinfo *)msgptr;
	msgptr->m_prev = NULL;
	splx(s);
#endif
#endif
}

/* Resource lock held by caller */
/* ARGSUSED */
void
insert_mdb_inuse(struct mdbblock *mdbptr)
{
#ifdef XXXrs_BOGUS
#if !_K64U64
	register struct mbinfo *msgptr = &(mdbptr->msgblk);
	register struct dbinfo *datptr = &(mdbptr->datblk);
	register int s = splstr();

	msgptr->m_next = (struct mbinfo *)(Strinfo[DYN_MSGBLOCK].sd_head);
	Strinfo[DYN_MSGBLOCK].sd_head = (void *)(msgptr);
	if (msgptr->m_next)
		msgptr->m_next->m_prev = (struct mbinfo *)msgptr;
	msgptr->m_prev = NULL;

	datptr->d_next = (struct dbinfo *)(Strinfo[DYN_MDBBLOCK].sd_head);
	Strinfo[DYN_MDBBLOCK].sd_head = (void *)(datptr);
	if (datptr->d_next)
		datptr->d_next->d_prev = (struct dbinfo *)datptr;
	datptr->d_prev = NULL;
	splx(s);
#endif
#endif
}

# endif /* DEBUG */

#ifdef STREAMS_MEM_TRACE
static void
str_mem_trace(struct str_mem_trace_log *logp,
	      int t, void *a1, void *a2)
{
	register int idx;

	idx = logp->woff % MEM_TRACE_LOG_SIZE;
	logp->woff++;
	logp->entry[idx].type = t;
	logp->entry[idx].arg1 = a1;
	logp->entry[idx].arg2 = a2;
}

void
str_mem_trace_print(struct str_mem_trace_log *logp, char *pfx)
{
	register int i, j;

	for (i = 0, j = logp->woff % MEM_TRACE_LOG_SIZE;
	     i < MEM_TRACE_LOG_SIZE;
	     i++, j = ++j % MEM_TRACE_LOG_SIZE) {
		qprintf("%stracelog[%d]: type %d a1 0x%x a2 0x%x\n",
			pfx, j,
		        logp->entry[j].type,
		        logp->entry[j].arg1,
		        logp->entry[j].arg2);
	}
}
#endif /* STREAMS_MEM_TRACE */



/*
 * NOTE: Some code in streamio.c relies on knowledge of the underlying
 * message-data-buffer pool structure.
 *
 * 'pri' is no longer used, but is retained for compatibility.
 */
/* ARGSUSED */
mblk_t *
allocb(register int size, uint pri)
{
	register int s, s2, notfast;
	register struct mdbblock *mdbptr;
	register unchar *buf = NULL;


	/* allocate the combined message/data block + FASTBUF sized buffer
	 * from the freelist.
	 */

	notfast = (size > (int)FASTBUF);

	s = splstr();
	LOCK_STR_RESOURCE(s2);
	if ( mdbfreelist ) { 
		ALLOCMDBBLOCK(mdbptr);
		UNLOCK_STR_RESOURCE(s2);
	} else {
		UNLOCK_STR_RESOURCE(s2);
		if (!(mdbptr = xmdballoc())) {
			splx(s); 
			return(NULL);
		}
		STR_MEM_TRACE(((struct msgb *)(mdbptr)), ALLOCEDBY,
			      (inst_t *)__return_address, 0);
	}
	splx(s);

	/* save return address */
	saveaddr(&(((struct mbinfo *)mdbptr)->m_func));

	((struct msgb *)mdbptr)->b_datap = 
		(struct datab *)(((char *)mdbptr) + sizeof(struct mbinfo)); 
	STR_MEM_TRACE((((struct msgb *)mdbptr)->b_datap),
		      ALLOCEDBY, (inst_t *)__return_address, 0);
	((struct msgb *)mdbptr)->b_next = NULL;
	((struct msgb *)mdbptr)->b_prev = NULL;
	((struct msgb *)mdbptr)->b_cont = NULL;
	mdbptr->datblk.d_dbfrtnp = NULL;
	((struct msgb *)mdbptr)->b_band = 0;
	((struct msgb *)mdbptr)->b_flag = 0;

	(mdbptr->datblk).d_dbtype = M_DATA;
	(mdbptr->datblk).d_dbref = 1;
	(mdbptr->datblk).d_dbmbufref = 0;
	(mdbptr->datblk).d_dbmsgaddr = (caddr_t)mdbptr;
	(mdbptr->datblk).d_dbiswhat = _MDBBLK;

	if (notfast) {
		buf = str_mem_alloc(size, &((mdbptr->datblk).d_dbztype));
		if (buf == NULL) {
			LOCK_STR_RESOURCE(s2);
			FREEMDBBLOCK(mdbptr);
			UNLOCK_STR_RESOURCE(s2);
			return(NULL);
		}
		((struct msgb *)mdbptr)->b_rptr = 
		((struct msgb *)mdbptr)->b_wptr = 
				(mdbptr->datblk).d_dbbase = buf;
		(mdbptr->datblk).d_dbsize = size;	
		(mdbptr->datblk).d_dblim = ((unsigned char *)buf) + size;
		Strcount += size;
	}
	else	{

		((struct msgb *)mdbptr)->b_rptr = 
			((struct msgb *)mdbptr)->b_wptr = 
				(mdbptr->datblk).d_dbbase = 
				 	(unsigned char *)(mdbptr->databuf);
		if (size == 0 )
			(mdbptr->datblk).d_dbsize = 4;
		else
			(mdbptr->datblk).d_dbsize = size;

		(mdbptr->datblk).d_dblim = 
			(unsigned char *)mdbptr + sizeof(struct mdbblock);
#if DEBUG
		(mdbptr->datblk).d_dbztype = STR_NUM_DZONE;	
#endif
	}

	return((struct msgb *)mdbptr);
			
} 


/* 
 * Allocate a class zero data block and attach an externally-supplied
 * data buffer pointed to by base to it.
 */
/* ARGSUSED */
mblk_t *
esballoc(unsigned char *base, int size, int pri, frtn_t *fr_rtn)
{
	register struct msgb *mp;

	if (!base || !fr_rtn)
		return(NULL);
	if (!(mp = allocb(-1, pri)))
		return(NULL);
	
	STR_MEM_TRACE((mp->b_datap), ALLOCEDBY, (inst_t *)__return_address, 0);
	mp->b_datap->db_frtnp = fr_rtn;
	mp->b_datap->db_base = base;
	mp->b_datap->db_lim = base + size;
	mp->b_datap->db_size = size;
	mp->b_rptr = base;
	mp->b_wptr = base;
	return(mp);
}

/*
 * Call function 'func' with 'arg' when a class zero block can 
 * be allocated with priority 'pri'.
 */
/* ARGSUSED */
toid_t
esbbcall(int pri, void (*func)(), long arg)
{
	register int s, s2;
	struct strevent *sep;

	/*
	 * allocate new stream event to add to linked list 
	 */
	if (!(sep = sealloc(SE_NOSLP))) {
		cmn_err(CE_WARN, "esbbcall: could not allocate stream event\n");
		return(0);
	}

	sep->se_next = NULL;
	sep->se_func = func;
	sep->se_arg = arg;
	sep->se_size = -1;

	s = splstr();
	/*
	 * add newly allocated stream event to existing
	 * linked list of events.
	 */
	LOCK_STR_RESOURCE(s2);
	if ( strbcalls.bc_head == NULL ) {
		strbcalls.bc_head = strbcalls.bc_tail = sep;
	} else {
		strbcalls.bc_tail->se_next = sep;
		strbcalls.bc_tail = sep;
	}
	UNLOCK_STR_RESOURCE(s2);
	splx(s);

	strbcwait = 1;	/* so kmem_free() will know to call setqsched() */
	return((toid_t)(__psint_t)sep);
}

/*
 * test if block of given size can be allocated with a request of
 * the given priority.
 * 'pri' is no longer used, but is retained for compatibility.
 */
/* ARGSUSED */
int
testb(register int size, uint pri)
{
	return(size <= kmem_avail());
}

/*
 * Call function 'func' with argument 'arg' when there is a reasonably
 * good chance that a block of size 'size' can be allocated.
 * 'pri' is no longer used, but is retained for compatibility.
 */
/* ARGSUSED */
toid_t
bufcall(uint size, int pri, void (*func)(), long arg)
{
	register int s, s2;
	struct strevent *sep;

	if (!(sep = sealloc(SE_NOSLP))) {
		cmn_err(CE_WARN, "bufcall: could not allocate stream event\n");
		return(0);
	}

	sep->se_next = NULL;
	sep->se_func = func;
	sep->se_arg = arg;
	sep->se_size = size;

#ifdef MP_STREAMS
	/*
	 * Stash a pointer to a pointer to the currently held monitor
	 * so that we can reacquire the correct monitor before we call
	 * se_func.  The double-indirection is necessary because the monitor
	 * might get swapped.
	 */
	sep->se_monpp = private.p_activemonpp;
#endif /* MP_STREAMS */

	s = splstr();
	/*
	 * add newly allocated stream event to existing
	 * linked list of events.
	 */
	LOCK_STR_RESOURCE(s2);
        if ( strbcalls.bc_head == NULL ) {
		strbcalls.bc_head = strbcalls.bc_tail = sep;
	} else {
                strbcalls.bc_tail->se_next = sep;
		strbcalls.bc_tail = sep;
	}
	UNLOCK_STR_RESOURCE(s2);
	splx(s);

	strbcwait = 1;	/* so kmem_free() will know to call setqsched() */
	return((toid_t)(__psint_t)sep);
}

/*
 * Cancel a bufcall request.
 */

void
unbufcall(register toid_t id)
{
	register int s, s2;
	register struct strevent *sep;
	register struct strevent *psep;

	s = splstr();
	psep = NULL;
	LOCK_STR_RESOURCE(s2);
	for (sep = strbcalls.bc_head; sep; sep = sep->se_next) {
		if (id == (toid_t)(__psint_t)sep)
			break;
		psep = sep;
	}
	if (sep) {
		if (psep)
			psep->se_next = sep->se_next;
		else
			strbcalls.bc_head = sep->se_next;
		if (sep == strbcalls.bc_tail)
			strbcalls.bc_tail = psep;
		sefree(sep);
	}
	UNLOCK_STR_RESOURCE(s2);
	splx(s);
}

/*
 * Free a message block.
 */

void
freeb(register mblk_t *bp)
{
	register int s, s2, i;
	register struct datab *dbp;

	ASSERT(bp);
	
	LOCK_STR_RESOURCE(s);
	dbp = bp->b_datap;

	ASSERT( dbp->db_ref > 0 );

	if (dbp->db_msgaddr != (caddr_t)bp) {
		FREEMSGBLOCK(bp);
	}
	else {
		if (((struct mdbblock *)bp)->datblk.d_dbiswhat == _NOT_D) {
			FREEMDBBLOCK((struct mdbblock *)bp);
		}
		else 
			((struct mdbblock *)bp)->datblk.d_dbiswhat = _NOT_M;
	}

	if ( --(dbp->db_ref) > 0 ) {
		UNLOCK_STR_RESOURCE(s);
		return; 
	}

	STR_MEM_TRACE(dbp, FREEDBY, (inst_t *)__return_address, 0);
	ASSERT(dbp->db_mbufref == 0);

	/* seperate databuffer? */
	if ((char *)dbp + sizeof(struct dbinfo) != (char *)(dbp->db_base)) {
		if (dbp->db_frtnp == NULL) {
			UNLOCK_STR_RESOURCE(s);
			i = str_mem_free(dbp);
			LOCK_STR_RESOURCE(s);
			Strcount -= i;
		} else {
			if (dbp->db_frtnp->free_func) {
				s2 = splstr();
				UNLOCK_STR_RESOURCE(s2);
				/*
				 * XXXrs - I used to single-thread calls to
				 * free functions on the grounds that they
				 * can't be guaranteed MP-safe.  This
				 * caused ABBA locking problems.  Free
				 * functions need to be MP-safe (if used with
				 * MP drivers).
				 */
				(*dbp->db_frtnp->free_func)(dbp->db_frtnp->free_arg);
				LOCK_STR_RESOURCE(s2);
				splx(s2);
			}
		}
	}

	/*
	 * and free the triplet so long as the message block part is
	 * not in use
	 */
	if (dbp->db_iswhat == _NOT_M) {
		bp = (struct msgb *)((char *)dbp - sizeof(struct mbinfo));
		FREEMDBBLOCK((struct mdbblock *)bp);
	}
	else 	/*
		 * update the flag to denote that the data block is not
		 * in use
		 */
		dbp->db_iswhat = _NOT_D;

	UNLOCK_STR_RESOURCE(s);
	return;
} 

/*
 * Non-locking version of freeb(), resource lock and/or appropriate
 * spl assumed held by caller.
 */
void
nlfreeb(register mblk_t *bp)
{
	register int s;
	register struct datab *dbp;

	ASSERT(bp);
	
	dbp = bp->b_datap;

	ASSERT( dbp->db_ref > 0 );

	if (dbp->db_msgaddr != (caddr_t)bp) {
		FREEMSGBLOCK(bp);
	}
	else {
		if (((struct mdbblock *)bp)->datblk.d_dbiswhat == _NOT_D) {
			FREEMDBBLOCK((struct mdbblock *)bp);
		}
		else 
			((struct mdbblock *)bp)->datblk.d_dbiswhat = _NOT_M;
	}

	if ( --(dbp->db_ref) > 0 ) {
		return; 
	}

	STR_MEM_TRACE(dbp, FREEDBY, (inst_t *)__return_address, 0);
	ASSERT(dbp->db_mbufref == 0);

	/* seperate databuffer? */
	if ((char *)dbp + sizeof(struct dbinfo) != (char *)(dbp->db_base)) {
		if (dbp->db_frtnp == NULL) 
			Strcount -= str_mem_free(dbp);
		else {
			if (dbp->db_frtnp->free_func) {
				s = splhi();
				UNLOCK_STR_RESOURCE(s);
				/*
				 * XXXrs - I used to single-thread calls to
				 * free functions on the grounds that they
				 * can't be guaranteed MP-safe.  This caused
				 * ABBA locking problems.  Free functions need
				 * to be MP-safe (if used with/by MP drivers).
				 */
				(*dbp->db_frtnp->free_func)(dbp->db_frtnp->free_arg);
				LOCK_STR_RESOURCE(s);
				splx(s);
			}
		}
	}

	/*
	 * and free the triplet so long as the message block part is
	 * not in use
	 */
	if (dbp->db_iswhat == _NOT_M) {
		bp = (struct msgb *)((char *)dbp - sizeof(struct mbinfo));
		FREEMDBBLOCK((struct mdbblock *)bp);
	}
	else 	/*
		 * update the flag to denote that the data block is not
		 * in use
		 */
		dbp->db_iswhat = _NOT_D;

	return;
} 

/*
 * Free a dblk_t that has been severed from all mblk_ts
 */
void
freedatab(register struct datab *dbp)
{
	struct mdbblock mdb;
	register int s, s2;

	ASSERT(dbp);

        s = splstr();
	LOCK_STR_RESOURCE(s2);

        if (--dbp->db_ref != 0) {
		UNLOCK_STR_RESOURCE(s2);
		splx(s);
		return;
	}
	STR_MEM_TRACE(dbp, FREEDBY, (inst_t *)__return_address, 0);
	mdb.datblk.d_dbiswhat = _NOT_M;         /* so mdb  won't get freed */
	mdb.msgblk.m_mblock.b_datap = dbp;	
	dbp->db_msgaddr = (caddr_t)&mdb.msgblk.m_mblock; /* don't free mblk */
	dbp->db_ref = 1;			/* so freeb will free it */

	UNLOCK_STR_RESOURCE(s2);
        splx(s);

	freeb(&mdb.msgblk.m_mblock);
}


/*
 * Free all message blocks in a message using freeb().  
 * The message may be NULL.
 */
void
freemsg(register mblk_t *bp)
{
	register mblk_t *tp;
	register int s;

	s = splstr();
	while (bp) {
		tp = bp->b_cont;
		_FREEB(bp);
		bp = tp;
	}
	splx(s);
}

/*
 * Duplicate a message block
 *
 * Allocate a message block and assign proper
 * values to it (read and write pointers)
 * and link it to existing data block.
 * Increment reference count of data block.
 */
mblk_t *
dupb(register mblk_t *bp)
{
	register int s, s2;
	register mblk_t *nbp;

	ASSERT(bp);

	s = splstr();
	LOCK_STR_RESOURCE(s2);
	if (bp->b_datap->db_ref == 255) {
		UNLOCK_STR_RESOURCE(s2);
		splx(s);
		return (NULL);		/* don't wrap */
	}
	if (msgfreelist) {
		ALLOCMSGBLOCK(nbp);
	} else {
		UNLOCK_STR_RESOURCE(s2);
	 	if ((nbp = xmsgalloc()) == NULL) {
			splx(s);
			return(NULL);
		}
		LOCK_STR_RESOURCE(s2);
	}

	(nbp->b_datap = bp->b_datap)->db_ref++;
	UNLOCK_STR_RESOURCE(s2);
	splx(s);

	/* save return address */
	saveaddr(&(((struct mbinfo *)nbp)->m_func));

	nbp->b_next = NULL;
	nbp->b_prev = NULL;
	nbp->b_cont = NULL;
	nbp->b_rptr = bp->b_rptr;
	nbp->b_wptr = bp->b_wptr;
	nbp->b_flag = bp->b_flag;
	nbp->b_band = bp->b_band;
	STR_MEM_TRACE(nbp, DUPEDBY, (inst_t *)__return_address, bp);
	STR_MEM_TRACE(bp, DUPEDBY, (inst_t *)__return_address, nbp);
	return(nbp);
} 

/*
 * Duplicate a message block by block (uses dupb), returning
 * a pointer to the duplicate message.
 * Returns a non-NULL value only if the entire message
 * was dup'd.
 */

mblk_t *
dupmsg(register mblk_t *bp)
{
	register mblk_t *head, *nbp;
	register int s;

	s = splstr();
	if (!bp) {
		splx(s);
		return NULL;
	}
	else	{
		_DUPB(nbp, bp);
		if ( !nbp ) {
			splx(s);
			return NULL;
		}
		else	
			head = nbp;
	}

	while (bp->b_cont) {
		_DUPB( nbp->b_cont, bp->b_cont);
		if (!(nbp->b_cont)) {
			freemsg(head);
			splx(s);
			return(NULL);
		}
		nbp = nbp->b_cont;
		bp = bp->b_cont;
	}
	splx(s);
	return(head);
}

/*
 * Copy data from message block to newly allocated message block and
 * data block.  The copy is rounded out to full word boundaries so that
 * the (usually) more efficient word copy can be done.
 * Returns new message block pointer, or NULL if error.
 */

mblk_t *
copyb(register mblk_t *bp)
{
	register mblk_t *nbp;
	register dblk_t *dp, *ndp;
	caddr_t base;

	ASSERT(bp);
	ASSERT(bp->b_wptr >= bp->b_rptr);

	dp = bp->b_datap;
	if (!(nbp = allocb(dp->db_lim - dp->db_base, BPRI_MED)))
		return(NULL);
	nbp->b_flag = bp->b_flag;
	nbp->b_band = bp->b_band;
	ndp = nbp->b_datap;	
	ndp->db_type = dp->db_type;
	nbp->b_rptr = ndp->db_base + (bp->b_rptr - dp->db_base);
	nbp->b_wptr = ndp->db_base + (bp->b_wptr - dp->db_base);
	base = straln(nbp->b_rptr);
	strbcpy(straln(bp->b_rptr), base,
		straln(nbp->b_wptr + (sizeof(__psint_t)-1)) - base);
	return(nbp);
}


/*
 * Copy data from message to newly allocated message using new
 * data blocks.  Returns a pointer to the new message, or NULL if error.
 */
mblk_t *
copymsg(register mblk_t *bp)
{
	register mblk_t *head, *nbp;
	register int s;

	if (!bp || !(nbp = head = copyb(bp)))
		return(NULL);

	s = splstr();
	while (bp->b_cont) {
		if (!(nbp->b_cont = copyb(bp->b_cont))) {
			freemsg(head);
			splx(s);
			return(NULL);
		}
		nbp = nbp->b_cont;
		bp = bp->b_cont;
	}
	splx(s);
	return(head);
}

/* 
 * link a message block to tail of message
 */
void
linkb(register mblk_t *mp, register mblk_t *bp)
{
	register int s;

	ASSERT(mp && bp);

	s = splstr();
	for (; mp->b_cont; mp = mp->b_cont)
		;
	mp->b_cont = bp;
	splx(s);
}

/*
 * unlink a message block from head of message
 * return pointer to new message.
 * NULL if message becomes empty.
 */
mblk_t *
unlinkb(register mblk_t *bp)
{
	register mblk_t *bp1;
	register int s;

	ASSERT(bp);

	s = splstr();
	bp1 = bp->b_cont;
	bp->b_cont = NULL;
	splx(s);
	return(bp1);
}

/* 
 * remove a message block "bp" from message "mp"
 *
 * Return pointer to new message or NULL if no message remains.
 * Return -1 if bp is not found in message.
 */
mblk_t *
rmvb( register mblk_t *mp, register mblk_t *bp)
{
	register mblk_t *tmp;
	register mblk_t *lastp = NULL;
	register int s;

	ASSERT(mp && bp);
	s = splstr();
	for (tmp = mp; tmp; tmp = tmp->b_cont) {
		if (tmp == bp) {
			if (lastp)
				lastp->b_cont = tmp->b_cont;
			else
				mp = tmp->b_cont;
			tmp->b_cont = NULL;
			splx(s);
			return(mp);
		}
		lastp = tmp;
	}
	splx(s);
	return((mblk_t *)-1L);
}



/*
 * Concatenate and align first len bytes of common message type.
 *
 * Len == -1  means concatenate everything.
 *
 * Returns 1 on success, 0 on failure.
 *
 * After pullup, mp points to the pulled up data.
 */

int
pullupmsg(register mblk_t *mp, register int len)
{
	register struct datab	*dbp;
	register struct msgb	*bp;
	register int n;
	int s, s2;
	register struct msgb	*bcont;	
	register struct msgb *newmp;
	int free_newmp = 0;

	ASSERT(mp != NULL);

	/*
	 * special cases
	 */

	if (len == -1)	{
		if (mp->b_cont == NULL) {
			if (str_aligned(mp->b_rptr)) 
				return(1);
			/* else */
			len = (int) (mp->b_wptr - mp->b_rptr);
		}
		else {
			len = xmsgsize(mp, 1);
		}
	}
	else {
		ASSERT( mp->b_wptr >= mp->b_rptr );
		if (mp->b_wptr - mp->b_rptr >= len) {
			if (str_aligned(mp->b_rptr))
				return (1);
			/* Stretch len to include what's in the first
			 * mblock, so we don't worry about residue in
			 * the first block after copying. The idea is
			 * to avoid having to set aside a new mblock for
			 * that residue. Also, this means, if db_ref==1, we
			 * can free the data buffer.
			 */
			len = mp->b_wptr - mp->b_rptr;
		}
		else  {
			if (xmsgsize(mp, 1) < len)
				return (0);
		}
	}

	/* Get a new "data block" if the current one is shared. 
	 * Otherwise, get only a new databuffer of size len.
	 */

	/*
	 * XXXrs - not optimal, but I had to rework things because we can't
	 * hold the resource lock when we call allocb.
	 */

	newmp = allocb(len, BPRI_MED);

	LOCK_STR_RESOURCE(s);
	ASSERT( mp->b_datap->db_ref > 0 );
	if ((dbp = mp->b_datap)->db_ref > 1) {

		if (newmp == NULL) {
			UNLOCK_STR_RESOURCE(s);
			return(0);
		}
					
		/* save return address */
		saveaddr(&(((struct mbinfo *)newmp)->m_func));

		((struct mdbblock *)newmp)->datblk.d_dbtype = dbp->db_type;
		--( dbp->db_ref );	

		/*--- copy the data into the new data buffer ---*/
		bp = mp;	
		while (len && bp) {
			ASSERT( bp->b_wptr >= bp->b_rptr );
			n = min((int)(bp->b_wptr - bp->b_rptr), len);
			bcopy(bp->b_rptr, newmp->b_wptr, n);
			newmp->b_wptr += n;
			len -= n;
			bp->b_rptr += n;
			if (bp->b_wptr != bp->b_rptr) 
				break;
			bcont = bp->b_cont;
			if (bp != mp)
				nlfreeb(bp); /* free if not head block */
			bp = bcont;
		} /* while */

		/* and update the message block and the (new) data
		 * block headers.
		 */
		newmp->b_band = mp->b_band; /* we don't care about message */
					    /* header *newmp; but the band */
					    /* and flag fields are in the  */
		newmp->b_flag = mp->b_flag; /* data block in some cases,   */
					    /* so we copy them             */

		if (dbp->db_msgaddr == (caddr_t)mp) {
			((struct mdbblock *)newmp)->datblk.d_dbmsgaddr = (caddr_t)mp;
			if (((struct mdbblock *)mp)->datblk.d_dbiswhat != _NOT_D)
			   ((struct mdbblock *)mp)->datblk.d_dbiswhat = _XCONN;
		}
		((struct mdbblock *)newmp)->datblk.d_dbiswhat = _NOT_M;

		mp->b_datap = newmp->b_datap;
		mp->b_rptr = newmp->b_rptr;
		mp->b_wptr = newmp->b_wptr;
		mp->b_cont = bp;

		UNLOCK_STR_RESOURCE(s);

		return(1);
	}
	else {
		register caddr_t newbuf_wp;
		caddr_t	newbuf;
		int savelen;
		unsigned char ztype;

		if (newmp != NULL)
			free_newmp++;
		if ((newbuf = str_mem_alloc(len, &ztype)) == NULL) {
			UNLOCK_STR_RESOURCE(s);
			if (free_newmp)
				freeb(newmp);
			return(0);
		}
		/*
		 * we can't use the db_size field to remember the new
		 * data buffer size. Because we use db_size to detect if
		 * stream does weird things with the buffer that can screw
		 * up zone when we free memory.
		 */
		savelen = len;
		Strcount += len;

		/*
		 * Copy data into the new data buffer.
		 */
		newbuf_wp = newbuf;	
		bp = mp;
		while (len && bp ) {
			ASSERT(bp->b_wptr >= bp->b_rptr );
			n = min((int)(bp->b_wptr - bp->b_rptr), len);
			bcopy( bp->b_rptr, newbuf_wp, n );
			newbuf_wp += n;
			len -= n;
			bp->b_rptr += n;
			if (bp->b_wptr != bp->b_rptr)
				break;
			bcont = bp->b_cont;
			if (bp != mp)
				nlfreeb(bp); /* free if not head block */
			bp = bcont;
		} /* end while */
		mp->b_cont = bp;

		/* Release previous data buffer is it was seperate
		 * from the data header.
		 */
		if ((unsigned char *)dbp + sizeof(struct dbinfo)
				!= dbp->db_base) {
			if (dbp->db_frtnp == NULL) 
				Strcount -= str_mem_free(dbp);
			else {
				/*
				 * XXXrs - free_func better be MP-safe!
				 */
				if (dbp->db_frtnp->free_func) {
					s2 = splhi();
					UNLOCK_STR_RESOURCE(s2);
					(*dbp->db_frtnp->free_func)(dbp->db_frtnp->free_arg);
					LOCK_STR_RESOURCE(s2);
					splx(s2);
				}
				dbp->db_frtnp = NULL;
			}
		}
		
		/*
		 * Update header information.
		 */
		mp->b_rptr = (unsigned char *)newbuf;
		mp->b_wptr = (unsigned char *)newbuf_wp;
		dbp->db_base = (unsigned char *)newbuf;
		dbp->db_size = savelen;
		dbp->db_ztype = ztype;
		dbp->db_lim = (unsigned char *)newbuf + dbp->db_size;
		UNLOCK_STR_RESOURCE(s);
		if (free_newmp)
			freeb(newmp);
		return(1);
	}

} 

/*
 * Trim bytes from message
 *  len > 0, trim from head
 *  len < 0, trim from tail
 * Returns 1 on success, 0 on failure.
 */
int
adjmsg(mblk_t *mp, register int len)
{
	register mblk_t *bp;
	register int n;
	int fromhead;

	ASSERT(mp != NULL);

	fromhead = 1;
	if (len < 0) {
		fromhead = 0;
		len = -len;
	}
	if (xmsgsize(mp, fromhead) < len)
		return(0);

	if (fromhead) {
		bp = mp;
		while (len) {
			n = min(bp->b_wptr - bp->b_rptr, len);
			bp->b_rptr += n;
			len -= n;
			bp = bp->b_cont;
		}
	} else {
		register mblk_t *endbp = NULL;

		while (len) {
			for (bp = mp; bp->b_cont != endbp; bp = bp->b_cont)
				;
			n = min(bp->b_wptr - bp->b_rptr, len);
			bp->b_wptr -= n;
			len -= n;
			endbp = bp;
		}
	}
	return(1);
}

/*
 * get number of data bytes in message
 */
int
msgdsize(register mblk_t *bp)
{
	register int count = 0;
	register int s;

	s = splstr();
	for (; bp; bp = bp->b_cont)
		if (bp->b_datap->db_type == M_DATA) {
			ASSERT(bp->b_wptr >= bp->b_rptr);
			count += bp->b_wptr - bp->b_rptr;
		}
	splx(s);
	return(count);
}

/*
 * Get a message off head of queue
 *
 * If queue has no buffers then mark queue
 * with QWANTR. (queue wants to be read by
 * someone when data becomes available)
 *
 * If there is something to take off then do so.
 * If queue falls below hi water mark turn off QFULL
 * flag.  Decrement weighted count of queue.
 * Also turn off QWANTR because queue is being read.
 *
 * The queue count is maintained on a per-band basis.
 * Priority band 0 (normal messages) uses q_count,
 * q_lowat, etc.  Non-zero priority bands use the
 * fields in their respective qband structures
 * (qb_count, qb_lowat, etc.)  All messages appear
 * on the same list, linked via their b_next pointers.
 * q_first is the head of the list.  q_count does
 * not reflect the size of all the messages on the
 * queue.  It only reflects those messages in the
 * normal band of flow.  The one exception to this
 * deals with high priority messages.  They are in
 * their own conceptual "band", but are accounted
 * against q_count.
 *
 * If queue count is below the lo water mark and QWANTW
 * is set, enable the closest backq which has a service 
 * procedure and turn off the QWANTW flag.
 *
 * getq could be built on top of rmvq, but isn't because
 * of performance considerations.
 */
mblk_t *
getq(register queue_t *q)
{
	register mblk_t *bp;
	register mblk_t *tmp;
	register int s;
	register qband_t *qbp = (qband_t *) NULL;
	int backenable = 0;
	int wantw = 0;
	queue_t *nq;

	ASSERT(q);
	ASSERT(STREAM_LOCKED(q));

	s = splstr();
	if (!(bp = q->q_first)) {
		q->q_flag |= QWANTR;
		backenable = 1;		/* we might backenable */
		wantw = q->q_flag & QWANTW;
	} else {
		if (bp->b_flag & MSGNOGET) {	/* hack hack hack */
			while (bp && (bp->b_flag & MSGNOGET))
				bp = bp->b_next;
			if (bp)
				rmvq(q, bp);
			splx(s);
			return (bp);
		}
		if (!(q->q_first = bp->b_next))
			q->q_last = NULL;
		else
			q->q_first->b_prev = NULL;

		if (bp->b_band == 0) {
			wantw = q->q_flag & QWANTW;
			for (tmp = bp; tmp; tmp = tmp->b_cont)
				q->q_count -= (tmp->b_wptr - tmp->b_rptr);
			if (q->q_count < q->q_hiwat)
				q->q_flag &= ~QFULL;
			if (q->q_count <= q->q_lowat)
				backenable = 1;
		} else {
			unsigned char i;

			ASSERT(bp->b_band <= q->q_nband);
			ASSERT(q->q_bandp != NULL);
			qbp = q->q_bandp;
			i = bp->b_band;
			while (--i > 0)
				qbp = qbp->qb_next;
			if (qbp->qb_first == qbp->qb_last) {
				qbp->qb_first = NULL;
				qbp->qb_last = NULL;
			} else {
				qbp->qb_first = bp->b_next;
			}
			wantw = qbp->qb_flag & QB_WANTW;
			for (tmp = bp; tmp; tmp = tmp->b_cont)
				qbp->qb_count -= (tmp->b_wptr - tmp->b_rptr);
			if ((qbp->qb_count < qbp->qb_hiwat) &&
			    (qbp->qb_flag & QB_FULL)) {
				qbp->qb_flag &= ~QB_FULL;
				q->q_blocked--;
			}
			if (qbp->qb_count <= qbp->qb_lowat)
				backenable = 1;
		}
		q->q_flag &= ~QWANTR;
		bp->b_next = NULL;
		bp->b_prev = NULL;
	}
	if (backenable) {
		/* find nearest back queue with service proc */
		for (nq = backq(q); nq && !nq->q_qinfo->qi_srvp; nq = backq(nq))
			;
		if (wantw) {
			if (bp && bp->b_band != 0)
				qbp->qb_flag &= ~QB_WANTW;
			else
				q->q_flag &= ~QWANTW;
			if (nq) {
				qenable(nq);
				setqback(nq, bp ? bp->b_band : 0);
			}
		}
		if ((q->q_flag & QWANTW) && (q->q_blocked == 0) &&
		    !(q->q_flag & QFULL) && nq) {
			qenable(nq);
			setqback(nq, 0);
		}
	}
	splx(s);
	return(bp);
}

/*
 * Remove a message from a queue.  The queue count and other
 * flow control parameters are adjusted and the back queue
 * enabled if necessary.
 */
void
rmvq(register queue_t *q, register mblk_t *mp)
{
	register int s;
	register mblk_t *tmp;
	register int i;
	register qband_t *qbp = NULL;
	int backenable = 0;
	int wantw = 0;
	queue_t *nq;

	ASSERT(q);
	ASSERT(mp);
	ASSERT(STREAM_LOCKED(q));

	s = splstr();
	ASSERT(mp->b_band <= q->q_nband);
	if (mp->b_band != 0) {		/* Adjust band pointers */
		ASSERT(q->q_bandp != NULL);
		qbp = q->q_bandp;
		i = mp->b_band;
		while (--i > 0)
			qbp = qbp->qb_next;
		if (mp == qbp->qb_first) {
			if (mp->b_next && mp->b_band == mp->b_next->b_band)
				qbp->qb_first = mp->b_next;
			else
				qbp->qb_first = NULL;
		}
		if (mp == qbp->qb_last) {
			if (mp->b_prev && mp->b_band == mp->b_prev->b_band)
				qbp->qb_last = mp->b_prev;
			else
				qbp->qb_last = NULL;
		}
	}

	/*
	 * Remove the message from the list.
	 */
	if (mp->b_prev)
		mp->b_prev->b_next = mp->b_next;
	else
		q->q_first = mp->b_next;
	if (mp->b_next)
		mp->b_next->b_prev = mp->b_prev;
	else
		q->q_last = mp->b_prev;
	mp->b_next = NULL;
	mp->b_prev = NULL;

	if (mp->b_band == 0) {		/* Perform q_count accounting */
		wantw = q->q_flag & QWANTW;
		for (tmp = mp; tmp; tmp = tmp->b_cont)
			q->q_count -= (tmp->b_wptr - tmp->b_rptr);
		if (q->q_count < q->q_hiwat)
			q->q_flag &= ~QFULL;
		if (q->q_count <= q->q_lowat)
			backenable = 1;
	} else {			/* Perform qb_count accounting */
		wantw = qbp->qb_flag & QB_WANTW;
		for (tmp = mp; tmp; tmp = tmp->b_cont)
			qbp->qb_count -= (tmp->b_wptr - tmp->b_rptr);
		if ((qbp->qb_count < qbp->qb_hiwat) &&
		    (qbp->qb_flag & QB_FULL)) {
			qbp->qb_flag &= ~QB_FULL;
			q->q_blocked--;
		}
		if (qbp->qb_count <= qbp->qb_lowat)
			backenable = 1;
	}

	if (backenable) {
		/* find nearest back queue with service proc */
		for (nq = backq(q); nq && !nq->q_qinfo->qi_srvp; nq = backq(nq))
			;
		if (wantw) {
			if (mp->b_band != 0)
				qbp->qb_flag &= ~QB_WANTW;
			else
				q->q_flag &= ~QWANTW;
			if (nq) {
				qenable(nq);
				setqback(nq, mp->b_band);
			}
		}
		if ((q->q_flag & QWANTW) && (q->q_blocked == 0) &&
		    !(q->q_flag & QFULL) && nq) {
			qenable(nq);
			setqback(nq, 0);
		}
	}
	splx(s);
}

/*
 * Empty a queue.  
 * If flag is set, remove all messages.  Otherwise, remove
 * only non-control messages.  If queue falls below its low
 * water mark, and QWANTW is set , enable the nearest upstream
 * service procedure.
 */
void
flushq(register queue_t *q, int flag)
{
	register mblk_t *mp, *nmp;
	register int s;
	register qband_t *qbp;
	int backenable = 0;
	unsigned char bpri, i;
	unsigned char qbf[NBAND];


	ASSERT(q);
	ASSERT(STREAM_LOCKED(q));

	s = splstr();
	mp = q->q_first;
	q->q_first = NULL;
	q->q_last = NULL;
	q->q_count = 0;
	for (qbp = q->q_bandp; qbp; qbp = qbp->qb_next) {
		qbp->qb_first = NULL;
		qbp->qb_last = NULL;
		qbp->qb_count = 0;
		qbp->qb_flag &= ~QB_FULL;
	}
	q->q_blocked = 0;
	q->q_flag &= ~QFULL;
	while (mp) {
		nmp = mp->b_next;
		if (flag || datamsg(mp->b_datap->db_type))
			freemsg(mp);
		else
			putq(q, mp);
		mp = nmp;
	}
	bzero((caddr_t)qbf, NBAND);
	bpri = 1;
	for (qbp = q->q_bandp; qbp; qbp = qbp->qb_next) {
		if ((qbp->qb_flag & QB_WANTW) && (qbp->qb_count <= qbp->qb_lowat)) {
			qbp->qb_flag &= ~QB_WANTW;
			backenable = 1;
			qbf[bpri] = 1;
		}
		bpri++;
	}
	if ((q->q_flag & QWANTW) && (q->q_count <= q->q_lowat)) {
		q->q_flag &= ~QWANTW;
		backenable = 1;
		qbf[0] = 1;
	}

	/*
	 * If any band can now be written to, and there is a writer
	 * for that band, then backenable the closest service procedure.
	 */
	if (backenable) {
		/* find nearest back queue with service proc */
		for (q = backq(q); q && !q->q_qinfo->qi_srvp; q = backq(q))
			;
		if (q) {
			qenable(q);
			for (i = 0; i < bpri; i++)
				if (qbf[i])
					setqback(q, i);
		}
	}
	splx(s);
}

/*
 * Flush the queue of messages of the given priority band.
 * There is some duplication of code between flushq and flushband.
 * This is because we want to optimize the code as much as possible.
 * The assumption is that there will be more messages in the normal
 * (priority 0) band than in any other.
 */
void
flushband(register queue_t *q, unsigned char pri, int flag)
{
	register int s;
	register mblk_t *mp;
	register mblk_t *nmp;
	register mblk_t *last;
	register qband_t *qbp;
	int band;

	ASSERT(q);
	ASSERT(STREAM_LOCKED(q));

	s = splstr();
	if (pri > q->q_nband) {
		splx(s);
		return;
	}
	if (pri == 0) {
		mp = q->q_first;
		q->q_first = NULL;
		q->q_last = NULL;
		q->q_count = 0;
		for (qbp = q->q_bandp; qbp; qbp = qbp->qb_next) {
			qbp->qb_first = NULL;
			qbp->qb_last = NULL;
			qbp->qb_count = 0;
			qbp->qb_flag &= ~QB_FULL;
		}
		q->q_blocked = 0;
		q->q_flag &= ~QFULL;
		while (mp) {
			nmp = mp->b_next;
			if ((mp->b_band == 0) &&
			    (mp->b_datap->db_type < QPCTL) &&
			    (flag || datamsg(mp->b_datap->db_type)))
				freemsg(mp);
			else
				putq(q, mp);
			mp = nmp;
		}
		if ((q->q_flag & QWANTW) && (q->q_count <= q->q_lowat)) {
			/* find nearest back queue with service proc */
			q->q_flag &= ~QWANTW;
			for (q = backq(q); q && !q->q_qinfo->qi_srvp; q = backq(q))
				;
			if (q) {
				qenable(q);
				setqback(q, pri);
			}
		}
	} else {	/* pri != 0 */
		ASSERT(q->q_bandp != NULL);
		band = pri;
		qbp = q->q_bandp;
		while (--band > 0)
			qbp = qbp->qb_next;
		mp = qbp->qb_first;
		if (mp == NULL) {
			splx(s);
			return;
		}
		last = qbp->qb_last;
		if (mp == last)		/* only message in band */
			last = mp->b_next;
		while (mp != last) {
			nmp = mp->b_next;
			if (mp->b_band == pri) {
				if (flag || datamsg(mp->b_datap->db_type)) {
					rmvq(q, mp);
					freemsg(mp);
				}
			}
			mp = nmp;
		}
		if (mp && mp->b_band == pri) {
			if (flag || datamsg(mp->b_datap->db_type)) {
				rmvq(q, mp);
				freemsg(mp);
			}
		}
	}
	splx(s);
}

/*
 * Return 1 if the queue is not full.  If the queue is full, return
 * 0 (may not put message) and set QWANTW flag (caller wants to write
 * to the queue).
 */
int
canput(register queue_t *q)
{
	register int s;

	if (!q)
		return(0);
	s = splstr();
	while (q->q_next && !q->q_qinfo->qi_srvp)
		q = q->q_next;
	if ((q->q_flag & QFULL) || (q->q_blocked != 0)) {
		q->q_flag |= QWANTW;
		splx(s);
		return(0);
	}
	splx(s);
	return(1);
}

/*
 * This is the new canput for use with priority bands.  Return 1 if the
 * band is not full.  If the band is full, return 0 (may not put message)
 * and set QWANTW(QB_WANTW) flag for zero(non-zero) band (caller wants to
 * write to the queue).
 */
int
bcanput(register queue_t *q, unsigned char pri)
{
	register int s;
	register qband_t *qbp;

	if (!q)
		return(0);
	s = splstr();
	while (q->q_next && !q->q_qinfo->qi_srvp)
		q = q->q_next;
	if (pri == 0) {
		if (q->q_flag & QFULL) {
			q->q_flag |= QWANTW;
			splx(s);
			return(0);
		}
	} else {	/* pri != 0 */
		if (pri > q->q_nband) {
			/*
			 * No band exists yet, so return success.
			 */
			splx(s);
			return(1);
		}
		qbp = q->q_bandp;
		while (--pri)
			qbp = qbp->qb_next;
		if (qbp->qb_flag & QB_FULL) {
			qbp->qb_flag |= QB_WANTW;
			splx(s);
			return(0);
		}
	}
	splx(s);
	return(1);
}

/*
 * Put a message on a queue.  
 *
 * Messages are enqueued on a priority basis.  The priority classes
 * are HIGH PRIORITY (type >= QPCTL), PRIORITY (type < QPCTL && band > 0),
 * and B_NORMAL (type < QPCTL && band == 0). 
 *
 * Add appropriate weighted data block sizes to queue count.
 * If queue hits high water mark then set QFULL flag.
 *
 * If QNOENAB is not set (putq is allowed to enable the queue),
 * enable the queue only if the message is PRIORITY,
 * or the QWANTR flag is set (indicating that the service procedure 
 * is ready to read the queue.  This implies that a service 
 * procedure must NEVER put a priority message back on its own
 * queue, as this would result in an infinite loop (!).
 */
int
putq(register queue_t *q, register mblk_t *bp)
{
	register int s;
	register mblk_t *tmp;
	register qband_t *qbp = NULL;
	int mcls = (int)queclass(bp);

	ASSERT(q && bp);
	s = splstr();

	/*
	 * Make sanity checks and if qband structure is not yet
	 * allocated, do so.
	 */
	if (mcls == QPCTL) {
		if (bp->b_band != 0)
			bp->b_band = 0;		/* force to be correct */
	} else if (bp->b_band != 0) {
		register int i;
		qband_t **qbpp;

		if (bp->b_band > q->q_nband) {

			/*
			 * The qband structure for this priority band is
			 * not on the queue yet, so we have to allocate
			 * one on the fly.  It would be wasteful to
			 * associate the qband structures with every
			 * queue when the queues are allocated.  This is
			 * because most queues will only need the normal
			 * band of flow which can be described entirely
			 * by the queue itself.
			 */
			qbpp = &q->q_bandp;
			while (*qbpp)
				qbpp = &(*qbpp)->qb_next;
			while (bp->b_band > q->q_nband) {
				if ((*qbpp = allocband()) == NULL) {
					splx(s);
					return(0);
				}
				(*qbpp)->qb_hiwat = q->q_hiwat;
				(*qbpp)->qb_lowat = q->q_lowat;
				q->q_nband++;
				qbpp = &(*qbpp)->qb_next;
			}
		}
		qbp = q->q_bandp;
		i = bp->b_band;
		while (--i)
			qbp = qbp->qb_next;
	}

	/*
	 * If queue is empty, add the message and initialize the pointers.
	 * Otherwise, adjust message pointers and queue pointers based on
	 * the type of the message and where it belongs on the queue.  Some
	 * code is duplicated to minimize the number of conditionals and
	 * hopefully minimize the amount of time this routine takes.
	 */
	if (!q->q_first) {
		bp->b_next = NULL;
		bp->b_prev = NULL;
		q->q_first = bp;
		q->q_last = bp;
		if (qbp) {
			qbp->qb_first = bp;
			qbp->qb_last = bp;
		}
	} else if (!qbp) {	/* bp->b_band == 0 */

		/* 
		 * If queue class of message is less than or equal to
		 * that of the last one on the queue, tack on to the end.
		 */
		tmp = q->q_last;
		if (mcls <= (int)queclass(tmp)) {
			bp->b_next = NULL;
			bp->b_prev = tmp;
			tmp->b_next = bp;
			q->q_last = bp;
		} else {
			tmp = q->q_first;
			while ((int)queclass(tmp) >= mcls)
				tmp = tmp->b_next;
			ASSERT(tmp != NULL);

			/*
			 * Insert bp before tmp.
			 */
			bp->b_next = tmp;
			bp->b_prev = tmp->b_prev;
			if (tmp->b_prev)
				tmp->b_prev->b_next = bp;
			else
				q->q_first = bp;
			tmp->b_prev = bp;
		}
	} else {		/* bp->b_band != 0 */
		if (qbp->qb_first) {
			ASSERT(qbp->qb_last != NULL);
			tmp = qbp->qb_last;

			/*
			 * Insert bp after the last message in this band.
			 */
			bp->b_next = tmp->b_next;
			if (tmp->b_next)
				tmp->b_next->b_prev = bp;
			else
				q->q_last = bp;
			bp->b_prev = tmp;
			tmp->b_next = bp;
		} else {
			tmp = q->q_last;
			if ((mcls < (int)queclass(tmp)) ||
			    (bp->b_band <= tmp->b_band)) {

				/*
				 * Tack bp on end of queue.
				 */
				bp->b_next = NULL;
				bp->b_prev = tmp;
				tmp->b_next = bp;
				q->q_last = bp;
			} else {
				tmp = q->q_first;
				while (tmp->b_datap->db_type >= QPCTL)
					tmp = tmp->b_next;
				while (tmp->b_band >= bp->b_band)
					tmp = tmp->b_next;
				ASSERT(tmp != NULL);

				/*
				 * Insert bp before tmp.
				 */
				bp->b_next = tmp;
				bp->b_prev = tmp->b_prev;
				if (tmp->b_prev)
					tmp->b_prev->b_next = bp;
				else
					q->q_first = bp;
				tmp->b_prev = bp;
			}
			qbp->qb_first = bp;
		}
		qbp->qb_last = bp;
	}

	if (qbp) {
		for (tmp = bp; tmp; tmp = tmp->b_cont)
			qbp->qb_count += (tmp->b_wptr - tmp->b_rptr);
		if ((qbp->qb_count >= qbp->qb_hiwat) &&
		    !(qbp->qb_flag & QB_FULL)) {
			qbp->qb_flag |= QB_FULL;
			q->q_blocked++;
		}
	} else {
		for (tmp = bp; tmp; tmp = tmp->b_cont)
			q->q_count += (tmp->b_wptr - tmp->b_rptr);
		if (q->q_count >= q->q_hiwat)
			q->q_flag |= QFULL;
	}
	if ((mcls > QNORM) || (canenable(q) && (q->q_flag & QWANTR)))
		qenable(q);
	splx(s);
	return(1);
}

/*
 * Put stuff back at beginning of Q according to priority order.
 * See comment on putq above for details.
 */
int
putbq(register queue_t *q, register mblk_t *bp)
{
	register int s;
	register mblk_t *tmp;
	register qband_t *qbp = NULL;
	int mcls = (int)queclass(bp);

	ASSERT(q && bp);
	ASSERT(bp->b_next == NULL);
	ASSERT(STREAM_LOCKED(q));

	s = splstr();

	/*
	 * Make sanity checks and if qband structure is not yet
	 * allocated, do so.
	 */
	if (mcls == QPCTL) {
		if (bp->b_band != 0)
			bp->b_band = 0;		/* force to be correct */
	} else if (bp->b_band != 0) {
		register int i;
		qband_t **qbpp;

		if (bp->b_band > q->q_nband) {
			qbpp = &q->q_bandp;
			while (*qbpp)
				qbpp = &(*qbpp)->qb_next;
			while (bp->b_band > q->q_nband) {
				if ((*qbpp = allocband()) == NULL) {
					splx(s);
					return(0);
				}
				(*qbpp)->qb_hiwat = q->q_hiwat;
				(*qbpp)->qb_lowat = q->q_lowat;
				q->q_nband++;
				qbpp = &(*qbpp)->qb_next;
			}
		}
		qbp = q->q_bandp;
		i = bp->b_band;
		while (--i)
			qbp = qbp->qb_next;
	}

	/* 
	 * If queue is empty or if message is high priority,
	 * place on the front of the queue.
	 */
	tmp = q->q_first;
	if ((!tmp) || (mcls == QPCTL)) {
		bp->b_next = tmp;
		if (tmp)
			tmp->b_prev = bp;
		else
			q->q_last = bp;
		q->q_first = bp;
		bp->b_prev = NULL;
		if (qbp) {
			qbp->qb_first = bp;
			qbp->qb_last = bp;
		}
	} else if (qbp) {	/* bp->b_band != 0 */
		tmp = qbp->qb_first;
		if (tmp) {

			/*
			 * Insert bp before the first message in this band.
			 */
			bp->b_next = tmp;
			bp->b_prev = tmp->b_prev;
			if (tmp->b_prev)
				tmp->b_prev->b_next = bp;
			else
				q->q_first = bp;
			tmp->b_prev = bp;
		} else {
			tmp = q->q_last;
			if ((mcls < (int)queclass(tmp)) ||
			    (bp->b_band < tmp->b_band)) {

				/*
				 * Tack bp on end of queue.
				 */
				bp->b_next = NULL;
				bp->b_prev = tmp;
				tmp->b_next = bp;
				q->q_last = bp;
			} else {
				tmp = q->q_first;
				while (tmp->b_datap->db_type >= QPCTL)
					tmp = tmp->b_next;
				while (tmp->b_band > bp->b_band)
					tmp = tmp->b_next;
				ASSERT(tmp != NULL);

				/*
				 * Insert bp before tmp.
				 */
				bp->b_next = tmp;
				bp->b_prev = tmp->b_prev;
				if (tmp->b_prev)
					tmp->b_prev->b_next = bp;
				else
					q->q_first = bp;
				tmp->b_prev = bp;
			}
			qbp->qb_last = bp;
		}
		qbp->qb_first = bp;
	} else {		/* bp->b_band == 0 && !QPCTL */

		/*
		 * If the queue class or band is less than that of the last
		 * message on the queue, tack bp on the end of the queue.
		 */
		tmp = q->q_last;
		if ((mcls < (int)queclass(tmp)) || (bp->b_band < tmp->b_band)) {
			bp->b_next = NULL;
			bp->b_prev = tmp;
			tmp->b_next = bp;
			q->q_last = bp;
		} else {
			tmp = q->q_first;
			while (tmp->b_datap->db_type >= QPCTL)
				tmp = tmp->b_next;
			while (tmp->b_band > bp->b_band)
				tmp = tmp->b_next;
			ASSERT(tmp != NULL);

			/*
			 * Insert bp before tmp.
			 */
			bp->b_next = tmp;
			bp->b_prev = tmp->b_prev;
			if (tmp->b_prev)
				tmp->b_prev->b_next = bp;
			else
				q->q_first = bp;
			tmp->b_prev = bp;
		}
	}

	if (qbp) {
		for (tmp = bp; tmp; tmp = tmp->b_cont)
			qbp->qb_count += (tmp->b_wptr - tmp->b_rptr);
		if ((qbp->qb_count >= qbp->qb_hiwat) &&
		    !(qbp->qb_flag & QB_FULL)) {
			qbp->qb_flag |= QB_FULL;
			q->q_blocked++;
		}
	} else {
		for (tmp = bp; tmp; tmp = tmp->b_cont)
			q->q_count += (tmp->b_wptr - tmp->b_rptr);
		if (q->q_count >= q->q_hiwat)
			q->q_flag |= QFULL;
	}
	if ((mcls > QNORM) || (canenable(q) && (q->q_flag & QWANTR)))
		qenable(q);
	splx(s);
	return(1);
}

/*
 * Insert a message before an existing message on the queue.  If the
 * existing message is NULL, the new messages is placed on the end of
 * the queue.  The queue class of the new message is ignored.  However,
 * the priority band of the new message must adhere to the following
 * ordering:
 *
 *	emp->b_prev->b_band >= mp->b_band >= emp->b_band.
 *
 * All flow control parameters are updated.
 */
int
insq(register queue_t *q, register mblk_t *emp, register mblk_t *mp)
{
	register mblk_t *tmp;
	register int s;
	register qband_t *qbp = NULL;
	int mcls = (int)queclass(mp);

	ASSERT(STREAM_LOCKED(q));

	s = splstr();
	if (mcls == QPCTL) {
		if (mp->b_band != 0)
			mp->b_band = 0;		/* force to be correct */
		if (emp && emp->b_prev &&
		    (emp->b_prev->b_datap->db_type < QPCTL))
			goto badord;
	}
	if (emp) {
		if (((mcls == QNORM) && (mp->b_band < emp->b_band)) ||
		    (emp->b_prev && (emp->b_prev->b_datap->db_type < QPCTL) &&
		    (emp->b_prev->b_band < mp->b_band))) {
			goto badord;
		}
	} else {
		tmp = q->q_last;
		if (tmp && (mcls == QNORM) && (mp->b_band > tmp->b_band)) {
badord:
			splx(s);
			cmn_err(CE_WARN, "insq: attempt to insert message out of order on q %x\n", q);
			return(0);
		}
	}

	if (mp->b_band != 0) {
		register int i;
		qband_t **qbpp;

		if (mp->b_band > q->q_nband) {
			qbpp = &q->q_bandp;
			while (*qbpp)
				qbpp = &(*qbpp)->qb_next;
			while (mp->b_band > q->q_nband) {
				if ((*qbpp = allocband()) == NULL) {
					splx(s);
					return(0);
				}
				(*qbpp)->qb_hiwat = q->q_hiwat;
				(*qbpp)->qb_lowat = q->q_lowat;
				q->q_nband++;
				qbpp = &(*qbpp)->qb_next;
			}
		}
		qbp = q->q_bandp;
		i = mp->b_band;
		while (--i)
			qbp = qbp->qb_next;
	}

	if (mp->b_next = emp) {
		if (mp->b_prev = emp->b_prev) 
			emp->b_prev->b_next = mp;
		else
			q->q_first = mp;
		emp->b_prev = mp;
	} else {
		if (mp->b_prev = q->q_last) 
			q->q_last->b_next = mp;
		else 
			q->q_first = mp;
		q->q_last = mp;
	}

	if (qbp) {	/* adjust qband pointers and count */
		if (!qbp->qb_first) {
			qbp->qb_first = mp;
			qbp->qb_last = mp;
		} else {
			if (qbp->qb_first == emp)
				qbp->qb_first = mp;
			else if (mp->b_next && (mp->b_next->b_band !=
			    mp->b_band))
				qbp->qb_last = mp;
		}
		for (tmp = mp; tmp; tmp = tmp->b_cont)
			qbp->qb_count += (tmp->b_wptr - tmp->b_rptr);
		if ((qbp->qb_count >= qbp->qb_hiwat) &&
		    !(qbp->qb_flag & QB_FULL)) {
			qbp->qb_flag |= QB_FULL;
			q->q_blocked++;
		}
	} else {
		for (tmp = mp; tmp; tmp = tmp->b_cont)
			q->q_count += (tmp->b_wptr - tmp->b_rptr);
		if (q->q_count >= q->q_hiwat)
			q->q_flag |= QFULL;
	}
	if (canenable(q) && (q->q_flag & QWANTR))
		qenable(q);
	splx(s);
	return(1);
}

/*
 * Create and put a control message on queue.
 */
int
putctl(queue_t *q, int type)
{
	register mblk_t *bp;

	if ((datamsg(type) && (type != M_DELAY)) || !(bp = allocb(0, BPRI_HI)))
		return(0);
	bp->b_datap->db_type = type;
	(*q->q_qinfo->qi_putp)(q, bp);
	return(1);
}

/*
 * Control message with a single-byte parameter
 */
int
putctl1(queue_t *q, int type, int param)
{
	register mblk_t *bp;

	if ((datamsg(type) && (type != M_DELAY)) || !(bp = allocb(1, BPRI_HI)))
		return(0);
	bp->b_datap->db_type = type;
	*bp->b_wptr++ = param;
	(*q->q_qinfo->qi_putp)(q, bp);
	return(1);
}

/*
 * Return the queue upstream from this one
 */
queue_t *
backq(register queue_t *q)
{
	ASSERT(q);

	q = OTHERQ(q);
	if (q->q_next) {
		q = q->q_next;
		return(OTHERQ(q));
	}
	return(NULL);
}



/*
 * Send a block back up the queue in reverse from this
 * one (e.g. to respond to ioctls)
 */
void
qreply(register queue_t *q, mblk_t *bp)
{

	ASSERT(q && bp);

	q = OTHERQ(q);
	ASSERT(q->q_next);
	(*q->q_next->q_qinfo->qi_putp)(q->q_next, bp);
}

/*
 * Streams Queue Scheduling
 * 
 * Queues are enabled through qenable() when they have messages to 
 * process.  They are serviced by queuerun(), which runs each enabled
 * queue's service procedure.  The call to queuerun() is processor
 * dependent - the general principle is that it be run whenever a queue
 * is enabled but before returning to user level.  For system calls,
 * the function runqueues() is called if their action causes a queue
 * to be enabled.  For device interrupts, queuerun() should be
 * called before returning from the last level of interrupt.  Beyond
 * this, no timing assumptions should be made about queue scheduling.
 */

/*
 * Enable a queue: put it on list of those whose service procedures are
 * ready to run and set up the scheduling mechanism.
 */
void
qenable(register queue_t *q)
{
	extern int privenable;
	register int s, s2;

	ASSERT(q);
	/*
	 * The following ASSERT is a feeble attempt to check the validity
	 * of the queue.  If the condition is false, it proves that the
	 * queue is bad (already freed, garbage queue pointer, etc.), but
	 * if the condition is true, it is no real guarantee that the queue
	 * is good.  It's better than nothing though.
	 */
	ASSERT(q->q_flag & QUSE);
	ASSERT(STREAM_LOCKED(q));

	if (!q->q_qinfo->qi_srvp)
		return;

	s = splstr();

	/*
	 * Only enable once.
	 */
	LOCK_STR_RESOURCE(s2);
	if (q->q_pflag & QENAB) {
		UNLOCK_STR_RESOURCE(s2);
		splx(s);
		return;
	}

	/*
	 * Mark queue enabled and put on run list.
	 */
	q->q_pflag |= QENAB;

#ifdef MP_STREAMS
	UNLOCK_STR_RESOURCE(s2);
        s2 = STREAMS_ENAB_INTERRUPT(q->q_monpp, q->q_qinfo->qi_srvp, q, 0, 0);
        ASSERT(s2 == 0);
#else
	ASSERT(*qtailp == NULL);
	*qtailp = q;
	q->q_link = NULL;
	qtailp = &q->q_link;

	UNLOCK_STR_RESOURCE(s2);

	/*
	 * set up scheduling mechanism
	 */
	setqsched();

#endif /* MP_STREAMS */

	splx(s);
}

/* 
 * Return number of messages on queue
 */
int
qsize(register queue_t *qp)
{
	register int count = 0;
	register mblk_t *mp;
	register int s;

	ASSERT(qp);

	s = splstr();
	for (mp = qp->q_first; mp; mp = mp->b_next)
		count++;
	splx(s);
	return(count);
}

/*
 * noenable - set queue so that putq() will not enable it.
 * enableok - set queue so that putq() can enable it.
 */
void
noenable(queue_t *q)
{ 
	register int s;

	s = splstr();
	q->q_flag |= QNOENB;
	splx(s);
}

void
enableok(queue_t *q)
{
	register int s;

	s = splstr();
	q->q_flag &= ~QNOENB;
	splx(s);
}

/*
 * Set queue fields.
 */
int
strqset(register queue_t *q,
	qfields_t what,
	register unsigned char pri,
	long val)
{
	register int s;
	register qband_t *qbp = NULL;
	int error = 0;

	if (what >= QBAD)
		return (EINVAL);
	if (pri != 0) {
		register int i;
		qband_t **qbpp;

		s = splstr();
		if (pri > q->q_nband) {
			qbpp = &q->q_bandp;
			while (*qbpp)
				qbpp = &(*qbpp)->qb_next;
			while (pri > q->q_nband) {
				if ((*qbpp = allocband()) == NULL) {
					splx(s);
					return (EAGAIN);
				}
				(*qbpp)->qb_hiwat = q->q_hiwat;
				(*qbpp)->qb_lowat = q->q_lowat;
				q->q_nband++;
				qbpp = &(*qbpp)->qb_next;
			}
		}
		qbp = q->q_bandp;
		i = pri;
		while (--i)
			qbp = qbp->qb_next;
		splx(s);
	}
	switch (what) {
	case QHIWAT:
		if (qbp)
			qbp->qb_hiwat = (ulong)val;
		else
			q->q_hiwat = (ulong)val;
		break;

	case QLOWAT:
		if (qbp)
			qbp->qb_lowat = (ulong)val;
		else
			q->q_lowat = (ulong)val;
		break;

	case QMAXPSZ:
		if (qbp)
			error = EINVAL;
		else
			q->q_maxpsz = val;
		break;

	case QMINPSZ:
		if (qbp)
			error = EINVAL;
		else
			q->q_minpsz = val;
		break;

	case QCOUNT:
	case QFIRST:
	case QLAST:
	case QFLAG:
		error = EPERM;
		break;

	default:
		error = EINVAL;
		break;
	}
	return (error);
}

/*
 * Get queue fields.
 */
int
strqget(register queue_t *q,
	qfields_t what,
	register unsigned char pri,
	long *valp)
{
	register int s;
	register qband_t *qbp = NULL;
	int error = 0;

	if (what >= QBAD)
		return (EINVAL);
	if (pri != 0) {
		register int i;
		qband_t **qbpp;

		s = splstr();
		if (pri > q->q_nband) {
			qbpp = &q->q_bandp;
			while (*qbpp)
				qbpp = &(*qbpp)->qb_next;
			while (pri > q->q_nband) {
				if ((*qbpp = allocband()) == NULL) {
					splx(s);
					return (EAGAIN);
				}
				(*qbpp)->qb_hiwat = q->q_hiwat;
				(*qbpp)->qb_lowat = q->q_lowat;
				q->q_nband++;
				qbpp = &(*qbpp)->qb_next;
			}
		}
		qbp = q->q_bandp;
		i = pri;
		while (--i)
			qbp = qbp->qb_next;
		splx(s);
	}
	switch (what) {
	case QHIWAT:
		if (qbp)
			*(ulong *)valp = qbp->qb_hiwat;
		else
			*(ulong *)valp = q->q_hiwat;
		break;

	case QLOWAT:
		if (qbp)
			*(ulong *)valp = qbp->qb_lowat;
		else
			*(ulong *)valp = q->q_lowat;
		break;

	case QMAXPSZ:
		if (qbp)
			error = EINVAL;
		else
			*(long *)valp = q->q_maxpsz;
		break;

	case QMINPSZ:
		if (qbp)
			error = EINVAL;
		else
			*(long *)valp = q->q_minpsz;
		break;

	case QCOUNT:
		if (qbp)
			*(ulong *)valp = qbp->qb_count;
		else
			*(ulong *)valp = q->q_count;
		break;

	case QFIRST:
		if (qbp)
			*(mblk_t **)valp = qbp->qb_first;
		else
			*(mblk_t **)valp = q->q_first;
		break;

	case QLAST:
		if (qbp)
			*(mblk_t **)valp = qbp->qb_last;
		else
			*(mblk_t **)valp = q->q_last;
		break;

	case QFLAG:
		if (qbp)
			*(ulong *)valp = qbp->qb_flag;
		else
			*(ulong *)valp = q->q_flag;
		break;

	default:
		error = EINVAL;
		break;
	}
	return (error);
}

/* 
 * Call the put routine of the given queue.
 */

void
put(register queue_t *q, register mblk_t *mp)
{
        (*q->q_qinfo->qi_putp)(q, mp);
}

/*
 * Test if message is a "priority" type.
 */

int
pcmsg(unsigned char db_type)
{
        return(db_type >= QPCTL);
}

/*
 * DDI/DKI "MP safe" version of bcanput(q->q_next).  Since IRIX streams use
 * the streams monitor, this is trivial.
 */
int
bcanputnext(register queue_t *q, unsigned char pri)
{
	return(bcanput(q->q_next, pri));
}

/*
 * DDI/DKI "MP safe" version of canput(q->q_next).  Since IRIX streams use
 * the streams monitor, this is trivial.
 */
int
canputnext(register queue_t *q)
{
	return(canput(q->q_next));
}

/*
 * DDI/DKI "MP safe" version of putctl(q->q_next, type).  Since IRIX streams
 * use the streams monitor, this is trivial.
 */
int
putnextctl(register queue_t *q, int type)
{
	return(putctl(q->q_next, type));
}

/*
 * DDI/DKI "MP safe" version of putctl1(q->q_next, type, param).  Since IRIX
 * streams use the streams monitor, this is trivial.
 */
int
putnextctl1(register queue_t *q, int type, int param)
{
	return(putctl1(q->q_next, type, param));
}

/*
 * DDI/DKI function to enable put/service procedures.  Allows open routine
 * to set up resources in MP environment without having to worry about an
 * "early" call to any put/service procedure.  This is unnecessary for IRIX
 * because of the streams monitor.
 *
 * XXXrs Will any brain-dead test actually try to call the put/service
 *	 procedure before/after calling qprocson?
 */
/* ARGSUSED */
void
qprocson(register queue_t *rq)
{
	return;
}

/*
 * DDI/DKI function to disable put/service procedures.  Allows close routine
 * to tear down resources in MP environment without having to worry about a
 * "late" call to any put/service procedure.  This is unnecessary for IRIX
 * because of the streams monitor.
 *
 * XXXrs Will any brain-dead test actually try to call the put/service
 *	 procedure before/after calling qprocsoff?
 */
/* ARGSUSED */
void
qprocsoff(register queue_t *rq)
{
	return;
}

/*
 * DDI/DKI function to "freeze" all activity on a queue and set the
 * interrupt level to splstr.  The former part is unnecessary for IRIX
 * because of the streams monitor.
 */
/* ARGSUSED */
int	/* XXXrs should be pl_t? */
freezestr(queue_t *q)
{
	return(splstr());
}

/*
 * DDI/DKI function to "unfreeze" all activity on a queue and set the
 * interrupt level to s.  The former part is unnecessary for IRIX
 * because of the streams monitor.
 */
/* ARGSUSED */
void
unfreezestr(queue_t *q, int s) /* XXXrs s should be pl_t? */
{
	splx(s);
}

/*
 * DDI/DKI replacement for pullupmsg.
 *
 * Concatenate and align first len bytes of common message type.
 *
 * Len == -1  means concatenate everything.
 * 
 * Returns pointer to new pulled-up message, or NULL on failure.
 */
mblk_t *
msgpullup(register mblk_t *mp, int len)
{
	register mblk_t *bp, *nmp, *tmp;
	register int n, partial;

	/*
	 * special cases
	 */

	ASSERT(mp != NULL);

	if (len == -1) { /* cat all so set len to proper size */
		if (mp->b_cont == NULL) {
			if (str_aligned(mp->b_rptr)) {
				return(copyb(mp));
			}
			len = (int) (mp->b_wptr - mp->b_rptr);
		}
		else {
			len = xmsgsize(mp, 1);
		}
	}
	else {
		ASSERT(mp->b_wptr >= mp->b_rptr);
		if (mp->b_wptr - mp->b_rptr >= len) {
			if (str_aligned(mp->b_rptr)) {
				return(copymsg(mp));
			}
		}
		else  {
			if (xmsgsize(mp, 1) < len) {
				return(NULL);
			}
		}
	}

	/*
	 * Get new mblk.
	 */
	if ((nmp = allocb(len, BPRI_MED)) == NULL)
		return(NULL);

	/*
	 * Copy over important info.
	 */
	nmp->b_datap->db_type = mp->b_datap->db_type;
	nmp->b_flag = mp->b_flag;
	nmp->b_band = mp->b_band;

	/*
	 * Save return address.
	 */
	saveaddr(&(((struct mbinfo *)nmp)->m_func));

	/*
	 * Copy the data into the new message.
	 */

	bp = mp;	
	partial = 0;
	while (len) {
		ASSERT(bp);
		ASSERT(bp->b_wptr >= bp->b_rptr);
		n = min((int)(bp->b_wptr - bp->b_rptr), len);
		bcopy(bp->b_rptr, nmp->b_wptr, n);
		nmp->b_wptr += n;
		len -= n;
		if (!len && (bp->b_rptr + n) != bp->b_wptr) {
			partial++;
			break;
		}
		bp = bp->b_cont;
	}
	/*
	 * Everything the caller wanted "pulled-up" has been "pulled-up"
	 * into the new message block.  If there is any residual data left
	 * over, we need to make a copy of it and attach it to the new
	 * message block as a continuation.
	 *
	 * If bp != NULL, it points to a mblk that was not or was only
	 * partially "pulled-up".  In partial case, "partial" will be set
	 * non-zero, and n is the number of bytes already "pulled-up".
	 */
	if (bp != NULL) {
		if ((tmp = copymsg(bp)) == NULL) {
			freeb(nmp);
			return(NULL);
		}
		if (partial) {
			tmp->b_rptr += n;
		}
		nmp->b_cont = tmp;
	}
	return(nmp);
}

/*
 * freeing data buffer separate from fastbuf
 * make sure that the buffer is not part of a zone
 */
static int
str_mem_free(struct datab *dbp)
{
	register int size;
	unsigned char ztype;

	ztype = dbp->db_ztype;
	size = (char *)dbp->db_lim - (char *)dbp->db_base;
	ASSERT(size == dbp->db_size);
	if (size > MAX_STRZONE_SZ) {
		ASSERT(ztype == (unsigned char)-1);
		kmem_free(dbp->db_base, size);
	} else {
		ASSERT(ztype <= STR_NUM_DZONE);
		ASSERT(strdzone[ztype].zonesz >= size);
		kmem_zone_free(strdzone[ztype].zonep, dbp->db_base);
	}
	return(size);
}

/*
 * allocate data buffer using zone if small enough
 */ 
static void * 
str_mem_alloc(int size, unsigned char *ztype)
{
	void *buf;

	/* doesn't fit into zones */
	if (size > MAX_STRZONE_SZ) {
		buf = kmem_alloc(size, KM_NOSLEEP);
#if DEBUG
		*ztype = (unsigned char)-1;
#endif
	} else {
		register int i;

		for (i=0; i<STR_NUM_DZONE; i++) 
			if (strdzone[i].zonesz >= size)
				break;
		ASSERT(i<STR_NUM_DZONE);
		buf = kmem_zone_alloc(strdzone[i].zonep, KM_NOSLEEP);
#if DEBUG
		strdzone[i].usecnt++;
#endif
		*ztype = i;
	}
	return(buf);
}
