/*	Copyright (c) 1990, 1991 UNIX System Laboratories, Inc.	*/
/*	Copyright (c) 1984, 1986, 1987, 1988, 1989, 1990 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF     	*/
/*	UNIX System Laboratories, Inc.                     	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

/* #ident	"@(#)uts-comm:io/strsubr.c	1.19" */

#ident "$Revision: 1.65 $"

#include <sys/types.h>
#include <sys/sysmacros.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/ksignal.h>
#include <sys/pda.h>
#include <sys/proc.h>
#include <sys/session.h>
#include <sys/cred.h>
#include <sys/user.h>
#include <sys/conf.h>
#include <sys/debug.h>
#include <sys/vnode.h>
#include <sys/file.h>
#include <sys/stream.h>
#include <sys/strsubr.h>
#include <sys/stropts.h>
#include <sys/strstat.h>
#include <sys/poll.h>
#include <sys/sema.h>
#ifdef _IRIX_LATER
#include <sys/inline.h>
#endif /* _IRIX_LATER */
#include <sys/systm.h>
#include <sys/uio.h>
#include <sys/cmn_err.h>
#include <sys/strmp.h>
#include <limits.h>
#include <fifofs/fifonode.h>
#include <sys/ddi.h>
#include <sys/runq.h>
#include <sys/idbgactlog.h>
#include <sys/mload.h>

extern int mon_runq(mon_t *, int);


#define	szlong_m1		(sizeof(long)-1)
#define SALIGN(p)               (char *)(((__psint_t)p+1) & ~1)
#define IALIGN(p)               (char *)(((__psint_t)p+3) & ~3)
#define LALIGN(p)               (char *)(((__psint_t)p+szlong_m1) & ~szlong_m1)
#define SNEXT(p)                (char *)((__psint_t)p + sizeof (short))
#define INEXT(p)                (char *)((__psint_t)p + sizeof (int))
#define LNEXT(p)                (char *)((__psint_t)p + sizeof (long))

/* KPERF is for kernel perf. measurment tools */

#ifdef KPERF  
#include <sys/disp.h>
#endif /* KPERF */

/*
 * WARNING:
 * The variables and routines in this file are private, belonging
 * to the STREAMS subsystem.  These should not be used by modules
 * or drivers.  Compatibility will not be guaranteed. 
 */

#define ncopyin(A, B, C, D)	copyin(A, B, C)		/* temporary */
#define ncopyout(A, B, C, D)	copyout(A, B, C)	/* temporary */

/*
 * =======================================================================
 * Many of the following fields must be protected on MP systems by the
 * global resource spinlock (see str_resource_lock in strmp.h/strmp.c).
 * "!" marks the protected fields.
 */

/*
 * Id value used to distinguish between different multiplexor links.
 */
STATIC int lnk_id;		/* ! */

/*
 * Queue scheduling control variables.
 */
char qrunflag;			/* ! set iff queues are enabled */
				/* XXXrs - all races appear benign? */
char queueflag;			/* ! set iff inside queuerun() */
				/* XXXrs there's some ugly code in tpisocket.c
				 * that tries to use this flag.  It's currently
				 * ifdef-ed out.  It should stay that way!
				 */
struct queue *qhead = NULL;	/* ! head of queues to run */
struct queue **qtailp = &qhead; /* !  pointer to tail of queues to run */

char strbcwait;			/* ! bufcall functions waiting */
char strbcflag;			/* ! bufcall functions ready to go */
				/* XXXrs - all races appear benign? */
struct bclist strbcalls;	/* ! list of waiting bufcalls */

/*
 * STRHOLD feature control variables.  XXXrs (protected?!?)
 */
struct stdata *strholdhead = NULL;		/* !streams with held msgs */
struct stdata **strholdtailp = &strholdhead;	/*  !pointer to tail */
int strholdflag;				/* strhold timeout pending */

struct	strstat strst;		/* ! Streams statistics structure */

/* the following 2 free lists are now referenced elsewhere
 * via macros for improved performance, so they are nolonger STATIC.
 */
struct	msgb		*msgfreelist;	/* ! */
struct	mdbblock	*mdbfreelist;	/* ! */

STATIC struct mux_node *mux_nodes;	/* ! mux info for cycle checking */
					/* XXXrs - read only structure */

#define SECACHE 10
STATIC struct seinfo *sefreelist; 	/* ! */
STATIC struct seinfo Secache[SECACHE];	/* ! emergency cache of seinfo's */
					/*   in case memory runs out	 */
STATIC struct seinfo *secachep;		/* ! cache list head		 */

struct strinfo Strinfo[NDYNAMIC];	/* ! dynamic resource info	 */
long Strcount;				/* ! count of streams resources	 */
					/* in bytes			 */

/* ======================================================================= */

struct strzne strdzone[] = {{512,NULL}, {1024,NULL}, {MAX_STRZONE_SZ, NULL}}; 

static void strgiveback(void);
static void mux_rmvedge(stdata_t *, int);
static void strunbcall(int, struct proc *);
static void strqbuf(struct stdata *);
static void adjfmtp(char **, mblk_t *, int);
static void freeband(qband_t *);

extern int insert_msg_inuse(struct mbinfo *);
extern int insert_mdb_inuse(struct mdbblock *);

#ifdef MP_STREAMS
void streams_mpdetach(queue_t *);
#endif /* MP_STREAMS */

#define MAXCPUS	128
struct queue *cpu_cur_q[MAXCPUS];

/*
 * When stream is a CTTY we need to make sure that a process exit does not
 * tear down the stream head & vnode before we're awakened with a STRHUP.
 * The strbegin/strend pair will bump se_ttycnt during this critical operation.
 */
void
strbegin(struct stdata *stp)
{
	session_t *sp = stp->sd_session;
	int s;

	if (sp) {
		ASSERT(&stp->sd_session == sp->se_sessionptr);
		s = mp_mutex_spinlock(&sp->se_lock);
		/* Make sure that session wasn't just exit-ed */
		if (stp->sd_session)
			sp->se_ttycnt++;
		mp_mutex_spinunlock(&sp->se_lock, s);
	}
}

void
strend(struct stdata *stp)
{
	session_t *sp = stp->sd_session;
	int s;

	if (sp) {
		ASSERT(&stp->sd_session == sp->se_sessionptr);
		s = mp_mutex_spinlock(&sp->se_lock);
		if (--sp->se_ttycnt == 0)
			sv_signal(&sp->se_wait);
		mp_mutex_spinunlock(&sp->se_lock, s);
	}
}

/*
 * Init routine run from main at boot time.  This contains some 
 * machine dependent code.  spl's probably are not necessary, but
 * they don't cost anything here and they can't hurt.
 */
static int strinited = 0;
static int strgivingback = NULL;

void
strinit(void)
{
	register int i;
	register int s;

	/*
	 * Duart code may call strinit() "early", before the callout
	 * tables are ready.  Make sure the following itimeout() is
	 * set up when strinit() gets called later from main.
	 */
	if (strgivingback == NULL) {
		strgivingback = itimeout((strtimeoutfunc_t)strgiveback,
					 0, 300 * HZ, splstr);
	}
	if (strinited++ != 0)
		return;

#ifdef MP_STREAMS
	str_mp_init();
#endif /* MP_STREAMS */

	s = splstr();

	/*
	 * Set up the freelists.
	 */
	mdbfreelist = NULL;
	msgfreelist = NULL;

	/*
	 * set up seinfo cache (if memory runs out, we need a few of
	 * these so things have a chance to recover)
	 */
	sefreelist = NULL;
	for (i = 0; i < SECACHE; ++i) {
		Secache[i].s_next = secachep;
		secachep = &Secache[i];
	}

	/*
	 * Set up mux_node structures.
	 */
	if ((mux_nodes = kmem_alloc((sizeof(struct mux_node) * cdevmax),
				    KM_NOSLEEP)) == NULL)
		cmn_err(CE_PANIC, "Could not allocate space for mux_nodes\n");
	for (i = 0; i < cdevmax; i++) {
		mux_nodes[i].mn_imaj = i;
		mux_nodes[i].mn_indegree = 0;
		mux_nodes[i].mn_originp = NULL;
		mux_nodes[i].mn_startp = NULL;
		mux_nodes[i].mn_outp = NULL;
		mux_nodes[i].mn_flags = 0;
	}
	for (i=0; i<STR_NUM_DZONE; i++) 
		strdzone[i].zonep = kmem_zone_init(strdzone[i].zonesz,
								"Streams");
	splx(s);
	return;
}


/*
 * see about giving back streams resources
 */

static void
strgiveback(void)
{
	static int mdbavguse = 0;
	static int msgavguse = 0;
	register int mdbfree;
	register int msgfree;
	register int n, j, s;
	register struct msgb *msgp;
	register struct mdbblock *mdbp;
	register int s2;

	LOCK_STR_RESOURCE(s2);
	mdbavguse = (mdbavguse + strst.mdbblock.use) >> 1;
	msgavguse = (msgavguse + strst.msgblock.use) >> 1;
	mdbfree	= Strinfo[DYN_MDBBLOCK].sd_cnt - strst.mdbblock.use;
	msgfree = Strinfo[DYN_MSGBLOCK].sd_cnt - strst.msgblock.use;
	UNLOCK_STR_RESOURCE(s2);
	if (mdbfree > mdbavguse)  {
		/* give back up to 1/8 of free list (easy to calculate) */
		s = splstr();
		n = mdbfree >> 3;
		for (j = 0; j < n; ++j) {
			LOCK_STR_RESOURCE(s2);
			mdbp = mdbfreelist;
			if (mdbp == NULL) {
				UNLOCK_STR_RESOURCE(s2);
				break;
			}
			mdbfreelist = 
		  	   (struct mdbblock *)( mdbp->msgblk.m_mblock.b_next);
			UNLOCK_STR_RESOURCE(s2);
			kmem_free(mdbp, sizeof(struct mdbblock));
			LOCK_STR_RESOURCE(s2);
			Strinfo[DYN_MDBBLOCK].sd_cnt--;
			Strcount -= sizeof(struct mdbblock);
			UNLOCK_STR_RESOURCE(s2);
		}
		splx(s);
	}
	if (msgfree > msgavguse) {
		/* give back up to 1/8 of free list (easy to calculate) */
		s = splstr();	
		n = msgfree >> 3;
		for (j = 0; j < n; ++j) {
			LOCK_STR_RESOURCE(s2);
			msgp = msgfreelist;
			if (msgp == NULL) {
				UNLOCK_STR_RESOURCE(s2);
				break;
			}
			msgfreelist = msgp->b_next;
			UNLOCK_STR_RESOURCE(s2);
			kmem_free(msgp, sizeof(struct mbinfo));
			LOCK_STR_RESOURCE(s2);
			Strinfo[DYN_MSGBLOCK].sd_cnt--;
			Strcount -= sizeof(struct mbinfo);
			UNLOCK_STR_RESOURCE(s2);
		}
		splx(s);
	}

	n = itimeout(strgiveback, 0, 60 * HZ, splstr);
	ASSERT(n != NULL);
}




/*
 * Send SIGPOLL signal to all processes registered on the given signal
 * list that want a signal for the specified event.  Always called at
 * splstr().
 */
void
strsendsig(register struct strevent *siglist, register int event, long data)
{
	struct strevent *sep;
	k_siginfo_t info;

	info.si_signo = SIGPOLL;
	info.si_errno = 0;
	for (sep = siglist; sep; sep = sep->se_next) {
		if (sep->se_events & event) {
			switch (event) {
			case S_INPUT:
				info.si_code = POLL_IN;
				info.si_band = data;
				goto addq;

			case S_OUTPUT:
				info.si_code = POLL_OUT;
				info.si_band = data;
				goto addq;

			case S_HIPRI:
				info.si_code = POLL_PRI;
				info.si_band = 0;
				goto addq;

			case S_MSG:
				info.si_code = POLL_MSG;
				info.si_band = data;
				goto addq;

			case S_ERROR:
				info.si_code = POLL_ERR;
				info.si_band = 0;
				info.si_errno = (int)data;
				goto addq;

			case S_HANGUP:
				info.si_code = POLL_HUP;
				info.si_band = 0;
addq:
				/*
				 * Add the info structure to se_procp's
				 * sv_sigqueue.  Fall through and actually
				 * send the signal below.
				 */
				sigaddq(sep->se_procp, &info, ASYNC_SIGNAL);

			case S_RDBAND:
				if (sep->se_events & S_BANDURG) {
					psignal(sep->se_procp, SIGURG);
					break;
				}
				/* Fall through */

			case S_RDNORM:
			case S_WRBAND:
				psignal(sep->se_procp, SIGPOLL);
				break;

			default:
				cmn_err(CE_PANIC,
				    "strsendsig: unknown event %x\n", event);
			}
		}
	}
}

/*
 * Attach a stream device or module.
 *
 * qp is a read queue; the new queue pair goes in "below" (downstream
 * from) qp so that the new queue's next read ptr points to qp, and the
 * next write ptr of the write queue associated with qp points to the
 * write queue of the new queue pair.
 *
 * Return 0 on success, or a non-zero errno on failure.
 */
int
qattach(register queue_t *qp,
	dev_t *devp,
	int flag,
	int table,
	int idx,
	cred_t *crp)
{
	register queue_t *rq;
	register int s;
	struct streamtab *qinfop;
	int sflg;
	dev_t idev;
	int error = 0;

	if ((rq = allocq()) == NULL)
		return (ENXIO);

	idev = *devp;
	if (table == CDEVSW) {
		if (*cdevsw[idx].d_flags & D_OLD) {
			rq->q_flag |= QOLD;
			WR(rq)->q_flag |= QOLD;
		} else
			/* If its a loadable module, increment the refcnt. */
			cdhold (idx);
		qinfop = cdevsw[idx].d_str;
		sflg = 0;
	} else {
		ASSERT(table == FMODSW);
		if (*fmodsw[idx].f_flags & D_OLD) {
			rq->q_flag |= QOLD;
			WR(rq)->q_flag |= QOLD;
		} else
			/* If its a loadable module, increment the refcnt. */
			fmhold (idx);
		qinfop = fmodsw[idx].f_str;
		sflg = MODOPEN;
	}


	s = splstr();
	rq->q_next = qp;
	WR(rq)->q_next = WR(qp)->q_next;
	if (WR(qp)->q_next)
		OTHERQ(WR(qp)->q_next)->q_next = rq;
	WR(qp)->q_next = WR(rq);
	rq->q_monpp = WR(rq)->q_monpp = qp->q_monpp;
	setq(rq, qinfop->st_rdinit, qinfop->st_wrinit);
 	rq->q_flag |= QWANTR;
	WR(rq)->q_flag |= QWANTR;

	/*
	 * Open the attached module or driver.
	 * The open may sleep, but it must always return here.  Therefore
	 * all sleeps must set PCATCH or ignore all signals to avoid a 
	 * longjmp if a signal arrives.
	 */
	if (rq->q_flag & QOLD) {	/* Old style interface */

		int (*olqopen)(queue_t *, dev_t, int, int, struct cred *);

		olqopen = (int (*)())rq->q_qinfo->qi_qopen;
		if ((*olqopen)(rq, idev, flag, sflg, crp) == OPENFAIL) {
			qdetach(rq, 0, 0, crp);
			splx(s);
			if ((error = curthreadp->k_error) == 0)	/* XXX */
				error = ENXIO;
			return (error);
		}
	} else {
		if (error=(*rq->q_qinfo->qi_qopen)(rq, devp, flag, sflg, crp)) {
			qdetach(rq, 0, 0, crp);
			splx(s);
			return (error);
		}
		idx = getmajor(*devp);
		/*
		 * Was the device modified on open?  If so, re-validate the
		 * major device number, get the "new" streamtab, and reset
		 * the queue information.
		 */ 
		if ((table == CDEVSW) && (idx != getmajor(idev))) {
			if (!cdvalid(idx) || !(qinfop = cdevsw[idx].d_str)) {
				qdetach(rq, 0, 0, crp);
				splx(s);
				return (ENXIO);
			}
			setq(rq, qinfop->st_rdinit, qinfop->st_wrinit);
		}
	}
	splx(s);
	return (0);
}

/*
 * Detach a stream module or device.
 * If clmode == 1 then the module or driver was opened and its
 * close routine must be called.  If clmode == 0, the module
 * or driver was never opened or the <open failed, and so its close
 * should not be called.
 */
void
qdetach(register queue_t *rq, int clmode, int flag, cred_t *crp)
{
	register int s, s2;
	register queue_t *wq;
	int (*olqclose)(queue_t *, int, struct cred *);

        ASSERT(rq->q_flag&QREADR);
        wq = WR(rq);

	if (clmode) {
#ifdef _EAGER_RUNQUEUES
		if (qready())
			runqueues();
#endif /* _EAGER_RUNQUEUES */
		s = splstr();
		if (rq->q_flag & QOLD) {
			olqclose = rq->q_qinfo->qi_qclose;
			(*olqclose)(rq, flag, crp);
		} else {
			(*rq->q_qinfo->qi_qclose)(rq, flag, crp);
		}
		flushq(rq, FLUSHALL);
		flushq(wq, FLUSHALL);
	} else 
		s = splstr();

	/* If this is a loadable module, decrement the refcnt */
	mlqdetach (rq);

	/*
	 * Check if queues are still enabled, and remove from 
	 * runlist if necessary.
	 */
	LOCK_STR_RESOURCE(s2);

#ifndef MP_STREAMS
	if ((rq->q_pflag | wq->q_pflag) & QENAB) {
                register queue_t *q;
                register queue_t **prevp;

                q = qhead;
                prevp = &qhead;
                while (q) {
                        if (q == rq || q == wq)
                                *prevp = q->q_link;
                        else
                                prevp = &q->q_link;
                        q = q->q_link;
                }
                qtailp = prevp;
	}
#endif

        /*
         * If this queue has held messages, remove its associated
         * stream head from the list of streams with held messages.
         */
        if (wq->q_flag & QHLIST) {
                register struct stdata *sp = strholdhead;
                register struct stdata **prevp = &strholdhead;
                struct stdata *stp = (struct stdata *)wq->q_ptr;

                while (sp) {
                        if (sp == stp) {
                                *prevp = sp->sd_hnext;
                                wq->q_flag &= ~QHLIST;
                        }
                        else
                                prevp = &sp->sd_hnext;
                        sp = sp->sd_hnext;
                }
                strholdtailp = prevp;

                ASSERT(!(wq->q_flag & QHLIST));
        }

	UNLOCK_STR_RESOURCE(s2);

	if (wq->q_next)
		backq(rq)->q_next = rq->q_next;
	if (rq->q_next)
		backq(wq)->q_next = wq->q_next;
	splx(s);
#ifdef MP_STREAMS
	streams_mpdetach(rq);
#endif /* MP_STREAMS */
	freeq(rq);
}

/*
 * This function is placed in the callout table when messages have been
 * held in strwrite, in case the hope of being able to add more data is
 * unrealized.
 */
void
strrelease()  {
	int s;

	LOCK_STR_RESOURCE(s);
	setqsched();
	strholdflag = 0;
	UNLOCK_STR_RESOURCE(s);
}

/*
 * This function is placed in the callout table to wake up a process
 * waiting to close a stream that has not completely drained.
 */
void
strtime(struct stdata *stp)
{
	register int s;

	if (stp->sd_flag & STRTIME) {
		s = splstr();
		stp->sd_flag &= ~STRTIME;
		splx(s);
		wakeup(stp->sd_wrq);
	}
}

/*
 * This function is placed in the callout table to wake up all
 * processes waiting to send an ioctl down a particular stream,
 * as well as the process whose ioctl is still outstanding.  The
 * process placing this function in the callout table will remove
 * it if he gets control of the ioctl mechanism for the stream -
 * this should only run if there is a failure.  This wakes up
 * the same processes as str3time below.
 */
void
str2time(struct stdata *stp)
{
	register int s;

	if (stp->sd_flag & STR2TIME) {
		s = splstr();
		stp->sd_flag &= ~STR2TIME;
		splx(s);
		wakeup(&stp->sd_iocwait);
	}
}

/*
 * This function is placed in the callout table to wake up the
 * process that has an outstanding ioctl waiting acknowledgement
 * on a stream, as well as any processes waiting to send their
 * own ioctl messages.  It should be removed from the callout table
 * when the acknowledgement arrives.  If this function runs, it
 * is the result of a failure.  This wakes up the same processes
 * as str2time above.
 */
void
str3time(struct stdata *stp)
{
	register int s;

	if (stp->sd_flag & STR3TIME) {
		s = splstr();
		stp->sd_flag &= ~STR3TIME;
		splx(s);
		wakeup(stp);
	}
}

/*
 * Put ioctl data from user land to ioctl buffers.  Return non-zero
 * errno for failure, 1 for success.
 */
int
putiocd(register mblk_t *bp,
	mblk_t *ebp,
	caddr_t arg,
	int copymode,
	int flag,
	char *fmt)
{
	register mblk_t *tmp;
	register int count, n;
	mblk_t *obp = bp;
	int error = 0;

	/* XXXrs - needed? ASSERT(STREAMS_LOCKED); */

	if (bp->b_datap->db_type == M_IOCTL)
		count = ((struct iocblk *)bp->b_rptr)->ioc_count;
	else {
		ASSERT(bp->b_datap->db_type == M_COPYIN);
		count = ((struct copyreq *)bp->b_rptr)->cq_size;
	}
	/*
	 * strdoioctl validates ioc_count, so if this assert fails it
	 * cannot be due to user error.
	 */
	ASSERT(count >= 0);

	while (count) {
		n = MIN(MAXIOCBSZ, count);
		if (flag == SE_SLEEP) {
			while (!(tmp = allocb(n, BPRI_HI))) {
				if (error = strwaitbuf(n, BPRI_HI))
					return (error);
			}
		} else if (!(tmp = allocb(n, BPRI_HI)))
			return (EAGAIN);
		error = STREAMS_COPYIN((caddr_t)arg, (caddr_t)tmp->b_wptr,
				       n, fmt, copymode);
		if (error) {
			freeb(tmp);
			return (error);
		}
		if (fmt && (count > MAXIOCBSZ) && (copymode == U_TO_K))
			adjfmtp(&fmt, tmp, n);
		arg += n;
		tmp->b_datap->db_type = M_DATA;
		tmp->b_wptr += n;
		count -= n;
		bp = (bp->b_cont = tmp);
	}

	/*
	 * If ebp was supplied, place it between the
	 * M_IOCTL block and the (optional) M_DATA blocks.
	 */
	if (ebp) {
		ebp->b_cont = obp->b_cont;
		obp->b_cont = ebp;
	}
	return (0);
}

/*
 * Copy ioctl data to user-land.  Return non-zero errno on failure,
 * 0 for success.
 */
int
getiocd(register mblk_t *bp, caddr_t arg, int copymode, char *fmt)
{
	register int count, n;
	int error;

	/* XXXrs - needed? ASSERT(STREAMS_LOCKED); */

	if (bp->b_datap->db_type == M_IOCACK)
		count = ((struct iocblk *)bp->b_rptr)->ioc_count;
	else {
		ASSERT(bp->b_datap->db_type == M_COPYOUT);
		count = ((struct copyreq *)bp->b_rptr)->cq_size;
	}
	ASSERT(count >= 0);

	for (bp = bp->b_cont; bp && count; count -= n, bp = bp->b_cont, arg += n) {
		n = MIN(count, bp->b_wptr - bp->b_rptr);
		error = STREAMS_COPYOUT((caddr_t)bp->b_rptr, (caddr_t)arg,
					n, fmt, copymode);
		if (error)
			return (error);
		if (fmt && bp->b_cont && (copymode == U_TO_K))
			adjfmtp(&fmt, bp, n);
	}
	ASSERT(count == 0);
	return (0);
}

/* 
 * Allocate a linkinfo table entry given the write queue of the
 * bottom module of the top stream and the write queue of the
 * stream head of the bottom stream.
 *
 * linkinfo table entries are freed by nulling the li_lblk.l_qbot field.
 */
struct linkinfo *
alloclink(queue_t *qup, queue_t *qdown, file_t *fpdown)
{
	register struct linkinfo *linkp, *lp;
	register int s, s2;

	s = splstr();
	if ((linkp = kmem_zalloc(sizeof(struct linkinfo),
				 KM_NOSLEEP)) == NULL) {
		LOCK_STR_RESOURCE(s2);
		strst.linkblk.fail++;
		UNLOCK_STR_RESOURCE(s2);
		splx(s);
		return(NULL);
	}
	linkp->li_lblk.l_qtop = qup;
	linkp->li_lblk.l_qbot = qdown;

	/*
	 * Assign link id, being careful to watch out
	 * for wrap-around leading to id clashes.  Unlike
	 * ioctl ids, link ids can be long-lived.
	 * If we're concerned about covert channels, start the id search
	 * at a random place rather than where the previous search stopped.
	 * If the random() routine is stubbed out, it will return 0,
	 * in which case we want to revert to the sequential method.
	 */
	LOCK_STR_RESOURCE(s2);
	do {
		linkp->li_lblk.l_index = ++lnk_id;
		if (lnk_id <= 0)
			linkp->li_lblk.l_index = lnk_id = 1;
		for (lp = (struct linkinfo *)Strinfo[DYN_LINKBLK].sd_head;
		     lp; lp = lp->li_next)
			if (lp->li_lblk.l_index == lnk_id)
				break;
	} while (lp);

	linkp->li_fpdown = fpdown;
	BUMPUP(strst.linkblk);
	Strinfo[DYN_LINKBLK].sd_cnt++;
	Strcount += sizeof(struct linkinfo);
	linkp->li_next = (struct linkinfo *) Strinfo[DYN_LINKBLK].sd_head;
	if (linkp->li_next)
		linkp->li_next->li_prev = linkp;
	linkp->li_prev = NULL;
	Strinfo[DYN_LINKBLK].sd_head = (void *) linkp;
	UNLOCK_STR_RESOURCE(s2);
	splx(s);
	return(linkp);
}

/*
 * Free a linkinfo entry.
 */
void
lbfree(register struct linkinfo *linkp)
{
	register int s;
	register int s2;

	s = splstr();
	LOCK_STR_RESOURCE(s2);
	if (linkp->li_prev == NULL) {
		if (linkp->li_next)
			linkp->li_next->li_prev = NULL;
		Strinfo[DYN_LINKBLK].sd_head = (void *) linkp->li_next;
	}
	else {
		if (linkp->li_next)
			linkp->li_next->li_prev = linkp->li_prev;
		linkp->li_prev->li_next = linkp->li_next;
	}
	UNLOCK_STR_RESOURCE(s2);
	kmem_free(linkp, sizeof(struct linkinfo));
	LOCK_STR_RESOURCE(s2);
	Strinfo[DYN_LINKBLK].sd_cnt--;
	Strcount -= sizeof(struct linkinfo);
	strst.linkblk.use--;
	UNLOCK_STR_RESOURCE(s2);
	splx(s);
	return;
}

/*
 * Check for a potential linking cycle.
 * Return 1 if a link will result in a cycle,
 * and 0 otherwise.
 */
int
linkcycle(stdata_t *upstp, stdata_t *lostp)
{
	register struct mux_node *np;
	register struct mux_edge *ep;
	register int i;
	long lomaj;
	long upmaj;
	/*
	 * if the lower stream is a pipe/FIFO, return, since link
	 * cycles can not happen on pipes/FIFOs
	 */
	if (lostp->sd_vnode->v_type == VFIFO)
		return(0);

	for (i = 0; i < cdevmax; i++) {
		np = &mux_nodes[i];
		MUX_CLEAR(np);
	}
	lomaj = getmajor(lostp->sd_vnode->v_rdev);
	upmaj = getmajor(upstp->sd_vnode->v_rdev);
	np = &mux_nodes[lomaj];
	for ( ; ; ) {
		if (!MUX_DIDVISIT(np)) {
			if (np->mn_imaj == upmaj)
				return (1);
			if (np->mn_outp == NULL) {
				MUX_VISIT(np);
				if (np->mn_originp == NULL)
					return (0);
				np = np->mn_originp;
				continue;
			}
			MUX_VISIT(np);
			np->mn_startp = np->mn_outp;
		} else {
			if (np->mn_startp == NULL) {
				if (np->mn_originp == NULL)
					return (0);
				else {
					np = np->mn_originp;
					continue;
				}
			}
			ep = np->mn_startp;
			np->mn_startp = ep->me_nextp;
			ep->me_nodep->mn_originp = np;
			np = ep->me_nodep;
		}
	}
	/* NOTREACHED */
}

/* 
 * Find linkinfo table entry corresponding to the parameters.
 */
struct linkinfo *
findlinks(stdata_t *stp, int index, int type)
{
	register struct linkinfo *linkp;
	register struct mux_edge *mep;
	register int s2;
	struct mux_node *mnp;
	queue_t *qup;

	if ((type & LINKTYPEMASK) == LINKNORMAL) {
		qup = getendq(stp->sd_wrq);
		LOCK_STR_RESOURCE(s2);
		for (linkp = (struct linkinfo *) Strinfo[DYN_LINKBLK].sd_head;
		     linkp; linkp = linkp->li_next) {
			if ((qup == linkp->li_lblk.l_qtop) &&
		    	    (!index || (index == linkp->li_lblk.l_index))) {
				UNLOCK_STR_RESOURCE(s2);
				return (linkp);
			}
		}
		UNLOCK_STR_RESOURCE(s2);
	} else {
		ASSERT((type & LINKTYPEMASK) == LINKPERSIST);
		mnp = &mux_nodes[getmajor(stp->sd_vnode->v_rdev)];
		mep = mnp->mn_outp;
		while (mep) {
			if ((index == 0) || (index == mep->me_muxid))
				break;
			mep = mep->me_nextp;
		}
		if (!mep)
			return (NULL);
		LOCK_STR_RESOURCE(s2);
		for (linkp = (struct linkinfo *) Strinfo[DYN_LINKBLK].sd_head;
		     linkp; linkp = linkp->li_next) {
			if ( (!linkp->li_lblk.l_qtop) &&
			    (mep->me_muxid == linkp->li_lblk.l_index)) {
				UNLOCK_STR_RESOURCE(s2);
				return (linkp);
			}
		}
		UNLOCK_STR_RESOURCE(s2);
	}
	return (NULL);
}

/* 
 * Given a queue ptr, follow the chain of q_next pointers until you reach the
 * last queue on the chain and return it.
 */
queue_t *
getendq(register queue_t *q)
{
	ASSERT( q!= NULL);
	while (SAMESTR(q))
		q = q->q_next;
	return (q);
}

/*
 * Unlink a multiplexor link.  Stp is the controlling stream for the
 * link, fpdown is the file pointer for the lower stream, and
 * linkp points to the link's entry in the linkinfo table.
 */
int
munlink(struct stdata **stpp,
	struct linkinfo *linkp,
	int flag,
	cred_t *crp,
	int *rvalp)
{
	register int s;
	struct strioctl strioc;
	struct stdata *stp, *stpdown;
	queue_t *rq;
	int error = 0;
	struct file *fp;

	stp = *stpp; /* safe since we're locked */
	ASSERT(stp);
	if ((flag & LINKTYPEMASK) == LINKNORMAL)
		strioc.ic_cmd = I_UNLINK;
	else
		strioc.ic_cmd = I_PUNLINK;
	strioc.ic_timout = 0;
	strioc.ic_len = sizeof(struct linkblk);
	strioc.ic_dp = (char *)&linkp->li_lblk;
	
	error = strdoioctl(stp, &strioc, NULL, K_TO_K, STRLINK, crp, rvalp);

	/*
	 * If there was an error and this is not called via strclose, 
	 * return to the user.  Otherwise, pretend there was no error 
	 * and close the link.  
	 */
	if (error) {
		if (flag & LINKCLOSE) {
			cmn_err(CE_CONT, "KERNEL: munlink: could not perform unlink ioctl, closing anyway\n");
			s = splstr();
			stp->sd_flag &= ~(STRDERR|STWRERR); /* allows strdoioctl() to work */
			splx(s);
		} else
			return (error);
	}

	stpdown = linkp->li_fpdown->f_vnode->v_stream;
	s = splstr();
	stpdown->sd_flag &= ~STPLEX;
	mux_rmvedge(stp, linkp->li_lblk.l_index);
	rq = RD(stpdown->sd_wrq);
	setq(rq, &strdata, &stwdata);
	rq->q_ptr = WR(rq)->q_ptr = (caddr_t)stpdown;
	splx(s);
	fp = linkp->li_fpdown;
	lbfree(linkp);
	STRHEAD_UNLOCK(stp);
	closef(fp); /* XXXrs is this really safe? */
	STRHEAD_LOCK(stpp, stp);
	ASSERT(stp); /* XXXrs really? */
	return (0);
}

/*
 * Unlink all multiplexor links for which stp is the controlling stream.
 * Return 0, or a non-zero errno on failure.
 */
int
munlinkall(struct stdata **stpp, int flag, cred_t *crp, int *rvalp)
{
	struct stdata *stp;
	struct linkinfo *linkp;
	int error = 0;

	stp = *stpp; /* safe since we're locked */
	ASSERT(stp);
	while (linkp = findlinks(stp, 0, flag)) {
		if (error = munlink(stpp, linkp, flag, crp, rvalp))
			return (error);
	}
	return (0);
}

/*
 * A multiplexor link has been made.  Add an
 * edge to the directed graph.  Returns 0 on success
 * and an error number on failure.
 */
int
mux_addedge(stdata_t *upstp, stdata_t *lostp, int muxid)
{
	register struct mux_node *np;
	register struct mux_edge *ep;
	long upmaj;
	long lomaj;

	upmaj = getmajor(upstp->sd_vnode->v_rdev);
	lomaj = getmajor(lostp->sd_vnode->v_rdev);
	np = &mux_nodes[upmaj];
	if (np->mn_outp) {
		ep = np->mn_outp;
		while (ep->me_nextp)
			ep = ep->me_nextp;
		if ((ep->me_nextp = (struct mux_edge *)kmem_alloc(sizeof(struct mux_edge), 0)) == NULL) {
			cmn_err(CE_WARN, "Can not allocate memory for mux_edge\n");
			return (EAGAIN);
		}
		ep = ep->me_nextp;
	} else {
		if ((np->mn_outp = (struct mux_edge *)kmem_alloc(sizeof(struct mux_edge), 0)) == NULL) {
			cmn_err(CE_WARN, "Can not allocate memory for mux_edge\n");
			return (EAGAIN);
		}
		ep = np->mn_outp;
	}
	ep->me_nextp = NULL;
	ep->me_muxid = muxid;
	ep->me_nodep = &mux_nodes[lomaj];
	return (0);
}

/*
 * A multiplexor link has been removed.  Remove the
 * edge in the directed graph.
 */
static void
mux_rmvedge(stdata_t *upstp, int muxid)
{
	register struct mux_node *np;
	register struct mux_edge *ep;
	register struct mux_edge *pep = NULL;
	long upmaj;

	upmaj = getmajor(upstp->sd_vnode->v_rdev);
	np = &mux_nodes[upmaj];
	ASSERT (np->mn_outp != NULL);
	ep = np->mn_outp;
	while (ep) {
		if (ep->me_muxid == muxid) {
			if (pep)
				pep->me_nextp = ep->me_nextp;
			else
				np->mn_outp = ep->me_nextp;
			kmem_free(ep, sizeof(struct mux_edge));
			return;
		}
		pep = ep;
		ep = ep->me_nextp;
	}
	ASSERT(0);	/* should not reach here */
}

/*
 * Set the interface values for a pair of queues (qinit structure,
 * packet sizes, water marks).
 */
void
setq(queue_t *rq, struct qinit *rinit, struct qinit *winit)
{
	register queue_t  *wq;
	register int s;

	ASSERT(STREAM_LOCKED(rq));

	wq = WR(rq);
	s = splstr();
	rq->q_qinfo = rinit;
	rq->q_hiwat = rinit->qi_minfo->mi_hiwat;
	rq->q_lowat = rinit->qi_minfo->mi_lowat;
	rq->q_minpsz = rinit->qi_minfo->mi_minpsz;
	rq->q_maxpsz = rinit->qi_minfo->mi_maxpsz;
	wq->q_qinfo = winit;
	wq->q_hiwat = winit->qi_minfo->mi_hiwat;
	wq->q_lowat = winit->qi_minfo->mi_lowat;
	wq->q_minpsz = winit->qi_minfo->mi_minpsz;
	wq->q_maxpsz = winit->qi_minfo->mi_maxpsz;
	splx(s);
}

/*
 * Make a protocol message given control and data buffers.
 *
 * Caller should be holding the streams monitor.
 */
int
strmakemsg(register struct strbuf *mctl,
	   int count,
	   struct uio *uiop,
	   struct stdata *stp,
	   long flag,
	   mblk_t **mpp)
{
	register mblk_t *mp = NULL;
	register mblk_t *bp;
	caddr_t base;
	int pri;
	int msgtype;
	int wroff = (int)stp->sd_wroff;
	int offlg = 0;
	int error = 0;

	ASSERT(STREAM_LOCKED(RD(stp->sd_wrq)));

	*mpp = NULL;
	if (flag & RS_HIPRI)
		pri = BPRI_MED;
	else
		pri = BPRI_LO;

	/*
	 * Create control part of message, if any.
	 */
	if ((mctl != NULL) && (mctl->len >= 0)) {
		register int ctlcount;
		int allocsz;

		if (flag & RS_HIPRI) 
			msgtype = M_PCPROTO;
		else 
			msgtype = M_PROTO;

		ctlcount = mctl->len;
		base = mctl->buf;

		/*
		 * Give modules a better chance to reuse M_PROTO/M_PCPROTO
		 * blocks by increasing the size to something more usable.
		 */
		allocsz = MAX(ctlcount, 64);

		/*
		 * Range checking has already been done; simply try
		 * to allocate a message block for the ctl part.
		 */
		while (!(bp = allocb(allocsz, pri))) {
			if (uiop->uio_fmode  & (FNDELAY|FNONBLOCK))
				return (EAGAIN);
			if (error = strwaitbuf(allocsz, pri))
				return (error);
		}

		bp->b_datap->db_type = msgtype;
		if (STREAMS_COPYIN(base, (caddr_t)bp->b_wptr, ctlcount,
				   "XXXrs fmt", U_TO_K)) {
			freeb(bp);
			return (EFAULT);
		}

		/*
		 * We could have slept copying in user pages.
		 * Recheck the stream head state (the other end
		 * of a pipe could have gone away).
		 */
		if (stp->sd_flag & (STRHUP|STWRERR|STPLEX)) {
			error = ((stp->sd_flag & STPLEX) ? EINVAL :
			    stp->sd_werror);
			freeb(bp);
			return (error);
		}
		bp->b_wptr += ctlcount;
		mp = bp;
	}

	/*
	 * Create data part of message, if any.
	 */
	if (count >= 0) {
		register int size;

		size = count + (offlg ? 0 : wroff);
		while ((bp = allocb(size, pri)) == NULL) {
			if (uiop->uio_fmode  & (FNDELAY|FNONBLOCK))
				return (EAGAIN);
			if (error = strwaitbuf(size, pri)) {
				freemsg(mp);
				return (error);
			}
		}
		if (wroff && !offlg++ &&
		    (wroff < bp->b_datap->db_lim - bp->b_wptr)) {
			bp->b_rptr += wroff;
			bp->b_wptr += wroff;
		}
		if ((size = MIN(count, bp->b_datap->db_lim - bp->b_wptr)) &&
		    (error = STREAMS_UIOMOVE((caddr_t)bp->b_wptr, size,
					     UIO_WRITE, uiop))) {
			freeb(bp);
			freemsg(mp);
			return (error);
		}

		/*
		 * We could have slept copying in user pages.
		 * Recheck the stream head state (the other end
		 * of a pipe could have gone away).
		 */
		if (stp->sd_flag & (STRHUP|STWRERR|STPLEX)) {
			error = ((stp->sd_flag & STPLEX) ? EINVAL :
			    stp->sd_werror);
			freeb(bp);
			freemsg(mp);
			return (error);
		}
		bp->b_wptr += size;
		count -= size;
		if (!mp)
			mp = bp;
		else
			linkb(mp, bp);

	}
	*mpp = mp;
	return (0);
}

/*
 * Wait for a buffer to become available.  Return non-zero errno
 * if not able to wait, 0 if buffer is probably there.
 */
int
strwaitbuf(int size, int pri)
{
	register int s;

	s = splstr();
	if (!bufcall(size,pri,(void(*)(long))wakeup,(long)&curprocp->p_flag)) {
		splx(s);
		return (ENOSR);
	}
	if (sleep((caddr_t)&(curprocp->p_flag), STOPRI|PCATCH)) {
		strunbcall(size, curprocp);
		splx(s);
		return (EINTR);
	}
	strunbcall(size, curprocp);
	splx(s);
	return (0);
}

/*
 * Remove a wakeup for the given process from the bufcall list for
 * the given buffer size.  'size' can be -1 for externally-supplied buffers.
 */
static void
strunbcall(int size, struct proc *p)
{
	register int s;
	register struct strevent *sep, *prevsep;
	
	LOCK_STR_RESOURCE(s);
	sep = strbcalls.bc_head;
	prevsep = NULL;
	while (sep) {
		if ((sep->se_arg == (long)&p->p_flag) &&
		    (sep->se_func == (void (*)(long))wakeup) &&
		    (sep->se_size == size)) {
			if (prevsep == NULL)
				strbcalls.bc_head = sep->se_next;
			else
				prevsep->se_next = sep->se_next;
			if (sep == strbcalls.bc_tail)
				strbcalls.bc_tail = prevsep;
			sefree(sep);
			UNLOCK_STR_RESOURCE(s);
			return;
		}
		prevsep = sep;
		sep = sep->se_next;
	}
	UNLOCK_STR_RESOURCE(s);
}


static int
strwaitsleep(struct stdata *stp, queue_t *q, int flag, int pri)
{
        struct vnode *vp;

        vp = stp->sd_vnode;
        if (vp->v_type == VFIFO) {
		sv_t *svp = (flag & READWAIT || flag & GETWAIT) ?
				&VTOF(vp)->fn_empty : &VTOF(vp)->fn_full;
		mutex_t *mp;

		ASSERT(VTOF(vp)->fn_lockp);
		mp = &VTOF(vp)->fn_lockp->fn_lock;
		mutex_lock(mp, PZERO);
		if ((pri & PMASK) > PZERO)
			return sv_wait_sig(svp, pri, mp, 0);
		sv_wait(svp, pri, mp, 0);
		return 0;
        } else {
		return(sleep(q, pri));
        }
}

/*
 * This function waits for a read or write event to happen on a stream.
 */
int
strwaitq(register struct stdata *stp,
	 int flag,
	 off_t count,
	 int fmode,
#ifndef _IRIX_LATER
	 int vtime,
#endif /* !_IRIX_LATER */
	 int *done)
{
	extern void strwakeq(struct stdata *stp, queue_t *q, int flag);
	register int s, s2;
	int slpflg, slppri, errs;
	int error = 0;
	queue_t *slpq;
	mblk_t *mp;
	long *rd_count;
	int bid;
#ifndef _IRIX_LATER
	int id;
#endif /* !_IRIX_LATER */

#ifdef _IRIX_LATER
	if (fmode & (FNDELAY|FNONBLOCK)) {
		if (!(flag & NOINTR))
			error = EAGAIN;
		*done = 1;
		return (error);
	}
#else /* !_IRIX_LATER */
	/* if POSIX no-delay and it's a tty, return error */
	if ((fmode & FNONBLK) && (stp->sd_flag & STRISTTY)) {
		*done = 1;
                return(EAGAIN);
        }

	if (((fmode&(FNONBLK|FNDELAY))          /* if no delay on, */
	    && (!(stp->sd_flag & STRISTTY)     /* and not a tty or reading */
		 || (flag & (READWAIT|NOINTR))))
	    || (stp->sd_flag & STFIONBIO)) {    /* or 4.2 non-delay on */
		if (flag & NOINTR) {    /* quit */
			;               /* without error if 2nd time around */

		} else if (stp->sd_flag & STFIONBIO) {
			error = EWOULDBLOCK;        /* with 4.2 error */

		} else if (!(stp->sd_flag & STRISTTY)) {
			error = EAGAIN;     /* no error if tty */
		}
		*done = 1;
                return(error);
	}
#endif /* _IRIX_LATER */

	if ((flag & READWAIT) || (flag & GETWAIT)) {
		slpflg = RSLEEP;
		slpq = RD(stp->sd_wrq);
		slppri = STIPRI;
		errs = STRDERR|STPLEX;

		/*
		 * If any module downstream has requested read notification
		 * by setting SNDMREAD flag using M_SETOPTS, send a message
		 * down stream.
		 */
		if ((flag & READWAIT) && (stp->sd_flag & SNDMREAD)) {
			while ((mp = allocb(sizeof(long), BPRI_MED)) == NULL) {
				s = splstr();
				bid = bufcall(sizeof(long), BPRI_MED,
					      (void(*)(long))strqbuf,
					      (long)stp);
				if (bid == 0) {
					splx(s);
					*done = 1;
					return (EAGAIN);
				}
				if (stp->sd_sid)
					strbegin(stp);
				stp->sd_flag |= RDBUFWAIT;
				if (strwaitsleep(stp, slpq,
						 flag, slppri|PCATCH)) {
					if (stp->sd_flag & RDBUFWAIT) {
						/* interrupted sleep */
						stp->sd_flag &= ~RDBUFWAIT;
						unbufcall(bid);
						if (stp->sd_sid)
							strend(stp);
						splx(s);
						*done = 1;
						return (EINTR);
					}
				}
				if (stp->sd_sid)
					strend(stp);
				splx(s);
			}
			mp->b_datap->db_type = M_READ;
			rd_count = (long *)mp->b_wptr;
			*rd_count = count;
			mp->b_wptr += sizeof(long);
			/*
			 * Send the number of bytes requested by the
			 * read as the argument to M_READ.
			 */
			putnext(stp->sd_wrq, mp);
#ifdef _EAGER_RUNQUEUES
			if (qready())
				runqueues();
#endif /* _EAGER_RUNQUEUES */
			/*
			 * If any data arrived due to inline processing
			 * of putnext(), don't sleep.
			 */
			mp = RD(stp->sd_wrq)->q_first;
			while (mp) {
				if (!(mp->b_flag & MSGNOGET))
					break;
				mp = mp->b_next;
			}
			if (mp != NULL) {
				*done = 0;
				return (error);
			}
		}
	} else {
		slpflg = WSLEEP;
		slpq = stp->sd_wrq;
		slppri = STOPRI;
		errs = STWRERR|STRHUP|STPLEX;
	}
	
	s = splstr();
#ifndef _IRIX_LATER
	if (vtime)
		id = MP_STREAMS_TIMEOUT2(stp->sd_wrq->q_monpp, strwakeq, stp,
					 vtime*(short)(HZ/10), slpq,
					 (void *)(__psunsigned_t)flag);
#endif /* !_IRIX_LATER */
	if (stp->sd_sid)
		strbegin(stp);
	LOCK_STR_RESOURCE(s2);
	stp->sd_pflag |= slpflg;
	UNLOCK_STR_RESOURCE(s2);
	if (strwaitsleep(stp, slpq, flag, slppri|PCATCH)) {
		LOCK_STR_RESOURCE(s2);
		stp->sd_pflag &= ~slpflg;
		UNLOCK_STR_RESOURCE(s2);
		if (stp->sd_sid)
			strend(stp);
		splx(s);
		strwakeq(stp, slpq, flag);
#ifndef _IRIX_LATER
		if (vtime)
			untimeout(id);
#endif /* !_IRIX_LATER */
		if (!(flag & NOINTR))
			error = EINTR;
		*done = 1;
		return (error);
	}
	if (stp->sd_sid)
		strend(stp);
	splx(s);
#ifndef _IRIX_LATER
	if (vtime)
		untimeout(id);
#endif /* !_IRIX_LATER */
	if (stp->sd_flag & errs) {
		*done = 1;
		return ((stp->sd_flag & STPLEX) ? EINVAL :
		    ((errs & STWRERR) ? stp->sd_werror : stp->sd_rerror));
	}
	*done = 0;
	return (error);
}

static int
str2num(register char **str)
{
	register int n;

	n = 0;
	for (; **str >= '0' && **str <= '9'; (*str)++)
		n = 10 * n + **str - '0';
	return (n);
}

/*
 * Update canon format pointer with "bytes" worth of data.
 *
 * Note that adjfmtp is only called when copymode == U_TO_K.
 * Hence it is OK for adjfmtp to access curprocp->p_abi.
 */
static void
adjfmtp(register char **str, register mblk_t *bp, int bytes)
{
	register caddr_t addr;
	register caddr_t lim;
	long num;
#if _K64U64
	/* This depends on being called only when copymode is U_TO_K */
	register int abi32 = ! ABI_IS_IRIX5_64(curprocp->p_abi);
#endif

	addr = (caddr_t)bp->b_rptr;
	lim = addr + bytes;
	while (addr < lim) {
		switch (*(*str)++) {
		case 's':			/* short */
			addr = SALIGN(addr);
			addr = SNEXT(addr);
			break;
		case 'i':			/* integer */
			addr = IALIGN(addr);
			addr = INEXT(addr);
			break;
		case 'l':			/* long */
#if _K64U64
			if (abi32) {
				addr = IALIGN(addr);
				addr = INEXT(addr);
				break;
			}
#endif
			addr = LALIGN(addr);
			addr = LNEXT(addr);
			break;
		case 'b':			/* byte */
			addr++;
			break;
		case 'c':			/* character */
			if ((num = str2num(str)) == 0) {
				while (*addr++)
					;
			} else
				addr += num;
			break;
		case 0:
			return;
		default:
			break;
		}
	}
}

static void
strqbuf(register struct stdata *stp)
{
	register int s;
	 
	if (stp->sd_flag & RDBUFWAIT) {
		s = splstr();
		stp->sd_flag &= ~RDBUFWAIT;
		splx(s);
		wakeup(RD(stp->sd_wrq));
	}
}

/*
 * Return size of message of block type (bp->b_datap->db_type)
 * If fromhead is non-zero, then start at beginning of message.
 * If fromhead is zero, then start at end of message.
 */
int
xmsgsize(mblk_t *mp, int fromhead)
{
	register mblk_t *bp;
	register unsigned char type;
	register int count;
	register int s;
	mblk_t *endbp;
	
	s = splstr();
	if (fromhead) {
		bp = mp;
		type = bp->b_datap->db_type;
		count = 0;
		for (; bp; bp = bp->b_cont) {
			if (type != bp->b_datap->db_type)
				break;
			ASSERT(bp->b_wptr >= bp->b_rptr);
			count += bp->b_wptr - bp->b_rptr;
		}
	} else {
		for (bp = mp; bp->b_cont; bp = bp->b_cont)
			;
		type = bp->b_datap->db_type;
		ASSERT(bp->b_wptr >= bp->b_rptr);
		count = bp->b_wptr - bp->b_rptr;
		endbp = bp;
		while (bp != mp) {
			for (bp = mp; bp->b_cont != endbp; bp = bp->b_cont)
				;
			if (type != bp->b_datap->db_type)
				break;
			ASSERT(bp->b_wptr >= bp->b_rptr);
			count += bp->b_wptr - bp->b_rptr;
			endbp = bp;
		}
	}
	splx(s);
	return(count);
}

/*
 * Allocate a stream head.
 */
struct stdata *
shalloc(queue_t *qp)
{
	register int s;
	register int s2;
	register stdata_t *stp;
	register struct shinfo *shp;

	s = splstr();
	if ((shp = kmem_zalloc(sizeof(struct shinfo), KM_SLEEP)) == NULL) {
		LOCK_STR_RESOURCE(s2);
		strst.stream.fail++;
		UNLOCK_STR_RESOURCE(s2);
		splx(s);
		return(NULL);
	}
	stp = (stdata_t *) shp;
	stp->sd_wrq = WR(qp);
	initpollhead(&stp->sd_pollist);
	LOCK_STR_RESOURCE(s2);
	BUMPUP(strst.stream);
	Strinfo[DYN_STREAM].sd_cnt++;
	Strcount += sizeof(struct shinfo);
	shp->sh_next = (struct shinfo *) Strinfo[DYN_STREAM].sd_head;
	if (shp->sh_next)
		shp->sh_next->sh_prev = shp;
	shp->sh_prev = NULL;
	Strinfo[DYN_STREAM].sd_head = (void *) shp;
	UNLOCK_STR_RESOURCE(s2);
	splx(s);
	return(stp);
}

/*
 * Free a stream head.
 */
void
shfree(register struct stdata *stp)
{
	register int s;
	register int s2;
	register struct shinfo *shp;

	while (stp->sd_pollist.ph_list != NULL)
		/* 
		 * allow the processes that slep on stp->sd_pollist
		 * to clean up polldat
 		 */
		delay(10);
	spinlock_destroy(&stp->sd_pollist.ph_lock);
	s = splstr();
	LOCK_STR_RESOURCE(s2);
	shp = (struct shinfo *) stp;
	if (shp->sh_prev == NULL) {
		if (shp->sh_next)
			shp->sh_next->sh_prev = NULL;
		Strinfo[DYN_STREAM].sd_head = (void *) shp->sh_next;
	}
	else {
		if (shp->sh_next)
			shp->sh_next->sh_prev = shp->sh_prev;
		shp->sh_prev->sh_next = shp->sh_next;
	}
	UNLOCK_STR_RESOURCE(s2);
	kmem_free(shp, sizeof(struct shinfo));
	LOCK_STR_RESOURCE(s2);
	Strinfo[DYN_STREAM].sd_cnt--;
	Strcount -= sizeof(struct shinfo);
	strst.stream.use--;
	UNLOCK_STR_RESOURCE(s2);
	splx(s);
	return;
}


/*
 * Allocate a message block.
 *
 * Should be protected by appropriate spl's from the calling code.
 */
struct msgb *
xmsgalloc(void)
{
	register struct msgb *mp;
	register int s2;

	if ((mp	= kmem_zalloc(sizeof(struct mbinfo), KM_NOSLEEP)) == NULL) {
		LOCK_STR_RESOURCE(s2);
		strst.msgblock.fail++;
		UNLOCK_STR_RESOURCE(s2);
		return NULL;
	}
	LOCK_STR_RESOURCE(s2);
	BUMPUP(strst.msgblock);
	Strinfo[DYN_MSGBLOCK].sd_cnt++;
	Strcount += sizeof(struct mbinfo);
	UNLOCK_STR_RESOURCE(s2);
	_INSERT_MSG_INUSE((struct mbinfo *)mp);
	return((mblk_t *)mp);
}

/*
 * Allocate a message/data/buffer triplet.
 *
 * Should be protected by appropriate spl's from the calling code.
 */
struct mdbblock *
xmdballoc(void)
{
	register struct mdbblock *mp;
	register int s;

	if ((mp	= kmem_zalloc(sizeof(struct mdbblock), KM_NOSLEEP)) == NULL) {
		LOCK_STR_RESOURCE(s);
		strst.mdbblock.fail++;
		UNLOCK_STR_RESOURCE(s);
		return NULL;
	}

	LOCK_STR_RESOURCE(s);
	BUMPUP(strst.mdbblock);
	Strinfo[DYN_MDBBLOCK].sd_cnt++;
	Strcount += sizeof(struct mdbblock);
	UNLOCK_STR_RESOURCE(s);
	_INSERT_MDB_INUSE(mp);
	return(mp);
}

/*
 * Allocate a pair of queues
 */
queue_t *
allocq(void)
{
	register int s;
	register int s2;
	register queue_t *qp;
	register struct queinfo *qip;

	s = splstr();
	/* allocate a queinfo struct (which contains two queues) */
	if ((qip = kmem_zalloc(sizeof(struct queinfo), KM_SLEEP)) == NULL) {
		LOCK_STR_RESOURCE(s2);
		strst.queue.fail++;
		UNLOCK_STR_RESOURCE(s2);
		splx(s);
		return(NULL);
	}
	qp = (queue_t *) qip;
	qp->q_qinfo = NULL;
	qp->q_first = NULL;
	qp->q_last = NULL;
	qp->q_next = NULL;
	qp->q_ptr = NULL;
	qp->q_count = 0;
	qp->q_flag = QUSE | QREADR;
	qp->q_minpsz = 0;
	qp->q_maxpsz = 0;
	qp->q_hiwat = 0;
	qp->q_lowat = 0;
	qp->q_link = NULL;
	qp->q_bandp = NULL;
	qp->q_nband = 0;
	WR(qp)->q_qinfo = NULL;
	WR(qp)->q_first = NULL;
	WR(qp)->q_last = NULL;
	WR(qp)->q_next = NULL;
	WR(qp)->q_ptr = NULL;
	WR(qp)->q_count = 0;
	WR(qp)->q_flag = QUSE;
	WR(qp)->q_minpsz = 0;
	WR(qp)->q_maxpsz = 0;
	WR(qp)->q_hiwat = 0;
	WR(qp)->q_lowat = 0;
	WR(qp)->q_link = NULL;
	WR(qp)->q_bandp = NULL;
	WR(qp)->q_nband = 0;
	/* for accounting purposes, count as 2 */
	LOCK_STR_RESOURCE(s2);
	BUMPUP(strst.queue);
	BUMPUP(strst.queue);
	Strinfo[DYN_QUEUE].sd_cnt += 2;
	Strcount += sizeof(struct queinfo);
	qip->qu_next = (struct queinfo *) Strinfo[DYN_QUEUE].sd_head;
	if (qip->qu_next)
		qip->qu_next->qu_prev = qip;
	qip->qu_prev = NULL;
	Strinfo[DYN_QUEUE].sd_head = (void *) qip;
	UNLOCK_STR_RESOURCE(s2);
	splx(s);
	return(qp);
}

/*
 * Free a pair of queues.
 */
void
freeq(register queue_t *qp)
{
	register int s;
	register int s2;
	register struct queinfo *qip;
	register qband_t *qbp;
	register qband_t *nqbp;

	s = splstr();
#ifdef DEBUG
	qp->q_flag &= ~QUSE;
	WR(qp)->q_flag &= ~QUSE;
#endif /* DEBUG */
	qbp = qp->q_bandp;
	while (qbp) {
		nqbp = qbp->qb_next;
		freeband(qbp);
		qbp = nqbp;
	}
	qbp = WR(qp)->q_bandp;
	while (qbp) {
		nqbp = qbp->qb_next;
		freeband(qbp);
		qbp = nqbp;
	}
	qip = (struct queinfo *) qp;
	LOCK_STR_RESOURCE(s2);
	if (qip->qu_prev == NULL) {
		if (qip->qu_next)
			qip->qu_next->qu_prev = NULL;
		Strinfo[DYN_QUEUE].sd_head = (void *) qip->qu_next;
	}
	else {
		if (qip->qu_next)
			qip->qu_next->qu_prev = qip->qu_prev;
		qip->qu_prev->qu_next = qip->qu_next;
	}
	Strcount -= sizeof(struct queinfo);
	/* for accounting purposes, count individually */
	strst.queue.use -= 2;
	Strinfo[DYN_QUEUE].sd_cnt -= 2;
	UNLOCK_STR_RESOURCE(s2);
	kmem_free(qip, sizeof(struct queinfo));
	splx(s);
	return;
}

/*
 * Allocate a qband structure.
 */
qband_t *
allocband(void)
{
	register int s;
	register int s2;
	register struct qbinfo *qbip;

	s = splstr();
	if ((qbip = kmem_zalloc(sizeof(struct qbinfo), KM_NOSLEEP)) == NULL) {
		splx(s);
		return(NULL);
	}
	LOCK_STR_RESOURCE(s2);
	Strinfo[DYN_QBAND].sd_cnt++;
	Strcount += sizeof(struct qbinfo);
	qbip->qbi_next = (struct qbinfo *) Strinfo[DYN_QBAND].sd_head;
	if (qbip->qbi_next)
		qbip->qbi_next->qbi_prev = qbip;
	qbip->qbi_prev = NULL;
	Strinfo[DYN_QBAND].sd_head = (void *) qbip;
	UNLOCK_STR_RESOURCE(s2);
	splx(s);
	return((qband_t *) qbip);
}

/*
 * Free a qband structure.
 */
static void
freeband(register qband_t *qbp)
{
	register int s;
	register int s2;
	register struct qbinfo *qbip;

	s = splstr();
	qbip = (struct qbinfo *) qbp;
	if (qbip->qbi_prev == NULL) {
		if (qbip->qbi_next)
			qbip->qbi_next->qbi_prev = NULL;
		LOCK_STR_RESOURCE(s2);
		Strinfo[DYN_QBAND].sd_head = (void *) qbip->qbi_next;
		UNLOCK_STR_RESOURCE(s2);
	}
	else {
		if (qbip->qbi_next)
			qbip->qbi_next->qbi_prev = qbip->qbi_prev;
		qbip->qbi_prev->qbi_next = qbip->qbi_next;
	}
	kmem_free(qbip, sizeof(struct qbinfo));
	LOCK_STR_RESOURCE(s2);
	Strcount -= sizeof(struct qbinfo);
	Strinfo[DYN_QBAND].sd_cnt--;
	UNLOCK_STR_RESOURCE(s2);
	splx(s);
	return;
}

/*
 * Allocate a stream event cell.
 */
struct strevent *
sealloc(int slpflag)
{
	register int s, s2;
	register struct seinfo *sep;

	s = splstr();
	LOCK_STR_RESOURCE(s2);
	if (sefreelist) {
		sep = sefreelist;
		sefreelist = sep->s_next;
		sep->s_strevent.se_procp = NULL;
		sep->s_strevent.se_events = 0;
		sep->s_strevent.se_next = NULL;
	}
	else {
		UNLOCK_STR_RESOURCE(s2);
		if (sep = kmem_zalloc(sizeof(struct seinfo),
				   (slpflag == SE_SLEEP) ?
					KM_SLEEP : KM_NOSLEEP)) {
			LOCK_STR_RESOURCE(s2);
			Strinfo[DYN_STREVENT].sd_cnt++;
			Strcount += sizeof(struct seinfo);
		} else {
			LOCK_STR_RESOURCE(s2);
			if (slpflag == SE_NOSLP) {
				/* use the cache for these */
				if (secachep) {
					sep = secachep;
					secachep = secachep->s_next;
					sep->s_strevent.se_procp = NULL;
					sep->s_strevent.se_events = 0;
					sep->s_strevent.se_next = NULL;
				}
			}
			if (sep == NULL) {
				/* failed anyhow */
				strst.strevent.fail++;
				UNLOCK_STR_RESOURCE(s2);
				splx(s);
				return(NULL);
			}
		}
	}
	BUMPUP(strst.strevent);
	sep->s_next = (struct seinfo *) Strinfo[DYN_STREVENT].sd_head;
	if (sep->s_next)
		sep->s_next->s_prev = sep;
	sep->s_prev = NULL;
	Strinfo[DYN_STREVENT].sd_head = (void *) sep;
	UNLOCK_STR_RESOURCE(s2);
	splx(s);
	return((struct strevent *) sep);
}

/*
 * Free a stream event cell.
 * Caller must hold resource lock.
 */
void
sefree(struct strevent *sep)
{
	register int s;
	register struct seinfo *seip;

	s = splstr();
	seip = (struct seinfo *) sep;
	if (seip->s_prev == NULL) {
		if (seip->s_next)
			seip->s_next->s_prev = NULL;
		Strinfo[DYN_STREVENT].sd_head = (void *) seip->s_next;
	}
	else {
		if (seip->s_next)
			seip->s_next->s_prev = seip->s_prev;
		seip->s_prev->s_next = seip->s_next;
	}
	if (seip >= &Secache[0] && seip <= &Secache[SECACHE-1]) {
		/* it's from the cache */
		seip->s_next = secachep;
		secachep = seip;
	}
	else {
		seip->s_next = sefreelist;
		sefreelist = seip;
	}
	strst.strevent.use--;
	splx(s);
	return;
}


/*
 * Check linked list of strhead write queues with held messages;
 * put downstream those that have been held long enough.
 *
 * Run the service procedures of each enabled queue
 *	-- must not be reentered
 *
 * Called by service mechanism (processor dependent) if there
 * are queues to run.  The mechanism is reset.
 */

/* XXXrs */
static int queuerun_gen = 0;

void
#ifdef MP_STREAMS
_queuerun(void)
#else
queuerun(void)
#endif /* MP_STREAMS */
{
	register queue_t *q;
	register int s, s2;
	register int count;
	register struct strevent *sep;
	register int nevent;
	register struct stdata *stp;
	mblk_t *mp;

	/* XXXrs - kludge */
	extern int strwsrv(queue_t *);
	/* XXXrs - kludge */

	s = splstr();

#ifdef MP_STREAMS
	qqrun = 0;
#endif

	qrunflag = 0;	/*
			 * Clear at top.  Nothing prevents it from getting
			 * set while queuerun runs.  Worst that could happen
			 * is that someone calls setqsched() between now and
			 * when we run the queue, and that should only result
			 * in an unnecessary pass through here.  Since here
			 * is the only place it gets cleared, and since we are
			 * about to run the queue, we should never miss a
			 * queue run.
			 */

	LOCK_STR_RESOURCE(s2);
        while ((stp = strholdhead) && (stp->sd_rtime <= lbolt)) {

		if ((strholdhead = stp->sd_hnext) == NULL)
			strholdtailp = &strholdhead;

		q = stp->sd_wrq;
		ASSERT(q->q_flag & QHLIST);
		q->q_flag &= ~QHLIST;

		if (mp = q->q_first)  {
			q->q_first = NULL;
			if (stp->sd_flag & (STRHUP|STWRERR))  {
				UNLOCK_STR_RESOURCE(s2);
				freemsg(mp);
				LOCK_STR_RESOURCE(s2);
			} else {
				UNLOCK_STR_RESOURCE(s2);
				splx(s);
				ASSERT(q->q_next);
				MP_STREAMS_INTERRUPT(q->q_monpp, putnext,
						     (void *)q, (void *)mp, 0);
				s = splstr();
				LOCK_STR_RESOURCE(s2);
			}
		}
	}
	do {
		if (strbcflag) {
			strbcwait = strbcflag = 0;

			/*
			 * count how many events are on the list
			 * now so we can check to avoid looping
			 * in low memory situations
			 */
			nevent = 0;
			for (sep = strbcalls.bc_head; sep; sep = sep->se_next)
				nevent++;
			/*
			 * get estimate of available memory from kmem_avail().
			 * awake all bufcall functions waiting for
			 * memory whose request could be satisfied 
			 * by 'count' memory and let 'em fight for it.
			 */
			count = kmem_avail();
			queuerun_gen++;
			while ((sep = strbcalls.bc_head) && nevent) {
				--nevent;
				if (sep->se_size <= count) {
				    strbcalls.bc_head = sep->se_next;
				    UNLOCK_STR_RESOURCE(s2);

				    MP_STREAMS_INTERRUPT(sep->se_monpp,
							 sep->se_func,
							 (void *)sep->se_arg,
							 0, 0);

				    LOCK_STR_RESOURCE(s2);
				    sefree(sep);
				} else {
				    /*
				     * too big, try again later - note that
				     * nevent was decremented above so we
				     * won't retry this one on this iteration
				     * of the loop
				     */
				    if (sep->se_next == NULL) {
					ASSERT(strbcalls.bc_tail == sep);
					strbcalls.bc_head = sep;
				    } else {
					ASSERT(strbcalls.bc_tail != sep);
					ASSERT(!strbcalls.bc_tail->se_next);
					strbcalls.bc_head = sep->se_next;
					sep->se_next = NULL;
					strbcalls.bc_tail->se_next = sep;
					strbcalls.bc_tail = sep;
				    }
				}
			}
			if (strbcalls.bc_head)
				/*
				 * still some bufcalls we couldn't do
				 * let kmem_free know
				 */
				strbcwait = 1;
			else
				strbcalls.bc_tail = NULL;
		}

		while (q = qhead) {
			/*
			 * XXXrs - we aren't holding the queue's monitor
			 *         here so it's not safe to fiddle with
			 *         the flags.  To make things safe means
			 *	   holding some common lock.  Most other
			 *	   q_flag manipulations are done under the
			 *	   monitor, so perhaps it would be possible
			 *	   to just attempt to grab the monitor
			 *	   semaphore spinlock???  Either that, or all
			 *	   other q_flag manipulations will need to
			 *	   be under the resource lock.  Ugh.
			 *	   Could make special cases of the QENAB and
			 *	   QBACK flags and move them to their own
			 *	   flags fields?  That probably violates some
			 *	   stinking ABI?
			 */
			ASSERT(q->q_pflag & QENAB);
			ASSERT(0 != q->q_ptr);
			if (!(qhead = q->q_link))
				qtailp = &qhead;
			if (q->q_qinfo->qi_srvp) {
				UNLOCK_STR_RESOURCE(s2);
				splx(s);

#ifdef KPERF
				if(kpftraceflg)
					kperf_write(KPT_INTR,
					    q->q_qinfo->qi_srvp, curproc);
#endif /* KPERF */

				STREAMS_ENAB_INTERRUPT(q->q_monpp,
						       q->q_qinfo->qi_srvp,
						       q, 0, 0);

#ifdef KPERF
				if(kpftraceflg)
					kperf_write(KPT_INT_KRET,
					    q->q_qinfo->qi_srvp, curproc);
#endif /* KPERF */
				s = splstr();
				LOCK_STR_RESOURCE(s2);
				/*
				 * XXXrs - QBACK/QB_BACK are now cleared
				 *         as part of the new
				 *	   STREAMS_ENAB_INTERRUPT mechanism.
				 *         See the comments in
				 *         streams_service().
				 */
			}
		}
	} while (strbcflag);

        if (strholdhead && !strholdflag)  {
                strholdflag++;
                MP_STREAMS_TIMEOUT(&streams_monp, strrelease, 0, STRHOLDTIME);
        }

	UNLOCK_STR_RESOURCE(s2);

	splx(s);
}

#ifdef _EAGER_RUNQUEUES
/*
 * Function to kick off queue scheduling for those system calls
 * that cause queues to be enabled (read, recv, write, send, ioctl).
 */
void
runqueues(void)
{
	register int s, s2;

	s = splstr();
	LOCK_STR_RESOURCE(s2);
	if (qrunflag && !queueflag) {
		queueflag = 1;
		UNLOCK_STR_RESOURCE(s2);
		queuerun();
		LOCK_STR_RESOURCE(s2);
		queueflag = 0;
		UNLOCK_STR_RESOURCE(s2);
		splx(s);
		return;
	}
	UNLOCK_STR_RESOURCE(s2);
	splx(s);
}
#else
/*
 * Otherwise, this function is handled by the macro setqsched().
 * See the definition in sys/strsubr.h
 */
#endif /* _EAGER_RUNQUEUES */

/* 
 * Find module
 * 
 * return index into fmodsw
 * or -1 if not found
 */
int
findmod(register char *name)
{
	register int i, j;

	for (i = 0; i < fmodmax; i++) {
		if (fmhold(i))		/* invalid fmodsw index */
			continue;
		for (j = 0; j < FMNAMESZ + 1; j++) {
			if (fmodsw[i].f_name[j] != name[j]) 
				break;
			if (name[j] == '\0') {
				fmrele(i);
				return(i);
			}
		}
		fmrele(i);
	}
	return(-1);
}

/* 
 * Same as findmod() but holds a reference to the module so it can't
 * be unloaded.  Reference can be given up later by calling fmrele().
 */
int
fmfind(register char *name)
{
	register int i, j;

	for (i = 0; i < fmodmax; i++) {
		if (fmhold(i))		/* invalid fmodsw index */
			continue;
		for (j = 0; j < FMNAMESZ + 1; j++) {
			if (fmodsw[i].f_name[j] != name[j]) 
				break;
			if (name[j] == '\0')
				return(i);
		}
		fmrele(i);
	}
	return(-1);
}


/*
 * Set the QBACK or QB_BACK flag in the given queue for
 * the given priority band.
 */
void
setqback(register queue_t *q, register unsigned char pri)
{
	register int i;
	qband_t *qbp;
	qband_t **qbpp;

	if (pri != 0) {
		if (pri > q->q_nband) {
			qbpp = &q->q_bandp;
			while (*qbpp)
				qbpp = &(*qbpp)->qb_next;
			while (pri > q->q_nband) {
				if ((*qbpp = allocband()) == NULL) {
					cmn_err(CE_WARN,
					    "setqback: can't allocate qband\n");
					return;
				}
				(*qbpp)->qb_hiwat = q->q_hiwat;
				(*qbpp)->qb_lowat = q->q_lowat;
				q->q_nband++;
				qbpp = &(*qbpp)->qb_next;
			}
		}
		qbp = q->q_bandp;
		i = pri;
		while (--i)
			qbp = qbp->qb_next;

		qbp->qb_flag |= QB_BACK;
	} else {
		q->q_flag |= QBACK;
	}
}

/* ARGSUSED */
int
strcopyin(caddr_t from, caddr_t to, unsigned int len, char *fmt, int copyflag)
{
	if (copyflag == U_TO_K) {
		if (ncopyin(from, to, len, fmt))
			return (EFAULT);
	} else {
		ASSERT(copyflag == K_TO_K);
		bcopy(from, to, len);
	}
	return (0);
}

/* ARGSUSED */
int
strcopyout(
	caddr_t from,
	caddr_t to,
	unsigned int len,
	char *fmt,
	int copyflag)
{
	if (copyflag == U_TO_K) {
		if (ncopyout(from, to, len, fmt))
			return (EFAULT);
	} else {
		ASSERT(copyflag == K_TO_K);
		bcopy(from, to, len);
	}
	return (0);
}

void
strsignal(struct stdata *stp, int sig, long band)
{
	register int s;

	switch (sig) {
	case SIGPOLL:				
		s = splstr();
		if (stp->sd_sigflags & S_MSG) 
			strsendsig(stp->sd_siglist, S_MSG, band);
		splx(s);
		break;

	default:
		if (stp->sd_pgrp)
			signal(stp->sd_pgrp, sig);
		break;
	}
}

void
strhup(struct stdata *stp)
{
	int s;

	s = splstr();
	if (stp->sd_sigflags & S_HANGUP) 
		strsendsig(stp->sd_siglist, S_HANGUP, 0L);
	pollwakeup(&stp->sd_pollist, POLLHUP);
	splx(s);
}

/*
 * Called by module open routines that want to become the
 * controlling tty of the process.
 */
void
strctltty(queue_t *rq)
{

	register struct stdata *stp;

	/*
	 * Make sure it's a read queue.
	 */
	if (!(rq->q_flag & QREADR))
		rq = RD(rq);

	/*
	 * Chain up to the stream head.
	 */
	while (rq->q_next != NULL)
		rq = rq->q_next;

	stp = (struct stdata *) rq->q_ptr;
	ASSERT(stp);

	stp->sd_flag |= STRISTTY;
}

void
stralloctty(
        register proc_t *pp,
	register struct vnode *vp,
        register struct stdata *stp)
{
	if (!(stp->sd_flag & STRISTTY) ||  /* Have we requested to be ctty? */
	    vp->v_type != VCHR)		   /* No pipes allowed! */
		return;
	/*
	 * If the caller is a session leader, doesn't have
	 * a controlling tty already and the terminal isn't
	 * already a controlling tty for some other process,
	 * then the current process becomes the terminal process
	 * group leader.
	 */
	/*
	 * Get semaphore pglobal to protect sid
	 */
	mutex_lock(&pglobal, PZERO);
	ASSERT(pp->p_session);
	if (pp->p_session->se_sid == pp->p_pid &&
	    pp->p_session->se_vp == NULL &&
	    stp->sd_pgrp == 0) {
		alloctty(pp, vp, &stp->sd_pgrp,
			 &stp->sd_sid, &stp->sd_session);

	}
	mutex_unlock(&pglobal);
}


/*
 * Unlink "all" persistant links.
 * Currently called only from spec_sync on final system shutdown
 */
void
strpunlink(struct cred *crp)
{
	struct shinfo  *shp;
	int rval;
	int s;

	/*
	 * for each allocated stream head, call munlinkall()
	 * with flag of LINKPERSIST to unlink any/all persistant
	 * links for the device.
	 */

	STREAMS_LOCK();
	LOCK_STR_RESOURCE(s);
	shp = (struct shinfo *) Strinfo[DYN_STREAM].sd_head;
	UNLOCK_STR_RESOURCE(s);
	while(shp)
	{
	    (void)munlinkall((struct stdata **)&(Strinfo[DYN_STREAM].sd_head),
			     LINKIOCTL|LINKPERSIST, crp, &rval);
	    LOCK_STR_RESOURCE(s);
	    shp = shp->sh_next;
	    UNLOCK_STR_RESOURCE(s);
	}
	STREAMS_UNLOCK();
}
