/*
 * |-----------------------------------------------------------|
 * | Copyright (c) 1991, 1990 MIPS Computer Systems, Inc.      |
 * | All Rights Reserved                                       |
 * |-----------------------------------------------------------|
 * |          Restricted Rights Legend                         |
 * | Use, duplication, or disclosure by the Government is      |
 * | subject to restrictions as set forth in                   |
 * | subparagraph (c)(1)(ii) of the Rights in Technical        |
 * | Data and Computer Software Clause of DFARS 252.227-7013.  |
 * |         MIPS Computer Systems, Inc.                       |
 * |         950 DeGuigne Avenue                               |
 * |         Sunnyvale, California 94088-3650, USA             |
 * |-----------------------------------------------------------|
 */
#ident "$Revision: 1.69 $"

/*
 *	tpisocket
 *		provide TPI interface above sockets-based network
 *		protocol stacks
 *
 *	tpisocket is a library module used by protocol-specific
 *	streams pseudo-drivers, such as tpitcp, tpiudp, and so on.
 *	Almost all of the functionality of such pseudo-drivers
 *	is present in tpisocket, with only protocol-specific data
 *	conversions and the actual streams module descriptions
 *	defined outside tpisocket.
 */

#undef _KERNEL_SVR4
#define _KERNEL_SVR4

#ifndef COMPAT_43
#define COMPAT_43 1
#endif

#include "tpisocket.h"
#include "sys/user.h"
#include "sys/sema.h"
#include "bsd/sys/hashing.h"

#include "sys/callo.h"
#include "sys/conf.h"
#include "sys/sysmacros.h"
#include "sys/domain.h"
#include "sys/types.h"
#include "sys/socket.h"
#include "bsd/net/if.h"
#include "bsd/netinet/in.h"
#include "bsd/netinet/in_var.h"

#include "bsd/net/soioctl.h"
/* #include "sys/svr4_ioctl.h" */
#include <sys/cred.h>
#include <sys/ddi.h>	/* for geteminor(), etc. */
#include <sys/major.h>
#include <sys/systm.h>
#include <sys/sesmgr.h>
#include <sys/atomic_ops.h>	/* atomic ops prototypes */

#ifdef _MP_NETLOCKS
#include <sys/cmn_err.h>
#endif /*_MP_NETLOCKS */

#include <sys/strsubr.h>
#include "net/net64.h"
#include <sys/sat.h>			/* XXX gl */

#include <sys/so_dac.h>
#ifndef IOCGROUP
#define IOCGROUP(x) (((x) >> 8) & 0xFF)
#endif /* IOCGROUP */

#define TSO_HOLD(x) (atomicAddUlong(&(x)->tso_count,1))
#define TSO_RELE(x) {if ((x)->tso_count == 1) \
				tpisocket_dismantle((x)); \
			    else \
				atomicAddUlong(&(x)->tso_count, -1); }

/* external declarations */

#ifdef TPISOCKET_DAEMON
extern struct callout *kernel_process_init_list;
#endif

extern int queueflag;
extern struct fileops vnodefops;
extern struct streamtab clninfo;

extern struct ifnet *ifunit(register char *);
extern struct vnode *makespecvp();
extern int soclose(struct socket *);
extern struct mbuf *mblk_to_mbuf( mblk_t *, int, int, int);
extern mblk_t *mbuf_to_mblk( struct mbuf *, int);

/* forward declarations */

void	tpisocket_conn_req_done( struct tpisocket *);
void	tpisocket_convert_error( int *, int *);
int	tpisocket_convert_socket( struct file *, struct cred *);
void	tpisocket_copy_to_mblk_chain( caddr_t, int, mblk_t **, int *);
struct mbuf *tpisocket_copy_to_mbuf( void *, int, int, int*);
void	tpisocket_daemon(void);
void	tpisocket_discon_req_done( struct tpisocket *);
static void	tpisocket_discard(void);
static void	tpisocket_dismantle( struct tpisocket *);
struct mbuf *tpisocket_dup_mbuf( struct mbuf *, int, queue_t *, int);
int	tpisocket_event( struct socket *, int, struct tpisocket *);
int	tpihold(void *, void **, void *);
int	tpisocket_ifconf(  struct tpisocket *, mblk_t *);
int	tpisocket_ifiocdata( struct tpisocket *, mblk_t *);
int	tpisocket_ifioctl( struct tpisocket *, mblk_t *);
void	tpisocket_init(void);
struct mbuf *tpisocket_mblk_to_mbuf( mblk_t *, int, int, int, queue_t *);
void	tpisocket_mbufcall_wakeup( struct tpisocket *);
int	tpisocket_open( queue_t *, dev_t *, int, int, cred_t *,
			struct tpisocket_control *);
void	tpisocket_optmgmt_req_done( struct tpisocket *);
void	tpisocket_release_recv_resources( struct tpisocket *);
void	tpisocket_release_send_resources( struct tpisocket *);
void	tpisocket_request_mbufcall( queue_t *);
int	tpisocket_rtiocdata( struct tpisocket *, mblk_t *);
int	tpisocket_rtioctl( struct tpisocket *, mblk_t *);
void	tpisocket_scan_conn_q( struct tpisocket *, struct socket *);
int	tpisocket_sioc_name( struct tpisocket *, mblk_t *, int);
mblk_t *tpisocket_allocb( queue_t *, int, int);
mblk_t *tpisocket_mbuf_to_mblk( queue_t *, struct mbuf *, int);
int	tpisocket_soreceive( struct socket *, struct mbuf **, struct uio *,
			struct mbuf **, struct mbuf **, int *);
int	tpisocket_sosend( struct socket *, struct mbuf *, struct uio *,
			struct mbuf *, struct mbuf *, int);
int	tpisocket_ti_name( struct tpisocket *, mblk_t *);
struct mbuf *tpisocket_trim_mbuf_chain( struct mbuf *, int);
int	tpisocket_wact( struct tpisocket *, mblk_t *);
int	tpisocket_wreject( struct tpisocket *, mblk_t *, int, int, int);

static int tpisocket_send_check( struct tpisocket *);
static int tpisocket_send_iocack( queue_t *, mblk_t *, mblk_t *);
static int tpisocket_send_iocnak( queue_t *, mblk_t *, int);
static int tpisocket_send_ok_message( struct tpisocket *,
			struct socket *, mblk_t *);

static int _tpisocket_rsrv(queue_t *, void *, struct tpisocket *);
static int _tpisocket_wput(queue_t *, mblk_t *, struct tpisocket *);
static int _tpisocket_wsrv(queue_t *, void *, struct tpisocket *);

/* Global variables */

int	tpisocket_clone_major = -1;
int	tpisocket_initialized = 0;
struct callout tpisocket_process_init;
static struct socket *tpisocket_discard_list = NULL;
struct tpisocket_control *tpisocket_protocols = NULL;

static int      opt_off = 0;
static struct linger lingerdef = {0, 0};
static struct optdefault sockdefault[] = {
	SO_LINGER,	(char *)&lingerdef,	sizeof(struct linger),
	SO_DEBUG,	(char *)&opt_off,	sizeof(int),
	SO_KEEPALIVE,	(char *)&opt_off,	sizeof(int),
	SO_DONTROUTE,	(char *)&opt_off,	sizeof(int),
	SO_USELOOPBACK,	(char *)&opt_off,	sizeof(int),
	SO_BROADCAST,	(char *)&opt_off,	sizeof(int),
	SO_REUSEADDR,	(char *)&opt_off,	sizeof(int),
	SO_OOBINLINE,	(char *)&opt_off,	sizeof(int),
	SO_IMASOCKET,	(char *)&opt_off,	sizeof(int),
	0, 		(char *)0, 		0
};

struct tsockd_entry {
        struct tsockd_entry *next;
        void    (*func)(void *, void *, void *);
        void    *arg1, *arg2, *arg3;
};

/*
 *	tpisocket_init
 *
 *	Called via the io_init[] list at system startup time.
 *	Adds an entry to the list of kernel processes to be started.
 */
#ifdef _MP_NETLOCKS
static lock_t tpisockd_lock;
static lock_t tpisocket_discard_lock;
static sv_t tpisockd_wait, tpisockd_sync, tpisockd_inuse;
static struct tsockd_entry *tpisockd_head, *tpisockd_tail, *tpisockd_free;
static int tpisockd_running;
static zone_t *tpisockd_zone;
static void tpisockd_call(void (*)(), void *, void *, void *);
static void    tpisockd(void *);
#define TPISOCKD_FREEMAX	20
int tpisockd_freecnt = TPISOCKD_FREEMAX;
int tpisockd_freemax = TPISOCKD_FREEMAX;

/*
 * tpisockd_call()
 */
static void
tpisockd_call(void (*fun)(), void *arg1, void *arg2, void *arg3)
{
	int s;
        struct tsockd_entry *entry = kmem_zone_alloc(tpisockd_zone, KM_NOSLEEP);

	s = mp_mutex_spinlock(&tpisockd_lock);
	if (!entry) {
		entry = tpisockd_free;
		if (!entry)
                	cmn_err(CE_PANIC, "tpisockd_call: could not allocate sockd entry");
		else {
			tpisockd_free = entry->next;
			tpisockd_freecnt--;
		}
	}

        entry->func = fun;
        entry->arg1 = arg1;
        entry->arg2 = arg2;
	entry->arg3 = arg3;
        entry->next = 0;
        if (!tpisockd_tail) {
                ASSERT(!tpisockd_head);
                tpisockd_head = tpisockd_tail = entry;
        }
        else {
                tpisockd_tail->next = entry;
                tpisockd_tail = entry;
        }
        sv_signal(&tpisockd_wait);
        mp_mutex_spinunlock(&tpisockd_lock, s);
}

void
tpisockd_init(void)
{
	register int i;

	tpisockd_zone = 
		kmem_zone_init(sizeof(struct tsockd_entry), "tpisockd entry");
        spinlock_init(&tpisockd_lock, "tpisockd_lock");
	spinlock_init(&tpisocket_discard_lock, "tpisocket_discard_lock");
        sv_init(&tpisockd_wait, SV_LIFO, "tpisockd_wait");
        sv_init(&tpisockd_sync, SV_DEFAULT, "tpisockd_sync");
	sv_init(&tpisockd_inuse, SV_DEFAULT, "tpisockd_inuse");
	/* create back up free list */
	for (i=tpisockd_freemax; i; i--) {
        	struct tsockd_entry *entry;

		entry = kmem_zone_alloc(tpisockd_zone, KM_NOSLEEP);
		entry->next = tpisockd_free;
		tpisockd_free = entry;
	}
	kthread_create("tpisockd", 0, 8192, (void (*)(void *))tpisockd,
			(void *)(__psint_t)39, 0);
}

/*
 * tpisockd()
 */
static void
tpisockd(void *pri)
{
	int s;
        struct tsockd_entry *entry;

	curprocp->p_pri = curprocp->p_rtpri = (__psint_t)pri;

        STREAMS_LOCK();
	s = mp_mutex_spinlock(&tpisockd_lock);
        tpisockd_running++;
again:
        if (!(entry = tpisockd_head)) {
                tpisockd_running--;
                if (tpisockd_running == 0)
                        sv_signal(&tpisockd_sync);
                mp_sv_wait(&tpisockd_wait, PZERO-1, &tpisockd_lock, s);
		s = mp_mutex_spinlock(&tpisockd_lock);
                tpisockd_running++;
                goto again;
        }
        tpisockd_head = entry->next;
        if (tpisockd_head == 0) {
                ASSERT(entry == tpisockd_tail);
                tpisockd_tail = 0;
        }
	mp_mutex_spinunlock(&tpisockd_lock, s);
        entry->func(entry->arg1, entry->arg2, entry->arg3);
	s = mp_mutex_spinlock(&tpisockd_lock);
        if (tpisockd_freecnt < tpisockd_freemax) {
                entry->next = tpisockd_free;
                tpisockd_free = entry;
                tpisockd_freecnt++;
        }
        else
        	kmem_zone_free(tpisockd_zone, entry);
        goto again;
}

#endif /* _MP_NETLOCKS */

void
tpisocket_init(void)
{

	if (! tpisocket_initialized) {
		tpisocket_initialized = 1;
		tpisocket_clone_major = CLN_MAJOR;
#ifdef TPISOCKET_DAEMON
		bzero((caddr_t) &tpisocket_process_init,
		      sizeof(tpisocket_process_init));
		tpisocket_process_init.c_func = tpisocket_daemon;
		tpisocket_process_init.c_next = kernel_process_init_list;
		kernel_process_init_list = &tpisocket_process_init;
#endif

	}
}


/*
 *	tpisocket_register
 *
 *	Register a protocol, for use in converting sockets to TPI sockets
 */
int
tpisocket_register(struct tpisocket_control *tpictl)
{
	struct	tpisocket_control *ti;

	for (ti = tpisocket_protocols; ti != NULL; ti = ti->tsc_next)
		if (ti == tpictl)
			return 0;
	tpictl->tsc_next = tpisocket_protocols;
	tpisocket_protocols = tpictl;
	return 1;
}

/*
 *	tpisocket_discard
 *
 *		Discard sockets discarded via the TPI accept procedure.
 *		Since soclose sleeps, this cannot occur in a service routine.
 */
static void
tpisocket_discard(void)
{
	struct	socket *so;
	int s;

	s = mp_mutex_spinlock(&tpisocket_discard_lock);
	while (tpisocket_discard_list != NULL) {
		so = tpisocket_discard_list;
		tpisocket_discard_list = (struct socket *)so->so_callout_arg;
		mp_mutex_spinunlock(&tpisocket_discard_lock, s);
		SOCKET_LOCK(so, s);
		so->so_state &= ~(SS_TPISOCKET | SS_WANT_CALL);
		SOCKET_UNLOCK(so, s);
		soclose(so);
		s = mp_mutex_spinlock(&tpisocket_discard_lock);
	}
	mp_mutex_spinunlock(&tpisocket_discard_lock, s);
}

#ifdef TPISOCKET_DAEMON
/* 
 *	tpisocket_daemon
 *
 *	Started at system startup time as a kernel daemon.
 *	Lives only to discard sockets discarded via the TPI accept
 *	procedure.
 */
void
tpisocket_daemon(void)
{
	struct	socket *so;

	strcpy(u.u_comm,"tpisocketd");
	strcpy(u.u_psargs,"tpisocketd");
	splnet();
	splstr(); /* whichever is higher */
	while (1) {
		tpisocket_discard();
		sleep(&tpisocket_discard_list, PRIBIO);
	}
}
#endif

/*
 *	tpisocket_open
 *
 *	Called to do the stream open
 */
int
tpisocket_open(
	register queue_t *rq,
	dev_t	*devp,
	int	flag,
	int	sflag,
	cred_t	*crp,
	struct	tpisocket_control *tpictl)
{
	struct tpisocket *tpiso;
	int	s;
	int	d;

#define FAILOPEN(x) { /* curthreadp->k_error = (x); */ splx(s); return (x); }

	s = splstr();
	if (TSC_MAXDEVICES == 0) {
		FAILOPEN(ENODEV);
	}
	if (TSC_DEVICES == NULL) {
		struct tpisocket **tsod;

#define TSODSIZE (sizeof(struct tpisocket *) * TSC_MAXDEVICES)
		tsod = (struct tpisocket **)kmem_zalloc(TSODSIZE, KM_SLEEP);
		ASSERT(tsod != NULL);
		if (TSC_DEVICES == NULL) {
			TSC_DEVICES = tsod;
		} else {
			kmem_free(tsod, TSODSIZE);
		}
	}
#undef TSODSIZE

	if (sflag) {			/* do 'clone' open */
		for (d = 0; d < TSC_MAXDEVICES && NULL != TSC_DEVICES[d]; d++)
			continue;
	} else {
		d = geteminor(*devp);
	}
	if (d >= TSC_MAXDEVICES) {
		FAILOPEN(ENODEV);
	}

#define TSO_WAITING ((struct tpisocket *)((__psunsigned_t)1))
again:
	if (TSC_DEVICES[d] == NULL || TSC_DEVICES[d] == TSO_WAITING) {
		TSC_DEVICES[d] = TSO_WAITING;
		tpiso = (struct tpisocket *)kmem_zalloc(sizeof *tpiso, 
				KM_SLEEP);
		ASSERT(tpiso != NULL);
		if (TSC_DEVICES[d] == TSO_WAITING) {
			tpiso->tso_count = 1;
			tpiso->tso_index = d;
			tpiso->tso_control = tpictl;
			TSC_DEVICES[d] = tpiso;
		} else {
			kmem_free(tpiso, sizeof *tpiso);
		}
	}

	tpiso = TSC_DEVICES[d];
	if (tpiso->tso_flags & TSOF_EXCLUSIVE && drv_priv(crp) != 0) {
		FAILOPEN(EBUSY);
	}
	if (tpiso->tso_flags & TSOF_OPENING) {
		if (flag & (FNDELAY | FNONBLOCK)) {
			FAILOPEN(EAGAIN);
		}
		tpiso->tso_flags |= TSOF_OPEN_WAIT;
		TSO_HOLD(tpiso);
		if (sleep((caddr_t)tpiso, STOPRI | PCATCH)) {
#ifdef _MP_NETLOCKS
			/* wake up tpisocket_close if waiting */
			if ((tpiso->tso_count == 2) &&
			    (tpiso->tso_flags & TSOF_CLOSING)) {
				sv_broadcast(&tpisockd_inuse);
			}
#endif
			TSO_RELE(tpiso);
			FAILOPEN(EINTR);
		}
		TSO_RELE(tpiso);
		goto again;
	}

	if (tpiso->tso_flags & TSOF_CLOSING) {
		FAILOPEN(EBUSY);
	}

	if (tpiso->tso_flags & TSOF_ISOPEN) {
		if (tpiso->tso_rq != rq) {
			FAILOPEN(EBUSY);
		}
	} else {
		int err;

		rq->q_ptr = (caddr_t)tpiso; /* connect new device to stream */
		WR(rq)->q_ptr = (caddr_t)tpiso;
		tpiso->tso_rq = rq;
		tpiso->tso_wq = WR(rq);
		tpiso->tso_flags |= TSOF_OPENING;
		if (sflag)
			tpiso->tso_flags |= TSOF_CLONE_OPEN;

		ASSERT(!tpiso->tso_socket);
		/* allocate a socket structure */
		err = socreate(TSC_DOMAIN, &tpiso->tso_socket,
			       TSC_SOCKET_TYPE, TSC_PROTOCOL);
		if (! err) {
			SOCKET_LOCK(tpiso->tso_socket, s);
			tpiso->tso_socket->so_callout_arg = (caddr_t)tpiso;
			tpiso->tso_socket->so_callout = tpisocket_event;
			tpiso->tso_socket->so_state |=
				(SS_TPISOCKET | SS_NBIO | SS_WANT_CALL);
			tpiso->tso_socket->so_snd.sb_flags |= SB_NOTIFY;
			SOCKET_UNLOCK(tpiso->tso_socket, s);
		}
		tpiso->tso_flags &= ~TSOF_OPENING;
		if (tpiso->tso_flags & TSOF_OPEN_WAIT) {
			tpiso->tso_flags &= ~TSOF_OPEN_WAIT;
			wakeup((caddr_t)tpiso);
		}
		if (err) {
close_on_error:
			tpiso->tso_flags |= TSOF_CLOSING;
			TSO_RELE(tpiso);
#ifdef _MP_NETLOCKS
			/* wake up tpisocket_close if waiting */
			if ((tpiso->tso_count == 1) &&
			    (tpiso->tso_flags & TSOF_CLOSING)) {
				sv_broadcast(&tpisockd_inuse);
			}
#endif
			FAILOPEN(err);
		}
		tpiso->tso_flags |= (TSOF_ISOPEN | TSOF_UNUSED);
		tpiso->tso_cred = crp;
		crhold(crp);
#ifdef sgi /* Audit */
#if 0
		/*
		 * XXX:SAT_macro this - casey
		 */
		if (u.u_sat_procp) {
			tpiso->tso_fd = u.u_sat_procp->sat_openfd;
			tpiso->tso_procp = curprocp;
		}
#endif
#endif
		MGET(tpiso->tso_temporary_mbuf, M_WAIT, MT_DATA);
		err = TSC_OPEN(tpiso);
		if (err)
			goto close_on_error;
	}

	if (sflag) {			/* do 'clone' open */
		*devp = makedevice(getemajor(*devp), d);
	}
	splx(s);

	/* Create audit record */
#ifdef sgi /* Audit */
#if 0
	_SAT_BSDIPC_CREATE(u.u_sat_procp->sat_openfd, tpiso->tso_socket,
		 TSC_DOMAIN, TSC_PROTOCOL, 0);
#endif
#endif
	return 0;
}


/*
 *	deallocate a tpisocket structure
 */
static void
tpisocket_dismantle( struct tpisocket *tpiso)
{
	struct tpisocket_control *tpictl = tpiso->tso_control;
	struct	socket *so;
	NETSPL_SPDECL(s)

#define RELEASE_MBUF(field) { \
	if (tpiso->field) { \
		m_freem(tpiso->field); \
		tpiso->field = NULL; \
	} }
#define RELEASE_MBLK(field) { \
	if (tpiso->field != NULL) { \
		freemsg(tpiso->field); \
		tpiso->field = NULL; \
	} }

	ASSERT((!(tpiso->tso_flags & (TSOF_OPEN_WAIT))));

	TSC_DEVICES[tpiso->tso_index] = NULL;
	if (tpiso->tso_flags & TSOF_ANY_MBUFCALL) {
		tpiso->tso_flags &= ~TSOF_ANY_MBUFCALL;
		untimeout(tpiso->tso_mbufcall_timeout);
	}
	so = tpiso->tso_socket;
	if (so != NULL) {
#ifdef _MP_NETLOCKS
		mon_state_t ms;
		proc_t *pp;
		mon_t *hadmonitor = NULL;
#endif
		tpiso->tso_socket = NULL;
		SOCKET_LOCK(so, s);
		/* only should get here if we are closing */
		ASSERT(tpiso->tso_flags & TSOF_CLOSING);
		so->so_state &= ~(SS_TPISOCKET | SS_WANT_CALL);
		so->so_callout = NULL;
		SOCKET_UNLOCK(so, s);

#ifdef _MP_NETLOCKS
		/* soclose can call qswtch so we free the monitor here */
		pp = private.p_curproc;
		ASSERT(pp->p_kthread.k_monitor);
		/*
		 * soclose may have released monitor. In this case *monpp
		 * is 0. This would cause amonitor to panic.
		 */
		if (hadmonitor = pp->p_kthread.k_monitor) {
			rmonitor(pp->p_kthread.k_monitor, &ms);
		}
#endif

		soclose(so);

#ifdef _MP_NETLOCKS
		/* recover monitor here */
		if (hadmonitor)
			amonitor(&ms);
#endif
	}
	if (tpiso->tso_cred) {
		crfree(tpiso->tso_cred);
		tpiso->tso_cred = NULL;
	}
	tpisocket_release_send_resources(tpiso);
	tpisocket_release_recv_resources(tpiso);
	RELEASE_MBLK(tso_pending_ioctl_req);
	RELEASE_MBLK(tso_pending_request);
	RELEASE_MBUF(tso_temporary_mbuf);
	RELEASE_MBLK(tso_pending_optmgmt_req);
	RELEASE_MBUF(tso_pending_optmgmt_mbuf);
	kmem_free(tpiso, sizeof *tpiso);

#undef RELEASE_MBLK
#undef RELEASE_MBUF
}

/*	tpisocket_close
 *		close a tpisocket connection
 */
/*ARGSUSED*/
int		/* this should be void, but USL's prototypes are wrong */
tpisocket_close(
	queue_t	*rq,
	int	flag,
	cred_t *crp)
{
	struct	tpisocket *tpiso = (struct tpisocket *)rq->q_ptr;
	struct	tpisocket_control *tpictl;
	int	s;
	struct  socket *so;
	NETSPL_SPDECL(ss)

	if (tpiso == NULL)
		return 0;
	tpictl = tpiso->tso_control;

	s = splstr();

	/*
	 * ensure no more entries placed on tpisockd queue.
	 */
	ASSERT(!(tpiso->tso_flags & TSOF_OPENING));
	so = tpiso->tso_socket;
	if (so != NULL)
		SOCKET_LOCK(so, ss);
	tpiso->tso_flags |= TSOF_CLOSING;
	if (so != NULL)
		SOCKET_UNLOCK(so, ss);

#ifdef _MP_NETLOCKS

	/*
	 * while in use, suspend tpisocket_close (may need mutex lock).
	 */
	while (tpiso->tso_count > 1) {
		sv_wait(&tpisockd_inuse, PZERO-1, NULL, 0);
	}
#endif

	TSC_CLOSE(tpiso);

	str_unbcall(rq);		/* stop waiting for buffers */

#if 0
	_SAT_BSDIPC_SHUTDOWN(u.u_sat_procp->sat_openfd,tpiso->tso_socket,
		flag,0);
#endif
	TSO_RELE(tpiso);
#ifndef TPISOCKET_DAEMON
	tpisocket_discard();	/* HACK - should be asynchronous process */
#endif
	splx(s);
	return(0);
}


/*
 *	tpisocket_event
 *		Called by socket stack when it becomes possible to send to
 *		or receive from the socket (when formerly it was not).  
 *		Enables call on tpisocket_wsrv() or tpisocket_rsrv(), 
 *		which will check for queued output or input packets and 
 *		transfer them between the streams queue and the socket.
 *
 *		Entered at splnet().
 */
int
tpisocket_event(
	struct socket *so,
	int	event,
	struct tpisocket *tpiso)
{
	int	s;
	int	flag;

	flag = event & 0x10;
	event &= ~0x10;

	if (tpiso->tso_socket != so)
		goto done;

	/*
	 * If we are closing then nothing to do. The queue may have already
	 * been dismantled by a strclose since we may have slept waiting
	 * for the STREAM monitor to become free.
	 */
	if (flag && (tpiso->tso_flags & TSOF_CLOSING)) {
		goto done;
	}

	s = splstr();

	switch (event) {
	case SSCO_OUTPUT_AVAILABLE:
		if (!(tpiso->tso_flags & TSOF_WSRV_QUEUED) &&
		    (tpiso->tso_flags & TSOF_ISOPEN) &&
		    !(tpiso->tso_flags & TSOF_CLOSING)) {
			tpiso->tso_flags |= TSOF_WSRV_QUEUED;
#ifdef notyet
			/*
			 * XXXrs - Not ever if you ask me.  This code is
			 * trying to "outsmart" the streams scheduling
			 * mechanism.  There may be some additional
			 * MP considerations w.r.t. the queuerun flag and
			 * blocking queueruns.  Think carefully before
			 * trying to turn something like this on.
			 *
			 * See the queuerun() code.
			 */
			if (! queueflag) {
				queuerun_block();
				splx(s);
				tpisocket_wsrv(tpiso->tso_wq);
				queuerun_unblock();
				return;
			} else
#endif /* notyet */
				qenable(tpiso->tso_wq);
		}
		break;

	case SSCO_OOB_INPUT_ARRIVED:
		if (tpiso->tso_flags & TSOF_OOB_INPUT_AVAILABLE)
			break;
		tpiso->tso_flags |= TSOF_OOB_INPUT_AVAILABLE;
		/* fall through */

	case SSCO_INPUT_ARRIVED:
		if (!(tpiso->tso_flags & TSOF_RSRV_QUEUED) &&
		    canenable(tpiso->tso_rq)) {
			tpiso->tso_flags |= TSOF_RSRV_QUEUED;
#ifdef notyet
			/*
			 * XXXrs - Not ever if you ask me.  This code is
			 * trying to "outsmart" the streams scheduling
			 * mechanism.  There may be some additional
			 * MP considerations w.r.t. the queuerun flag and
			 * blocking queueruns.  Think carefully before
			 * trying to turn something like this on.
			 *
			 * See the queuerun() code.
			 */
			if (! queueflag) {
				queuerun_block();
				splx(s);
				tpisocket_rsrv(tpiso->tso_rq);
				queuerun_unblock();
				return;
			} else
#endif /* notyet */
				qenable(tpiso->tso_rq);
		}
		break;

	default:
		break;
	}
	splx(s);
done:
	if (flag) {
		TSO_RELE(tpiso);
#ifdef _MP_NETLOCKS
		if ((tpiso->tso_count == 1) &&
		    (tpiso->tso_flags & TSOF_CLOSING)) {
			sv_broadcast(&tpisockd_inuse);
		}
#endif
	}

	return 0;
}

/*
 * function call from streams_interrupt to lock tpiso
 */
int
tpihold(void * func, void **arg2, void *tpiso)
{
	long a;
	/*
	 * kludge to lock tpisocket in case we defer operation
	 * because STREAMS monitor already in use.
	 */
	if ((func == (void *)tpisocket_event) && tpiso) {
		if (((struct tpisocket *)tpiso)->tso_flags & TSOF_CLOSING) {
			return 1;
		}
		TSO_HOLD(((struct tpisocket *)tpiso));
		a = (long)*arg2;
		a |= 0x10;
		*arg2 = (void *)a;
	}
	return 0;
}

/*
 *	tpisocket_allocb
 *		Allocates a buffer.  
 *		Queues a bufcall if necessary, returning NULL.
 *
 *		Entered at splstr().
 */
mblk_t *
tpisocket_allocb(
	queue_t	*qp,
	int	size,
	int	pri)
{
	mblk_t	*bp;

	bp = allocb(size, pri);
	if (!bp) {			/* if we fail, set to try again */
		struct	tpisocket *tpiso;
		int	flag;

		tpiso = (struct tpisocket *)qp->q_ptr;
		flag = (qp == tpiso->tso_rq ? TSOF_INPUT_BUFCALL
					    : TSOF_OUTPUT_BUFCALL);
		if (!(tpiso->tso_flags & flag) &&
		    bufcall(size, pri, qenable, (long)qp)) {
			tpiso->tso_flags |= flag;
			noenable(qp);
		}
	}
	return bp;
}


/*
 *	tpisocket_copy_to_mblk_chain
 *
 *	Copy a data buffer to an offset in an mblk chain
 */
void
tpisocket_copy_to_mblk_chain(
	caddr_t	datap,
	int	len,
	mblk_t	**bpp,
	int	*offsetp)
{
	int	count;
	mblk_t	*bp;

	bp = *bpp;
	while (len > 0 && bp != NULL) {
		count = (bp->b_wptr - bp->b_rptr) - *offsetp;
		if (count <= 0) {
			*offsetp = 0;
			bp = bp->b_cont;
			*bpp = bp;
			continue;
		}
		if (count > len)
			count = len;
		bcopy(datap, bp->b_rptr + *offsetp, count);
		(*offsetp) += count;
		len -= count;
		datap += count;
	}
}


/*
 *	tpisocket_mbufcall_wakeup
 *
 *		Wakes up streams waiting for mbufs
 */
void
tpisocket_mbufcall_wakeup( struct tpisocket *tpiso)
{
	int	s;

	s = splstr();
	if (tpiso->tso_flags & TSOF_INPUT_MBUFCALL) {
		enableok(tpiso->tso_rq);
		qenable(tpiso->tso_rq);
	}
	if (tpiso->tso_flags & TSOF_OUTPUT_MBUFCALL) {
		enableok(tpiso->tso_wq);
		qenable(tpiso->tso_wq);
	}
	tpiso->tso_flags &= ~TSOF_ANY_MBUFCALL;
	splx(s);
}


/*
 *	tpisocket_request_mbufcall
 *
 *		Request an mbuf available qenable in the specified qp.
 *
 *		Entered at splstr().
 */
void
tpisocket_request_mbufcall( queue_t *qp)
{
	struct	tpisocket *tpiso;

	tpiso = (struct tpisocket *)qp->q_ptr;
	if (!(tpiso->tso_flags & TSOF_ANY_MBUFCALL))
		tpiso->tso_mbufcall_timeout =
			STREAMS_TIMEOUT((strtimeoutfunc_t)tpisocket_mbufcall_wakeup,
					tpiso, 10);
		 tpiso->tso_flags |= (qp == tpiso->tso_rq ? TSOF_INPUT_MBUFCALL : 
						   TSOF_OUTPUT_MBUFCALL);
	noenable(qp);
}


/*
 *	tpisocket_mbuf_to_mblk
 *		Converts an mbuf chain to an mblk chain
 *
 *		Entered at splstr().
 */
mblk_t *
tpisocket_mbuf_to_mblk(
	queue_t	*qp,
	struct	mbuf *m,
	int	pri)
{
	mblk_t	*bp;

	bp = mbuf_to_mblk(m, pri);
	if (!bp)			/* if we fail, set to try again */
		tpisocket_request_mbufcall(qp);
	return bp;
}

/*
 *	tpisocket_mblk_to_mbuf
 *		Converts an mblk chain to an mbuf chain.
 *
 *		Entered at splstr().
 */
struct mbuf *
tpisocket_mblk_to_mbuf(
	mblk_t	*mp,		/* pointer to message being passed */
	int	alloc_flags,
	int	mbuf_type,
	int	mbuf_flags,
	queue_t	*qp)
{
	struct	mbuf *m;

	m = mblk_to_mbuf(mp, alloc_flags, mbuf_type, mbuf_flags);
	if (m == NULL)
		tpisocket_request_mbufcall(qp);
	return m;
}


/*
 *	tpisocket_trim_mbuf_chain
 *
 *		Trim len bytes from the front of an mbuf chain
 *		Similar to m_adj(), but discards empty mbufs, and
 *		returns pointer to beginning of shortened chain.
 */
struct mbuf *
tpisocket_trim_mbuf_chain(
	struct	mbuf *m,
	int	len)
{
	if (m == NULL)
		return NULL;

	while (len > 0 && m != NULL) {
		struct	mbuf *n;

		if (len < m->m_len) {
			m->m_off += len;
			m->m_len -= len;
			break;
		}
		len -= m->m_len;
		MFREE(m, n);
		m = n;
	}
	return m;
}

/*
 *	tpisocket_dup_mbuf
 *
 *		Duplicate mbuf chain.  
 *		Requests mbuf wakeup if mbufs not available.
 *
 *		Entered at splstr().
 */
/*ARGSUSED*/
struct mbuf *
tpisocket_dup_mbuf(
	struct mbuf	*src,
	int	alloc_flags,
	queue_t	*qp,
	int	len)
{
	struct mbuf	*first;

	ASSERT(alloc_flags == M_DONTWAIT);	/* m_copy assumes DONTWAIT */
	first = m_copy(src, 0, len);
	if (first == NULL)
		tpisocket_request_mbufcall(qp);
	return first;
}


/*
 *	tpisocket_copy_to_mbuf
 *		Copy data buffer to an mbuf
 *
 */
struct mbuf *
tpisocket_copy_to_mbuf(
	void 	*addr,
	int	len,
	int	mbuf_type,
	int	*error_p)
{
	struct	mbuf *m;

	if (((u_int)len) > MLEN) {	/* also catches len < 0 */
		*error_p = EINVAL;
		return NULL;
	}

#if defined(COMPAT_43) && defined(WRONG) 
	if (mbuf_type == MT_SONAME && ((u_int)len < MLEN))
		len = MLEN;	/* unix domain compat. hack */
#endif /* COMPAT_43 */

	MGET(m, M_DONTWAIT, mbuf_type);
	if (m == NULL) {
		*error_p = ENOBUFS;
		return NULL;
	}
	m->m_len = len;
	bcopy((caddr_t)addr, mtod(m, caddr_t), (u_int)len);
#ifdef RENO
	if (mbuf_type == MT_SONAME) 
		mtod(m, struct sockaddr *)->sa_len = len;
#endif
	*error_p = 0;
	return m;
}



/*
 *	tpisocket_send_iocack
 *		Converts a buffer from an M_IOCTL to an M_IOCACK message,
 *		and appends a supplied data buffer.
 *
 *		Entered at splstr().
 */
static int
tpisocket_send_iocack(
	queue_t	*qp,
	mblk_t	*bp,
	mblk_t	*rbp)
{
	struct iocblk *iocp;

	freemsg(bp->b_cont);
	bp->b_cont = rbp;
	bp->b_datap->db_type = M_IOCACK;
	iocp = (struct iocblk *)bp->b_rptr;
	iocp->ioc_count = msgdsize(rbp);
	qreply(qp, bp);
	return 1;
}


/*
 *	tpisocket_send_iocnak
 *		Converts a buffer from an M_IOCTL to an M_IOCNAK message,
 *		and supplies a given error code.
 *
 *		Entered at splstr().
 */
static int
tpisocket_send_iocnak(
	queue_t *qp,
	mblk_t	*bp,
	int	unix_error)
{
	struct iocblk *iocp;

	freemsg(bp->b_cont);
	bp->b_cont = NULL;
	bp->b_datap->db_type = M_IOCNAK;
	iocp = (struct iocblk *)bp->b_rptr;
	iocp->ioc_count = 0;
	iocp->ioc_error = unix_error;
	qreply(qp, bp);
	return 1;
}

/*
 *	tpisocket_send_ok_ack
 *
 *		Turn bp into a T_OK_ACK for the primitive correct_prim, 
 *		and send upstream.
 */
static void
tpisocket_send_ok_ack(
	struct tpisocket *tpiso,
	mblk_t	*bp,
	int	correct_prim)
{
	struct T_ok_ack *oa;

	bp->b_datap->db_type = M_PCPROTO;
	bp->b_wptr = bp->b_rptr + sizeof(*oa);
	oa = (struct T_ok_ack *)bp->b_rptr;
	oa->PRIM_type = T_OK_ACK;
	oa->CORRECT_prim = correct_prim;

	qreply(tpiso->tso_wq, bp);
}

/*
 *
 */
void
tpisocket_send_discon_ind(
	struct tpisocket *tpiso,
	mblk_t	*bp,
	int	unix_error)
{
	union	T_primitives *pi;

	tpiso->tso_flags |= TSOF_RECV_DISCON;

	bp->b_datap->db_type = M_PROTO;
	bp->b_wptr = bp->b_rptr + sizeof(struct T_discon_ind);

	pi = (union T_primitives *)bp->b_rptr;
	pi->type = T_DISCON_IND;
	pi->discon_ind.DISCON_reason = unix_error;
	pi->discon_ind.SEQ_number = -1;

	qreply(tpiso->tso_wq, bp);
}


/*
 *	tpisocket_wreject
 *		Rejects a TPI request.
 *
 *		Entered at splstr().
 */
#define TSO_PUTWQ(tpiso, bp) { \
		tpiso->tso_flags |= TSOF_QUEUE_OUTPUT; \
		putq(tpiso->tso_wq,(bp)); }
#define TSO_PUTBWQ(tpiso, bp) { \
		tpiso->tso_flags |= TSOF_QUEUE_OUTPUT; \
		putbq(tpiso->tso_wq,(bp)); }

int
tpisocket_wreject(
	struct tpisocket *tpiso,
	mblk_t	*bp,
	int	tli_prim,
	int	tli_error,
	int	unix_error)
{
	mblk_t	*nbp;
	struct	T_error_ack *terror;

	if (bp != NULL &&
	    (bp->b_datap->db_type == M_PROTO ||
	     bp->b_datap->db_type == M_PCPROTO) &&
	    (bp->b_datap->db_lim - bp->b_datap->db_base) > 
			sizeof(struct T_error_ack)) {
		bp->b_rptr = bp->b_datap->db_base;
		nbp = bp;
		bp = NULL;
	} else {
		nbp = tpisocket_allocb(tpiso->tso_wq,
				       sizeof(struct T_error_ack),
				       BPRI_HI);
		if (nbp == NULL) {
			if (bp)
				TSO_PUTBWQ(tpiso, bp);
			return 0;
		}
	}
	terror = (struct T_error_ack *)nbp->b_rptr;
	nbp->b_wptr = nbp->b_rptr + sizeof(struct T_error_ack);
	nbp->b_datap->db_type = M_PCPROTO;
	terror->ERROR_prim = tli_prim;
	terror->PRIM_type = T_ERROR_ACK;
	terror->TLI_error = tli_error;
	terror->UNIX_error = unix_error;
	qreply(tpiso->tso_wq, nbp);
	if (bp)
		freemsg(bp);
	return 1;
}

static int
tpisocket_send_ok_message(
	struct tpisocket *tpiso,
	struct socket *so,
	mblk_t	*bp)
{
	union	T_primitives *pi = (union T_primitives *)bp->b_rptr;

	if (bp->b_cont != NULL) {
		freemsg(bp->b_cont);
		bp->b_cont = NULL;
	}
	tpisocket_send_ok_ack(tpiso, bp, pi->type);

	/*
	 * get tpisocket_rsrv() to recheck the connection queue 
	 */
	tpisocket_event(so, SSCO_INPUT_ARRIVED, tpiso);

	return 1;
}

/*
 *	tpisocket_convert_error
 *
 *	Convert a unix error code to a tli error code
 */
void
tpisocket_convert_error(
	int	*unix_error_p,
	int	*tli_error_p)
{
	int	unix_error = *unix_error_p;

	if (unix_error < 0) {
		*tli_error_p = - unix_error;
		*unix_error_p = (*tli_error_p == TSYSERR ? EINVAL : 0);
		return;
	}

	switch (unix_error) {
	case EADDRNOTAVAIL:
	case EADDRINUSE:
#ifdef _IRIX_LATER	/* XTI wants TADDRBUSY, TLI wants TNOADDR */
		*tli_error_p = TADDRBUSY;
#else
		*tli_error_p = TNOADDR;
#endif
		break;
	case EINVAL:
		*tli_error_p = TBADOPT;
		break;
	case EACCES:
		*tli_error_p = TACCES;
		break;
	case ENOTCONN:
	case EALREADY:
		*tli_error_p = TOUTSTATE;
		break;
	default:
		*tli_error_p = TSYSERR;
		break;
	}
}


/*
 *	tpisocket_wact
 *		Acts on an output packet.  
 *		If the packet cannot be processed immediately, 
 *		it is put back on the output queue.
 *	Returns:
 *		0 if message not completely processed.
 *		non-0 if message finished, ready for next message.
 *
 *		Entered at splstr().
 */
int
tpisocket_wact(
	struct tpisocket *tpiso,
	mblk_t	*bp)
{
	struct	tpisocket_control *tpictl = tpiso->tso_control;
	struct	socket *so = tpiso->tso_socket;
	struct	mbuf *m;
	union	T_primitives *pi;
	cred_t	*crp;
	mblk_t	*rbp;
	int	tli_prim;
	int	tli_error;
	int	unix_error;
	NETSPL_SPDECL(s)

	switch (bp->b_datap->db_type) {
	case M_DATA:
		tpiso->tso_flags &= ~TSOF_UNUSED;
		tpiso->tso_pending_send_flags = 0;
		tpiso->tso_pending_send_addr.mblk = NULL;
		tpiso->tso_pending_send_control.mblk = NULL;
		tpiso->tso_pending_send_header = NULL;
		tpiso->tso_pending_send.mblk = bp;
		goto setup_send_message_3;

	case M_PCPROTO:
	case M_PROTO:
		break;	/* No need to nest switch statements */

	case M_IOCTL:
	{
		struct	iocblk *iocp;

		iocp = (struct iocblk *)bp->b_rptr;
		switch (iocp->ioc_cmd) {
		case SIOCGETNAME:
			return tpisocket_sioc_name(tpiso, bp, PRU_SOCKADDR);

		case SIOCGETPEER:
			if (!(so->so_state &
					(SS_ISCONNECTED|SS_ISCONFIRMING))) {
				return tpisocket_send_iocnak(tpiso->tso_wq, bp,
						      ENOTCONN);
			}
			return tpisocket_sioc_name(tpiso, bp, PRU_PEERADDR);

#ifdef _SVR4_NOTYET
		case INITQPARMS:
			/* XXX */
			return 1;
#endif /* _SVR4_NOTYET */

		case TI_GETMYNAME:
			return tpisocket_ti_name(tpiso, bp);

		case TI_GETPEERNAME:
			return tpisocket_ti_name(tpiso, bp);

		case SIOCGETLABEL:
			{
			/* force copy of label to be atomic */
			SOCKET_LOCK(so, s);

			bcopy((void *)sesmgr_get_label(so),
				bp->b_cont->b_rptr,
				sizeof(mac_label *));
			SOCKET_UNLOCK(so, s);

			bp->b_datap->db_type = M_IOCACK;
			(void)qreply(tpiso->tso_wq,bp);
			return(0);
			}

		case SIOCSETLABEL:
			{
			mac_label * new_label;
			int error;
                        if (iocp->ioc_count != sizeof(mac_label *))
                                error = EINVAL;
                        else {
                                bcopy(bp->b_cont->b_rptr,
                                        (char *)&new_label,
                                        sizeof(mac_label *));

				SOCKET_LOCK(so, s);
				error = (*so->so_proto->pr_usrreq)(so,
						PRU_SOCKLABEL,
						(struct mbuf *)0,
						(struct mbuf *) new_label,
						(struct mbuf *)0);
				SOCKET_UNLOCK(so, s);
			}
			iocp->ioc_error = error;
                        bp->b_datap->db_type = M_IOCACK;
                        (void)qreply(tpiso->tso_wq,bp);
#if 0
			_SAT_BSDIPC_MAC_CHANGE(tpiso->tso_fd,
					       tpiso->tso_socket,
					       new_label,
					       tpiso->tso_procp,error);
#endif
                        return(0);
			}

		case SIOCGETACL:
			{
			/* force copy of label to be atomic */
			SOCKET_LOCK(so, s);

			bcopy((void *)sesmgr_get_soacl(so),
				bp->b_cont->b_rptr,
				sizeof(struct soacl *));
			SOCKET_UNLOCK(so, s);

			bp->b_datap->db_type = M_IOCACK;
                        (void)qreply(tpiso->tso_wq,bp);
                        return(0);
			}

		case SIOCSETACL:
			{
			struct soacl * new_soacl;

			bcopy(bp->b_cont->b_rptr, (char *)&new_soacl,
				sizeof(struct soacl *));
			sesmgr_set_soacl(so,new_soacl);
			bp->b_datap->db_type = M_IOCACK;
			(void)qreply(tpiso->tso_wq,bp);
#if 0
			_SAT_BSDIPC_DAC_CHANGE(tpiso->tso_fd,
					       tpiso->tso_socket,
					       tpiso->tso_procp,
					       0);
#endif
			return(0);
			}

		case SIOCGETUID:
			{
			uid_t tmp_uid = sesmgr_get_uid(so);
			bcopy((char *)&tmp_uid,
				bp->b_cont->b_rptr,
				sizeof(uid_t));
			bp->b_datap->db_type = M_IOCACK;
			(void)qreply(tpiso->tso_wq,bp);
			return(0);
			}

		case SIOCSETUID:
			{
			uid_t new_uid;
			bcopy(bp->b_cont->b_rptr, (char *)&new_uid,
				sizeof(uid_t));
			sesmgr_set_uid(so,new_uid);
			bp->b_datap->db_type = M_IOCACK;
			(void)qreply(tpiso->tso_wq,bp);
#if 0
			_SAT_BSDIPC_DAC_CHANGE(tpiso->tso_fd, tpiso->tso_socket,
				tpiso->tso_procp,0);
#endif
			return(0);
			}

		case SIOCGETRCVUID:
			{
			uid_t tmp_uid = sesmgr_get_rcvuid(so);
			bcopy((char *)&tmp_uid,
				bp->b_cont->b_rptr,
				sizeof(uid_t));
			bp->b_datap->db_type = M_IOCACK;
			(void)qreply(tpiso->tso_wq,bp);
			return(0);
			}

		default:
			break;
		}
		if (IOCGROUP(iocp->ioc_cmd) == 'i')
			return tpisocket_ifioctl(tpiso, bp);
		if (IOCGROUP(iocp->ioc_cmd) == 'r')
			return tpisocket_rtioctl(tpiso, bp);
		return tpisocket_send_iocnak(tpiso->tso_wq, bp, EINVAL);
	}

	case M_IOCDATA:
		if (!(tpiso->tso_flags & TSOF_IOCTL_PENDING)) {
			freemsg(bp);
			return 1;
		}
		switch (tpiso->tso_pending_ioctl) {
		case TSOI_TI_NAME:
			return tpisocket_ti_name(tpiso, bp);
		case TSOI_IFIOCTL:
			return tpisocket_ifiocdata(tpiso, bp);
		case TSOI_RTIOCTL:
			return tpisocket_rtiocdata(tpiso, bp);
		default:
			break;
		}
		freemsg(bp);
		return 1;

	case M_CTL:
	default:
		freemsg(bp);
		return 1;
	}

	/*
	 * Process M_PROTO and M_PCPROTO
	 */
	pi = (union T_primitives *)bp->b_rptr;
	switch (pi->type) {

	case T_CONN_REQ:	/* connection request     */

		if ((so->so_state & SS_ISCONNECTING) ||
		    tpiso->tso_pending_request != NULL) {

			return tpisocket_wreject(tpiso, bp, pi->type, 
				TOUTSTATE, 0);
		}
		rbp = tpisocket_allocb(tpiso->tso_wq, sizeof(struct T_ok_ack), 
				BPRI_MED);
		if (rbp == NULL) {
			TSO_PUTBWQ(tpiso, bp);
			return 0;
		}
		m = tpisocket_copy_to_mbuf(
			bp->b_rptr + pi->conn_req.DEST_offset,
			min(pi->conn_req.DEST_length,
			    (bp->b_wptr - bp->b_rptr) -
			       pi->conn_req.DEST_offset),
			MT_SONAME, &unix_error);

		if (m == NULL) {
			freemsg(rbp);
			if (unix_error == ENOBUFS) {
				tli_error = TSYSERR;
				unix_error = ENOSR;
			} else {
				tli_error = TBADADDR;
				unix_error = 0;
			}
			return tpisocket_wreject(tpiso, bp, pi->type, 
				tli_error, unix_error);
		}
		SOCKET_LOCK(so, s);
		unix_error = soconnect(so, m);
		SOCKET_UNLOCK(so, s);
		m_freem(m);

		tpiso->tso_flags &= ~TSOF_UNUSED;
		/*
		 * send connection request ack message back
		 */
		tpisocket_send_ok_ack(tpiso, rbp, pi->type);

		if (!unix_error) { /* Connection request succeeded */
			/* waiting for a real response */
			tpiso->tso_flags |= TSOF_CONN_REQ_PENDING;
			tpiso->tso_pending_request = bp;
			tpisocket_conn_req_done(tpiso);
			return 1;
		}

		/* Connection request failed */
		tpiso->tso_flags &= ~TSOF_CONN_REQ_PENDING;
		tpiso->tso_flags |= TSOF_CONN_REQ_FAILED;
		tpiso->tso_pending_request = (mblk_t *)NULL;

		if (tpiso->tso_flags & TSOF_RECV_DISCON) {
			freemsg(bp); /* discard the message */
			return 1;
		}

		/*
		 * Received a socket error.  So return a disconn message 
		 * with the errno inside when we haven't already
		 * received a disconnect indication.
		 */
		tpisocket_send_discon_ind(tpiso, bp, unix_error);
		return 1;

	case T_CONN_RES:	/* connection response    */
	{
		int	i;
		struct	tpisocket *ntpiso;
		struct	socket *nso;
		struct	socket *oso;
		struct	socket *head;
		int ss;

		/* tpiso is the listen TPIsocket, its so_q holds the socket
		 * 	for the not-yet-accepted connection.
		 * ntpiso is the TPIsocket newly created to become the
		 *	endpoint for the not-yet-accepted connection. 
		 * oso is the socket that was attached to ntpiso when
		 * 	ntipso was created.  It will be discarded.
		 * nso is the socket on tpiso's so_q.  It will be removed
		 *	from the queue, and attached to ntpiso.
		 */

		/*
		 * This is a response to a T_CONN_IND, so presume it is
		 * no longer queued.
		 */
		ASSERT(tpiso->tso_outcnt < 2);
		tpiso->tso_outcnt = 0;

		/* check for a valid tpisocket queue */
		ntpiso = NULL;
		for (i = 0; i < TSC_MAXDEVICES; i++) {
			if (TSC_DEVICES[i] == NULL ||
			    TSC_DEVICES[i]->tso_rq == NULL)
				continue;
			if (TSC_DEVICES[i]->tso_rq == (queue_t*)
					NET64_TOPTR(pi->conn_res.QUEUE_ptr)) {
				ntpiso = TSC_DEVICES[i];
				break;
			}
		}
		/* check queue state */
		if (ntpiso == NULL ||
		    ntpiso->tso_flags & (TSOF_OPENING|TSOF_CLOSING)) {
			return tpisocket_wreject(tpiso, bp, pi->type, 
				TBADF, 0);
		}

		/*
		 * If we're already using this for a connection or fd is
		 * not a listening endpoint, fail.
		 */
		if ((ntpiso->tso_socket->so_state &
			(SS_ISCONNECTED | SS_ISCONNECTING |
			 SS_ISDISCONNECTING | SS_ISCONFIRMING)) ||
		    !(tpiso->tso_socket->so_options & SO_ACCEPTCONN)) {
			return tpisocket_wreject(tpiso, bp, pi->type, 
				TOUTSTATE, 0);
		}

		/* if resfd is not the same as fd, it can't be listening */
		if ((tpiso != ntpiso) && 
		    (ntpiso->tso_socket->so_options & SO_ACCEPTCONN)) {
			return tpisocket_wreject(tpiso, bp, pi->type, 
				TBADF, 0);
		}

		/* resfd must be unbound or idle */
		if (!(ntpiso->tso_flags & TSOF_UNUSED) &&
		    !(ntpiso->tso_flags & TSOF_ADDRESS_BOUND)) {
  			return tpisocket_wreject(tpiso, bp, pi->type, 
  				TOUTSTATE, 0);
  		}

		/*
		 * If ntpiso == tpiso, then there can be only one connection
		 * pending.  If not equal, then there must be no connections
		 * pending on ntpiso (all pending connections are on tpiso).
		 */
		head = ntpiso->tso_socket;
		SOCKET_LOCK(head, s);

		if (ntpiso->tso_socket->so_qlen != (ntpiso == tpiso)) {
			SOCKET_UNLOCK(head, s);
			return tpisocket_wreject(tpiso, bp, pi->type, TBADF,0);
		}
		SOCKET_UNLOCK(head, s);

		/*
		 * Check if the specified connection is still there
		 */
restart_conn:
		head = tpiso->tso_socket;
		SOCKET_LOCK(head, s);

		for (nso = head->so_q; nso != NULL && nso != head;
		     nso = nso->so_q) {

#ifdef _MP_NETLOCKS
			if (!SOCKET_CMPLOCK(nso)) {
				mon_state_t ms;
				proc_t *pp;

				SOCKET_MPUNLOCK(head);

				/*
				 * since we call qswtch directly, we
				 * must release and reaquire the the
				 * streams monitor as we yield the cpu.
				 */
				pp = private.p_curproc;
				ASSERT(pp->p_kthread.k_monitor);
				rmonitor(pp->p_kthread.k_monitor, &ms);

				qswtch(RESCHED_Y);

				amonitor(&ms);

				goto restart_conn;
			}
#endif
			if (nso == (struct socket *)NET64_TOPTR(pi->conn_res.SEQ_number)) {
				break;
			}
			SOCKET_MPUNLOCK(nso);
		}

		/*
		 * We arrive here with the so and nso locks held
		 */
		if (nso == NULL || nso == head ||
		   !(nso->so_state & SS_TPISOCKET)) {
#ifdef _MP_NETLOCKS
			if ((nso != NULL) && (nso != head)) {
				SOCKET_MPUNLOCK(nso);
			}
#endif
			SOCKET_UNLOCK(head, s);
			return tpisocket_wreject(tpiso, bp,pi->type,TBADSEQ,0);
		}

		/*
		 * remove the socket from the queue
		 */
		{
			register s1;
			SOCKET_MPQLOCK(head->so_qlock, s1);
			soqremque(nso, 1);
			SOCKET_MPQUNLOCK(head->so_qlock, s1);
		}
		tpiso->tso_flags &= ~TSOF_UNUSED;

		/* swap the new socket for the old one */
		oso = ntpiso->tso_socket;
		oso->so_state        &= ~SS_WANT_CALL;
		oso->so_snd.sb_flags &= ~SB_NOTIFY;
		ntpiso->tso_socket    = nso;
		ntpiso->tso_flags    &= ~TSOF_UNUSED;
		nso->so_callout       = tpisocket_event;
		nso->so_callout_arg   = (caddr_t)ntpiso;
		/*
		 * Note: this implementation uses the same bit value
		 * for SS_CONN_IND_SENT and SS_TPISOCKET, since the
		 * field is only a short. Once the socket is attached to
		 * a TPI data structure, it takes on the second meaning.
		 */
#if (SS_TPISOCKET != SS_CONN_IND_SENT)
		nso->so_state        &= ~SS_CONN_IND_SENT;
#endif
		nso->so_state        |= (SS_WANT_CALL | SS_TPISOCKET);
		nso->so_snd.sb_flags |= SB_NOTIFY;
		nso->so_state        &= ~SS_NOFDREF;

		/* queue old socket for discard */
		ss = mp_mutex_spinlock(&tpisocket_discard_lock);
		oso->so_callout_arg = (caddr_t)tpisocket_discard_list; 
		tpisocket_discard_list = oso;
		mp_mutex_spinunlock(&tpisocket_discard_lock, ss);
#ifdef TPISOCKET_DAEMON
		wakeup(&tpisocket_discard_list);
#endif
		/*
		 * wakeup the new socket to look at the current state
		 */
		tpisocket_event(nso, SSCO_INPUT_ARRIVED, ntpiso);
		tpisocket_event(nso, SSCO_OOB_INPUT_ARRIVED, ntpiso);
		tpisocket_event(nso, SSCO_OUTPUT_AVAILABLE, ntpiso);

		SOCKET_MPUNLOCK(nso);
		SOCKET_UNLOCK(head, s);

		/* send the "ok" message */
		/* (reuse the old message) */
		return (tpisocket_send_ok_message(tpiso, so, bp));
	}

	case T_DISCON_REQ:	/* disconnect request     */
	{
		struct	socket *oso;

		/* which kind of disconnect? */
		if (so->so_options & SO_ACCEPTCONN) {
			NETSPL_SPDECL(s)

			/*
			 * This is a response to a T_CONN_IND, 
			 * so presume it is no longer queued.
			 */
			ASSERT(tpiso->tso_outcnt < 2);
			tpiso->tso_outcnt = 0;

			/*
			 * need a valid sequence number
			 */
			SOCKET_LOCK(so, s);
restart_disconn:
			for (oso = so->so_q;
			     oso != NULL && oso != so;
			     oso = oso->so_q) {
#ifdef _MP_NETLOCKS
				if (!SOCKET_CMPLOCK(oso)) { /* failed restart*/
					mon_state_t ms;
					proc_t *pp;

					SOCKET_MPUNLOCK(so);

					/*
					 * since we call qswtch directly, we
					 * must release and reaquire the the
					 * streams monitor as we yield the cpu.
					 */
					pp = private.p_curproc;
					ASSERT(pp->p_kthread.k_monitor);
					rmonitor(pp->p_kthread.k_monitor, &ms);

					qswtch(RESCHED_Y);

					amonitor(&ms);
					SOCKET_MPLOCK(so);
					goto restart_disconn;
				}
#endif /* _MP_NETLOCKS */
				if (!(oso->so_state & SS_TPISOCKET)) {
					SOCKET_MPUNLOCK(oso);
					continue;
				}
				if (oso == (struct socket *) 
				NET64_TOPTR(pi->discon_req.SEQ_number))
					break;
				SOCKET_MPUNLOCK(oso);
			}
			if (oso == NULL || oso == so ||
			  !(oso->so_state & SS_TPISOCKET)) {
				SOCKET_UNLOCK(so, s);
				return tpisocket_wreject(tpiso, bp, pi->type, 
					TBADSEQ, 0);
			}
			/* XXX: for the time being, ignore the user data */
			tpiso->tso_flags &= ~TSOF_UNUSED;
			/*
			 * dequeue the socket
			 */
			{
				register s1;
				struct socket *h = oso->so_head;
				SOCKET_MPQLOCK(h->so_qlock, s1);
				soqremque(oso, 1);
				SOCKET_MPQUNLOCK(h->so_qlock, s1);
			}

			oso->so_state &= ~SS_CONN_IND_SENT;
 
			/* discard the socket, releases lock on oso */
			if (soabort(oso)) {
				SOCKET_MPUNLOCK(oso);
			}

			/*
			 * get tpisocket_rsrv() to recheck the connection queue 
			 */
			tpisocket_event(so, SSCO_INPUT_ARRIVED, tpiso);

			SOCKET_UNLOCK(so, s);

			/* send the "ok" message */
			return (tpisocket_send_ok_message(tpiso, so, bp));
		}
		rbp = tpisocket_allocb(tpiso->tso_wq, sizeof(struct T_ok_ack),
			       BPRI_MED);
		if (rbp == NULL) {
			TSO_PUTBWQ(tpiso, bp);
			return 0;
		}

		SOCKET_LOCK(so, s);
		unix_error = sodisconnect(so);
		SOCKET_UNLOCK(so, s);

		if (unix_error == ENOTCONN || unix_error == EALREADY) {
			freemsg(rbp);
			return tpisocket_wreject(tpiso, bp, pi->type, 
				TOUTSTATE, 0);
		}
		tpiso->tso_flags &= ~TSOF_UNUSED;
		tpiso->tso_flags |= TSOF_DISCON_REQ_PENDING;

#if defined(WRONG) && (WRONG == RIGHT)
		/* send ok acknowledgement */
		tpisocket_send_ok_ack(tpiso, rbp, pi->type);
#else
		freemsg(rbp);	/* why do we have an rbp ?? */
#endif

		freemsg(bp);
		tpisocket_discon_req_done(tpiso);
		return 1;
	}

	case T_DATA_REQ:
		tpiso->tso_pending_send_flags =
			(pi->data_req.MORE_flag ? 0 : MSG_EOR);
		goto setup_send_message;

	case T_EXDATA_REQ:	/* expedited data request */
		tpiso->tso_pending_send_flags =
			(pi->exdata_req.MORE_flag ? 0 : MSG_EOR) | MSG_OOB;
setup_send_message:
		tpiso->tso_pending_send_addr.mblk = NULL;
		tpiso->tso_pending_send_control.mblk = NULL;
setup_send_message_2:
		tpiso->tso_pending_send_header = bp;
		tpiso->tso_pending_send.mblk = bp->b_cont;
		bp->b_cont = NULL;
setup_send_message_3:
		tpiso->tso_pending_send_error = 0;
		tpiso->tso_flags |= TSOF_SEND_PENDING;
		tpiso->tso_flags &= ~TSOF_UNUSED;
		return tpisocket_send_check(tpiso);

	case T_INFO_REQ:	/* information request    */
	{
		struct T_info_ack *ia;

		rbp = tpisocket_allocb(tpiso->tso_wq,
			       sizeof(struct T_info_ack),
			       BPRI_MED);
		if (rbp == NULL) {
			TSO_PUTBWQ(tpiso, bp);
			return 0;
		}
		ia = (struct T_info_ack *)rbp->b_rptr;
		rbp->b_wptr = rbp->b_rptr + sizeof(struct T_info_ack);
		rbp->b_datap->db_type = M_PCPROTO;
		ia->PRIM_type  = T_INFO_ACK;
		ia->TSDU_size  = TSC_TSDU_SIZE; 
		ia->ETSDU_size = TSC_ETSDU_SIZE;
		ia->CDATA_size = TSC_CDATA_SIZE;
		ia->DDATA_size = TSC_DDATA_SIZE;
		ia->ADDR_size  = TSC_ADDRESS_SIZE(tpiso, 
					TSCA_REMOTE_ADDRESS, NULL);
		if (ia->ADDR_size == 0)
			ia->ADDR_size = -1;
		ia->OPT_size = TSC_ADDRESS_SIZE(tpiso, TSCA_OPTIONS, 
					NULL);
		if (ia->OPT_size == 0)
			ia->OPT_size = -1;
		ia->TIDU_size = TSC_TIDU_SIZE;
		ia->SERV_type = TSC_TRANSPORT_TYPE;
		if (tpiso->tso_flags & TSOF_ADDRESS_BOUND)
			ia->CURRENT_state = TS_IDLE;
		else
			ia->CURRENT_state = TS_UNBND;
		if (tpiso->tso_flags & TSOF_CONN_REQ_PENDING)
			ia->CURRENT_state = TS_WACK_CREQ;
		else if (tpiso->tso_flags & TSOF_DISCON_REQ_PENDING)
			ia->CURRENT_state = TS_WACK_DREQ6;
		else if (tpiso->tso_flags & TSOF_OPENING)
			ia->CURRENT_state = TS_UNBND;
		else if (so->so_state & SS_ISDISCONNECTING)
			ia->CURRENT_state = TS_WACK_DREQ7;
		else if (so->so_state & SS_CANTRCVMORE)
			ia->CURRENT_state = TS_WIND_ORDREL;
		else if (so->so_state & SS_CANTSENDMORE)
			ia->CURRENT_state = TS_WREQ_ORDREL;
		else if (tpiso->tso_outcnt > 0)
			ia->CURRENT_state = TS_WRES_CIND;
		else if (so->so_state & SS_ISCONNECTING) {
			if (so->so_options & SO_ACCEPTCONN) 
				ia->CURRENT_state = TS_WACK_CRES;
			else
				ia->CURRENT_state = TS_WCON_CREQ;
		} else if (so->so_state & SS_ISCONNECTED)
			ia->CURRENT_state = TS_DATA_XFER;
		ia->PROVIDER_flag = 0;
		putnext(tpiso->tso_rq, rbp);
		break;
	}

	case T_BIND_REQ:	/* bind request		  */
	{
		mblk_t	*ap;
		int	size;

		if (tpiso->tso_flags & TSOF_ADDRESS_BOUND ||
		    (so->so_state & (SS_ISCONNECTED | SS_ISCONFIRMING |
				     SS_ISCONNECTING))) {
			return tpisocket_wreject(tpiso, bp, pi->type, 
				TOUTSTATE, 0);
			}

		size = TSC_ADDRESS_SIZE(tpiso, TSCA_LOCAL_ADDRESS, NULL)
			+ sizeof(struct T_bind_ack);
		rbp = tpisocket_allocb(tpiso->tso_wq,
			max(sizeof(struct T_error_ack), size), BPRI_MED);
		if (rbp == NULL) {
			TSO_PUTBWQ(tpiso, bp);
			return 0;
		}
		rbp->b_datap->db_type = M_PCPROTO;
		if (pi->bind_req.ADDR_length > 0) {
			ap = dupb(bp);
			if (ap == NULL) {
				TSO_PUTBWQ(tpiso, bp);
				freemsg(rbp);
				return 0;
			}
			ap->b_rptr += pi->bind_req.ADDR_offset;
			ap->b_wptr = ap->b_rptr + pi->bind_req.ADDR_length;
			m = tpisocket_mblk_to_mbuf(ap, M_DONTWAIT,
					MT_DATA, 0, tpiso->tso_wq);
			if (m == NULL) {
				freemsg(ap);
				freemsg(rbp);
				TSO_PUTBWQ(tpiso, bp);
				return 0;
			}
                } else if (TSC_FLAGS & TSCF_DEFAULT_ADDRESS) {
                        MGET(m, M_DONTWAIT, MT_DATA);
                        if (m == NULL) {
                                freemsg(rbp);
                                TSO_PUTBWQ(tpiso, bp);
                                return 0;
                        }
                        if (TSC_CONVERT_ADDRESS(tpiso, NULL,
                                                TSCA_DEFAULT_ADDRESS |
                                                TSCA_MODE_TO_SOCKETS,
                                                -1, m, NULL)) {
                                m_freem(m);
                                m = NULL;
                        }
		} else
			m = NULL;

/* if kthread, cred will be pointing at sys_cred */
#define KTOP_GET_CURRENT_CRED() (curprocp->p_cred)
#define KTOP_SET_CURRENT_CRED(credp) (curprocp->p_cred = credp)
		/* switch credentials so that we can validate correctly */
		crp = KTOP_GET_CURRENT_CRED();
		KTOP_SET_CURRENT_CRED(tpiso->tso_cred);
		unix_error = sobind(tpiso->tso_socket, m);
		KTOP_SET_CURRENT_CRED(crp);
#if 0
		_SAT_BSDIPC_ADDR(tpiso->tso_fd, tpiso->tso_socket, m,
			tpiso->tso_procp, unix_error);
#endif
		if (m != NULL)
			m_freem(m);
		if (unix_error) {
			tpisocket_convert_error(&unix_error, &tli_error);
			(void)tpisocket_wreject(tpiso, rbp, T_BIND_REQ,
					tli_error, unix_error);
			break;
		}
		if (pi->bind_req.ADDR_offset > 0 ||
		    pi->bind_req.CONIND_number > 0)
			tpiso->tso_flags &= ~TSOF_UNUSED;
		tpiso->tso_flags |= TSOF_ADDRESS_BOUND;

		if (pi->bind_req.CONIND_number > 0) {
			(void)solisten(so, pi->bind_req.CONIND_number);
			tpiso->tso_flags &= ~TSOF_UNUSED;
		}
		rbp->b_wptr = rbp->b_rptr + size;
		pi = (union T_primitives *)rbp->b_rptr;
		pi->bind_ack.PRIM_type = T_BIND_ACK;
		pi->bind_ack.ADDR_length = size - sizeof(struct T_bind_ack);
		pi->bind_ack.ADDR_offset = sizeof(struct T_bind_ack);
		pi->bind_ack.CONIND_number = so->so_qlimit;
		if (pi->bind_ack.ADDR_length > 0)
			(void)TSC_CONVERT_ADDRESS(tpiso,
				(char *)rbp->b_rptr + pi->bind_ack.ADDR_offset,
				TSCA_LOCAL_ADDRESS | TSCA_MODE_READ,
				pi->bind_ack.ADDR_length,
				NULL, NULL);
		putnext(tpiso->tso_rq, rbp);
		break;
	}

	case T_UNBIND_REQ:	/* unbind request	  */

		SOCKET_LOCK(so, s);

		if (!(tpiso->tso_flags & TSOF_ADDRESS_BOUND) ||
		    (so->so_state & (SS_ISCONNECTED | SS_ISCONFIRMING |
				     SS_ISCONNECTING))) {
			SOCKET_UNLOCK(so, s);
			return tpisocket_wreject(tpiso, bp, pi->type, 
				TOUTSTATE, 0);
		}
		SOCKET_UNLOCK(so, s);

		rbp = tpisocket_allocb(tpiso->tso_wq,
				sizeof(struct T_error_ack), BPRI_MED);
		if (rbp == NULL) {
			TSO_PUTBWQ(tpiso, bp);
			return 0;
		}
		rbp->b_datap->db_type = M_PCPROTO;
		freemsg(bp);
		bp = rbp;
		pi = (union T_primitives *)bp->b_rptr;
		pi->type = T_UNBIND_REQ;
		tli_prim = pi->type;
		unix_error = TSC_UNBIND(tpiso);
		if (unix_error) {
			if (unix_error == EINVAL) {
				tli_error = TOUTSTATE;
				unix_error = 0;
			} else {	
				tli_error = TSYSERR;
			}
			return tpisocket_wreject(tpiso, bp, tli_prim,
					tli_error, unix_error);
		}
		tpiso->tso_flags &= ~TSOF_ADDRESS_BOUND;
		return (tpisocket_send_ok_message(tpiso, so, bp));

	case T_UNITDATA_REQ:	/* unitdata request       */
	{
		mblk_t *ap;
		mblk_t *cp;

		if (pi->unitdata_req.DEST_length > 0) {
			ap = dupb(bp);
			if (ap == NULL) {
				TSO_PUTBWQ(tpiso, bp);
				return 0;
			}
			ap->b_rptr += pi->unitdata_req.DEST_offset;
			ap->b_wptr = ap->b_rptr +
				pi->unitdata_req.DEST_length;
		} else
			ap = NULL;
#ifdef RENO
		if (pi->unitdata_req.OPT_length > 0) {
			cp = dupb(bp);
			if (cp == NULL) {
				if (ap != NULL)
					freeb(ap);
				TSO_PUTBWQ(tpiso, bp);
				return 0;
			}
			cp->b_rptr += pi->unitdata_req.OPT_offset;
			cp->b_wptr = cp->b_rptr +
				pi->unitdata_req.OPT_length;
		} else
#endif
			cp = NULL;
		tpiso->tso_pending_send_flags = MSG_EOR;
		tpiso->tso_pending_send_addr.mblk = ap;
		tpiso->tso_pending_send_control.mblk = cp;
		goto setup_send_message_2;
	}

	case T_OPTMGMT_REQ:	/* manage options req     */
		if (tpiso->tso_flags & TSOF_OPTMGMT_REQ_PENDING) {
			return tpisocket_wreject(tpiso, bp, pi->type, 
				TOUTSTATE, 0);
		}

		tpiso->tso_flags &= ~TSOF_UNUSED;
		if (tpiso->tso_pending_optmgmt_req != NULL)
			freemsg(tpiso->tso_pending_optmgmt_req);
		tpiso->tso_pending_optmgmt_req = bp;
		tpiso->tso_flags |= TSOF_OPTMGMT_REQ_PENDING;
		tpisocket_optmgmt_req_done(tpiso);
		return 1;

	case T_ORDREL_REQ:       /* orderly release req    */
		tpiso->tso_flags &= ~TSOF_UNUSED;
		soshutdown(so, FWRITE);
		break;

	default:
		return tpisocket_wreject(tpiso, bp, pi->type, TNOTSUPPORT, 0);
	}
	freemsg(bp);
	return 1;
}


/*
 *	tpisocket_wput
 *		Called for each output packet presented to this module.
 *		If the socket can accept another packet, and output need
 *		not be queued, transmits the packet directly.  Otherwise,
 *		queues the packet for later processing.
 *
 *		Entered at splstr().
 */
int
tpisocket_wput(queue_t *wq, mblk_t *bp)
{
#ifdef _MP_NETLOCKS
	struct tpisocket *tpiso = (struct tpisocket *)wq->q_ptr;

	if (! tpiso ||
            !(tpiso->tso_flags & TSOF_ISOPEN) ||
            tpiso->tso_flags & TSOF_CLOSING) {
                sdrv_error(wq, bp);
                return 0;
        }
	TSO_HOLD(tpiso);
	tpisockd_call((void (*)())_tpisocket_wput, wq, bp, tpiso);
	return 0;
}

static int
_tpisocket_wput(queue_t	*wq, mblk_t *bp, struct tpisocket *tpiso)
{
#else
	struct tpisocket *tpiso = (struct tpisocket *)wq->q_ptr;
#endif

	if ((tpiso == NULL) || (tpiso->tso_flags & TSOF_CLOSING)) {
		goto done;
	}

	ASSERT(tpiso == (struct tpisocket *)wq->q_ptr);

	if (!(tpiso->tso_flags & TSOF_ISOPEN)) {
		sdrv_error(wq, bp);
		goto done;
	}

	switch (bp->b_datap->db_type) {
	case M_FLUSH:
		if (*bp->b_rptr & FLUSHBAND) {
			if (*bp->b_rptr & FLUSHW)
				flushband(wq, FLUSHDATA, *(bp->b_rptr + 1));
			if (*bp->b_rptr & FLUSHR)
				flushband(RD(wq), FLUSHDATA, *(bp->b_rptr + 1));
		} else {
			if (*bp->b_rptr & FLUSHW) {
				flushq(wq, FLUSHDATA);
				/* XXX: flush socket output? */
			}
			if (*bp->b_rptr & FLUSHR) {
				flushq(RD(wq), FLUSHDATA);
				/* XXX: flush socket input? */
			}
		}
		freemsg(bp);
		break;

	case M_DATA:
	case M_DELAY:
	case M_PROTO:
	case M_PCPROTO:
	case M_IOCTL:
	case M_IOCDATA:
	case M_CTL:
		if (tpiso->tso_flags & (TSOF_QUEUE_OUTPUT |
					TSOF_SEND_PENDING)) {
			TSO_PUTWQ(tpiso, bp);
			goto done;
		}
		(void)tpisocket_wact(tpiso, bp);
		break;
	default:
		sdrv_error(wq, bp);
		break;
	}
done:
#ifdef _MP_NETLOCKS
	TSO_RELE(tpiso);
	if ((tpiso->tso_count == 1) &&
	    (tpiso->tso_flags & TSOF_CLOSING)) {
		sv_broadcast(&tpisockd_inuse);
	}
#endif
	return(0);
}


/*
 *	tpisocket_wsrv
 *		Process queued output messages
 */
int
tpisocket_wsrv(queue_t *wq)
{
#ifdef _MP_NETLOCKS
	struct tpisocket *tpiso = (struct tpisocket *)wq->q_ptr;

	if (! tpiso)
		return 0;

	if (!(tpiso->tso_flags & TSOF_ISOPEN) ||
	    tpiso->tso_flags & TSOF_CLOSING) {
		flushq(wq, FLUSHDATA);
		tpiso->tso_flags &= ~TSOF_WSRV_QUEUED;
		return 0;
	}

	TSO_HOLD(tpiso);
	tpisockd_call((void (*)())_tpisocket_wsrv, wq, NULL, tpiso);
	return 0;
}

/* ARGSUSED */
static int
_tpisocket_wsrv(queue_t *wq, void *arg2, struct tpisocket *tpiso)
{
#else
	struct tpisocket *tpiso = (struct tpisocket *)wq->q_ptr;
#endif
	mblk_t *bp;

	/*
	 * stream may be in process of being dismantled. Just quit then.
	 */
	if ((tpiso == NULL) || (tpiso->tso_flags & TSOF_CLOSING))
		goto done;

	ASSERT(tpiso == (struct tpisocket *)wq->q_ptr);

	tpiso->tso_flags &= ~TSOF_WSRV_QUEUED;

	if (!(tpiso->tso_flags & TSOF_ISOPEN)) {
		flushq(wq, FLUSHDATA);
		goto done;
	}

	if (tpiso->tso_flags & TSOF_SEND_PENDING &&
	    ! tpisocket_send_check(tpiso)) {
		goto done;
	}

	enableok(wq);
	while ((bp = getq(wq)) != NULL) 
		if (! tpisocket_wact(tpiso, bp)) {
			goto done;
		}
	tpiso->tso_flags &= ~TSOF_QUEUE_OUTPUT;
done:
#ifdef _MP_NETLOCKS
	TSO_RELE(tpiso);
	if ((tpiso->tso_count == 1) &&
	    (tpiso->tso_flags & TSOF_CLOSING)) {
		sv_broadcast(&tpisockd_inuse);
	}
#endif
	return(0);
}


/*
 *	tpisocket_release_send_resources
 *
 *	Releases data blocks used in sending 
 */
void
tpisocket_release_send_resources( struct tpisocket *tpiso)
{
	struct	mbuf *mp;
	mblk_t	*bp;

#define RELEASE_BUFFER_BLK(field, flag) { \
	if (tpiso->tso_pending_send_flags & flag) { \
		mp = tpiso->field.mbuf; \
		if (mp != NULL) { \
			m_freem(mp); \
			tpiso->field.mbuf = NULL; \
		} \
	} else { \
		bp = tpiso->field.mblk; \
		if (bp != NULL) { \
			freemsg(bp); \
			tpiso->field.mblk = NULL; \
		} \
	} }

	RELEASE_BUFFER_BLK(tso_pending_send,         TSOM_MSG_CONVERTED);
	RELEASE_BUFFER_BLK(tso_pending_send_addr,    TSOM_ADDR_CONVERTED);
	RELEASE_BUFFER_BLK(tso_pending_send_control, TSOM_CONTROL_CONVERTED);
	bp = tpiso->tso_pending_send_header;
	if (bp != NULL) {
		freemsg(bp);
		tpiso->tso_pending_send_header = NULL;
	}
}


/*
 *	tpisocket_release_recv_resources
 *
 *	Releases data blocks used in receiving
 */
void
tpisocket_release_recv_resources( struct tpisocket *tpiso)
{
	struct	mbuf *mp;
	mblk_t	*bp;

#define RELEASE_RECV_BUFFER_BLK(field, flag) { \
	if (!(tpiso->tso_pending_recv_flags & flag)) { \
		mp = tpiso->field.mbuf; \
		if (mp != NULL) { \
			m_freem(mp); \
			tpiso->field.mbuf = NULL; \
		} \
	} else { \
		bp = tpiso->field.mblk; \
		if (bp != NULL) { \
			freemsg(bp); \
			tpiso->field.mblk = NULL; \
		} \
	} }

	if (tpiso->tso_pending_recv_flags & TSOM_MSG_PRESENT) {
		RELEASE_RECV_BUFFER_BLK(tso_pending_recv, TSOM_MSG_CONVERTED);
		RELEASE_RECV_BUFFER_BLK(tso_pending_recv_addr, TSOM_ADDR_CONVERTED);
		RELEASE_RECV_BUFFER_BLK(tso_pending_recv_control, TSOM_CONTROL_CONVERTED);
	}
}


/*
 *	tpisocket_send_check
 *		Process pending mbuf send
 */
int
tpisocket_send_check( struct tpisocket *tpiso)
{
	struct mbuf *mp;
	mblk_t	*bp;
	int	error;
	int	pkt_type;
	struct mbuf *cp;
	int space, sendlen;
#ifdef RENO
	int first_time;
#endif

	if (!(tpiso->tso_flags & TSOF_SEND_PENDING))
		return 1;

	pkt_type = (tpiso->tso_pending_send_header != NULL 
			? ((union T_primitives *) 
				tpiso->tso_pending_send_header->b_rptr)->type
			: T_DATA_REQ);

	if (tpiso->tso_pending_send_error)
		goto process_send_error;

	/* convert buffers from mblk to mbuf */
#define CONVERT_SEND_BUFFER(ptr, flag, mbuf_type, mbuf_flags) { \
	if (!(tpiso->tso_pending_send_flags & flag) && \
	    tpiso->ptr.mblk != NULL) { \
		mp = tpisocket_mblk_to_mbuf(tpiso->ptr.mblk, \
			M_DONTWAIT, mbuf_type, mbuf_flags, tpiso->tso_wq); \
		if (mp == NULL) \
			return 0; \
		tpiso->ptr.mbuf = mp; \
		tpiso->tso_pending_send_flags |= flag; \
	} }

	CONVERT_SEND_BUFFER(tso_pending_send,         TSOM_MSG_CONVERTED, MT_DATA, 0);
	CONVERT_SEND_BUFFER(tso_pending_send_addr,    TSOM_ADDR_CONVERTED, MT_DATA, 0);
#ifdef RENO
	CONVERT_SEND_BUFFER(tso_pending_send_control, TSOM_CONTROL_CONVERTED, MT_CONTROL, 0);
#else
	ASSERT(tpiso->tso_pending_send_control.mblk == NULL);
#endif

	error = 0;

	cp = tpiso->tso_pending_send_addr.mbuf;
	if (cp != NULL &&
	    cp->m_next != NULL &&
	    cp->m_len < MLEN &&
	    (sendlen = m_length(cp)) > 0) {
		cp = tpisocket_dup_mbuf(cp, M_DONTWAIT, tpiso->tso_wq, sendlen);
		if (cp == NULL)
			return 0;
		cp = m_pullup(cp, min(MLEN, sendlen));
		if (cp == NULL) {
			tpisocket_request_mbufcall(tpiso->tso_wq);
			return 0;
		}
		m_freem(tpiso->tso_pending_send_addr.mbuf);
		tpiso->tso_pending_send_addr.mbuf = cp;
	}

#ifdef RENO
	cp = tpiso->tso_pending_send_control.mbuf;
	if (cp != NULL) {
		if (m_length(cp) > MLEN) {
			error = EMSGSIZE;
			tpiso->tso_pending_send_error = EMSGSIZE;
		} else {
			MGET(cp, M_DONTWAIT, tpiso->tso_pending_send_control.mbuf->m_type);
			if (cp == NULL) {
				tpisocket_request_mbufcall(tpiso->tso_wq);
				return 0;
			}
			cp->m_len = m_length(tpiso->tso_pending_send_control.mbuf),
			m_copydata(tpiso->tso_pending_send_control.mbuf,
				   0, cp->m_len, cp->m_data);
		}
	}
#else
	cp = NULL;
#endif

#ifdef RENO
	first_time = 1;
#endif
	while (error == 0) {
		int pkthdr_len;

		mp = tpiso->tso_pending_send.mbuf;
		if (mp != NULL) {
			pkthdr_len = m_length(mp);
			if (pkt_type == T_UNITDATA_REQ) {
				sendlen = pkthdr_len;
			} else {
				ASSERT(tpiso->tso_socket);
				space = sbspace(&tpiso->tso_socket->so_snd);
				if (space == 0) {
					m_freem(cp);
					return 0;
				}
				sendlen = min(space, pkthdr_len);
			}

			if (NULL == (mp = tpisocket_dup_mbuf(mp, M_DONTWAIT,
						    tpiso->tso_wq, sendlen))) {
				m_freem(cp);
				return 0;
			}
		} else if (cp == NULL && pkt_type != T_UNITDATA_REQ)
			break;

#ifdef RENO
		if (first_time  && tpiso->tso_pending_send_addr.mbuf) {
			mtod(tpiso->tso_pending_send_addr.mbuf,
				struct sockaddr *)->sa_family =
			    mtod(tpiso->tso_pending_send_addr.mbuf,
				    struct osockaddr *)->sa_family;
			mtod(tpiso->tso_pending_send_addr.mbuf,
				struct sockaddr *)->sa_len = 
			    tpiso->tso_pending_send_addr.mbuf->m_len;
		}
		first_time = 0;
#endif

		ASSERT(tpiso->tso_socket);
		/* send the message */
		SOCKET_MPLOCK(tpiso->tso_socket);
		error = tpisocket_sosend(
				tpiso->tso_socket,
				tpiso->tso_pending_send_addr.mbuf,
				NULL, mp, cp,
				tpiso->tso_pending_send_flags &
					TSOM_MSG_FLAGS_MASK);
		SOCKET_MPUNLOCK(tpiso->tso_socket);

		switch (error) {
		case EAGAIN:
		case EWOULDBLOCK:
			return 0;
		case ENOBUFS:
			tpisocket_request_mbufcall(tpiso->tso_wq);
			return 0;
		default:
			break;
		}
		tpiso->tso_pending_send_error = error;

		if ((pkt_type == T_UNITDATA_REQ) ||
		    (tpiso->tso_pending_send.mbuf == NULL) ||
		    (sendlen == pkthdr_len) ||
		    (NULL == (tpiso->tso_pending_send.mbuf = 
		          tpisocket_trim_mbuf_chain(
			      tpiso->tso_pending_send.mbuf, sendlen))))
			break;

		cp = NULL;
	}

process_send_error:
	if (tpiso->tso_pending_send_error &&
	    (tpiso->tso_flags & TSOF_SO_IMASOCKET ||
	     pkt_type == T_UNITDATA_REQ)) {
		union T_primitives *rpi;
		union T_primitives *pi;
		int	xsize;

		if (pkt_type == T_UNITDATA_REQ) {
			rpi = (union T_primitives *) 
				tpiso->tso_pending_send_header->b_rptr;
			xsize = rpi->unitdata_req.DEST_length +
				rpi->unitdata_req.OPT_length +
				sizeof(struct T_uderror_ind);
		} else
			xsize = sizeof(struct T_uderror_ind);

		bp = tpisocket_allocb(tpiso->tso_wq, xsize, BPRI_HI);
		if (bp == NULL)
			return 0;
		bp->b_datap->db_type = M_PROTO;
		bp->b_wptr = bp->b_rptr + xsize;
		pi = (union T_primitives *)bp->b_rptr;
		pi->type = T_UDERROR_IND;
		if (pkt_type != T_UNITDATA_REQ) {
			pi->uderror_ind.DEST_length = 0;
			pi->uderror_ind.DEST_offset = 0;
			pi->uderror_ind.OPT_length = 0;
			pi->uderror_ind.OPT_offset = 0;
		} else {
			pi->uderror_ind.DEST_length =
				rpi->unitdata_req.DEST_length;
			pi->uderror_ind.DEST_offset =
				rpi->unitdata_req.DEST_offset;
			bcopy(((caddr_t)rpi) + rpi->unitdata_req.DEST_offset,
			      ((caddr_t)pi) + pi->uderror_ind.DEST_offset,
			      pi->uderror_ind.DEST_length);
			pi->uderror_ind.OPT_length =
				rpi->unitdata_req.OPT_length;
			pi->uderror_ind.OPT_offset =
				rpi->unitdata_req.OPT_offset;
			bcopy(((caddr_t)rpi) + rpi->unitdata_req.OPT_offset,
			      ((caddr_t)pi) + pi->uderror_ind.OPT_offset,
			      pi->uderror_ind.OPT_length);
		}
		pi->uderror_ind.ERROR_type = tpiso->tso_pending_send_error;
		putnext(tpiso->tso_rq, bp);
	}
	tpiso->tso_pending_send_error = 0;

	/* release left-over resources */
	tpisocket_release_send_resources(tpiso);

	tpiso->tso_flags &= ~TSOF_SEND_PENDING;
	tpiso->tso_pending_send_flags = 0;
	return 1;

}




/*
 *	tpisocket_conn_req_done
 *		complete a pending connection request
 *
 *		Entered at splstr().
 */
void
tpisocket_conn_req_done( struct	tpisocket *tpiso)
{
#define tpictl tpiso->tso_control
	struct	socket *so = tpiso->tso_socket;
	struct	T_conn_con *cc;
	mblk_t	*bp;
	int	a_size;
	int	o_size;
	int	unix_error;

	if (!(tpiso->tso_flags & TSOF_CONN_REQ_PENDING))
		return; /* nothing pending */

	if ((so->so_state & SS_ISCONNECTING) && (so->so_error == 0))
		return; /* not done yet */

	unix_error = so->so_error;

	if (unix_error) { /* outstanding socket error */
		bp = tpiso->tso_pending_request;
		tpiso->tso_pending_request = NULL;
		tpiso->tso_flags &= ~TSOF_CONN_REQ_PENDING;
		tpiso->tso_flags |= TSOF_CONN_REQ_FAILED;

		if (tpiso->tso_flags & TSOF_RECV_DISCON) {
			freemsg(bp);	/* discard the message in this case */
			return;
		}

		/*
		 * When we haven't already received one,
		 * convert the pending request message into a
		 * discon_ind message containing the errno.
		 */
		tpisocket_send_discon_ind(tpiso, bp, unix_error);
		return;
	}

	a_size = TSC_ADDRESS_SIZE(tpiso, TSCA_REMOTE_ADDRESS, NULL);
	o_size = TSC_ADDRESS_SIZE(tpiso, TSCA_ACTUAL_OPTIONS, NULL);

	bp = tpisocket_allocb(tpiso->tso_rq,
			      sizeof(struct T_conn_con) + a_size + o_size,
			      BPRI_MED);
	if (bp == NULL)
		return;

	freemsg(tpiso->tso_pending_request);
	tpiso->tso_pending_request = NULL;
	tpiso->tso_flags &= ~TSOF_CONN_REQ_PENDING;

	bp->b_wptr = bp->b_rptr + (sizeof(struct T_conn_con)
				+ a_size + o_size);
	bp->b_datap->db_type = M_PROTO;
	cc = (struct T_conn_con *)bp->b_rptr;
	cc->PRIM_type = T_CONN_CON;
	cc->RES_length = a_size;
	cc->RES_offset = sizeof(struct T_conn_con) + o_size;
	cc->OPT_length = o_size;
	cc->OPT_offset = sizeof(struct T_conn_con);
	TSC_CONVERT_ADDRESS(tpiso,((caddr_t)cc) + cc->OPT_offset,
		    TSCA_OPTIONS | TSCA_MODE_READ, o_size, NULL, NULL);
	TSC_CONVERT_ADDRESS(tpiso,((caddr_t)cc) + cc->RES_offset,
		    TSCA_REMOTE_ADDRESS | TSCA_MODE_READ, a_size, NULL, NULL);
	putnext(tpiso->tso_rq, bp);

#undef tpictl
}


/*
 *	tpisocket_discon_req_done
 *
 *	Check if disconnection is complete.  If so, send "ok" message.
 */
void
tpisocket_discon_req_done( struct tpisocket *tpiso)
{
#define tpictl tpiso->tso_control
	struct	socket *so = tpiso->tso_socket;
	mblk_t	*bp;

	if (!(tpiso->tso_flags & TSOF_DISCON_REQ_PENDING))
		return; /* nothing pending */

	if (so->so_state & (SS_ISCONNECTING|SS_ISCONNECTED|SS_ISDISCONNECTING))
		return;

	bp = tpisocket_allocb(tpiso->tso_rq, sizeof(struct T_ok_ack), BPRI_MED);
	if (bp == NULL)
		return;
	tpisocket_send_ok_ack(tpiso, bp, T_DISCON_REQ);

	tpiso->tso_flags &= ~ TSOF_DISCON_REQ_PENDING;
#undef tpictl
}


/*
 *	tpisocket_optmgmt_req_done
 *
 *	Check if optmgmtnection is complete.  If so, send "ok" message.
 */

void
tpisocket_optmgmt_req_done( struct tpisocket *tpiso)
{
#define tpictl tpiso->tso_control
	mblk_t	*bp;
	struct T_optmgmt_ack *oma;
	int	more_to_do;
	int	unix_error;
	int	tli_error;

	if (!(tpiso->tso_flags & TSOF_OPTMGMT_REQ_PENDING))
		return; /* nothing pending */
	/*
	 * Note:  this code depends on T_optmgmt_ack and T_optmgmt_req
	 * being the same shape and size.
	 */
	more_to_do = TSCO_UNKNOWN;
	unix_error = TSC_OPTMGMT(tpiso, &more_to_do);
	bp = tpiso->tso_pending_optmgmt_req;
	if (unix_error) {
		tpisocket_convert_error(&unix_error, &tli_error);
		tpiso->tso_pending_optmgmt_req = NULL;
send_rejection:
		if (! tpisocket_wreject(tpiso, bp,
				  T_OPTMGMT_REQ, tli_error, unix_error))
			return;
	} else {
		switch (more_to_do) {
		default:
		case TSCO_UNKNOWN:
			tli_error = TSYSERR;
			unix_error = EINVAL;
			goto send_rejection;
		case TSCO_DONE:
			oma = (struct T_optmgmt_ack *)bp->b_rptr;
			oma->PRIM_type = T_OPTMGMT_ACK;
		case TSCO_DONE_DO_ACK:
			tpiso->tso_pending_optmgmt_req = NULL;
			putnext(tpiso->tso_rq, bp);
			break;
		case TSCO_DONE_NO_ACK:
			tpiso->tso_pending_optmgmt_req = NULL;
			freemsg(bp);
			break;
		case TSCO_MORE:
			return;
		}
	}
	tpiso->tso_flags &= ~TSOF_OPTMGMT_REQ_PENDING;

#undef tpictl
}


/*
 *	tpisocket_address_size
 *
 *	Return size of an address or control item.
 */
/*ARGSUSED*/
int
tpisocket_address_size(
	struct	tpisocket *tpiso,
	int	code,
	struct	socket *nso)
{
	switch (code) {
	case TSCA_LOCAL_ADDRESS:
	case TSCA_REMOTE_ADDRESS:
		return sizeof(struct sockaddr);
	case TSCA_OPTIONS:
		return(1024); /* SVR4 value */
	case TSCA_ACTUAL_OPTIONS:
		return(0);
	default:
		return 0;
	}
}


/*ARGSUSED*/
int
tpisocket_convert_address(
	struct	tpisocket *tpiso,
	char   *ptr,
	int	mode,
	int	size,
	struct	mbuf *m,
	struct	socket *nso)
{
	struct	socket *so;
	int	error;

	switch (mode & TSCA_CODE) {
	case TSCA_LOCAL_ADDRESS:
	case TSCA_REMOTE_ADDRESS:
		break;
	case TSCA_OPTIONS:
	default:
		return EINVAL;
	}

	if (nso) { /* non-zero means the 'nso' socket is already locked */
		so = nso;
	} else { /* NULL 'nso' means get the socket lock on 'so' */
		so = tpiso->tso_socket;
		SOCKET_MPLOCK(so);
	}

	switch (mode & TSCA_MODE) {
	case TSCA_MODE_READ:
		m = tpiso->tso_temporary_mbuf;
		error = (*so->so_proto->pr_usrreq)(so,
				((mode & TSCA_CODE) == TSCA_LOCAL_ADDRESS
					? PRU_SOCKADDR : PRU_PEERADDR), 
				0, m, 0);
		if (error) {
			if (!nso) {
				SOCKET_MPUNLOCK(so);
			}
			return error;
		}
		if (size > m->m_len)
			bzero(ptr, size);
		bcopy(mtod(m, caddr_t), ptr, min(size, m->m_len));
#ifdef RENO
		((struct sockaddr *)ptr)->sa_len = 0;
#endif
		break;

	case TSCA_MODE_TO_SOCKETS:
		m->m_len = 0;
		m->m_len = min(M_TRAILINGSPACE(m), size);
		bcopy(ptr, mtod(m, caddr_t), m->m_len);

		switch (mode & TSCA_CODE) {
		case TSCA_LOCAL_ADDRESS:
		case TSCA_REMOTE_ADDRESS:
#ifdef RENO
			mtod(m, struct sockaddr *)->sa_family =
				((struct osockaddr *)ptr)->sa_family;
			mtod(m, struct sockaddr *)->sa_len = m->m_len;
#endif
			break;

		case TSCA_OPTIONS:
			break;
		default:
			error = EINVAL;
			break;
		}
		break;

	case TSCA_MODE_TO_STREAMS:
		bcopy(mtod(m, caddr_t), ptr, min(m->m_len, size));
		switch (mode & TSCA_CODE) {
		case TSCA_LOCAL_ADDRESS:
		case TSCA_REMOTE_ADDRESS:
#ifdef RENO
			((struct osockaddr *)ptr)->sa_family =
				((struct sockaddr *)ptr)->sa_family;
#endif
			break;

		case TSCA_OPTIONS:
			break;
		default:
			error = EINVAL;
			break;
		}
		break;

	case TSCA_MODE_WRITE:
	default:
		error = EINVAL;
		break;
	}
	if (!nso) {
		SOCKET_MPUNLOCK(so);
	}
	return error;
}

/*
 * 'so' is head (listen) socket
 */
void
tpisocket_scan_conn_q(
	struct tpisocket *tpiso,
	struct socket *so)		/* head (listen) socket */
{
	struct socket *nso, *next;

#define tpictl (tpiso->tso_control)

	/*
	 * Although the TPI state machine explicitly allows more than
	 * one T_CONN_IND to be outstanding (sent up to the user process)
	 * at a time, the TLI code in libnsl fails if a second T_CONN_IND
	 * arrives at the stream head before the first one is completely
	 * processed.  So, if the count of "outstanding" T_CONN_INDs is
	 * non-zero, we won't send any more.
	 *
	 * Note that merely counting the sockets with the SS_CONN_IND_SENT
	 * bit set doesn't suffice, because if a RST arrives before the
	 * T_CONN_IND is replied to, the following chain of function calls
	 * dequeues and discards the socket with the SS_CONN_IND_SENT bit:
	 * tcp_input->tcp_close->in_pcbdetach->sofree->soqremque
	 */
	ASSERT(tpiso->tso_outcnt < 2);
	if (tpiso->tso_outcnt)
		return;

	ASSERT(SOCKET_ISLOCKED(so));

	/*
	 * scan the connection queue on the socket
	 */
restart_scan:

	for (nso = so->so_q; nso != NULL && nso != so; nso = next) {
		next = nso->so_q;

#ifdef _MP_NETLOCKS
		ASSERT(mutex_owner(&(nso)->so_sem) != private.p_curproc);

		if (!SOCKET_CMPLOCK(nso)) { /* failed to get child's lock */
			mon_state_t ms;
			proc_t *pp;

			SOCKET_MPUNLOCK(so);

			/*
			 * since we call qswtch directly, we
			 * must release and reaquire the the
			 * streams monitor as we yield the cpu.
			 */
			pp = private.p_curproc;
			ASSERT(pp->p_kthread.k_monitor);
			rmonitor(pp->p_kthread.k_monitor, &ms);

			qswtch(RESCHED_Y);

			amonitor(&ms);
			
			SOCKET_MPLOCK(so);
			goto restart_scan;
		}
#endif
		/*
		 * Arrive here with parent's 'so' and child's 'nso' locks held
		 * Check if connection broken on child before accept finished
		 */
		if (nso->so_state & SS_CANTRCVMORE) {
			/* connection broken before accept */

			if (!(nso->so_state & SS_CONN_IND_SENT)) {

				/*
				 * CONN_IND not yet sent: discard 
				 * connection (otherwise, wait for
				 * CONN_RES or DISCON_REQ from user)
				 */

				 /* dequeue the socket */
				{
					register s1;
					struct socket *h = nso->so_head;
					SOCKET_MPQLOCK(h->so_qlock, s1);
					soqremque(nso, 1);
					SOCKET_MPQUNLOCK(h->so_qlock, s1);
				}

				/* release parent lock to avoid deadlocks */
				SOCKET_MPUNLOCK(so);
				nso->so_state &= ~SS_CONN_IND_SENT;

				/*
				 * Discard the socket
				 * NOTE: lock on 'nso' MUST be held
				 * on the call to soabort()
				 * abort relases lock on nso
				 * as socket is destroyed
				 */
				if (soabort(nso)) {
					SOCKET_MPUNLOCK(nso);
				}

				/* reclaim parent lock for loop invariant */
				SOCKET_MPLOCK(so);
				continue;
			}
		}

		if (nso->so_state & SS_CONN_IND_SENT) {
			/* ASSERT(tpiso->tso_outcnt); */
			SOCKET_MPUNLOCK(nso);
			continue;
		}

		if (! tpiso->tso_outcnt &&
		   (nso->so_state & (SS_ISCONNECTED | SS_ISCONFIRMING))) {
			int	a_size;
			int	o_size;
			mblk_t	*bp;
			union	T_primitives *pi;

			/*
			 * connection with CONN_IND not yet sent 
			 * and no other CONN_IND is pending
			 */
			a_size = TSC_ADDRESS_SIZE(tpiso,
					TSCA_REMOTE_ADDRESS, nso);
			o_size = TSC_ADDRESS_SIZE(tpiso, 
					TSCA_ACTUAL_OPTIONS, nso);
			bp = tpisocket_allocb(tpiso->tso_rq,
				sizeof(struct T_conn_ind) + a_size + o_size,
				BPRI_MED);
			if (bp == NULL) {
				SOCKET_MPUNLOCK(nso);
				break;
			}
			bp->b_wptr = bp->b_rptr +
				(sizeof(struct T_conn_ind) + a_size + o_size);
			bp->b_datap->db_type = M_PROTO;
			pi = (union T_primitives *)bp->b_rptr;
			pi->type = T_CONN_IND;
			ASSERT(NET64_IN_RANGE(nso));
			pi->conn_ind.SEQ_number = NET64_TOINT(nso);
			pi->conn_ind.SRC_length = a_size;
			pi->conn_ind.SRC_offset =
				sizeof(struct T_conn_ind) + o_size;
			pi->conn_ind.OPT_length = o_size;
			pi->conn_ind.OPT_offset = sizeof(struct T_conn_ind);
			nso->so_state |= (SS_CONN_IND_SENT | SS_TPISOCKET);
			TSC_CONVERT_ADDRESS(tpiso,
				((char *)pi) + pi->conn_ind.OPT_offset,
				TSCA_OPTIONS | TSCA_MODE_READ,
				o_size, NULL, nso);
			TSC_CONVERT_ADDRESS(tpiso,
				((char *)pi) + pi->conn_ind.SRC_offset,
				TSCA_REMOTE_ADDRESS | TSCA_MODE_READ,
				a_size, NULL, nso);

			tpiso->tso_outcnt++;
			SOCKET_MPUNLOCK(nso);

			putnext(tpiso->tso_rq, bp);
		} else {
			/* some other state? at least release lock */
			SOCKET_MPUNLOCK(nso);
		}
	}
}

/*
 *	tpisocket_rsrv
 *
 *	Process incoming messages from the socket and post them
 *	to the stream.
 *
 *	Entered with queuerun blocked (including from queuerun).
 */
int
tpisocket_rsrv(queue_t	*rq)
{
#ifdef _MP_NETLOCKS
	struct tpisocket *tpiso = (struct tpisocket *)rq->q_ptr;

	if (! tpiso)
                return 0;

	if (!(tpiso->tso_flags & TSOF_ISOPEN) ||
	    !(tpiso->tso_socket) ||
	     (tpiso->tso_flags & TSOF_CLOSING)) {
		tpiso->tso_flags &= ~TSOF_RSRV_QUEUED;
		return 0;
	}

	TSO_HOLD(tpiso);
	tpisockd_call((void (*)())_tpisocket_rsrv, rq, NULL, tpiso);

	return 0;
}

/* ARGSUSED */
static int
_tpisocket_rsrv(queue_t *rq, void *arg2, struct tpisocket *tpiso)
{
#else
	struct tpisocket *tpiso = (struct tpisocket *)rq->q_ptr;
#endif
	struct	socket *so;
	mblk_t	*bp;
	union	T_primitives *pi;
	int	prim;
	int	size;
	int	pri;
	mblk_t	*ep;
	mblk_t	*cp;

	ASSERT(tpiso != 0);
	ASSERT(tpiso == (struct tpisocket *)rq->q_ptr);

	if (! tpiso)
		goto done;

	tpiso->tso_flags &= ~TSOF_RSRV_QUEUED;

	if (!(tpiso->tso_flags & TSOF_ISOPEN) ||
	    !(so = tpiso->tso_socket) ||
	     (tpiso->tso_flags & TSOF_CLOSING)) {
		goto done;
	}
   
	/* check if waiting for special state changes */
	while (tpiso->tso_flags & TSOF_SET_RECV_CHECK) {
		if (tpiso->tso_flags & TSOF_CONN_REQ_PENDING) {
			tpisocket_conn_req_done(tpiso);
			if (tpiso->tso_flags & TSOF_CONN_REQ_PENDING)
				goto done;
			continue;
		}
		if (tpiso->tso_flags & TSOF_DISCON_REQ_PENDING) {
			tpisocket_discon_req_done(tpiso);
			if (tpiso->tso_flags & TSOF_DISCON_REQ_PENDING)
				goto done;
			continue;
		}
		if (tpiso->tso_flags & TSOF_OPTMGMT_REQ_PENDING) {
			tpisocket_optmgmt_req_done(tpiso);
			if (tpiso->tso_flags & TSOF_OPTMGMT_REQ_PENDING)
				goto done;
			continue;
		}
		goto done;
	}

	SOCKET_MPLOCK(so);

	/*
	 * Bug# 438858
	 * We can block in SOCKET_MPLOCK() and during this time,
	 * tpisocket_close() can be called. We are holding the tpisocket
	 * structure so we are safe there. However, the stream will be
	 * dismantled and we will no longer be able to send a message
	 * up stream. Since the stream will have been dismantled, there
	 * is no reason to even continue processing and we will drop out
	 * the bottem (after releasing socket lock). The tpisocket_dismantle
	 * routine will clean up the socket if we are now the last user of
	 * the socket (by calling soclose).
	 */
	if (!(tpiso->tso_flags & TSOF_ISOPEN) ||
	    !(so = tpiso->tso_socket) ||
	     (tpiso->tso_flags & TSOF_CLOSING)) {
		goto exit;
	}

	if (!(so->so_state & SS_ISCONFIRMING))
	    /* pull off pending data */
	    while ((so->so_rcv.sb_cc > 0) || 
                   ((so->so_rcv.sb_cc ==0) && (tpiso->tso_flags & TSOF_OOB_INPUT_AVAILABLE))) {
		if (! canput(tpiso->tso_rq->q_next)) 
			goto exit;
		if (!(tpiso->tso_pending_recv_flags & TSOM_MSG_PRESENT)) {
		    int	error;
		    struct uio uio;

		    tpiso->tso_pending_recv.mbuf = NULL;
		    tpiso->tso_pending_recv_flags = 0;
		    uio.uio_resid = so->so_rcv.sb_cc;
		    if (tpiso->tso_flags & TSOF_OOB_INPUT_AVAILABLE)
			tpiso->tso_pending_recv_flags |= MSG_OOB;
#ifdef RENO
		    else {
			struct	mbuf *m;

			for (m = so->so_rcv.sb_mb; m != NULL; m = m->m_next)
				if (m->m_type != MT_SONAME)
					break;
			for (; m != NULL; m = m->m_next)
				if (m->m_type != MT_CONTROL)
					break;
			for (; m != NULL; m = m->m_next)
				if (m->m_len != 0 || m->m_flags & M_PKTHDR)
					break;
			for (; m != NULL; m = m->m_next)
				if (m->m_len != 0)
					break;

			if (TSC_SOCKET_TYPE == SOCK_DGRAM) {
				uio.uio_resid = m_length(m);
			} else { /* stream so gobble part of what's available*/
				if (m == NULL)
					uio.uio_resid = 0;
				else if (m->m_flags & M_PKTHDR &&
					 m->m_pkthdr.len > 0) {
					uio.uio_resid = m->m_pkthdr.len;
				} else
					uio.uio_resid = m->m_len;
			}
		    }
#endif
		    error = tpisocket_soreceive(so,
				      &tpiso->tso_pending_recv_addr.mbuf,
				      &uio,
				      &tpiso->tso_pending_recv.mbuf,
				      &tpiso->tso_pending_recv_control.mbuf,
				      &tpiso->tso_pending_recv_flags);
		    if (error) {
			if (error == EWOULDBLOCK || error == EAGAIN) {
				tpisocket_request_mbufcall(rq);
				error = 0;
				break;	/* return; ?? <<< */
			}
			if (tpiso->tso_flags & TSOF_OOB_INPUT_AVAILABLE) {
			    tpiso->tso_flags &= ~TSOF_OOB_INPUT_AVAILABLE;
			    continue;
			}
			break;
		    }
		    tpiso->tso_pending_recv_flags |= TSOM_MSG_PRESENT;
		} /* NB: this is _NOT_ where the breaks come. */

		/* convert the message and send up the stream */

#ifdef RENO
		if (tpiso->tso_pending_recv_addr.mbuf != NULL)
			mtod(tpiso->tso_pending_recv_addr.mbuf,
				struct osockaddr *)->sa_family =
			    mtod(tpiso->tso_pending_recv_addr.mbuf,
				    struct sockaddr *)->sa_family;
#endif

		if (tpiso->tso_pending_recv_flags & MSG_OOB) 
			pri = BPRI_MED;
		else
			pri = BPRI_LO;

#define CONVERT_RECV_BUFFER(ptr, flag) { \
		if (!(tpiso->tso_pending_recv_flags & flag) && \
		    tpiso->ptr.mbuf != NULL) { \
			bp = tpisocket_mbuf_to_mblk(tpiso->tso_rq, \
						    tpiso->ptr.mbuf, \
						    pri); \
			if (bp == NULL) \
				break; \
			tpiso->ptr.mblk = bp; \
			tpiso->tso_pending_recv_flags |= flag; \
		} }

		CONVERT_RECV_BUFFER(tso_pending_recv, TSOM_MSG_CONVERTED);
		CONVERT_RECV_BUFFER(tso_pending_recv_addr, TSOM_ADDR_CONVERTED);
		CONVERT_RECV_BUFFER(tso_pending_recv_control, TSOM_CONTROL_CONVERTED);

		if (tpiso->tso_pending_recv_flags & MSG_OOB) {
			size = sizeof(struct T_exdata_ind);
			prim = T_EXDATA_IND;
		} else if (tpiso->tso_pending_recv_addr.mblk != NULL ||
			   tpiso->tso_pending_recv_control.mblk != NULL) {
			size = sizeof(struct T_unitdata_ind);
			prim = T_UNITDATA_IND;
		} else {
			size = sizeof(struct T_data_ind);

			prim = T_DATA_IND;
		}

		ep = bp = tpisocket_allocb(tpiso->tso_rq, size, pri);
		if (bp == NULL) 
			goto exit;
		bp->b_datap->db_type = M_PROTO;
		bp->b_wptr += size;
		pi = (union T_primitives *)bp->b_rptr;
		pi->type = prim;
		if (prim == T_UNITDATA_IND) {
			pi->unitdata_ind.SRC_length = msgdsize(
				tpiso->tso_pending_recv_addr.mblk);
			pi->unitdata_ind.SRC_offset = size;
			pi->unitdata_ind.OPT_length = msgdsize(
				tpiso->tso_pending_recv_control.mblk);
			pi->unitdata_ind.OPT_offset = size +
				pi->unitdata_ind.SRC_length;
			if (tpiso->tso_pending_recv_addr.mblk) {
				for (cp = tpiso->tso_pending_recv_addr.mblk;
				     cp != NULL; cp = cp->b_cont)
					cp->b_datap->db_type = M_PROTO;
				str_conmsg(&bp, &ep,
					   tpiso->tso_pending_recv_addr.mblk);
			}
			if (tpiso->tso_pending_recv_control.mblk) {
				for (cp = tpiso->tso_pending_recv_addr.mblk;
				     cp != NULL; cp = cp->b_cont)
					cp->b_datap->db_type = M_PROTO;
				str_conmsg(&bp, &ep,
					   tpiso->tso_pending_recv_control.mblk);
			}
		} else {
			if (TSC_FLAGS & TSCF_NO_SEAMS) {
			    pi->data_ind.MORE_flag = 0;
			} else {
			    pi->data_ind.MORE_flag = 
				!(tpiso->tso_pending_recv_flags & MSG_EOR);
			}
			freemsg(tpiso->tso_pending_recv_addr.mblk);
			freemsg(tpiso->tso_pending_recv_control.mblk);
		}
		if (tpiso->tso_pending_recv.mblk) {
			str_conmsg(&bp, &ep, tpiso->tso_pending_recv.mblk);
		} else {
			if (bp == ep)
				tpiso->tso_pending_recv.mblk = bp;
		}

		tpiso->tso_pending_recv_flags = 0;
		putnext(tpiso->tso_rq, bp);
	    }

	if (! canput(tpiso->tso_rq->q_next)) 
		goto exit;

	/* look for connection state changes */
	if ((so->so_state & (SS_ISDISCONNECTING |
			     SS_CANTSENDMORE | SS_CANTRCVMORE)) ||
	    !(so->so_state & SS_ISCONNECTED) ||
	    (so->so_q != NULL && so->so_q != so)) {
		if ((so->so_state & SS_CANTRCVMORE) &&
		    !(tpiso->tso_flags & 
		       	    (TSOF_RECV_ORDREL | TSOF_CONN_REQ_FAILED))) {
			bp = tpisocket_allocb(tpiso->tso_rq,
					sizeof(struct T_ordrel_ind),
					BPRI_HI);
			if (bp == NULL) 
				goto exit;
			bp->b_datap->db_type = M_PROTO;
			bp->b_wptr = bp->b_rptr + sizeof(struct T_ordrel_ind);
			pi = (union T_primitives *)bp->b_rptr;
			pi->ordrel_ind.PRIM_type = T_ORDREL_IND;
			putnext(tpiso->tso_rq, bp);
			tpiso->tso_flags |= TSOF_RECV_ORDREL;
			goto exit;
		}
		if (((so->so_state & (SS_CANTSENDMORE | SS_CANTRCVMORE))
				== (SS_CANTSENDMORE | SS_CANTRCVMORE)) &&
		    !(tpiso->tso_flags & TSOF_RECV_DISCON)) {

			bp = tpisocket_allocb(tpiso->tso_rq,
			      sizeof(struct T_discon_ind), BPRI_HI);
			if (bp == NULL)
				goto exit;
			tpisocket_send_discon_ind(tpiso, bp, so->so_error);
			goto exit;
		}
		if (!(so->so_state & SS_ISCONNECTED) &&
		    (so->so_options & SO_ACCEPTCONN) &&
		    (so->so_q != NULL && so->so_q != so)) {
			tpisocket_scan_conn_q(tpiso, so);
		}
	}
exit:
	SOCKET_MPUNLOCK(so);
done:
#ifdef _MP_NETLOCKS
	TSO_RELE(tpiso);
	if ((tpiso->tso_count == 1) &&
	    (tpiso->tso_flags & TSOF_CLOSING)) {
		sv_broadcast(&tpisockd_inuse);
	}
#endif
	return(0);
#undef tpictl
}


/*
 *	tpisocket_do_optmgmt
 *
 *	This routine performs options processing for SVR4-style protocols
 */
int
tpisocket_do_optmgmt(
	struct	tpisocket *tpiso,
	int	*more_to_do_p,
	struct opproc  *functions)
{
	struct T_optmgmt_req *omr;
	int		flags;
	int		error = 0;
	struct opproc	*f;
	struct opthdr	*opt, *nopt, *eopt;
	int		level;
	mblk_t		*mp = NULL;
	mblk_t		*bp;

	bp = tpiso->tso_pending_optmgmt_req;
	if (bp == NULL ||
	    (bp->b_wptr - bp->b_rptr) < sizeof(struct T_optmgmt_req))
		return -TSYSERR;
	omr = (struct T_optmgmt_req *)bp->b_rptr;
	if ((omr->OPT_offset + omr->OPT_length) >
			(bp->b_wptr - bp->b_rptr))
		return -TBADOPT;

	switch (flags = omr->MGMT_flags) {
	case T_CHECK:
	case T_NEGOTIATE:
		if (!(mp = tpisocket_allocb(tpiso->tso_wq, 64, BPRI_MED))) {
			*more_to_do_p = TSCO_MORE;
			return 0;
		}
		mp->b_datap->db_type = M_PCPROTO;
		opt = (struct opthdr *)(bp->b_rptr + omr->OPT_offset);
		eopt = (struct opthdr *)((char *)opt + omr->OPT_length);
		if ((char *)eopt > (char *)bp->b_wptr) {
			error = -TSYSERR;
			goto done;
		}
		do {
			nopt = (struct opthdr *)((char *)(opt + 1) + opt->len);
			if (nopt > eopt) {
				error = -TBADOPT;
				goto done;
			}
			level = opt->level;
			for (f = functions; f->func; f++) {
				if (f->level == level) {
					if (error = (*f->func)(tpiso, omr, 
								opt, mp))
						goto done;
					break;
				}
			}
			if (!f->func) {
				if (flags == T_CHECK)
					omr->MGMT_flags = T_FAILURE;
				else
					error = -TBADOPT;
				goto done;
			}
			if (flags == T_CHECK && omr->MGMT_flags == T_FAILURE)
				goto done;
		} while ((opt = nopt) < eopt);
		if (flags == T_CHECK)
			omr->MGMT_flags = T_SUCCESS;
		break;

	case T_DEFAULT:
		if (!(mp = tpisocket_allocb(tpiso->tso_wq, 256, BPRI_MED))) {
			*more_to_do_p = TSCO_MORE;
			return 0;
		}
		mp->b_datap->db_type = M_PCPROTO;
		for (f = functions; f->func; f++) {
			if (error = (*f->func)(tpiso, omr, 0, mp))
				break;
		}
		break;

	default:
		error = - TBADFLAG;
		break;
	}

done:
	if (error && mp)
		freemsg(mp);
	if (error)
		return error;

	tpiso->tso_pending_optmgmt_req = NULL;
	omr->PRIM_type = T_OPTMGMT_ACK;
	omr->OPT_offset = sizeof(struct T_optmgmt_ack);
	omr->OPT_length = xmsgsize(mp, 1);
	bp->b_datap->db_type = M_PCPROTO;
	bp->b_wptr = bp->b_rptr + sizeof(struct T_optmgmt_ack);
	if (bp->b_cont)
		freemsg(bp->b_cont);
	bp->b_cont = mp;
	/*
	 * The types of other blocks in the message are irrelevant at this
	 * point.
	 */
	pullupmsg(bp, -1);
	putnext(tpiso->tso_rq, bp);
	*more_to_do_p = TSCO_DONE_NO_ACK;
	return 0;
}



/*
 *	tpisocket_makeopt
 *
 *	Append formatted option (opthdr + value) to mblk.
 *	Extends buffer if possible and necessary.  
 *	Returns 0 if unable to allocate space, 1 if successful.
 */
int
tpisocket_makeopt(
	mblk_t	*bp,		/* append formatted option to this mblk */
	int	level,
	int	name,
	void	*ptr,		/* buffer with option value */
	int	len)		/* length of buffer */
{
	struct opthdr  *opt;
	int             rlen, tlen;

	for (; bp->b_cont; bp = bp->b_cont)
		;
	rlen = OPTLEN(len);
	tlen = sizeof(struct opthdr) + rlen;
	if ((bp->b_datap->db_lim - bp->b_wptr) < tlen) {
		if (!(bp->b_cont = allocb(max(tlen, 64), BPRI_MED)))
			return 0;
		bp = bp->b_cont;
		bp->b_datap->db_type = M_PCPROTO;
	}
	opt = (struct opthdr *)bp->b_wptr;
	opt->level = level;
	opt->name = name;
	opt->len = rlen;
	if (rlen)
		bzero(OPTVAL(opt), rlen);
	if (len)
		bcopy(ptr, OPTVAL(opt), len);
	bp->b_wptr += tlen;
	return 1;
}

/*	tpisocket_optmgmt
 *		return 0 if not error, or errno, or negative Terrno.
 */
int
tpisocket_optmgmt(
	struct	tpisocket *tpiso,
	struct T_optmgmt_req *req,	/* from original request */
	struct opthdr  *opt,		/* from original request */
	mblk_t         *mp)		/* append answer to this mblk */
{
#define tpictl (tpiso->tso_control)
	struct	socket *so = tpiso->tso_socket;
	struct	mbuf	*m;
	struct	mbuf	*m2;
	int		error;

	switch (req->MGMT_flags) {

	case T_NEGOTIATE:
		/* m2 will hold copy of OPTVAL, discarded by sosetopt.  */
		MGET(m2, M_DONTWAIT, MT_SOOPTS);
		if (m2 == NULL)
			return ENOSR;
		m2->m_len = 0;
		if (opt->len > M_TRAILINGSPACE(m2)) {
			m_freem(m2);
			return -TBADOPT;
		}
		m2->m_len = opt->len;
		bcopy(OPTVAL(opt), mtod(m2, caddr_t), opt->len);

		/* m is allocated to hold response */
		m = tpiso->tso_pending_optmgmt_mbuf;
		tpiso->tso_pending_optmgmt_mbuf = NULL;
		if (m == NULL) {
			MGET(m, M_DONTWAIT, MT_SOOPTS);
			if (m == NULL) {
				m_free(m2);
				return ENOSR;
			}
		}

		switch (opt->name) {
		case SO_LINGER:
		case SO_DEBUG:
		case SO_KEEPALIVE:
		case SO_DONTROUTE:
		case SO_USELOOPBACK:
		case SO_BROADCAST:
		case SO_REUSEADDR:
		case SO_OOBINLINE:
		case SO_SNDLOWAT:
		case SO_RCVLOWAT:
		case SO_SNDBUF:
		case SO_RCVBUF:
			error = sosetopt(so, SOL_SOCKET, opt->name, m2);
			/* sosetopt frees m2 */
			break;

		case SO_IMASOCKET:
			if ((*(mtod(m2, int *))))
				tpiso->tso_flags |= TSOF_SO_IMASOCKET;
			else
				tpiso->tso_flags &= ~TSOF_SO_IMASOCKET;
			error = 0;
			m_free(m2);
			break;

		case SO_PROTOTYPE:
			/* The correct way to "negotiate" this parameter
			 * is to ignore what the user asks for, and 
			 * return the current value.
			 */
			error = 0;
			m_free(m2);
			break;

		default:
			error = EINVAL;
			m_free(m2);
			break;
		}

		if (error) {
			if (error == EINVAL)
				error = -TBADOPT;
			tpiso->tso_pending_optmgmt_mbuf = m;
			return error;
		}
		goto check_status;

	case T_CHECK:
		m = tpiso->tso_pending_optmgmt_mbuf;
		tpiso->tso_pending_optmgmt_mbuf = NULL;
		if (m == NULL) {
			MGET(m, M_DONTWAIT, MT_SOOPTS);
			if (m == NULL)
				return ENOSR;
		}
check_status:
		switch (opt->name) {
		case SO_LINGER:
		case SO_DEBUG:
		case SO_KEEPALIVE:
		case SO_DONTROUTE:
		case SO_USELOOPBACK:
		case SO_BROADCAST:
		case SO_REUSEADDR:
		case SO_OOBINLINE:
		case SO_SNDLOWAT:
		case SO_RCVLOWAT:
		case SO_SNDBUF:
		case SO_RCVBUF:
		case SO_TYPE:
			error = sogetopt(so, SOL_SOCKET, opt->name, &m);
			if (error) {
				ASSERT(error != ENOPROTOOPT);
				if (error == EINVAL)
					error = -TBADOPT;
				m_freem(m);
				return error;
			}
			/*
			 * for binary options, force "true" value to 1
			 */
			switch (opt->name) {
			case SO_DEBUG:
			case SO_KEEPALIVE:
			case SO_DONTROUTE:
			case SO_USELOOPBACK:
			case SO_BROADCAST:
			case SO_REUSEADDR:
			case SO_OOBINLINE:
				if (m->m_len == sizeof(int) &&
				    *mtod(m, int *) != 0)
					*mtod(m, int *) = 1;
				break;
			case SO_LINGER:
				if (m->m_len == sizeof(struct linger) &&
				    mtod(m, struct linger *)->l_onoff != 0)
					mtod(m,struct linger *)->l_onoff = 1;
				break;
  			}
			break;

		case SO_IMASOCKET:
			*mtod(m, int *) = 
				((tpiso->tso_flags & TSOF_SO_IMASOCKET) != 0);
			m->m_len = sizeof(int);
			break;

		case SO_PROTOTYPE:
			*mtod(m, int *) = TSC_PROTOCOL;
			m->m_len = sizeof(int);
			break;

		default:
			req->MGMT_flags = T_FAILURE;
			tpiso->tso_pending_optmgmt_mbuf = m;
			return 0;
		}
		if (! tpisocket_makeopt(mp, SOL_SOCKET, opt->name,
				mtod(m, caddr_t), m->m_len)) {
			tpiso->tso_pending_optmgmt_mbuf = m;
			return ENOSR;
		}
		tpiso->tso_pending_optmgmt_mbuf = m;
		break;

	case T_DEFAULT:
	{
		struct optdefault *o;
		int		val;

		/* get default values from table */
		for (o = sockdefault; o->optname; o++) {
			if (! tpisocket_makeopt(mp, SOL_SOCKET,
					     o->optname,
					     o->val, o->len))
				return ENOSR;
		}

		/* add default values that aren't in the table */
		val = TSC_DEFAULT_VALUE(tpiso, ((__psunsigned_t)TSCD_SNDBUF));
		if (! tpisocket_makeopt(mp, SOL_SOCKET, 
					SO_SNDBUF, &val, sizeof(int)))
			return ENOSR;
		val = TSC_DEFAULT_VALUE(tpiso, ((__psunsigned_t)TSCD_RCVBUF));
		if (! tpisocket_makeopt(mp, SOL_SOCKET, 
					SO_RCVBUF, &val, sizeof(int)))
			return ENOSR;
		val = TSC_DEFAULT_VALUE(tpiso, ((__psunsigned_t)TSCD_SNDLOWAT));
		if (! tpisocket_makeopt(mp, SOL_SOCKET,
					SO_SNDLOWAT, &val, sizeof(int)))
			return ENOSR;
		val = TSC_DEFAULT_VALUE(tpiso, ((__psunsigned_t)TSCD_RCVLOWAT));
		if (! tpisocket_makeopt(mp, SOL_SOCKET,
				        SO_RCVLOWAT, &val, sizeof(int)))
			return ENOSR;

		break;
	}

	default:
		break;
	}

	return 0;
#undef tpictl
}

#ifdef _IRIX_LATER
/*
 *	tpisocket_convert_socket
 *
 *	Convert a socket file descriptor to a TPI stream file descriptor
 */
int
tpisocket_convert_socket(
	struct	file *fp,
	struct cred *cred)
{
	struct	socket *so;
	struct	socket *oso;
	struct	tpisocket_control *tpictl;
	dev_t	newdev;
	struct	vnode *vp;
	int	saved_error;
	int	error;
	int	newmajor;
	int	newminor;
	struct	tpisocket *tpiso;

	if (fp->f_type != DTYPE_SOCKET)
		return EINVAL;

	so = (struct socket *)fp->f_data;

	/* calculate the clone device for the protocol */
	for (tpictl = tpisocket_protocols;
	     tpictl != NULL;
	     tpictl = tpictl->tsc_next)
		if (so->so_type == TSC_SOCKET_TYPE &&
		    so->so_proto->pr_protocol == TSC_PROTOCOL &&
		    so->so_proto->pr_domain->dom_family == TSC_DOMAIN)
			break;
	if (tpictl == NULL || tpisocket_clone_major < 0)
		return EPROTONOSUPPORT;
	newdev = makedevice(tpisocket_clone_major, TSC_MAJORS[0]);

	/* create a stream via a clone open */
	vp = makespecvp(newdev, VCHR, cred);
	ASSERT(vp != NULL);
	saved_error = curthreadp->k_error;
	error = VOP_OPEN(&vp, FWRITE|FREAD|FNDELAY, cred);
	if (error) {
		VN_RELE(vp);
		curthreadp->k_error = saved_error;
		return error;
	}

	/* locate the tpisocket structure for the device */
	newmajor = getemajor(vp->v_rdev);
	newminor = getminor(vp->v_rdev);
	if (TSC_MAJORS[0] != newmajor ||
	    newminor >= TSC_MAXDEVICES || 
	    NULL == (tpiso = TSC_DEVICES[newminor]) ||
	    !(tpiso->tso_flags & TSOF_UNUSED)) {
		VOP_CLOSE(vp, FWRITE|FREAD|FNDELAY, L_TRUE, (off_t)0, cred);
		VN_RELE(vp);
		curthreadp->k_error = saved_error;
		return EPROTONOSUPPORT;
	}

	/* change the file descriptor to a streams descriptor */
	s = splhi();
	fp->f_type = DTYPE_VNODE;
	fp->f_ops = &vnodefops;
	fp->f_vnode = vp;
	splx(s);

	/* swap the original socket for the unused new socket */
	s = splstr();
	oso = tpiso->tso_socket;
	oso->so_state        &= ~SS_WANT_CALL;
	oso->so_snd.sb_flags &= ~SB_NOTIFY;
	tpiso->tso_socket     = so;
	so->so_callout        = tpisocket_event;
	so->so_callout_arg    = (caddr_t)tpiso;
	so->so_state         |= (SS_TPISOCKET | SS_NBIO | SS_WANT_CALL);
	so->so_snd.sb_flags  |= SB_NOTIFY;
	so->so_state         &= ~SS_NOFDREF;
	tpiso->tso_flags     &= ~TSOF_UNUSED;
	tpiso->tso_flags     |= TSOF_ADDRESS_BOUND;

	/* queue old socket for discard */
	oso->so_callout_arg    = (caddr_t)tpisocket_discard_list; 
	tpisocket_discard_list = oso;
	wakeup(&tpisocket_discard_list);

	/* wakeup the new socket to look at the current state */
	tpisocket_event(so, SSCO_INPUT_ARRIVED,     tpiso);
	tpisocket_event(so, SSCO_OOB_INPUT_ARRIVED, tpiso);
	tpisocket_event(so, SSCO_OUTPUT_AVAILABLE,  tpiso);
	splx(s);

	curthreadp->k_error = saved_error;
	return 0;
}
#endif /* _IRIX_LATER */


/*
 *	tpisocket_sioc_name
 *		Process an SIOC*NAME ioctl
 */
int
tpisocket_sioc_name(
	struct tpisocket *tpiso,
	mblk_t	*bp,
	int	pru)
{
	struct socket	*so = tpiso->tso_socket;
	struct	mbuf 	*m = tpiso->tso_temporary_mbuf;
	mblk_t		*rbp;
	int		error;

	if ((so->so_state & (SS_ISCONNECTED|SS_ISCONFIRMING)) == 0)
		return tpisocket_send_iocnak(tpiso->tso_wq, bp, ENOTCONN);

	m->m_len = 0;
	rbp = tpisocket_allocb(tpiso->tso_wq, M_TRAILINGSPACE(m), BPRI_MED);
	if (rbp == NULL) {
		TSO_PUTBWQ(tpiso, bp);
		return 0;
	}
	SOCKET_MPLOCK(so);
 	error = (*so->so_proto->pr_usrreq)(so, pru, 0, m, 0);
	SOCKET_MPUNLOCK(so);

 	if (error) {
		freemsg(rbp);
		return tpisocket_send_iocnak(tpiso->tso_wq, bp, error);
	}
	rbp->b_wptr = rbp->b_rptr + m->m_len;
	bcopy(mtod(m, caddr_t), rbp->b_rptr, m->m_len);
#ifdef RENO
	((struct osockaddr *)rbp->b_rptr)->sa_family =
		((struct sockaddr *)rbp->b_rptr)->sa_family;
#endif
	return tpisocket_send_iocack(tpiso->tso_wq, bp, rbp);
}


/*
 *	tpisocket_ti_name
 *		Process a TI_*NAME ioctl
 */
int
tpisocket_ti_name(
	struct tpisocket *tpiso,
	mblk_t *bp)
{
	struct tpisocket_control *tpictl = tpiso->tso_control;
	mblk_t	*rbp;
	int	local_size;
	int	remote_size;
	int	error;

	local_size = TSC_ADDRESS_SIZE(tpiso, TSCA_LOCAL_ADDRESS, NULL);
	remote_size = TSC_ADDRESS_SIZE(tpiso, TSCA_REMOTE_ADDRESS, NULL);
	rbp = tpisocket_allocb(tpiso->tso_wq, local_size + remote_size, BPRI_MED);
	if (rbp == NULL) {
		TSO_PUTBWQ(tpiso, bp);
		return 0;
	}
	rbp->b_wptr = rbp->b_rptr + local_size + remote_size;
	error = TSC_CONVERT_ADDRESS(tpiso, (char *)rbp->b_rptr,
		    TSCA_LOCAL_ADDRESS | TSCA_MODE_READ, local_size, 
		    NULL, NULL);
	if (error) {
free_after_error:
		freemsg(rbp);
		switch (bp->b_datap->db_type) {
		case M_IOCTL:
			return tpisocket_send_iocnak(tpiso->tso_wq, bp, error);
		case M_IOCDATA:
			tpiso->tso_flags &= ~TSOF_IOCTL_PENDING;
			tpiso->tso_pending_ioctl = TSOI_NULL;
			break;
		default:
			break;
		}
		freemsg(bp);
		return 1;
	}
	error = TSC_CONVERT_ADDRESS(tpiso, (char *)rbp->b_rptr + local_size,
		    TSCA_REMOTE_ADDRESS | TSCA_MODE_READ, remote_size, 
		    NULL, NULL);
	if (error) 
		goto free_after_error;

	switch (bp->b_datap->db_type) {
	case M_IOCTL:
		tpiso->tso_flags |= TSOF_IOCTL_PENDING;
		tpiso->tso_pending_ioctl = TSOI_TI_NAME;
		break;

	case M_IOCDATA:
		if (tpiso->tso_flags & TSOF_IOCTL_PENDING &&
		    tpiso->tso_pending_ioctl == TSOI_TI_NAME) 
			break;	/* ok */
		/* else FALLTHRU */
	default:
		freemsg(bp);
		bp = 0;
		break;
	}
	if (bp &&
	    DONAME_CONT != ti_doname(tpiso->tso_wq, bp, 
				rbp->b_rptr,              local_size, 
				rbp->b_rptr + local_size, remote_size)) {
		tpiso->tso_flags &= ~TSOF_IOCTL_PENDING;
		tpiso->tso_pending_ioctl = TSOI_NULL;
	}
	freemsg(rbp);
	return 1;
}


/*
 *	tpisocket_ifioctl
 *
 *	process interface ioctl's
 */
int
tpisocket_ifioctl(
	struct	tpisocket *tpiso,
	mblk_t	*bp)
{
	struct iocblk *iocp;
	struct ifnet *ifp;
	struct ifreq *ifr;
	mblk_t	*nbp;
	struct socket *so;
	int	error;
	int	cmd;
	NETSPL_SPDECL(ss);  /* SOCKET_LOCK */

	if (tpiso->tso_flags & TSOF_IOCTL_PENDING) {
		if (tpiso->tso_pending_ioctl_req) {
			freemsg(tpiso->tso_pending_ioctl_req);
			tpiso->tso_pending_ioctl_req = NULL;
		}
		tpiso->tso_flags &= ~TSOF_IOCTL_PENDING;
		tpiso->tso_pending_ioctl = TSOI_NULL;
	}
	iocp = (struct iocblk *)bp->b_rptr;
	nbp = bp->b_cont;
	if (iocp->ioc_count > 0 &&
	    (nbp == NULL || (msgdsize(nbp) < iocp->ioc_count)))
		return tpisocket_send_iocnak(tpiso->tso_wq, bp, EINVAL);
	cmd = iocp->ioc_cmd;
	switch (cmd) {
	case SIOCGIFCONF:
	case SIOCGIFCONF_INTERNAL:
		return tpisocket_ifconf(tpiso, bp);

	case SIOCSARP:
	case SIOCDARP:
		if (tpiso->tso_cred->cr_uid != 0)
			return tpisocket_send_iocnak(tpiso->tso_wq, bp, EPERM);
		/* fall through */
	case SIOCGARP:
#ifdef notyet
		return tpisocket_arpioctl(tpiso, bp);
#else /* notyet */
		return tpisocket_send_iocnak(tpiso->tso_wq, bp, EOPNOTSUPP);
#endif /* notyet */
	case SIOCSIFFLAGS:
	case SIOCSIFMETRIC:
		if (tpiso->tso_cred->cr_uid != 0)
			return tpisocket_send_iocnak(tpiso->tso_wq, bp, EPERM);
		/* fall through */
	default:
		break;
	}

	if (iocp->ioc_count < sizeof(struct ifreq) ||
	    (nbp->b_wptr - nbp->b_rptr) < sizeof(struct ifreq))
		return tpisocket_send_iocnak(tpiso->tso_wq, bp, EINVAL);
	ifr = (struct ifreq *)nbp->b_rptr;
	ifp = ifunit(ifr->ifr_name);
	if (ifp == 0)
		return tpisocket_send_iocnak(tpiso->tso_wq, bp, ENXIO);
	so = tpiso->tso_socket;
	error = 0;

	switch (cmd) {
#ifdef _RISCOS
	case SIOCGCTLR:
#ifdef notyet
		if (ifp->if_ioctl)
			error = (*ifp->if_ioctl)(ifp, cmd, nbp->b_rptr);
#else /* notyet */
		error = EOPNOTSUPP;
#endif /* notyet */
		break;
#endif /* _RISCOS */

	case SIOCGIFFLAGS:
#ifdef _RISCOS
		ifr->ifr_flags = ifp->if_flags &
			~(IFF_OACTIVE | IFF_SIMPLEX | IFF_LANCE);
#else /* _RISCOS */
		ifr->ifr_flags = ifp->if_flags;
#endif /* _RISCOS */
		break;

	case SIOCGIFMETRIC:
		ifr->ifr_metric = ifp->if_metric;
		break;

	case SIOCSIFFLAGS:
#ifdef notyet
#ifdef _RISCOS

		if ((ifp->if_flags & IFF_UP) != (ifr->ifr_flags & IFF_UP)) {
			ifp->if_lastchange = time;
			timevalsub(&ifp->if_lastchange, &boottime);
		}
#endif /* _RISCOS */

		if (ifp->if_flags & IFF_UP && 
		    (ifr->ifr_flags & IFF_UP) == 0) {
			int s = splimp();
			if_down(ifp);
			splx(s);
		}
		ifp->if_flags = (ifp->if_flags & IFF_CANTCHANGE) |
			(ifr->ifr_flags &~ IFF_CANTCHANGE);
		if (ifp->if_ioctl)
			(void)(*ifp->if_ioctl)(ifp, cmd, nbp->b_rptr);
#else /* notyet */
		error = EOPNOTSUPP;
#endif /* notyet */
		break;

	case SIOCSIFMETRIC:
		ifp->if_metric = ifr->ifr_metric;
		break;

	default:
		if (so->so_proto == 0) {
			error = EOPNOTSUPP;
			break;
		}
#ifndef notyet
		switch (cmd) {
		case SIOCGIFDSTADDR:
		case SIOCGIFADDR:
		case SIOCGIFBRDADDR:
		case SIOCGIFNETMASK:
			break;
		default:
			return tpisocket_send_iocnak(tpiso->tso_wq, bp, 
				EOPNOTSUPP);
		}
#endif /* notyet */
#if ! (defined(COMPAT_43) && defined(RENO))
		SOCKET_LOCK(so, ss);
		error = ((*so->so_proto->pr_usrreq)(so, PRU_CONTROL,
				    (struct mbuf *)((__psunsigned_t)cmd),
				    (struct mbuf *)nbp->b_rptr,
				    (struct mbuf *) ifp));
		SOCKET_UNLOCK(so, ss);
#else
		{
			int ocmd = cmd;

			switch (cmd) {

			case SIOCSIFDSTADDR:
			case SIOCSIFADDR:
			case SIOCSIFBRDADDR:
			case SIOCSIFNETMASK:
#if BYTE_ORDER != BIG_ENDIAN
				if (ifr->ifr_addr.sa_family == 0 &&
				    ifr->ifr_addr.sa_len < 16) {
					ifr->ifr_addr.sa_family = ifr->ifr_addr.sa_len;
					ifr->ifr_addr.sa_len = 16;
				}
#else
				if (ifr->ifr_addr.sa_len == 0)
					ifr->ifr_addr.sa_len = 16;
#endif
				break;
			}
			SOCKET_LOCK(so, ss);
			error =  ((*so->so_proto->pr_usrreq)(so, PRU_CONTROL,
						cmd, nbp->b_rptr, ifp));
			SOCKET_UNLOCK(so, ss);
		}
		break;
#endif
	}
	if (error)
		return tpisocket_send_iocnak(tpiso->tso_wq, bp, error);
	bp->b_cont = NULL;
	return tpisocket_send_iocack(tpiso->tso_wq, bp, nbp);

}

/*
 *	tpisocket_ifiocdata
 *
 *		process M_IOCDATA continuation of ifioctl.
 */
int
tpisocket_ifiocdata(
	struct	tpisocket *tpiso,
	mblk_t	*bp)
{
	struct iocblk *iocp;
	struct copyresp *crs;
	mblk_t *rbp;
	int	error;

	if (tpiso->tso_pending_ioctl_req == NULL) {
		freemsg(bp);
		return 1;
	}

	iocp = (struct iocblk *)tpiso->tso_pending_ioctl_req->b_rptr;
	switch (iocp->ioc_cmd) {
	case SIOCGIFCONF:
	case SIOCGIFCONF_INTERNAL:
		break;

	default:
		freemsg(bp);
		return 1;
	}

	if ((bp->b_wptr - bp->b_rptr) < sizeof(struct copyresp)) {
		freemsg(bp);
		return 1;
	}
	crs = (struct copyresp *)bp->b_rptr;
	if (crs->cp_id != iocp->ioc_id) {
		freemsg(bp);
		return 1;
	}
	error = (__psunsigned_t)crs->cp_rval;
	freemsg(bp);
	bp = tpiso->tso_pending_ioctl_req;
	tpiso->tso_pending_ioctl_req = NULL;
	tpiso->tso_flags &= ~TSOF_IOCTL_PENDING;
	tpiso->tso_pending_ioctl = TSOI_NULL;
	if (error != 0) 
		return tpisocket_send_iocnak(tpiso->tso_wq, bp, error);
	rbp = bp->b_cont;
	bp->b_cont = NULL;
	return tpisocket_send_iocack(tpiso->tso_wq, bp, rbp);

}

/*
 *	tpisocket_ifconf
 *
 *	Process SIOCGIFCONF ioctl
 */
int
tpisocket_ifconf(
	struct tpisocket *tpiso,
	mblk_t	*bp)
{
	register struct ifconf *ifc;
	register struct ifnet *ifp = ifnet;
	struct in_ifaddr *ia;
	struct ifaddr *ifa;
	char *cp, *ep;
	struct iocblk *iocp;
	struct copyreq *crq;
	mblk_t *rbp;
	mblk_t *copybp;
	int	space;
	int	copyoffset;
	int	i, j;
	struct ifreq ifr;
	char	idstringp[10];

	iocp = (struct iocblk *)bp->b_rptr;
	if (iocp->ioc_cmd == SIOCGIFCONF_INTERNAL) {
		rbp = bp;
		space = iocp->ioc_count;
		if (space <= 0)
			return tpisocket_send_iocnak(tpiso->tso_wq, bp,EINVAL);
	} else {
		if (iocp->ioc_count < sizeof(struct ifconf) ||
		    (bp->b_cont->b_wptr - bp->b_cont->b_rptr) 
		      < sizeof(struct ifconf))
			return tpisocket_send_iocnak(tpiso->tso_wq, bp,EINVAL);

		ifc = (struct ifconf *)bp->b_cont->b_rptr;
		space = ifc->ifc_len;
		if (space <= 0)
			return tpisocket_send_iocnak(tpiso->tso_wq, bp,EINVAL);
		if (space > 0x8000) {
			ifc->ifc_len = 0x8000;
			space = 0x8000;
		}
		rbp = tpisocket_allocb(tpiso->tso_rq, sizeof(struct copyreq), 
			BPRI_LO);
		if (rbp == NULL) {
			TSO_PUTBWQ(tpiso, bp);
			return 0;
		}
		copyoffset = space;
		copybp = rbp;
		while (copyoffset > 0) {
			copybp->b_cont = tpisocket_allocb(tpiso->tso_rq, 
						min(copyoffset, 0x1000), 
						BPRI_LO);
			if (copybp->b_cont == NULL) {
				freemsg(rbp);
				TSO_PUTBWQ(tpiso, bp);
				return 0;
			}
			copybp = copybp->b_cont;
			copybp->b_wptr = copybp->b_datap->db_lim;
			copyoffset -= (copybp->b_wptr - copybp->b_rptr);
		}
		rbp->b_datap->db_type = M_COPYOUT;
		crq = ((struct copyreq *)rbp->b_rptr);
		rbp->b_wptr += sizeof(struct copyreq);
		bzero(rbp->b_rptr, sizeof(struct copyreq));
		crq->cq_cmd = iocp->ioc_cmd;
		crq->cq_uid = iocp->ioc_uid;
		crq->cq_gid = iocp->ioc_gid;
		crq->cq_id = iocp->ioc_id;
		crq->cq_addr = (caddr_t)ifc->ifc_req;
		crq->cq_flag = STRCOPYTRANS;
	}
	copybp = rbp->b_cont;
	copyoffset = 0;
	ep = ifr.ifr_name + sizeof(ifr.ifr_name) - 2;
	for (; space > sizeof(ifr) && ifp; ifp = ifp->if_next) {
		bcopy(ifp->if_name, ifr.ifr_name, 
			sizeof(ifr.ifr_name) - 2);
		for (cp = ifr.ifr_name; cp < ep && *cp; cp++)
			;
		for (i = ifp->if_unit, j = sizeof(idstringp);
		     j > 0;) {
			idstringp[--j] = '0' + (i % 10);
			i /= 10;
			if (i == 0)
				break;
		}
		while (cp <= ep && j < sizeof(idstringp))
			*cp++ = idstringp[j++];
		*cp = 0;

		ia = (struct in_ifaddr *)(ifp->in_ifaddr);
		ifa = (ia) ? &(ia->ia_ifa) : 0;

		if (ifa == 0) {
			bzero((caddr_t)&ifr.ifr_addr, sizeof(ifr.ifr_addr));
			tpisocket_copy_to_mblk_chain(
				(caddr_t)&ifr, sizeof(ifr),
				&copybp, &copyoffset);
			space -= sizeof(ifr);
		} else for ( ; space > sizeof(ifr) && ifa; 
			       ifa = ifa->ifa_next) {
#ifdef RENO
			register struct sockaddr *sa = ifa->ifa_addr;
#ifdef COMPAT_43
			if (iocp->ioc_cmd == SIOCGIFCONF_INTERNAL) {
				struct osockaddr *osa =
				    (struct osockaddr *)&ifr.ifr_addr;
				ifr.ifr_addr = *sa;
				osa->sa_family = sa->sa_family;
				tpisocket_copy_to_mblk_chain(
					(caddr_t)&ifr, sizeof(ifr), 
					&copybp, &copyoffset);
			} else
#endif
			if (sa->sa_len <= sizeof(*sa)) {
				ifr.ifr_addr = *sa;
				tpisocket_copy_to_mblk_chain(
					(caddr_t)&ifr, sizeof(ifr),
					&copybp, &copyoffset);
			} else {
				space -= sa->sa_len - sizeof(*sa);
				if (space < sizeof(ifr))
					break;
				tpisocket_copy_to_mblk_chain(
					(caddr_t)&ifr,
					sizeof(ifr.ifr_name),
					&copybp, &copyoffset);
				tpisocket_copy_to_mblk_chain(
					(caddr_t)sa, sa->sa_len,
					&copybp, &copyoffset);
			}
#else /* !RENO */
			register struct sockaddr *sa = ifa->ifa_addr;
			ifr.ifr_addr = *sa;
			tpisocket_copy_to_mblk_chain(
				(caddr_t)&ifr, sizeof(ifr),
				&copybp, &copyoffset);
#endif

			space -= sizeof(ifr);
		}
	}
	if (iocp->ioc_cmd == SIOCGIFCONF_INTERNAL) {
#ifdef nolonger
		if (space) {
			tpisocket_clear_mblk_chain(space, &copybp, 
				&copyoffset);
			space = 0;
		}
#endif /* nolonger */
		if (copybp != NULL) {
			copybp->b_wptr = copybp->b_rptr + copyoffset;
			if (copybp->b_cont != NULL) {
				freemsg(copybp->b_cont);
				copybp->b_cont = NULL;
			}
		}
		bp->b_datap->db_type = M_IOCACK;
		iocp->ioc_count -= space;
	} else {
		ifc->ifc_len -= space;
		rbp->b_cont->b_wptr += ifc->ifc_len;
		crq->cq_size = ifc->ifc_len;
		tpiso->tso_pending_ioctl_req = bp;
		tpiso->tso_flags |= TSOF_IOCTL_PENDING;
		tpiso->tso_pending_ioctl = TSOI_IFIOCTL;
	}
	qreply(tpiso->tso_wq, rbp);

	return 1;
}


/*
 *	tpisocket_rtioctl
 *
 *	process routing ioctl's		[STUB only for now]
 */
int
tpisocket_rtioctl(
	struct	tpisocket *tpiso,
	mblk_t	*bp)
{
	ASSERT (bp->b_datap->db_type == M_IOCTL);
	return tpisocket_send_iocnak(tpiso->tso_wq, bp, EOPNOTSUPP);
}

/* ARGSUSED */
int
tpisocket_rtiocdata(
	struct	tpisocket *tpiso,
	mblk_t	*bp)
{
	ASSERT (bp->b_datap->db_type == M_IOCDATA);
	freemsg(bp);
	return 1;
}

/*
 * XXX this seems to imply that there's no contention in turning on SB_LOCK
 */
int 
tpisocket_sblock(struct sockbuf *sb)
{
	int s = splimp();
	ASSERT((sb->sb_flags & SB_LOCK) == 0);
	sb->sb_flags |= SB_LOCK;
	splx(s);
	return 0;
}

void 
tpisocket_sbunlock(struct sockbuf *sb)
{
	int s = splimp();
	ASSERT((sb->sb_flags & SB_LOCK) != 0);
	sb->sb_flags &= ~SB_LOCK;
	splx(s);
}

/*
 * Send on a socket.
 *
 * This is an adaptation of the RISCOS/BSD_reno sosend for tpisocket.
 * It uses the reno feature of passing in an mbuf chain, 
 * and not using a uio.  To reduce bloat, much dead code was expunged.
 * Reno "control" buffers are not supported.
 * Formal parameters were kept the same as in reno.
 * 
 * If send must go all at once and message is larger than
 * send buffering, then hard error.
 * If must go all at once and not enough room now, then
 * inform user that this would block and do nothing.
 * Otherwise, if nonblocking, send as much as possible.
 * The data to be sent is described by the mbuf chain "top".  
 * Data provided in mbuf chain must be small enough to send all at once.
 *
 * Returns nonzero on error, timeout or signal; callers
 * must check for short counts if EINTR/ERESTART are returned.
 * Data and control buffers are freed on return.
 */
/* ARGSUSED */
int
tpisocket_sosend(
	register struct socket *so,
	struct mbuf 	*addr,
	struct uio	*uio,
	struct mbuf 	*top,
	struct mbuf	*control,
	int 		flags)
{
	register int	space;
	register int	resid;
	int 		atomic;
	int 		dontroute;
	int 		error;
	int 		s;

#define	snderr(errno)	{ error = errno; splx(s); goto release; }
#undef sblock
#define sblock(sb) tpisocket_sblock(sb)
#undef sbunlock
#define sbunlock(sb) tpisocket_sbunlock(sb)

	ASSERT(uio == NULL);
	ASSERT(control == NULL);

	atomic = sosendallatonce(so) || top;

	resid = m_length(top);
	dontroute = (flags & MSG_DONTROUTE) && 
		(so->so_options & SO_DONTROUTE) == 0 &&
		(so->so_proto->pr_flags & PR_ATOMIC);

	ASSERT((so->so_snd.sb_flags & SB_LOCK) == 0);
	if (error = sblock(&so->so_snd))
		goto out;

	s = splnet();
	if (so->so_state & SS_CANTSENDMORE)
		snderr(EPIPE);
	if (so->so_error)
		snderr(so->so_error);
	if ((so->so_state & SS_ISCONNECTED) == 0) {
		if (so->so_proto->pr_flags & PR_CONNREQUIRED) {
			if ((so->so_state & SS_ISCONFIRMING) == 0)
				snderr(ENOTCONN);
		} else if (addr == 0)
			snderr(EDESTADDRREQ);
	}
	space = sbspace(&so->so_snd);
	if (flags & MSG_OOB)
		space += 1024;
	if (space < resid &&
	    (atomic || space < so->so_snd.sb_lowat || space < 0)) {
		if (atomic && resid > so->so_snd.sb_hiwat)
			snderr(EMSGSIZE);
		snderr(EWOULDBLOCK);
	}
	if (dontroute)
		so->so_options |= SO_DONTROUTE;
	error = (*so->so_proto->pr_usrreq)(so,
			(flags & MSG_OOB) ? PRU_SENDOOB : PRU_SEND,
			top, addr, control);
	if (dontroute)
		so->so_options &= ~SO_DONTROUTE;
	splx(s);
	top = 0;

release:
	sbunlock(&so->so_snd);
out:
	if (top)
		m_freem(top);
	return error;
}

/*
 * Implement receive operations on a socket.
 *
 * Also adapted from RISCOS/BSD_reno.
 * Supports only non-uio style output (mp0 != 0);
 * Ignores contents of uio completely.
 * Returns the next message (mbuf chain) in the receive sockbuf (for DGRAMs)
 * or the next non-empty mbuf (for STREAM).
 * Since, at most, one mbuf is copied, and then data is only copied if it's
 * not a "cluster", we don't splx() during this code.
 * NB: flags will be either 0 or MSG_OOB, never MSG_PEEK.
 * 
 * The caller receives the data as a single mbuf chain by supplying
 * an mbuf **mp0 for use in returning the chain.  
 */
/*ARGSUSED*/
int
tpisocket_soreceive(
	register struct socket *so,
	struct mbuf 	**paddr,
	struct uio 	*uio,
	struct mbuf 	**mp0,
	struct mbuf 	**controlp,
	int 		*flagsp)
{
	register struct mbuf *m;
	register struct mbuf **mp;
	register int 	flags;
	register int 	len;
	register int 	error;
	register int 	s;
	struct protosw	*pr = so->so_proto;
	struct mbuf	*nextrecord;

	ASSERT(SOCKET_ISLOCKED(so));
	ASSERT(mp0    != NULL);
	ASSERT(paddr  != NULL);
	ASSERT(flagsp != NULL);

	mp = mp0;
	*mp = (struct mbuf *)0;
	*paddr = 0;
	if (controlp)
		*controlp = 0;
	flags = *flagsp & ~MSG_EOR;
	if (flags & MSG_OOB) {
		m = m_get(M_DONTWAIT, MT_DATA);
		if (m == NULL)
			return EWOULDBLOCK;
		error = (*pr->pr_usrreq)(so, PRU_RCVOOB, m, 
				(struct mbuf *)0, (struct mbuf *)0);
		if (error)
			m_freem(m);
		else
			*mp = m;
		return error;
	}
	if (so->so_state & SS_ISCONFIRMING)
		(*pr->pr_usrreq)(so, PRU_RCVD, (struct mbuf *)0,
		    (struct mbuf *)0, (struct mbuf *)0);

	if (error = sblock(&so->so_rcv))
		return error;
	s = splnet();
	m = so->so_rcv.sb_mb;
	ASSERT(m != NULL);
	nextrecord = m->m_act;
	/*
	 * NOTE: This code assumes that PR_ADDR and PR_ATOMIC are always
	 * identical (both true or both false).
	 */
	if (pr->pr_flags & PR_ADDR) {
		/* return entire next message. */
		ASSERT(m->m_type == MT_SONAME);
		sbfree(&so->so_rcv, m);
		*paddr = m;
		so->so_rcv.sb_mb = m->m_next;
		m->m_next = 0;
		m = so->so_rcv.sb_mb;
		if (m != NULL) {
			ASSERT(m->m_type == MT_DATA || m->m_type == MT_HEADER);
			len = m_length(m);
			so->so_rcv.sb_cc -= len;
			*mp = m;
			m = NULL;
		}
	} else {
		int oobmark;

		ASSERT(m != NULL);
		/* 
		 * return next non-empty mbuf, (up to oobmark) 
		 */
		so->so_state &= ~SS_RCVATMARK;
		mp = &so->so_rcv.sb_mb;
		while (m != NULL && m->m_len == 0) {
			*mp = m->m_next;
			m->m_next = NULL;
			m_free(m);
			m = *mp;
		}
		if (m == NULL)
			goto release;
		len = m->m_len;
		if ((oobmark = so->so_oobmark) != 0) {
			if (len > oobmark) { /* want partial mbuf */
				struct mbuf *mc;

				len = oobmark;
				if (NULL == (mc = m_copy(m, 0, len))) {
					error = EWOULDBLOCK;
					goto release;
				}
				mc->m_next = m;
				*mp = mc;
				m->m_off += len;
				m->m_len -= len;
			}
			if ((oobmark -= len) == 0)
				so->so_state |= SS_RCVATMARK;
			so->so_oobmark = oobmark;
		}
		*mp0 = m = so->so_rcv.sb_mb;
		sbfree(&so->so_rcv, m);
		so->so_rcv.sb_mb = m->m_next;
		m->m_next = 0;
		m = so->so_rcv.sb_mb;
	}

	if (m) {
		so->so_rcv.sb_mb = m;
		m->m_act = nextrecord;
	} else
		so->so_rcv.sb_mb = nextrecord;
	if (pr->pr_flags & PR_WANTRCVD && so->so_pcb)
		(*pr->pr_usrreq)(so, PRU_RCVD, (struct mbuf *)0,
		    (struct mbuf *)0, (struct mbuf *)0);
	*flagsp |= flags;
release:
	sbunlock(&so->so_rcv);
	splx(s);
	return error;
}
